#include <stdlib.h>
#include <endian.h>
#include <stddef.h>
#include <stdint.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdbool.h>
#include <assert.h>
#include "xfs.h"
#include "xfs-print.h"

struct xfs_info {
	int fd;
	uint32_t	sb_blocksize;
	xfs_agblock_t	sb_agblocks;
	xfs_agnumber_t	sb_agcount;
	uint16_t	sb_sectsize;
	int		sb_versionnum;
	xfs_fsblock_t	sb_logstart;
	xfs_extlen_t	sb_logblocks;
	/* log */
	uint8_t		sb_blocklog;
	uint8_t		sb_agblklog;
	uint8_t		m_blkbb_log;	/* blocklog - BBSHIFT */
	uint64_t	logBBsize;      /* size of log subvolume (BBs) */
					/* (blocks allocated for use as
					 * log is stored in mount structure) */
	uint64_t	logBBstart;     /* start block of log subvolume (BBs) */
};

#define MAX_AG_COUNT 10
struct xfs_debug_config_t {
	char		device_name[255];
	uint32_t	blocksize;
	xfs_agnumber_t	ag_count;
	uint64_t	ag_offset[MAX_AG_COUNT];
	uint64_t	log_offset;
	uint64_t	log_size;
};

static int _xfs_format_ver_num = 0;
static bool _print_config_action = false;

void print_usage(char *program_name) {
	printf("Usage: %s [DEVICE] [ACTION]\n", program_name);
	printf("ACTION:\n");
	printf("\t <NONE>    - print ondisk structures\n");
	printf("\t print-cfg - print configuration for xfs_diff script\n");
	printf("\n");
}

void print_superblock(xfs_sb_t *super, uint64_t offset)
{
	if (_print_config_action)
		return;
	PRE_XFS_PRINT(super, offset);

	printf("\n-----------------xfs_sb_t-----------------\n");
	printf("# description: superblock\n");
	printf("# offset: %"U64TF", size: %"SZOF"\n", offset, sizeof(xfs_sb_t));
	printf("offsetof(xfs_sb_t, sb_blocksize) = %"OFFTF"\n", offsetof(xfs_sb_t, sb_blocksize));
	XFS_PRINT(sb_magicnum, magic, "magic number == XFS_SB_MAGIC");
	XFS_PRINT(sb_blocksize, u32, "logical block size, bytes");
	XFS_PRINT(sb_dblocks, u64, "number of data blocks");
	XFS_PRINT(sb_rblocks, u64, "number of realtime blocks");
	XFS_PRINT(sb_rextents, u64, "number of realtime extents");
	XFS_PRINT(sb_uuid, uuid, "");
	XFS_PRINT(sb_logstart, u64, "starting block of log if internal");
	XFS_PRINT(sb_rootino, x64, "root inode number");
	XFS_PRINT(sb_rbmino, x64, "bitmap inode for realtime extents");
	XFS_PRINT(sb_rsumino, x64, "summary inode for rt bitmap");
	XFS_PRINT(sb_rextsize, u32, "realtime extent size, blocks");
	XFS_PRINT(sb_agblocks, u32, "size of an allocation group");
	XFS_PRINT(sb_agcount, u32, "number of allocation groups");
	XFS_PRINT(sb_rbmblocks, u32, "number of rt bitmap blocks");
	XFS_PRINT(sb_logblocks, u32, "number of log blocks");

	uint16_t versionnum = be16toh(super->sb_versionnum);
	XFS_PRINT(sb_versionnum, x16, "header version == XFS_SB_VERSION");
	printf("    VERSION_NUM: %u\n", versionnum & XFS_SB_VERSION_NUMBITS);
	printf("    ATTRBIT:     %u\n", !!(versionnum & XFS_SB_VERSION_ATTRBIT));
	printf("    NLINKBIT:    %u\n", !!(versionnum & XFS_SB_VERSION_NLINKBIT));
	printf("    QUOTABIT:    %u\n", !!(versionnum & XFS_SB_VERSION_QUOTABIT));
	printf("    ALIGNBIT:    %u\n", !!(versionnum & XFS_SB_VERSION_ALIGNBIT));
	printf("    DALIGNBIT:   %u\n", !!(versionnum & XFS_SB_VERSION_DALIGNBIT));
	printf("    SHAREDBIT:   %u\n", !!(versionnum & XFS_SB_VERSION_SHAREDBIT));
	printf("    LOGV2BIT:    %u\n", !!(versionnum & XFS_SB_VERSION_LOGV2BIT));
	printf("    SECTORBIT:   %u\n", !!(versionnum & XFS_SB_VERSION_SECTORBIT));
	printf("    EXTFLGBIT:   %u\n", !!(versionnum & XFS_SB_VERSION_EXTFLGBIT));
	printf("    DIRV2BIT:    %u\n", !!(versionnum & XFS_SB_VERSION_DIRV2BIT));
	printf("    BORGBIT:     %u\n", !!(versionnum & XFS_SB_VERSION_BORGBIT));
	printf("    MOREBITSBIT: %u\n", !!(versionnum & XFS_SB_VERSION_MOREBITSBIT));

	XFS_PRINT(sb_sectsize, u16, "volume sector size, bytes");
	XFS_PRINT(sb_inodesize, u16, "inode size, bytes");
	XFS_PRINT(sb_inopblock, u16, "inodes per block");
	xfs_print_str("sb_fname", super->sb_fname, "file system name",
		   offset + offsetof(xfs_sb_t, sb_fname));
	XFS_PRINT(sb_blocklog, u8, "log2 of sb_blocksize");
	XFS_PRINT(sb_sectlog, u8, "log2 of sb_sectsize");
	XFS_PRINT(sb_inodelog, u8, "log2 of sb_inodesize");
	XFS_PRINT(sb_inopblog, u8, "log2 of sb_inopblock");
	XFS_PRINT(sb_agblklog, u8, "log2 of sb_agblocks (rounded up)");
	XFS_PRINT(sb_rextslog, u8, "log2 of sb_rextents");
	XFS_PRINT(sb_inprogress, u8, "mkfs is in progress, don't mount");
	XFS_PRINT(sb_imax_pct, u8, "max percent of fs for inode space");

	printf("# statistics:\n");
	XFS_PRINT(sb_icount, u64, "allocated inodes");
	XFS_PRINT(sb_ifree, u64, "free inodes");
	XFS_PRINT(sb_fdblocks, u64, "free data blocks");
	XFS_PRINT(sb_frextents, u64, "free realtime extents");

	XFS_PRINT(sb_uquotino, x64, "user quota inode");
	XFS_PRINT(sb_gquotino, x64, "group quota inode");
	XFS_PRINT(sb_qflags, x16, "quota flags");
	XFS_PRINT(sb_flags, x8, "misc. flags");
	XFS_PRINT(sb_shared_vn, u8, "shared version number");
	XFS_PRINT(sb_inoalignmt, u32, "inode chunk alignment, fsblocks");
	XFS_PRINT(sb_unit, u32, "stripe or raid unit");
	XFS_PRINT(sb_width, u32, "stripe or raid width");
	XFS_PRINT(sb_dirblklog, u8, "log2 of dir block size (fsbs)");
	XFS_PRINT(sb_logsectlog, u8, "log2 of the log sector size");
	XFS_PRINT(sb_logsectsize, u16, "sector size for the log, bytes");
	XFS_PRINT(sb_logsunit, u32, "stripe unit size for the log");
	XFS_PRINT(sb_features2, x32, "additional feature bits");
	if (versionnum & XFS_SB_VERSION_MOREBITSBIT) {
		uint32_t sb_features2 = be32toh(super->sb_features2);
		printf("    LAZYSBCOUNTBIT: %u\n", !!(sb_features2 & XFS_SB_VERSION2_LAZYSBCOUNTBIT));
		printf("    ATTR2BIT:       %u\n", !!(sb_features2 & XFS_SB_VERSION2_ATTR2BIT));
		printf("    PARENTBIT:      %u\n", !!(sb_features2 & XFS_SB_VERSION2_PARENTBIT));
		printf("    PROJID32BIT:    %u\n", !!(sb_features2 & XFS_SB_VERSION2_PROJID32BIT));
		printf("    CRCBIT:         %u\n", !!(sb_features2 & XFS_SB_VERSION2_CRCBIT));
		printf("    FTYPE:          %u\n", !!(sb_features2 & XFS_SB_VERSION2_FTYPE));
	}
	XFS_PRINT(sb_bad_features2, x32, "");
	if (_xfs_format_ver_num > 4) {
		XFS_PRINT(sb_features_compat, u32, "");
		XFS_PRINT(sb_features_ro_compat, u32, "");
		XFS_PRINT(sb_features_incompat, u32, "");
		XFS_PRINT(sb_features_log_incompat, u32, "");
		XFS_PRINT(sb_crc, x32, "superblock crc");
		XFS_PRINT(sb_spino_align, u32, "sparse inode chunk alignment");
		XFS_PRINT(sb_pquotino, x64, "project quota inode");
		XFS_PRINT(sb_lsn, x64, "last write sequence");
		XFS_PRINT(sb_meta_uuid, uuid, "");
	}
}

void print_agf(xfs_agf_t *agf, uint64_t offset)
{
	if (_print_config_action)
		return;
	PRE_XFS_PRINT(agf, offset);

	printf("\n---------------------xfs_agf_t--------------------\n");
	printf("# description: AG free block info\n");
	printf("# offset: %"U64TF", size: %"SZOF"\n", offset, sizeof(xfs_agf_t));
	printf("# Common information:\n");
	XFS_PRINT(agf_magicnum, magic, "magic number == XFS_AGF_MAGIC");
	XFS_PRINT(agf_versionnum, u32, "header version == XFS_AGF_VERSION");
	XFS_PRINT(agf_seqno, u32, "sequence # starting from 0");
	XFS_PRINT(agf_length, u32, "size in blocks of a.g.");
	printf("# Freespace information:\n");
	printf("agf_roots:" VALUE_POS);
	for (int i = 0; i < XFS_BTNUM_AGF; i++)
		printf("%u ", be32toh(agf->agf_roots[i]));
	printf(COMMENT_POS "root blocks\n");
	printf("agf_levels:" VALUE_POS);
	for (int i = 0; i < XFS_BTNUM_AGF; i++)
		printf("%u ", be32toh(agf->agf_levels[i]));
	printf(COMMENT_POS "btree levels\n");

	XFS_PRINT(agf_flfirst, u32, "first freelist block's index");
	XFS_PRINT(agf_fllast, u32, "last freelist block's index");
	XFS_PRINT(agf_flcount, u32, "count of blocks in freelist");
	XFS_PRINT(agf_freeblks, u32, "total free blocks");

	XFS_PRINT(agf_longest, u32, "longest free space");
	XFS_PRINT(agf_btreeblks, u32, "# of blocks held in AGF btrees");
	XFS_PRINT(agf_uuid, uuid, "");

	printf("# unlogged fields, written during buffer writeback.\n");
	XFS_PRINT(agf_lsn, x64, "last write sequence");
	XFS_PRINT(agf_crc, x32, "crc of agf sector");
}

void print_agi(xfs_agi_t *agi, uint64_t offset)
{
	if (_print_config_action)
		return;
	PRE_XFS_PRINT(agi, offset);
	int ino_per_line = 4;

	printf("\n-------------------xfs_agi_t--------------------\n");
	printf("# description: AG inode B+tree info\n");
	printf("# offset: %"U64TF", size: %"SZOF"\n", offset, sizeof(xfs_agi_t));
	printf("# Common allocation group header information\n");
	XFS_PRINT(agi_magicnum, magic, "magic number == XFS_AGI_MAGIC");
	XFS_PRINT(agi_versionnum, u32, "header version == XFS_AGI_VERSION");
	XFS_PRINT(agi_seqno, u32, "sequence # starting from 0");
	XFS_PRINT(agi_length, u32, "size in blocks of a.g.");
	printf("# Inode information:\n");
	XFS_PRINT(agi_count, u32, "count of allocated inodes");
	XFS_PRINT(agi_root, u32, "root of inode btree");
	XFS_PRINT(agi_level, u32, "levels in inode btree");
	XFS_PRINT(agi_freecount, u32, "number of free inodes");
	XFS_PRINT(agi_newino, x32, "new inode just allocated");
	XFS_PRINT(agi_dirino, x32, "last directory inode chunk");
	printf("# Hash table of inodes which have been unlinked but are still being referenced:\n");

	uint64_t agi_unlinked_off = offset + offsetof(xfs_agi_t, agi_unlinked);
	printf("%.8"X64TF"-%.8"X64TF": agi_unlinked[%d]:\n", agi_unlinked_off,
	       agi_unlinked_off + sizeof(uint32_t) * (XFS_AGI_UNLINKED_BUCKETS - 1),
	       XFS_AGI_UNLINKED_BUCKETS);
	for (int i = 0; i < XFS_AGI_UNLINKED_BUCKETS / ino_per_line +
	     !!(XFS_AGI_UNLINKED_BUCKETS % ino_per_line); i++) {
		printf("\t\t ");
		for (int j = i * ino_per_line; j < i * ino_per_line +
		     ino_per_line; j++) {
			if (j >= XFS_AGI_UNLINKED_BUCKETS)
				break;
			uint32_t inonum = agi->agi_unlinked[j];
			if (inonum == 0xffffffff)
				printf("[%.2d]:-- ", j);
			else
				printf("[%.2d]:0x%x\t", j, be32toh(inonum));
		}
		printf("\n");
	}
	printf("\t\t (\'--\' equal to 0xffffffff)\n");
	printf("# This marks the end of logging region 1 and start of logging region 2.\n");
	XFS_PRINT(agi_uuid, uuid, "");
	XFS_PRINT(agi_crc, x32, "crc of agi sector");
	XFS_PRINT(agi_pad32, u32, "");
	XFS_PRINT(agi_lsn, x64, "last write sequence");
	XFS_PRINT(agi_free_root, u32, "root of the free inode btree");
	XFS_PRINT(agi_free_level, u32, "levels in free inode btree");
}

void print_agfl(xfs_agfl_t *agfl, uint64_t offset, struct xfs_info *fs_info)
{
	if (_print_config_action)
		return;
	PRE_XFS_PRINT(agfl, offset);
	int bno_count;
	int bno_per_line = 8;

	bno_count = (fs_info->sb_sectsize - sizeof(xfs_agfl_t))
		    / sizeof(xfs_agblock_t);

	printf("\n-------------------agfl--------------------\n");
	printf("# description: AG inode B+tree info\n");
	printf("# offset: %"U64TF", size: %u, bno_count = %u\n", offset,
	       fs_info->sb_sectsize, bno_count);

	XFS_PRINT(agfl_magicnum, magic, "");
	XFS_PRINT(agfl_seqno, u32, "");
	XFS_PRINT(agfl_uuid, uuid, "");
	XFS_PRINT(agfl_crc, x32, "");

	printf("# blocks reserved for growing the free space B+trees:\n");
	uint64_t agfl_bno_off = offset + offsetof(xfs_agfl_t, agfl_bno);
	printf("%.8"X64TF"-%.8"X64TF": agfl_bno[%d]:\n", agfl_bno_off,
	       agfl_bno_off + sizeof(xfs_agblock_t) * (bno_count - 1),
	       bno_count);
	for (int i = 0; i < bno_count / bno_per_line +
	     !!(bno_count % bno_per_line); i++) {
		for (int j = i * bno_per_line; j < i * bno_per_line + bno_per_line;
		     j++) {
			if (j >= bno_count)
				break;
			uint32_t bno = be32toh(agfl->agfl_bno[j]);
			if (bno == 0xffffffff)
				printf("[%.3d]:-- ", j);
			else
				printf("[%.3d]:%2u ", j, bno);
		}
		printf("\n");
	}
	printf("\t (\'--\' equal to 0xFFFFFFFF)\n");
}

void print_log_rec_header(xlog_rec_header_t* rec_header, uint64_t offset)
{
	if (_print_config_action)
		return;
	#define MIN(a,b) (((a)<(b))?(a):(b))
	PRE_XFS_PRINT(rec_header, offset);
	printf("\n-------------xlog_rec_header_t-------------\n");
	printf("# description: log header:\n");
	printf("# offset: %"U64TF", size: %"SZOF"\n", offset, sizeof(xlog_rec_header_t));

	XFS_PRINT(h_magicno, x32, "log record (LR) identifier");
	XFS_PRINT(h_cycle, u32, "write cycle of log");
	XFS_PRINT(h_version, u32, "LR version");
	XFS_PRINT(h_len, u32, "length of LR in bytes");
	XFS_PRINT(h_lsn, lsn, "lsn of this LR");
	XFS_PRINT(h_tail_lsn, lsn, "lsn of 1st LR w/ buffers not committed");
	XFS_PRINT(h_crc, le_x32, "crc of log record");
	XFS_PRINT(h_prev_block, d32, "block number to previous LR");
	XFS_PRINT(h_num_logops, u32, "number of log operations in this LR");
	int datalen, bbs;
	datalen = be32toh(rec_header->h_len);
	bbs=BTOBB(datalen);

	printf("h_cycle_data[] has %u element,  max : %u:" COMMENT_POS "cycle num overwrites\n",
	       bbs, XLOG_HEADER_CYCLE_SIZE / BBSIZE);
	for (int i = 0; i < MIN(bbs, XLOG_HEADER_CYCLE_SIZE / BBSIZE); i++)
		printf("%.2u: 0x%x ", i, be32toh(rec_header->h_cycle_data[i]));
	printf("\n");
	XFS_PRINT(h_fmt, x32, "format of log record");
	XFS_PRINT(h_fs_uuid, uuid, "uuid of FS");
	XFS_PRINT(h_size, u32, "iclog size");
}

int read_xfs_info(struct xfs_info *info, char *dev_name)
{
	xfs_sb_t super;

	info->fd = open(dev_name, O_RDONLY);
	if (info->fd == -1) {
		printf("Failed to open file %s\n", dev_name);
		return ENOENT;
	}

	if (pread64(info->fd, &super, sizeof(xfs_sb_t), SUPERBLOCK_OFFSET) == -1) {
		fprintf(stderr, "ERROR: Failed read superblock. %s [%d]",
			strerror(errno), errno);
		close(info->fd);
		return EIO;
	}

	info->sb_blocksize = be32toh(super.sb_blocksize);
	info->sb_agblocks = be32toh(super.sb_agblocks);
	info->sb_agcount = be32toh(super.sb_agcount);
	info->sb_sectsize = be16toh(super.sb_sectsize);
	info->sb_versionnum = be16toh(super.sb_versionnum) &
			      XFS_SB_VERSION_NUMBITS;
	info->sb_logstart = be64toh(super.sb_logstart);
	info->sb_logblocks = be32toh(super.sb_logblocks);
	info->sb_blocklog = super.sb_blocklog;
	info->sb_agblklog = super.sb_agblklog;
	info->m_blkbb_log = info->sb_blocklog - BBSHIFT;
	info->logBBsize = XFS_FSB_TO_BB(info, info->sb_logblocks);
	info->logBBstart = XFS_FSB_TO_DADDR(info, info->sb_logstart);

	if (_print_config_action)
		return 0;

	printf("\n");
	printf("xfs_info:\n");
	printf("info->sb_blocksize = %u\n", info->sb_blocksize);
	printf("info->sb_agblocks = %u\n", info->sb_agblocks);
	printf("info->sb_agcount = %u\n", info->sb_agcount);
	printf("info->sb_sectsize = %u\n", info->sb_sectsize);
	printf("info->sb_versionnum = %u\n", info->sb_versionnum);

	printf("info->sb_logstart = %"U64TF"\n", info->sb_logstart);
	printf("info->sb_logblocks = %u\n", info->sb_logblocks);
	printf("info->sb_blocklog = %u\n", info->sb_blocklog);
	printf("info->sb_agblklog = %u\n", info->sb_agblklog);
	printf("info->m_blkbb_log = %u\n", info->m_blkbb_log);
	printf("info->logBBsize = %"U64TF"\n", info->logBBsize);
	printf("info->logBBstart = %"U64TF"\n", info->logBBstart);
	printf("\n");

	return 0;
}

static inline uint64_t ag_block_offset(struct xfs_info *fs, uint32_t ag,
				       uint32_t block)
{
	return (fs->sb_agblocks * ag + block) * fs->sb_blocksize;
}

static void print_xfs_diff_config(struct xfs_debug_config_t *cfg)
{
	assert(MAX_AG_COUNT >= cfg->ag_count);

	printf("# %s configuration for xfs_diff.sh\n", cfg->device_name);
	printf("BLOCK_SIZE=%u\n", cfg->blocksize);
	printf("AG_COUNT=%u\n", cfg->ag_count);
	for (int i = 0; i < cfg->ag_count; i++)
		printf("ag_offsets[%i]=%"U64TF"\n", i, cfg->ag_offset[i]);
	printf("LOG_OFF=%"U64TF"\n", cfg->log_offset);
	printf("LOG_SIZE=%"U64TF"\n", cfg->log_size);
}

int main(int argc, char **argv)
{
	int ret = EXIT_SUCCESS;
	char *program_name = argv[0];
	char *dev_name;
	struct xfs_info fs_info;
	xfs_sb_t super;
	xfs_agf_t agf;
	xfs_agi_t agi;
	xfs_agfl_t *agfl;
	xlog_rec_header_t log_rec_header;
	uint64_t ag_offset;
	struct xfs_debug_config_t xfs_debug_cfg;

	if (argc < 2 || !strcmp(argv[1], "--help")) {
		print_usage(program_name);
		return EXIT_FAILURE;
	}
	dev_name = argv[1];

	if (argc > 2) {
		if (strcmp(argv[2], "print-cfg")) {
			print_usage(argv[0]);
			return EXIT_FAILURE;
		} else {
			_print_config_action = true;
		}
	}

	if (ret = read_xfs_info(&fs_info, dev_name))
		return ret;

	strcpy(xfs_debug_cfg.device_name, dev_name);
	xfs_debug_cfg.ag_count = fs_info.sb_agcount;
	xfs_debug_cfg.blocksize = fs_info.sb_blocksize;

	for (int i = 0; i < fs_info.sb_agcount; i++) {
		ag_offset = XFS_AG_DADDR(&fs_info, i) << 9;
		if (!_print_config_action) {
			printf("\n\n****************************************************\n");
			printf("\tAllocation Group #%d. Offset: %"U64TF"\n", i, ag_offset);
			printf("****************************************************\n");
		} else {
			xfs_debug_cfg.ag_offset[i] = ag_offset;
		}

		/* 1) Superblock */
		ret = pread64(fs_info.fd, &super, sizeof(xfs_sb_t),
			    ag_offset + SUPERBLOCK_OFFSET);
		if (ret == -1) {
			ret = errno;
			fprintf(stderr, "ERROR: Failed read superblock. %s [%d]\n",
				strerror(errno), errno);
			goto clean;
		}
		if (super.sb_magicnum != be32toh(XFS_SB_MAGIC)) {
			fprintf(stderr, "ERROR: Superblock magic number mismatched\n");
			ret = EXIT_FAILURE;
			goto clean;
		}
		print_superblock(&super, ag_offset + SUPERBLOCK_OFFSET);

		/* 2) AGF */
		uint64_t agf_offset = fs_info.sb_sectsize * AGF_SECTOR +
				      ag_offset;
		ret = pread64(fs_info.fd, &agf, sizeof(xfs_agf_t), agf_offset);
		if (ret == -1) {
			ret = errno;
			fprintf(stderr, "ERROR: Failed read agf. %s [%d]\n",
				strerror(errno), errno);
			goto clean;
		}
		if (agf.agf_magicnum != be32toh(XFS_AGF_MAGIC)) {
			fprintf(stderr, "ERROR: AGF magic number mismatched\n");
			ret = EXIT_FAILURE;
			goto clean;
		}
		print_agf(&agf, agf_offset);

		/* 3) AGI */
		uint64_t agi_offset = fs_info.sb_sectsize * AGI_SECTOR +
				      ag_offset;
		ret = pread64(fs_info.fd, &agi, sizeof(xfs_agi_t), agi_offset);
		if (ret == -1) {
			ret = errno;
			fprintf(stderr, "ERROR: Failed read agi. %s [%d]\n", strerror(errno),
			       errno);
			goto clean;
		}
		if (agi.agi_magicnum != be32toh(XFS_AGI_MAGIC)) {
			fprintf(stderr, "ERROR: AGI magic number mismatched\n");
			ret = EXIT_FAILURE;
			goto clean;
		}
		print_agi(&agi, agi_offset);

		/* 4) AGFL */
		if (fs_info.sb_versionnum > 4) {
			uint64_t agfl_offset = fs_info.sb_sectsize * AGFL_SECTOR +
					  ag_offset;
			agfl = malloc(fs_info.sb_sectsize);
			if (!agfl) {
				fprintf(stderr, "ERROR: %s [%d]",
					strerror(errno), errno);
				return errno;
			}
			ret = pread64(fs_info.fd, agfl, fs_info.sb_sectsize,
				    agfl_offset);
			if (ret == -1) {
				ret = errno;
				fprintf(stderr, "ERROR: Failed read agfl. %s [%d]\n",
					strerror(errno), errno);
				free(agfl);
				goto clean;
			}
			if (agfl->agfl_magicnum != be32toh(XFS_AGFL_MAGIC)) {
				fprintf(stderr, "ERROR: AGFL magic number mismatched\n");
				ret = EXIT_FAILURE;
				free(agfl);
				goto clean;
			}
			print_agfl(agfl, agfl_offset, &fs_info);
			free(agfl);
		}
	}

	if (fs_info.sb_logstart == 0)
		fprintf(stderr, "WARNING: Filesystem has external log\n");

	int64_t log_offset = BBTOOFF64(fs_info.logBBstart);
	int64_t log_size = BBTOOFF64(fs_info.logBBsize);

	ret = pread64(fs_info.fd, &log_rec_header, XLOG_HEADER_SIZE, log_offset);
	if (ret == -1) {
		ret = errno;
		fprintf(stderr, "ERROR: Failed read log rec header. %s [%d]\n",
			strerror(errno), errno);
		free(agfl);
		goto clean;
	}

	/* For some reason log record start is not always starts at the start of
	 * xfs log. Supposedly this is due to ring structure of xfs log.
	 */
	if (be32toh(log_rec_header.h_magicno) != XLOG_HEADER_MAGIC_NUM)
		fprintf(stderr, "ERROR: Log header magic mismatched\n");

	if (!_print_config_action) {
		printf("\n\n****************************************************\n");
		printf("\tLog. offset: %"U64TF", size = %"U64TF" (bytes)\n", log_offset, log_size);
		printf("****************************************************\n");
	} else {
		xfs_debug_cfg.log_offset = log_offset;
		xfs_debug_cfg.log_size = log_size;
		print_xfs_diff_config(&xfs_debug_cfg);
		goto clean;
	}

	print_log_rec_header(&log_rec_header, log_offset);

	if (be32toh(log_rec_header.h_version) == 2) {
		int num_hdrs;
		num_hdrs = be32toh(log_rec_header.h_size) / XLOG_HEADER_CYCLE_SIZE;
		printf("num_hdrs = %u\n", num_hdrs);
//		if (num_hdrs == 1)
	}

clean:
	close(fs_info.fd);
	return EXIT_SUCCESS;
}
