#!/bin/bash
PWD_DIR=$(cd $(dirname $0); pwd)
CFG_FILE=$PWD_DIR/xfs_diff.config

if [[ $1 == "clean" ]]; then
	rm -r $PWD_DIR/?_* $CFG_FILE 2> /dev/null
	echo "Cleaned."
	exit 0
fi

if [ $UID -ne 0 ]; then
	echo "$0: only root can perform"
	exit 1
fi
	
if [[ ! -b $1 ]]; then
	echo "You should pass a path to a xfs partition e.g. \"/dev/sdb1\""
	exit 1
fi

XFS_DEBUG_DIR=$PWD_DIR/../..
XFS_DEBUG=$XFS_DEBUG_DIR/xfs-debug
DEV=$1
MOUNT_POINT=/mnt/`basename $DEV`
METADATA_SIZE=8096

function exception()
{
	echo "ERROR: $*"; exit 1
}

function is_mounted()
{
	local dev_name=$1
	mount | grep $dev_name > /dev/null || return 1 && return 0
}

function DD()
{
	dd $* >& /dev/null || exception "dd $*"
}

function pre_action()
{
	local dir=$1
	while ( is_mounted $DEV )
	do
		umount $DEV || sleep 1 && break
	done


	for (( i=0; i < $AG_COUNT; i++ )); do
		mkdir -p $dir/ag"$i"
		DD if=$DEV ibs=$BLOCK_SIZE skip=${ag_offsets[i]}	\
		   iflag=skip_bytes of=$dir/ag"$i"/ag"$i"_image_pre	\
		   bs=$METADATA_SIZE count=1
	done	

	mkdir -p $dir/log
	DD if=$DEV ibs=$BLOCK_SIZE skip=$LOG_OFF iflag=skip_bytes	\
	   of=$dir/log/log_image_pre bs=$BLOCK_SIZE count=$LOG_SIZE
	mount $DEV $MOUNT_POINT || exception "Can't mount $DEV"
}

function make_diff()
{
	local dir=$1
	local image_name=$2
	local image_pre=$image_name"_pre"
	local image_post=$image_name"_post"
	xxd $dir/$image_pre $dir/$image_pre".hex"
	rm $dir/$image_pre
	xxd $dir/$image_post $dir/$image_post".hex"
	rm $dir/$image_post
	diff $dir/$image_pre".hex" $dir/$image_post".hex" > $dir/$image_name".diff" \
	     && echo "$image_name: FAILED" || echo "$image_name: OK" 
}

function post_action()
{
	local DIR=$1
	while ( is_mounted $DEV )
	do
		umount $DEV || sleep 1 && break
	done

	for (( i=0; i < $AG_COUNT; i++ )); do
		DD if=$DEV ibs=$BLOCK_SIZE skip=${ag_offsets[i]} 	\
		   iflag=skip_bytes of=$DIR/ag"$i"/ag"$i"_image_post 	\
		   bs=$METADATA_SIZE count=1
	done
	DD if=$DEV skip=$LOG_OFF iflag=skip_bytes of=$DIR/log/log_image_post \
	   bs=$BLOCK_SIZE count=$LOG_SIZE
	for (( i=0; i < $AG_COUNT; i++ )); do
		make_diff $DIR/ag"$i" ag"$i"_image
	done
	make_diff $DIR/log log_image
}

# ********************************BEGIN*************************************
mkdir -p $MOUNT_POINT

# make cfg file and include it
if [[ ! -e $XFS_DEBUG ]]
then
	make -C $XFS_DEBUG_DIR || exception "Can't locate xfs_debug utilite"
fi	
$XFS_DEBUG $DEV print-cfg > $CFG_FILE
. xfs_diff.config

let LOG_SIZE=$LOG_SIZE/$BLOCK_SIZE

cat $CFG_FILE


TEST_FILE="test_file1"
TEST_FILE_NEW_NAME="test_file_new_name"
TEST_DIR="test_dir"
TEST_DIR_NEW_NAME=$TEST_DIR"_new"

echo -e "\nAct 1. Add a non-empty file: $MOUNT_POINT/$TEST_FILE"
DIR="1_add_file $TEST_FILE"
pre_action $DIR
DD if=/dev/urandom of=$MOUNT_POINT/$TEST_FILE bs=1M count=1
post_action $DIR

echo -e "\nAct 2. Rename a file: $MOUNT_POINT/$TEST_FILE to $MOUNT_POINT/$TEST_FILE_NEW_NAME"
DIR="2_rename_file $TEST_FILE to $TEST_FILE_NEW_NAME"
pre_action $DIR
mv $MOUNT_POINT/$TEST_FILE $MOUNT_POINT/$TEST_FILE_NEW_NAME
post_action $DIR

echo -e "\nAct 3. Rewrite the middle of a file: $MOUNT_POINT/$TEST_FILE_NEW_NAME"
DIR="3_write_in_middle_of_file"
pre_action $DIR
DD if=/dev/zero of=$MOUNT_POINT/$TEST_FILE_NEW_NAME seek=524288 \
   oflag=seek_bytes bs=512 count=1 conv=notrunc
post_action $DIR

echo -e "\nAct 4. Delete file: $MOUNT_POINT/$TEST_FILE_NEW_NAME"
DIR="4_delete_file"
pre_action $DIR
rm -f $MOUNT_POINT/$TEST_FILE_NEW_NAME
post_action $DIR

echo -e "\nAct 5. Add folder: $MOUNT_POINT/$TEST_DIR"
DIR="5_add_folder"
pre_action $DIR
mkdir $MOUNT_POINT/$TEST_DIR
post_action $DIR

echo -e "\nAct 6. Rename folder: $MOUNT_POINT/$TEST_DIR to $MOUNT_POINT/$TEST_DIR_NEW_NAME"
DIR="6_rename_folder"
pre_action $DIR
mv $MOUNT_POINT/$TEST_DIR $MOUNT_POINT/$TEST_DIR_NEW_NAME
post_action $DIR

echo -e "\nAct 7. Delete empty folder: $MOUNT_POINT/$TEST_DIR_NEW_NAME"
DIR="7_delete_empty_folder"
pre_action $DIR
rmdir $MOUNT_POINT/$TEST_DIR_NEW_NAME
post_action $DIR

echo -e "\nAct 8. Touch an existing file: $MOUNT_POINT/$TEST_FILE"
DIR="8_touch_existing_file $TEST_FILE"
mount $DEV $MOUNT_POINT || exception "Can't mount $DEV"
DD if=/dev/zero of=$MOUNT_POINT/$TEST_FILE bs=1M count=1
pre_action $DIR
touch $MOUNT_POINT/$TEST_FILE
post_action $DIR

echo -e "\nAct 9. Just mount and unmount"
DIR="9_just_mount_unmount"
pre_action $DIR
post_action $DIR

# clean
mount $DEV $MOUNT_POINT || exception "Can't mount $DEV"
rm -f $MOUNT_POINT/$TEST_FILE
umount $DEV

exit 0
