#define _BSD_SOURCE
#include <endian.h>
#include <stdint.h>
#include <stdio.h>
#include <uuid/uuid.h>
#include "xfs.h"
#include "xfs-print.h"

/* get lsn fields */
#define CYCLE_LSN(lsn) ((uint32_t)((lsn)>>32))
#define BLOCK_LSN(lsn) ((uint32_t)(lsn))

static inline void __xfs_print(uint64_t offset, const char *field_name,
			       const char *value, const char *comment)
{
	/* value is written after comment for avoiding its overlapping */
	printf("%.8"X64TF": %s:" COMMENT_POS "%s" VALUE_POS "%s\n", offset,
	       field_name, comment, value);
}

void xfs_print_u8(char *field_name, const uint8_t *value, char *comment,
		  uint64_t offset)
{
	char value_str[3];	/* max: 255 */
	sprintf(value_str, "%u", *value);
	__xfs_print(offset, field_name, value_str, comment);
}

void xfs_print_x8(char *field_name, const uint8_t *value, char *comment,
		  uint64_t offset)
{
	char value_str[4];	/* max: 0x01 */
	sprintf(value_str, "0x%x", *value);
	__xfs_print(offset, field_name, value_str, comment);
}

void xfs_print_u16(char *field_name, const uint16_t *value, char *comment,
		   uint64_t offset)
{
	char value_str[5];	/* max: 65535 */
	sprintf(value_str, "%u", be16toh(*value));
	__xfs_print(offset, field_name, value_str, comment);
}

void xfs_print_x16(char *field_name, const uint16_t *value, char *comment,
		   uint64_t offset)
{
	char value_str[6];	/* max: 0x0102 */
	sprintf(value_str, "0x%x", be16toh(*value));
	__xfs_print(offset, field_name, value_str, comment);
}

void xfs_print_u32(char *field_name, const uint32_t *value, char *comment,
		   uint64_t offset)
{
	char value_str[10];	/* max: 4294967295 */
	sprintf(value_str, "%u", be32toh(*value));
	__xfs_print(offset, field_name, value_str, comment);
}

void xfs_print_d32(char *field_name, const uint32_t *value, char *comment,
		   uint64_t offset)
{
	char value_str[10];	/* max: 4294967295 */
	sprintf(value_str, "%d", be32toh(*value));
	__xfs_print(offset, field_name, value_str, comment);
}


void xfs_print_x32(char *field_name, const uint32_t *value, char *comment,
		   uint64_t offset)
{
	char value_str[10];	/* max: 0x01020304 */
	sprintf(value_str, "0x%x", be32toh(*value));
	__xfs_print(offset, field_name, value_str, comment);
}

void xfs_print_le_x32(char *field_name, const uint32_t *value, char *comment,
		   uint64_t offset)
{
	char value_str[10];	/* max: 0x01020304 */
	sprintf(value_str, "0x%x", *value);
	__xfs_print(offset, field_name, value_str, comment);
}

void xfs_print_u64(char *field_name, const uint64_t *value, char *comment,
		   uint64_t offset)
{
	char value_str[20];	/* max: 18446744073709551615 */
	sprintf(value_str, "%"U64TF, be64toh(*value));
	__xfs_print(offset, field_name, value_str, comment);
}

void xfs_print_x64(char *field_name, const uint64_t *value, char *comment,
		   uint64_t offset)
{
	char value_str[18];	/* max: 0x0102030405060708 */
	sprintf(value_str, "0x%"X64TF, be64toh(*value));
	__xfs_print(offset, field_name, value_str, comment);
}

void xfs_print_lsn(char *field_name, const uint64_t *value, char *comment,
		   uint64_t offset)
{
	char value_str[20];
	sprintf(value_str, "%u,%u",
		CYCLE_LSN(be64toh(*value)), BLOCK_LSN(be64toh(*value)));
	__xfs_print(offset, field_name, value_str, comment);
}

void xfs_print_magic(char *field_name, const uint32_t *magic, char *comment,
		     uint64_t offset)
{
	uint8_t *p_magic = (uint8_t*)magic;
	char magic_str[4];

	sprintf(magic_str, "%C%C%C%C", p_magic[0], p_magic[1],
				       p_magic[2], p_magic[3]);
	__xfs_print(offset, field_name, magic_str, comment);
}

void xfs_print_uuid(char *field_name, const uuid_t *uuid,
		    __attribute__((unused)) char *comment, uint64_t offset)
{
	char uuid_str[64];

	uuid_unparse(*uuid, uuid_str);

	__xfs_print(offset, field_name, uuid_str, "");
}

void xfs_print_str(char *field_name, const char *str, char *comment,
		   uint64_t offset)
{
	__xfs_print(offset, field_name, str, comment);
}
