#ifndef _XFS_H_
#define _XFS_H_

#include <stdint.h>

/* Offsets inside allocation group */
#define SUPERBLOCK_OFFSET	0
#define AGF_SECTOR		1
#define AGI_SECTOR		2
#define AGFL_SECTOR		3

/* Common types */
typedef	uint32_t	xfs_agblock_t;	/* blockno in alloc. group */
typedef	uint32_t	xfs_agnumber_t;	/* allocation group number */
typedef	uint32_t	xfs_extlen_t;	/* extent length in blocks */

typedef uint64_t	xfs_rtblock_t;	/* extent (block) in realtime area */

typedef	int64_t	xfs_lsn_t;	/* log sequence number */

typedef	uint64_t	xfs_fsblock_t;	/* blockno in filesystem (agno|agbno) */
typedef uint64_t	xfs_rfsblock_t;	/* blockno in filesystem (raw) */

typedef unsigned char uuid_t[16];

typedef uint64_t	xfs_ino_t;

typedef enum {
	XFS_BTNUM_BNOi, XFS_BTNUM_CNTi, XFS_BTNUM_BMAPi, XFS_BTNUM_INOi,
	XFS_BTNUM_FINOi, XFS_BTNUM_MAX
} xfs_btnum_t;

/*
 * Super block
 * Fits into a sector-sized buffer at address 0 of each allocation group.
 * Only the first of these is ever updated except during growfs.
 */
#define	XFS_SB_MAGIC		0x58465342	/* 'XFSB' */
#define	XFS_SB_VERSION_1	1		/* 5.3, 6.0.1, 6.1 */
#define	XFS_SB_VERSION_2	2		/* 6.2 - attributes */
#define	XFS_SB_VERSION_3	3		/* 6.2 - new inode version */
#define	XFS_SB_VERSION_4	4		/* 6.2+ - bitmask version */
#define	XFS_SB_VERSION_5	5		/* CRC enabled filesystem */
#define	XFS_SB_VERSION_NUMBITS		0x000f
#define	XFS_SB_VERSION_ALLFBITS		0xfff0
#define	XFS_SB_VERSION_ATTRBIT		0x0010
#define	XFS_SB_VERSION_NLINKBIT		0x0020
#define	XFS_SB_VERSION_QUOTABIT		0x0040
#define	XFS_SB_VERSION_ALIGNBIT		0x0080
#define	XFS_SB_VERSION_DALIGNBIT	0x0100
#define	XFS_SB_VERSION_SHAREDBIT	0x0200
#define XFS_SB_VERSION_LOGV2BIT		0x0400
#define XFS_SB_VERSION_SECTORBIT	0x0800
#define	XFS_SB_VERSION_EXTFLGBIT	0x1000
#define	XFS_SB_VERSION_DIRV2BIT		0x2000
#define	XFS_SB_VERSION_BORGBIT		0x4000	/* ASCII only case-insens. */
#define	XFS_SB_VERSION_MOREBITSBIT	0x8000

/*
 * There are two words to hold XFS "feature" bits: the original
 * word, sb_versionnum, and sb_features2.  Whenever a bit is set in
 * sb_features2, the feature bit XFS_SB_VERSION_MOREBITSBIT must be set.
 *
 * These defines represent bits in sb_features2.
 */
#define XFS_SB_VERSION2_LAZYSBCOUNTBIT	0x00000002	/* Superblk counters */
#define XFS_SB_VERSION2_ATTR2BIT	0x00000008	/* Inline attr rework */
#define XFS_SB_VERSION2_PARENTBIT	0x00000010	/* parent pointers */
#define XFS_SB_VERSION2_PROJID32BIT	0x00000080	/* 32 bit project id */
#define XFS_SB_VERSION2_CRCBIT		0x00000100	/* metadata CRCs */
#define XFS_SB_VERSION2_FTYPE		0x00000200	/* inode type in dir */

/*
 * Superblock - in core version.  Must match the ondisk version below.
 * Must be padded to 64 bit alignment.
 */
typedef struct xfs_sb {
	uint32_t	sb_magicnum;	/* magic number == XFS_SB_MAGIC */
	uint32_t	sb_blocksize;	/* logical block size, bytes */
	xfs_rfsblock_t	sb_dblocks;	/* number of data blocks */
	xfs_rfsblock_t	sb_rblocks;	/* number of realtime blocks */
	xfs_rtblock_t	sb_rextents;	/* number of realtime extents */
	uuid_t		sb_uuid;	/* user-visible file system unique id */
	xfs_fsblock_t	sb_logstart;	/* starting block of log if internal */
	xfs_ino_t	sb_rootino;	/* root inode number */
	xfs_ino_t	sb_rbmino;	/* bitmap inode for realtime extents */
	xfs_ino_t	sb_rsumino;	/* summary inode for rt bitmap */
	xfs_agblock_t	sb_rextsize;	/* realtime extent size, blocks */
	xfs_agblock_t	sb_agblocks;	/* size of an allocation group */
	xfs_agnumber_t	sb_agcount;	/* number of allocation groups */
	xfs_extlen_t	sb_rbmblocks;	/* number of rt bitmap blocks */
	xfs_extlen_t	sb_logblocks;	/* number of log blocks */
	uint16_t	sb_versionnum;	/* header version == XFS_SB_VERSION */
	uint16_t	sb_sectsize;	/* volume sector size, bytes */
	uint16_t	sb_inodesize;	/* inode size, bytes */
	uint16_t	sb_inopblock;	/* inodes per block */
	char		sb_fname[12];	/* file system name */
	uint8_t		sb_blocklog;	/* log2 of sb_blocksize */
	uint8_t		sb_sectlog;	/* log2 of sb_sectsize */
	uint8_t		sb_inodelog;	/* log2 of sb_inodesize */
	uint8_t		sb_inopblog;	/* log2 of sb_inopblock */
	uint8_t		sb_agblklog;	/* log2 of sb_agblocks (rounded up) */
	uint8_t		sb_rextslog;	/* log2 of sb_rextents */
	uint8_t		sb_inprogress;	/* mkfs is in progress, don't mount */
	uint8_t		sb_imax_pct;	/* max % of fs for inode space */
					/* statistics */
	/*
	 * These fields must remain contiguous.  If you really
	 * want to change their layout, make sure you fix the
	 * code in xfs_trans_apply_sb_deltas().
	 */
	uint64_t	sb_icount;	/* allocated inodes */
	uint64_t	sb_ifree;	/* free inodes */
	uint64_t	sb_fdblocks;	/* free data blocks */
	uint64_t	sb_frextents;	/* free realtime extents */
	/*
	 * End contiguous fields.
	 */
	xfs_ino_t	sb_uquotino;	/* user quota inode */
	xfs_ino_t	sb_gquotino;	/* group quota inode */
	uint16_t	sb_qflags;	/* quota flags */
	uint8_t		sb_flags;	/* misc. flags */
	uint8_t		sb_shared_vn;	/* shared version number */
	xfs_extlen_t	sb_inoalignmt;	/* inode chunk alignment, fsblocks */
	uint32_t	sb_unit;	/* stripe or raid unit */
	uint32_t	sb_width;	/* stripe or raid width */
	uint8_t		sb_dirblklog;	/* log2 of dir block size (fsbs) */
	uint8_t		sb_logsectlog;	/* log2 of the log sector size */
	uint16_t	sb_logsectsize;	/* sector size for the log, bytes */
	uint32_t	sb_logsunit;	/* stripe unit size for the log */
	uint32_t	sb_features2;	/* additional feature bits */

	/*
	 * bad features2 field as a result of failing to pad the sb structure to
	 * 64 bits. Some machines will be using this field for features2 bits.
	 * Easiest just to mark it bad and not use it for anything else.
	 *
	 * This is not kept up to date in memory; it is always overwritten by
	 * the value in sb_features2 when formatting the incore superblock to
	 * the disk buffer.
	 */
	uint32_t	sb_bad_features2;

	/* version 5 superblock fields start here */

	/* feature masks */
	uint32_t	sb_features_compat;
	uint32_t	sb_features_ro_compat;
	uint32_t	sb_features_incompat;
	uint32_t	sb_features_log_incompat;

	uint32_t	sb_crc;		/* superblock crc */
	xfs_extlen_t	sb_spino_align;	/* sparse inode chunk alignment */

	xfs_ino_t	sb_pquotino;	/* project quota inode */
	xfs_lsn_t	sb_lsn;		/* last write sequence */
	uuid_t		sb_meta_uuid;	/* metadata file system unique id */

	/* must be padded to 64 bit alignment */
} xfs_sb_t;

#define	XFS_SB_VERSION_NUM(sbp)	(be16toh((sbp)->sb_versionnum) & XFS_SB_VERSION_NUMBITS)

/*
 * V5 superblock specific feature checks
 */
static inline int xfs_sb_version_hascrc(struct xfs_sb *sbp)
{
	return XFS_SB_VERSION_NUM(sbp) == XFS_SB_VERSION_5;
}

/***************************** Allocation group *****************************/
/*
 * Allocation group header
 *
 * This is divided into three structures, placed in sequential 512-byte
 * buffers after a copy of the superblock (also in a 512-byte buffer).
 */

#define	XFS_AGF_MAGIC	0x58414746	/* 'XAGF' */
#define	XFS_AGI_MAGIC	0x58414749	/* 'XAGI' */
#define	XFS_AGFL_MAGIC	0x5841464c	/* 'XAFL' */

/*
 * Btree number 0 is bno, 1 is cnt.  This value gives the size of the
 * arrays below.
 */
#define	XFS_BTNUM_AGF	((int)XFS_BTNUM_CNTi + 1)

typedef struct xfs_agf {
	/*
	 * Common allocation group header information
	 */
	uint32_t	agf_magicnum;	/* magic number == XFS_AGF_MAGIC */
	uint32_t	agf_versionnum;	/* header version == XFS_AGF_VERSION */
	uint32_t	agf_seqno;	/* sequence # starting from 0 */
	uint32_t	agf_length;	/* size in blocks of a.g. */
	/*
	 * Freespace information
	 */
	uint32_t	agf_roots[XFS_BTNUM_AGF];	/* root blocks */
	uint32_t	agf_spare0;	/* spare field */
	uint32_t	agf_levels[XFS_BTNUM_AGF];	/* btree levels */
	uint32_t	agf_spare1;	/* spare field */

	uint32_t	agf_flfirst;	/* first freelist block's index */
	uint32_t	agf_fllast;	/* last freelist block's index */
	uint32_t	agf_flcount;	/* count of blocks in freelist */
	uint32_t	agf_freeblks;	/* total free blocks */

	uint32_t	agf_longest;	/* longest free space */
	uint32_t	agf_btreeblks;	/* # of blocks held in AGF btrees */
	uuid_t		agf_uuid;	/* uuid of filesystem */

	/*
	 * reserve some contiguous space for future logged fields before we add
	 * the unlogged fields. This makes the range logging via flags and
	 * structure offsets much simpler.
	 */
	uint64_t	agf_spare64[16];

	/* unlogged fields, written during buffer writeback. */
	uint64_t	agf_lsn;	/* last write sequence */
	uint32_t	agf_crc;	/* crc of agf sector */
	uint32_t	agf_spare2;

	/* structure must be padded to 64 bit alignment */
} xfs_agf_t;


/*
 * Size of the unlinked inode hash table in the agi.
 */
#define	XFS_AGI_UNLINKED_BUCKETS	64U
typedef struct xfs_agi {
	/*
	 * Common allocation group header information
	 */
	uint32_t	agi_magicnum;	/* magic number == XFS_AGI_MAGIC */
	uint32_t	agi_versionnum;	/* header version == XFS_AGI_VERSION */
	uint32_t	agi_seqno;	/* sequence # starting from 0 */
	uint32_t	agi_length;	/* size in blocks of a.g. */
	/*
	 * Inode information
	 * Inodes are mapped by interpreting the inode number, so no
	 * mapping data is needed here.
	 */
	uint32_t	agi_count;	/* count of allocated inodes */
	uint32_t	agi_root;	/* root of inode btree */
	uint32_t	agi_level;	/* levels in inode btree */
	uint32_t	agi_freecount;	/* number of free inodes */

	uint32_t	agi_newino;	/* new inode just allocated */
	uint32_t	agi_dirino;	/* last directory inode chunk */
	/*
	 * Hash table of inodes which have been unlinked but are
	 * still being referenced.
	 */
	uint32_t	agi_unlinked[XFS_AGI_UNLINKED_BUCKETS];
	/*
	 * This marks the end of logging region 1 and start of logging region 2.
	 */
	uuid_t		agi_uuid;	/* uuid of filesystem */
	uint32_t	agi_crc;	/* crc of agi sector */
	uint32_t	agi_pad32;
	uint64_t	agi_lsn;	/* last write sequence */

	uint32_t	agi_free_root; /* root of the free inode btree */
	uint32_t	agi_free_level;/* levels in free inode btree */

	/* structure must be padded to 64 bit alignment */
} xfs_agi_t;

/*
 * Size of the AGFL.  For CRC-enabled filesystes we steal a couple of
 * slots in the beginning of the block for a proper header with the
 * location information and CRC.
 */
#define XFS_AGFL_SIZE(sb) \
	((be16toh(sb.sb_sectsize) - \
	 (xfs_sb_version_hascrc(&sb) ? \
		sizeof(struct xfs_agfl) : 0)) / \
	  sizeof(xfs_agblock_t))

typedef struct xfs_agfl {
	uint32_t	agfl_magicnum;
	uint32_t	agfl_seqno;
	uuid_t		agfl_uuid;
	uint64_t	agfl_lsn;
	uint32_t	agfl_crc;
	uint32_t	agfl_bno[];	/* actually XFS_AGFL_SIZE(sb) */
} xfs_agfl_t;

/****************************** log ******************************/
#define XLOG_HEADER_SIZE	512
#define XLOG_HEADER_CYCLE_SIZE	(32*1024)	/* cycle data in header */
#define XLOG_HEADER_MAGIC_NUM	0xFEEDbabe	/* Invalid cycle number */

/*
 * Block I/O parameterization.	A basic block (BB) is the lowest size of
 * filesystem allocation, and must equal 512.  Length units given to bio
 * routines are in BB's.
 */
#define BBSHIFT		9
#define BBSIZE		(1<<BBSHIFT)
#define BTOBB(bytes)	(((uint64_t)(bytes) + BBSIZE - 1) >> BBSHIFT)

#define BBTOOFF64(bbs)	(((loff_t)(bbs)) << BBSHIFT)

typedef uint64_t	xfs_daddr_t;

typedef struct xlog_rec_header {
	uint32_t h_magicno;	/* log record (LR) identifier		:  4 */
	uint32_t h_cycle;	/* write cycle of log			:  4 */
	uint32_t h_version;	/* LR version				:  4 */
	uint32_t h_len;		/* len in bytes; should be 64-bit aligned: 4 */
	uint64_t h_lsn;		/* lsn of this LR			:  8 */
	uint64_t h_tail_lsn;	/* lsn of 1st LR w/ buffers not committed: 8 */
	uint32_t h_crc;		/* crc of log record                    :  4 */
	uint32_t h_prev_block;	/* block number to previous LR		:  4 */
	uint32_t h_num_logops;	/* number of log operations in this LR	:  4 */
	uint32_t h_cycle_data[XLOG_HEADER_CYCLE_SIZE / BBSIZE];
	/* new fields */
	uint32_t h_fmt;		/* format of log record                 :  4 */
	uuid_t	 h_fs_uuid;	/* uuid of FS                           : 16 */
	uint32_t h_size;	/* iclog size				:  4 */
} xlog_rec_header_t;


static inline uint32_t xfs_mask32lo(int n)
{
	return ((uint32_t)1 << (n)) - 1;
}

/*
 * File system block to basic block conversions.
 */
#define	XFS_FSB_TO_BB(inf,fsbno)	((fsbno) << (inf)->m_blkbb_log)

#define	XFS_FSB_TO_AGNO(inf,fsbno)	\
	((xfs_agnumber_t)((fsbno) >> (inf)->sb_agblklog))

#define	XFS_FSB_TO_AGBNO(inf,fsbno)	\
	((xfs_agblock_t)((fsbno) & xfs_mask32lo((inf)->sb_agblklog)))

#define	XFS_AGB_TO_DADDR(inf,agno,agbno)	\
	((xfs_daddr_t)XFS_FSB_TO_BB(inf, \
		(xfs_fsblock_t)(agno) * (inf)->sb_agblocks + (agbno)))

#define	XFS_FSB_TO_DADDR(inf,fsbno)	XFS_AGB_TO_DADDR(inf, \
			XFS_FSB_TO_AGNO(inf,fsbno), XFS_FSB_TO_AGBNO(inf,fsbno))

#define	XFS_AG_DADDR(inf,agno)		(XFS_AGB_TO_DADDR(inf, agno, 0))

#endif /* _XFS_H_ */
