

This document duscusses the changes that will be made to re-enable the bitmap feature.  We will do this by re-writing the bitmap feature

This document is NOT a complete requirements or design document for the entire blksnapshot system.



# Requirements:

## General Abilities
We need to be able to support the following features:
- create base backup (copy ALL data on the disk, except the blocks holding the bitmap and the COW data)
- create incremental backup if a previous backup was completed successfully, and no errors have occured since then
- track changes since the last base across a reboot - so that an incremental can be preformed after the reboot
- Durring a backup, preform Copy on Write (COW) so that the "original" data gets backed up, and not the newly written data.
	- When a change to a block comes in, we need to make a copy of the original block before the change goes through to the disk.  
	- When backing up, for every block that's requested we need to first check if there's a COW block for that block.  
		- if there is a corresponding COW block, read it and send it up.  
		- if there isn't a corresponding COW block, read the block of the disk and send it up.


We do NOT need to support the following features:
- track changes across a power event or a kernel panic.
- track changes when the system boots up and the bitmap is not valid - ie, Dirty mode



# Operation

## Startup
At startup, we will read the "bitmap_state" value out of the metadata.  (it can be either clean or dirty)

- If it is set to dirty, we will not track changed blocks (either in memory, or in the bitmap) until a snapshot is started.  (generally, we can do this setting a function pointer for the IOCTL to a no-op function, and letting all IO go straight through to disk)

- If it is set to clean, we will turn on changed block tracking.

- if the bitmap store is clean, copy all the data from the bitmapstore into the changelog tree.

- We will set a global flag to indicate that the bitmap is dirty.  This will be needed durring shutdown.  

- The metadata on disk will be set back to Dirty.  This will allow the system to default to Dirty on the next boot if there's a power event, or 

### Psudocode
```
def startup()
	g_bitmap_status = lookup_bitmap_status()
	if (g_bitmap_status == "dirty")
		g_is_tracking_changlog = false
		g_is_storing_bitmap = false
	else
		g_is_tracking_changelog = true
		g_is_storing_bitmap = true

		current_changelog = load_changelog_from_bitmap_store();

	update_bitmap_status("dirty")
end

```


## Start snapshot
We need to do some additional work durring snapshot start.

- if the bitmap store is dirty, and this is an incremental, fail.  (we have to do a full if the bitmap is dirty)

- Freeze the file system (just like we do now)

- If the bitmap store is clean, create a new block tree in memory.  Assign the old tree to a new var (old_changelog).  Assign the new tree to current tree var (current_changelog)

- If the bitmap store is dirty, create a new block tree in memory.  Assign the new tree to the current tree var.  The old tree var will remain null.

- if the bitmap store is dirty, and this is a full backup, turn on the bitmap store feature.  All writes going forward (in the current_changelog tree) should be written to disk.

- make sure COW data files are available

- set a global variable to indicate we are in a snapshot - in_snapshot = true

### Psudocode
```
def start_snapshot()
	if (g_bitmap_status == "dirty" && is_incremental_snapshot)
		fail ("we must do a base image")

	if (!COW_is_valid())
		fail ("COW data store isn't valid")

	freeze_filesystem()

	old_changelog = current_changelog
	current_changelog = allocate_new_changelog()

	if (g_bitmap_status == "dirty")
		g_is_tracking_changelog = true

	g_in_snapshot = true
end

```


## Get Changelog
This should remain basically the same, except the source of data for the changelog.

- Check that we are in a snapshot.  If not fail.

- The source of this data should now be the old_changelog list, not the current_changelog.

- otherwise, it should work as it did before.

### Psudocode
```
def get_changelog()
	if (!g_in_snapshot)
		fail("You must be in a snapshot to get the changelog")

	return get_intervals_from_tree(old_changelog)
end
```


## Read snapshot data
Because we are using 2 trees now, we need to change the code that reads the snapshot.

- Check that we are in a snapshot.  If not fail.

- if this is an incremental
	- if the old_changelog is not available, fail.
	- for each block requested  (This should be fundamentally unchanged)
		- check if the block is in the COW data.  
			- If so, return that data.
			- else, read the data from disk
		- QUESTION.  Do we need to lock the COW and Data writes durring this read, to make sure there isn't a write in the middle of this operation?

- if this is a full backup
	- behaves exactally like the incremental, except we don't care if there's an old_changelog.

### Psudocode
```
def read_snapshot_data(block_id)
	if (!g_in_snapshot)
		fail("You must be in a snapshot to get the get snapshot data")

	if (is_incremental_snapshot && !old_changelog)
		fail ("we can't preform an incremental snapshot at this time")

	lock(cow_list)	// ??? do we need to do this?
	if (block_is_in_cow_list(block_id)
		data = read_data_from_cow_store(block_id)
	else
		data = read_data_from_disk(block_id)
	unlock(cow_list)	// ??? do we need to do this?

	return data
end
```


## Complete Snapshot
After the snapshot is finished we need to cleanup the old_changelog and COW to free up memory.

- Check that we are in a snapshot.  If not fail.

- mark that we aren't in a snapshot anymore - in_snapshot = false

- if the old_changelog exists, delete it (in a background thread - no need to keep anything else waiting on this).

- if there was any COW, you can clean this up.

- if the bitmap was dirty, AND the backup was a full, we can now record that we had a successful snapshot.  This will be needed later when we shutdown.



## Fail Snapshot
If the snapshot didn't complete successfully, we have some additonal work because of the 2 lists.

- Check that we are in a snapshot.  If not fail.

- mark that we aren't in a snapshot anymore - in_snapshot = false

- if the bitmap status was clean and the old_changelog exists, we need to merge it into the current_changelog.  All the changed blocks from both lists should be marked as changed.  This can be done in a background thread so we don't need to wait.  We also need to make sure this is complete before we start a new snapshot, or shutdown.

- if the bitmap was dirty before the snapshot, we can turn off the bitmap storage feature and we don't need to keep any changelog data.  We will have to do a full backup before the snapshot can be considered successful, so there's no need to record the changed blocks.  The current_changelog can be cleaned up in a background thread because it isn't needed anymore.


## Shutdown
When the system shuts down, or we are unloaded we need to cleanup and update the bitmap_status metadata field.

- if a background job is running (merging old_changelog into the current_changelog, or cleaning up the old_changelog), wait for it to finish.

- if we are recording bitmap storage, wait for the thread to finish running and persisting to disk.

- if the bitmap status was clean, and no errors have occured with the bitmap storage, record the bitmap status as clean.

- in all other cases mark the bitmap status as dirty.



## Data write
This is still the most important part of the code because this will happen thousands of times a second.  It needs to be FAST.

- when an IO write comes in, check if we are recording changelogs.
	- if so, write the block to the current_changelog tree.  This write should return one of two values - NEW, indicating that it had to add the value to the tree, or EXISTING - indicating that the value was already in the tree.
	- are we also recording COW?
		- if so, and the tree write returned NEW, block the IO and copy the original block to the COW store.
			- if the tree write retuned EXISTING, the data should already be in the COW store.

	- if we are writing to the bitmap store, AND the tree write returned NEW, put the block on a queue to be persisted by another thread.

	- let the IO go through


## Background bitmap storage thread
This will be responsible for reading from a queue and writing the data to disk.

- check if there is anything on the queue.
- if there is, copy each value to a buffer.
- if the buffer is full
	- lock the buffer
	- issue an IOCTL to append the buffer to the existing data on disk.
	- record the location of the next "free" position on disk.
	- empty the buffer
	- unlock the buffer
- sleep

NOTE: because the changelog code will only put new blocks into the queue, we don't need to bother checking if that data is already in the bitmap store.  It won't be.

- if ANY errors occur durring this process, mark the bitmap as dirty, and clean everything up.  The changelog code should stop recording also.


## Optimizations:
- if we aren't recording changed blocks, we can use a function pointer to a no-opt function.  This will prevent us from needing to check if we are recording each IO.  Basically, this would act like a state machine.

- if we aren't persisting blocks to bitmap storage, we can also use a function pointer to a no-opt function.



# Changes to existing logic:

TODO: figure out if we need to have this section.



# General notes:

The existing code is a mess, in my opinion.  There is no real seperation of concerns - blksnap-main.c is over 2700 lines long!  The naming conventions are a mess.  There are virtually no comments about what things are doing.  We have virtually no unit tests.  And we don't have good "interfaces" for anything.  (ex, tree nodes are passed around a lot)

This needs to stop.  Our job as professionals is to make quality code that works AND is easy to understand and extend.  

Going forward I will be hitting these "non-functional" aspects of the code durring reviews.  If you have a function that is too long - even if you didn't write it, and are  - I expect it to be refactored into smaller functions.  I expect good names to be used.  I expect unit tests on any new code.  I expect comments.

