Tree of Arrays of Block flags

# Overview

Current implementation of block tree consumes a lot of memory - 72 bytes per block.
The new feature "instant rollback" requires to keep limited information of state of
each block on volume rollback runs to.
For example, a 1TB volume contains 2,7 * 10^8 blocks. The full block tree for this
volume consumes 19GB of RAM.
Tree of arrays propagate new strategy to block state store. State of block could be
packed in 7 bits. So all state flags could be packed in 1 byte (one bit is reserved
for future). 
In this case the full map of block state of 1TB volume consumes ~270MB of RAM.

# Advantages

- extremely reduction of memory use to store block status outside of snapshot process
- increase performance of block status change

# Disadvantages

- operate two rb-trees on snapshot process (block tree and cow tree)
- possible difficulties in solving race conditions when snapshot gets 

# Use cases

1. Operation during instant rollback
2. Operation on time period from attach till start of snapshot
3. Operation on time period from complete of snapshot till start of snapshot

# Technical specification

## Data structures

enum blksnap_block_flag_bits {
	__SNAP_THIS,
	__SNAP_NEXT,
	__SNAP_IN_FLIGHT,
	__CLSTORE_COMMITED,
	__HAS_COW_DATA,
	__RB_IN_FLIGHT,
	__RB_COMMITED
};

Where:
	__SNAP_THIS - block should be get in this snapshot
	__SNAP_NEXT - block should be get in next snapshot (this bit indicates that block
			has been modified AFTER snapshot is started)
	__SNAP_IN_FLIGHT - block is modified AFTER snapshot, presents in block tree,
			but it's data to read from snapshot device will be available later
	__CLSTORE_COMMITED - block is already commited to changelog-store
	__HAS_COW_DATA - indicates that block has a node in cow tree (probably it is possible
			to use __SNAP_NEXT instead of this flag)
	__RB_IN_FLIGHT - performs to read this block from rollback device, but it's data is
			not available yet
	__RB_COMMITED - block has been modified by user OR block has been read from repositiory and
			has been written to volume

struct blksnap_block_array {
	/* blocksize aligned sector address */
	sector_t start;

	/* number of blocks in current array */
	sector_t blocks;

	/* blksnap_block_array tree membership */
	struct rb_node rb_node;

	/* lock manipulations with block array */
	struct mutex lock;

	/* array of block flag */
	char *blocks;
};

struct blksnap_cow {
	/* Blocksize aligned sector address */
	sector_t addr;

	/* blksnap_cow tree membership */
	struct rb_node rb_node;

	/* Reference count. This is used to know when it is safe to free
	 * a claim object.  When the reference count reaches 0, the claim is
	 * safe to free.  The case for this is when a claim has been submitted
	 * to disk but not yet returned.  Upon submission the claim's reference
	 * count is incremented, so a call to free() will not destroy a claim
	 * for which the end_io callback has not yet been called. */
	atomic_t refcnt;

	/* data is committed (written) to disk */
	bool committed;

	/* offset in the extent map, if this claim is stored on disk */
	sector_t extent_offset;

	/* page data if this claim is stored in memory */
	struct page *page;

	/* Members of this structure are protected by this lock */
	rwlock_t lock;
};

## Public interface of subsystem

bool blksnap_bm_set_flag(struct blksnap_volume *volume, sector_t block, enum blksnap_block_flag_bits flag);
bool blksnap_bm_get_flag(struct blksnap_volume *volume, sector_t block, enum blksnap_block_flag_bits flag);
char blksnap_bm_set_and_clear_flag(struct blksnap_volume *volume, sector_t block, enum blksnap_block_flag_bits set, enum blksnap_block_flag_bits clear);
bool blksnap_bm_test_flag(char data, enum blksnap_block_flag_bits flag);

## Kernel functions to be used to implement "tree of arrays"

