#ifndef _BLKSNAP_H
#define _BLKSNAP_H

#include <linux/types.h>
#include <linux/fiemap.h>

#define BLKSNAP_NAME		"blksnap"
#define BLKSNAP_NAME_SIZE	128

/* NOTE!!! DON'T FORGET TO UPDATE THIS VERSION FOR ANY MAJOR
 * UPDATE OF INTERFACE!
 */
/* Version 3 includes new command --update-metadata-store that is not
 * compatible with older versions
 *
 * Version 4 uses path name instead of major/minor as input parameter of ioctl,
 * it changes size of struct blksnap_ioctl
 *
 * Version 5 has reduced size of struct blksnap_ioctl (528 bytes).
 *
 * Version 6 changes type of "ndevices" from _u8 to __u32 and
 * type of devices from "struct blksnap_ioctl_device *" to
 * "struct blksnap_ioctl_device[]
 *
 * Version 7 adds "fstype" and "getcrc" fields to blksnap_ioctl_attach
 * structure
 *
 * Version 8 adds "flags" to blksnap_ioctl_attach and blksnap_ioctl_start
 *
 * Version 9 adds BLKSNAP_IOCTL_CMD_GRAB_MAP ioctl and returns old semantics of
 * BLKSNAP_IOCTL_CMD_MAP ioctl: it's only receives extents, not grabs it. Also,
 * remove deprecated major and minor fields from struct blksnap_ioctl_map.
 *
 * Version 10 uses special extent structure for get_changelog ioctl which
 * has the same size as common extent structure, but has a name of length field
 * which points out that it is measured in blocks, not in sectors.
 * Also, rename 'start' field in extent structure to 'start_sect'.
 *
 * Version 11 adds new member fs_flags to blksnap_volume and uses 
 * variable flags from blksnap_ioctl_attach to pass XFS version number to 
 * blksnap_volume.
 *
 * Version 12 adds BLKSNAP_IOCTL_CMD_CHECK ioctl.
*/
#define BLKSNAP_IOCTL_VERSION	12

#ifndef U32_MAX
# define U32_MAX		((__u32)~0U)
#endif

#ifndef SIZE_MAX
# define SIZE_MAX		(~(size_t)0)
#endif

/* flag to tell to get just amount of attached devices
 * instead of devices list */
#define BLKSNAP_DL_REQ		U32_MAX

/* Signature to mark sectors to be added to extent map of store file
 */
#define BLKSNAP_MAP_SIGN	0xD6DC531A1A93DC11ULL

/* List of supported file systems and file systems to plan to be support
 */
enum blksnap_fs_type {
	BLKSNAP_FS_UNKNOWN,
	BLKSNAP_FS_EXT2,
	BLKSNAP_FS_EXT3,
	BLKSNAP_FS_EXT4,
	BLKSNAP_FS_XFS,
	BLKSNAP_FS_BTRFS,
	BLKSNAP_FS_VFAT
};

/* Additional errno values
 *
 * These should not collide with any system-defined values.  The exit(2) system
 * call will mask out all but the lower 8 bits. To determine what value you'll
 * see in userspace: (value & 0377).
 */
#define EVERSION		700	/* 188 */
#define EHASSNAP		701	/* 189 */
#define ENOSNAP			702	/* 190 */
#define EDIRTY			703	/* 191 */
#define EHASDATA		704	/* 192 */ /* Device has mapped data
						     store */
#define ENOTALLOC		705	/* 193 */ /* store is not allocated */
#define ECLSNOMAP		706	/* 194 */ /* cl-store is not mapped */
#define EDTSNOMAP		707	/* 195 */ /* datastore is not mapped */
#define ECLSFULL		708	/* 196 */ /* cl-store is full */
#define EMDSNOMAP		709	/* 197 */ /* metadata is not mapped */
#define EOFBITMAP		710	/* 198 */ /* end of bitmap reached */
#define ESIZESTORE		711	/* 199 */ /* datastore has the wrong size */
#define ESIZEVOLUME		712	/* 200 */ /* volume has the wrong size */

enum blksnap_device_flag_bits {
	__BLKSNAP_DEVICE_DIRTY,
	__BLKSNAP_DEVICE_CHLOG,
	__BLKSNAP_DEVICE_DATA,
	__BLKSNAP_DEVICE_METADATA,
	__BLKSNAP_DEVICE_INCR
};

#define BLKSNAP_DEVICE_DIRTY	(1 << __BLKSNAP_DEVICE_DIRTY)
#define BLKSNAP_DEVICE_CHLOG	(1 << __BLKSNAP_DEVICE_CHLOG)
#define BLKSNAP_DEVICE_DATA	(1 << __BLKSNAP_DEVICE_DATA)
#define BLKSNAP_DEVICE_METADATA	(1 << __BLKSNAP_DEVICE_METADATA)
#define BLKSNAP_DEVICE_INCR	(1 << __BLKSNAP_DEVICE_INCR)

/* Flags of attach ioctl command */
#define BLKSNAP_ATTACH_MOUNT_RW		1
#define BLKSNAP_ATTACH_XFS_VERSION_5	2

/* Flags of start snapshot ioctl command
 * bsctl-main.c: values for act.flags
 */
#define BLKSNAP_START_INCR 1
#define BLKSNAP_EXTENTS    2

/* used by _attach, _map, _get_map and grab ioctl commands */
struct blksnap_ioctl_file_extent {
	__u64 start_sect;	/* input: offset of extent, from volume start */
	size_t nsectors;	/* input: length of extent */
};

/* used by _changelog ioctl command  */
struct blksnap_ioctl_disk_extent {
	__u64 start_sect;	/* output: offset of extent, from device start */
	size_t nblocks;		/* output: length of extent */
};

struct blksnap_ioctl_empty {};

struct blksnap_ioctl_attach {
	char name[BLKSNAP_NAME_SIZE];			/* input */
	enum blksnap_fs_type fstype;			/* input */
	__u8  flags;					/* input */
	/* an area of disk to calculate checksum to detect
	 * if there were any changes on disk between detach/attach
	 * commands
	 */
	__u32 nextents;					/* input */
	struct blksnap_ioctl_file_extent hashblocks[];	/* input */
};

struct blksnap_ioctl_detach {
	char name[BLKSNAP_NAME_SIZE];	/* input */
};

struct blksnap_ioctl_device {
	char name[BLKSNAP_NAME_SIZE];		/* input (output for list) */
	char device_name[BLKSNAP_NAME_SIZE];	/* underlying physical name */
	char snapshot_name[BLKSNAP_NAME_SIZE];	/* snapshot device path */
	char link_name[BLKSNAP_NAME_SIZE];	/* link device path */
	__u32 flags;				/* device flags */
};

struct blksnap_ioctl_list {
	__u32 ndevices;	/* output */
	struct blksnap_ioctl_device devices[0];
};

struct blksnap_ioctl_start {
	char name[BLKSNAP_NAME_SIZE];		/* input */
	__u8 flags;				/* input/output */
	char snapshot_name[BLKSNAP_NAME_SIZE];	/* output */
};

struct blksnap_ioctl_stop {
	char name[BLKSNAP_NAME_SIZE];	/* input */

	/* What is done with the marked tree nodes when stopping a snapshot. By
	 * default, marked nodes are pruned (free'd and removed), leaving the
	 * tree containing only the blocks for the next snapshot. If something
	 * goes wrong and this snapshot must be stopped but blocks should be
	 * preserved, the merge flag instructs the driver to merge the marked
	 * nodes into the set of unmarked nodes, so all current blocks are
	 * included in the next snapshot.
	 */
	__u8 merge;	/* input */
};

struct blksnap_ioctl_show {
	char name[BLKSNAP_NAME_SIZE];		/* input */
	char snapshot_name[BLKSNAP_NAME_SIZE];	/* output */
	char link_name[BLKSNAP_NAME_SIZE];	/* output */
};

struct blksnap_ioctl_changelog {
	char name[BLKSNAP_NAME_SIZE];	/* input */
	__u64 offset;			/* input, sector from volume start */
	union {
		__u32 nextents;	/* output if 0, input otherwise */
		__u32 nblocks;	/* input (used to request changelog size only) */
	};
	struct blksnap_ioctl_disk_extent extents[];	/* output */
};

enum blksnap_map_type {
	BLKSNAP_MAP_TYPE_CHLOG,
	BLKSNAP_MAP_TYPE_DATA,
	BLKSNAP_MAP_TYPE_METADATA
};

enum blksnap_unmap_type {
	BLKSNAP_UNMAP_TYPE_DATA
};

/* Flags to process multiply ioctl requests to map file
 *
 * First request MUST be flagged as BLKSNAP_STORE_FIRST,
 * First request MUST include valid "npages" value
 * (see struct blksnap_ioctl_map)
 * Last request MUST be flagged as BLKSNAP_STORE_LAST,
 * Last request cause to mark store map as valid to use
 * BLKSNAP_STORE_NEXT is reserved to be used in future
 * If there is the one only ioctl request to map the
 * whole file, it MUST be flaged both as
 * BLKSNAP_STORE_FIRST and BLKSNAP_STORE_LAST
 *
 * Flag BLKSNAP_STORE_UPDATE_METADATA tells kernel to map
 * metadata in a special way: record superblock calculated
 * during attach and clean volume if it hadn't been mounted
 * during attach. See bsctl --help.
 */
enum blksnap_store_flag_bits {
	__BLKSNAP_STORE_FIRST,
	__BLKSNAP_STORE_NEXT,
	__BLKSNAP_STORE_LAST,
	__BLKSNAP_STORE_UPDATE_METADATA
};

#define BLKSNAP_STORE_FIRST		(1 << __BLKSNAP_STORE_FIRST)
#define BLKSNAP_STORE_NEXT		(1 << __BLKSNAP_STORE_NEXT)
#define BLKSNAP_STORE_LAST		(1 << __BLKSNAP_STORE_LAST)
#define BLKSNAP_STORE_UPDATE_METADATA	(1 << __BLKSNAP_STORE_UPDATE_METADATA)

struct blksnap_ioctl_map {
	char name[BLKSNAP_NAME_SIZE];	/* input */
	__u32 type;	/* input */ /* see blksnap_[un]map_type */
	__u32 npages;	/* input */
	__u32 nextents;	/* input */
	__u32 flags;	/* input */
	struct blksnap_ioctl_file_extent extents[];	/* input */
};

/* If FS doesn't support fiemap/fibmap ioctls, bsctl and blk-snapshot
 * emulates fiemap with mark/grab operations.
 * At the start bsctl tells BLKSNAP_GRAB_MAP_START and writes each sector
 * of file with fixed signature.
 * blk-snapshot tracks write requests to the volume, grab sector numbers
 * that start with given signature and adds them into the temporary extents
 * storage in volume structure. Original bio request to write a signature
 * is not really passed to the drive.
 * BLKSNAP_GRAB_MAP_END tells module to finish gathering file extents and
 * turn off a bio filter. It returns to userspace count of file extents
 * to allow user to allocate sufficient space for fetching extents.
 * BLKSNAP_GRAB_MAP_FETCH_EXT tells module to copy extents to userspace
 * and remove them from temporary storage in volume structure.
 * BLKSNAP_GRAB_MAP_CANCEL tells to turn off bio filter and remove
 * already gathered extents from temporary storage in volume structure.
 */
enum blksnap_grab_map_cmd {
	BLKSNAP_GRAB_MAP_START,
	BLKSNAP_GRAB_MAP_END,
	BLKSNAP_GRAB_MAP_FETCH_EXT,
	BLKSNAP_GRAB_MAP_CANCEL
};

struct blksnap_ioctl_grab_map {
	char name[BLKSNAP_NAME_SIZE];
	__u32 cmd;			/* input, see enum above */
	__u32 nextents;			/* input on fetch, output on end */
	struct blksnap_ioctl_file_extent extents[];	/* output */
};

struct blksnap_ioctl_unmap {
	char name[BLKSNAP_NAME_SIZE];	/* input */
	__u32 type;			/* input */
};

struct blksnap_ioctl_dump {
	char name[BLKSNAP_NAME_SIZE];	/* input */
};

struct blksnap_ioctl_restore {
	char source_name[BLKSNAP_NAME_SIZE];	/* input */
	char target_name[BLKSNAP_NAME_SIZE];	/* input */
};

struct blksnap_ioctl_reset {
	char name[BLKSNAP_NAME_SIZE];	/* input */
};

struct blksnap_ioctl_check {
	char name[BLKSNAP_NAME_SIZE];	/* input */
};

struct blksnap_ioctl {
	__u32 version;
	__u32 size;	/* total size including this struct */

	union {
		struct blksnap_ioctl_empty	end_of_header;
		struct blksnap_ioctl_attach	attach;
		struct blksnap_ioctl_detach	detach;
		struct blksnap_ioctl_device	device;
		struct blksnap_ioctl_list	list;
		struct blksnap_ioctl_start	start;
		struct blksnap_ioctl_stop	stop;
		struct blksnap_ioctl_show	show;
		struct blksnap_ioctl_changelog	changelog;
		struct blksnap_ioctl_map	map;
		struct blksnap_ioctl_grab_map	grab;
		struct blksnap_ioctl_unmap	unmap;
		struct blksnap_ioctl_dump	dump;
		struct blksnap_ioctl_restore	restore;
		struct blksnap_ioctl_reset	reset;
		struct blksnap_ioctl_check	check;
	};
};

#define BLKSNAP_IOCTL_HEADER_SIZE offsetof(struct blksnap_ioctl, end_of_header)

enum {
	BLKSNAP_IOCTL_CMD_ATTACH,	/* attach device */
	BLKSNAP_IOCTL_CMD_DETACH,	/* detach device */
	BLKSNAP_IOCTL_CMD_INFO,		/* get device info */
	BLKSNAP_IOCTL_CMD_LIST,		/* get list of attached devices */
	BLKSNAP_IOCTL_CMD_START,	/* start snapshot */
	BLKSNAP_IOCTL_CMD_STOP,		/* stop snapshot */
	BLKSNAP_IOCTL_CMD_SHOW,		/* show snapshot or source of restore device name */
	BLKSNAP_IOCTL_CMD_CHANGELOG,	/* get changelog blocks list */
	BLKSNAP_IOCTL_CMD_MAP,		/* map file extents into module */
	BLKSNAP_IOCTL_CMD_UNMAP,	/* unmap file from module */
	BLKSNAP_IOCTL_CMD_GRAB_MAP,	/* gather file extents map for module */
	BLKSNAP_IOCTL_CMD_GET_MAP,	/* get file extents map from module */
	BLKSNAP_IOCTL_CMD_DUMP,		/* dump into kmsg some changelog info */
	BLKSNAP_IOCTL_CMD_RESTORE,	/* restore data from src device to target */
	BLKSNAP_IOCTL_CMD_RESET,	/* stop restoring data src -> target */
	BLKSNAP_IOCTL_CMD_CHECK,	/* check attached device */
	__BLKSNAP_IOCTL_CMD_LAST
};

/*
 * Ioctls definition
 */
#define BLKSNAP_IOCTL_MAGIC 'B'

#define BLKSNAP_IOCTL_CMD_FIRST BLKSNAP_IOCTL_CMD_ATTACH
#define BLKSNAP_IOCTL_CMD_LAST (__BLKSNAP_IOCTL_CMD_LAST - 1)

#define BLKSNAP_IOC_ATTACH \
	_IOR(BLKSNAP_IOCTL_MAGIC, \
	     BLKSNAP_IOCTL_CMD_ATTACH, struct blksnap_ioctl)

#define BLKSNAP_IOC_DETACH \
	_IOR(BLKSNAP_IOCTL_MAGIC, \
	     BLKSNAP_IOCTL_CMD_DETACH, struct blksnap_ioctl)

#define BLKSNAP_IOC_INFO \
	_IOR(BLKSNAP_IOCTL_MAGIC, \
	     BLKSNAP_IOCTL_CMD_INFO, struct blksnap_ioctl)

#define BLKSNAP_IOC_LIST \
	_IOR(BLKSNAP_IOCTL_MAGIC, \
	     BLKSNAP_IOCTL_CMD_LIST, struct blksnap_ioctl)

#define BLKSNAP_IOC_START \
	_IOR(BLKSNAP_IOCTL_MAGIC, \
	     BLKSNAP_IOCTL_CMD_START, struct blksnap_ioctl)

#define BLKSNAP_IOC_STOP \
	_IOR(BLKSNAP_IOCTL_MAGIC, \
	     BLKSNAP_IOCTL_CMD_STOP, struct blksnap_ioctl)

#define BLKSNAP_IOC_SHOW \
	_IOR(BLKSNAP_IOCTL_MAGIC, \
	     BLKSNAP_IOCTL_CMD_SHOW, struct blksnap_ioctl)

#define BLKSNAP_IOC_CHANGELOG \
	_IOR(BLKSNAP_IOCTL_MAGIC, \
	     BLKSNAP_IOCTL_CMD_CHANGELOG, struct blksnap_ioctl)

#define BLKSNAP_IOC_MAP \
	_IOR(BLKSNAP_IOCTL_MAGIC, \
	     BLKSNAP_IOCTL_CMD_MAP, struct blksnap_ioctl)

#define BLKSNAP_IOC_UNMAP \
	_IOR(BLKSNAP_IOCTL_MAGIC, \
	     BLKSNAP_IOCTL_CMD_UNMAP, struct blksnap_ioctl)

#define BLKSNAP_IOC_GRAB_MAP \
	_IOR(BLKSNAP_IOCTL_MAGIC, \
	     BLKSNAP_IOCTL_CMD_GRAB_MAP, struct blksnap_ioctl)

#define BLKSNAP_IOC_GET_MAP \
	_IOR(BLKSNAP_IOCTL_MAGIC, \
	     BLKSNAP_IOCTL_CMD_GET_MAP, struct blksnap_ioctl)

#define BLKSNAP_IOC_DUMP \
	_IOR(BLKSNAP_IOCTL_MAGIC, \
	     BLKSNAP_IOCTL_CMD_DUMP, struct blksnap_ioctl)

#define BLKSNAP_IOC_RESTORE \
	_IOR(BLKSNAP_IOCTL_MAGIC, \
	     BLKSNAP_IOCTL_CMD_RESTORE, struct blksnap_ioctl)

#define BLKSNAP_IOC_RESET \
	_IOR(BLKSNAP_IOCTL_MAGIC, \
	     BLKSNAP_IOCTL_CMD_RESET, struct blksnap_ioctl)

#define BLKSNAP_IOC_CHECK \
	_IOR(BLKSNAP_IOCTL_MAGIC, \
	     BLKSNAP_IOCTL_CMD_CHECK, struct blksnap_ioctl)

#endif /* _BLKSNAP_H */
