#ifndef _BSCTL_H
#define _BSCTL_H

#include <linux/limits.h>

/* NOTE!!! DON'T FORGET TO UPDATE THIS VERSION FOR ANY MAJOR
 * UPDATE OF USER INTERFACE in libbsctl/bsctl!
 */
/* Version 1 is base version (5.2)
 *
 * Version 2 includes new command --update-metadata-store
 *
 * Version 3 is first version with new interface (libbsctl)
 *
 * Version 4 includes new option --incremental for command --start-snapshot,
 *           includes new command --api-version and
 *                    new API function bsctl_get_api_version()
 *
 * Version 5 includes new API functions bsctl_aafmap_get() and
 *           bsctl_aafmap_free()
 *
 * Version 6 includes new flag __BLKSNAP_DEVICE_INCR for struct
 *           bsctl_device_info and corresponding new field in output of
 *           --show-device-info command
 *
 * Version 7 includes new command --check-device
 *                    new API function bsctl_check()
 *
 */
#define BSCTL_API_VERSION   7

#include <blksnap.h>
#include "bsctl-error.h"

enum map_method {
	MAP_AUTO,
	MAP_FIEMAP,
	MAP_FIBMAP,
	MAP_EMULATION
};

/* external interface to C#, need to know pack value */
#pragma pack(push, 1)
struct bsctl_chlog_extent {
	uint64_t start;	/* start of extent in sectors */
	uint32_t len;	/* WARNING!!! length of extent in 4k blocks */
};

struct bsctl_chlog {
	uint64_t nextents;
	struct bsctl_chlog_extent extents[];
};

struct bsctl_aaf_extent {
	uint64_t start_block;
	uint64_t nblocks;
};

struct bsctl_aafmap {
	uint64_t nextents;
	struct bsctl_aaf_extent extents[];
};

struct bsctl_device_info {
	char name[BLKSNAP_NAME_SIZE];
	char device_name[BLKSNAP_NAME_SIZE];
	char snapshot_name[BLKSNAP_NAME_SIZE];
	char link_name[BLKSNAP_NAME_SIZE];
	uint32_t flags;
};

struct bsctl_device_list {
	uint8_t ndevices;
	struct bsctl_device_info list[] ;
};
#pragma pack(pop)

struct mapping_params {
	enum blksnap_map_type type;
	const char *dev_path;
	char file_path[PATH_MAX];
	uint32_t npages;
	uint32_t flags;
	enum map_method method;
	ssize_t bufsize; /* size of buffer to perform fiemap emulation */
};

/* external API */
int bsctl_api_version(struct bsctl_error *err);
int bsctl_version(char *buf, size_t len);

int bsctl_attach(const char *dev_path, struct bsctl_error *err);
int bsctl_detach(const char *dev_path, struct bsctl_error *err);

int bsctl_snapshot_get(const char *dev_path, char *snap,
		       struct bsctl_error *err);
int bsctl_snapshot_start(const char *dev_path, char *snap, int32_t flags,
			 struct bsctl_error *err);
int bsctl_snapshot_complete(const char *dev_path, struct bsctl_error *err);
int bsctl_snapshot_cancel(const char *dev_path, struct bsctl_error *err);

int bsctl_chlog_create(const char *dev_path, struct bsctl_error *err);
int bsctl_chlog_reset(const char *dev_path, struct bsctl_error *err);
int bsctl_chlog_check(const char *dev_path, struct bsctl_error *err);
int bsctl_chlog_map(const char *dev_path, struct bsctl_error *err);
int bsctl_chlog_map_custom(const char *dev_path, enum map_method method,
		           ssize_t bufsize, struct bsctl_error *err);
struct bsctl_chlog *bsctl_chlog_get(const char *dev_path,
				    struct bsctl_error *err);
void bsctl_chlog_free(struct bsctl_chlog *chlog);

int bsctl_data_create(const char *dev_path, uint64_t size, double percent,
		      struct bsctl_error *err);
int __bsctl_data_check(const char *dev_path, uint64_t *size, double percent,
		     struct bsctl_error *err);
int bsctl_data_check(const char *dev_path, uint64_t size, double percent,
		     struct bsctl_error *err);
int bsctl_data_map(const char *dev_path, struct bsctl_error *err);
int bsctl_data_unmap(const char *dev_path, struct bsctl_error *err);
int bsctl_data_map_custom(const char *dev_path, enum map_method method,
			  ssize_t bufsize, struct bsctl_error *err);

int bsctl_metadata_map(const char *dev_path, struct bsctl_error *err);
int bsctl_metadata_map_custom(const char *dev_path, enum map_method method,
			      ssize_t bufsize, uint32_t flags,
			      struct bsctl_error *err);
int bsctl_metadata_update(const char *dev_path, struct bsctl_error *err);
int bsctl_metadata_unmap(const char *dev_path, struct bsctl_error *err);

int bsctl_volume_restore(const char *source, const char *target,
			 struct bsctl_error *err);
int bsctl_volume_reset(const char *dev_path, struct bsctl_error *err);

int bsctl_check(const char *dev_path, struct bsctl_error *err);

int bsctl_dump_kernel(const char *dev_path, struct bsctl_error *err);

struct bsctl_device_info *bsctl_device_info_get(const char *dev_path,
						struct bsctl_error *err);
void bsctl_device_info_free(struct bsctl_device_info *info);

struct bsctl_device_list *bsctl_device_list_get(struct bsctl_error *err);
void bsctl_device_list_free(struct bsctl_device_list *list);

struct bsctl_aafmap *bsctl_aafmap_get(const char *dev_path,
				      const char *file_path,
				      struct bsctl_error *err);
void bsctl_aafmap_free(struct bsctl_aafmap *map);

/* Functions for utils use only */
dev_t bsctl_get_rdev(const char *dev_path, struct bsctl_error *err);
int bsctl_get_device_freespace(const char *path, uint64_t *free_bytes,
			       struct bsctl_error *err);
uint64_t bsctl_get_device_size(const char *dev_path, struct bsctl_error *err);
uint64_t bsctl_get_volume_start_sect(const char *dev_path);

/* For debug use only */
struct bsctl_extent {
	uint64_t start;	/* sectors */
	size_t nsectors;
};

struct bsctl_map {
	uint64_t nextents;
	struct bsctl_extent extents[];
};

struct bsctl_map *bsctl_chlog_get_map(const char *dev_path,
				      struct bsctl_error *err);
struct bsctl_map *bsctl_data_get_map(const char *dev_path,
				     struct bsctl_error *err);
struct bsctl_map *bsctl_metadata_get_map(const char *dev_path,
					 struct bsctl_error *err);
void bsctl_map_free(struct bsctl_map *map);

/* For reuse in bsrw */
int bsctl_file_reset(int fd, uint64_t size, struct bsctl_error *err);

#endif /* _BSCTL_H */
