#include <linux/kernel.h>
#include <linux/kthread.h>
#include <linux/blkdev.h>
#include <linux/bio.h>

#include "blksnap.h"
#include "blksnap-bio-queue-share.h"
#include "blksnap-hash.h"
#include "blksnap-volume.h"
#include "blksnap-compat.h"
#include "blksnap-shared.h"
#include "blksnap-bio-queue.h"

#define BIO_QUEUE_WAIT (0.1 * HZ) /* timeout to run bio queue thread */

static int bio_flush_fn(void *data)
{
	struct blksnap_bio_queue *bq = (struct blksnap_bio_queue *) data;

	set_user_nice(current, MIN_NICE);

	while (!kthread_should_stop() || !bio_list_empty(&bq->bios)) {
		wait_event_interruptible(bq->event,
					 kthread_should_stop() ||
					 !bio_list_empty(&bq->bios));

		if (bio_list_empty(&bq->bios))
			continue;

		bq->flush(bq);
	}

	return 0;
}

unsigned int blksnap_bio_kthread_init(struct blksnap_volume *volume,
				      struct blksnap_bio_queue *queue,
				      const char *name,
				      flush_func_t flush)
{
	spin_lock_init(&queue->lock);

	bio_list_init(&queue->bios);
	init_waitqueue_head(&queue->event);

	queue->volume	 = volume;
	queue->flush	 = flush;

	queue->task = kthread_run(bio_flush_fn, queue, "%s-%u:%u",
				  name,
				  MAJOR(volume->bdev->bd_dev),
				  MINOR(volume->bdev->bd_dev));

	if (!queue->task || IS_ERR(queue->task))
		return -EFAULT;

	return 0;
}

static void bio_flush_all(struct blksnap_bio_queue *bq)
{
	while (!bio_list_empty(&bq->bios))
		bq->flush(bq);
}

void blksnap_bio_kthread_destroy(struct blksnap_bio_queue *bq)
{
	BUG_ON(!bq);
	BUG_ON(!bq->task);

	kthread_stop(bq->task);
	bio_flush_all(bq);
}

void blksnap_bio_kthread_enqueue(struct blksnap_bio_queue *bq, struct bio *bio)
{
	unsigned long flags;

	spin_lock_irqsave(&bq->lock, flags);
	bio_list_add(&bq->bios, bio);
	spin_unlock_irqrestore(&bq->lock, flags);

	wake_up(&bq->event);
}
