#include <linux/gfp.h>
#include <linux/mm_types.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/bio.h>
#include <linux/blkdev.h>

#if defined __has_include
#  if __has_include (<linux/genhd.h>)
#    include <linux/genhd.h>
#  endif
#else
#  include <linux/genhd.h>
#endif

#include "blksnap.h"
#include "blksnap-bio-queue-share.h"
#include "blksnap-hash.h"
#include "blksnap-device.h"
#include "blksnap-compat.h"
#include "blksnap-shared.h"
#include "blksnap-bio.h"
#include "blksnap-bio-submit.h"

/* Try to process as many bytes as allowed for single bio request.
 * @bdev:	partition or disk if no parititons
 * @sector:	first disk sector for the IO
 * @buf:	preallocated buffer to WRITE from it or to READ to
 * @len:	len of operation, bytes, may NOT be aligned to SECTOR_SIZE
 * @op:		[READ|WRITE] operation
 * Return value: count of processed bytes or error code.
 */
static int submit_io(struct block_device *bdev, sector_t sector,
		     unsigned char *buf, uint64_t len, int op)
{
	gfp_t gfp_flags = GFP_KERNEL;
	struct request_queue *q = bdev_get_queue(bdev);
	int ret = 0;
	int pages_todo;
	int bytes_todo, bytes_remain;
	unsigned int vec_size;
	struct bio *bio;
	void *data_ptr = buf;

	if (!q) {
		ret = -ENODEV;
		bsmsg_bdev(bdev, "ERROR: Can not determine request queue for bdev [%d:%d]",
			   MAJOR(bdev->bd_dev), MINOR(bdev->bd_dev));
		goto out;
	}

	bytes_todo = bytes_remain = MIN(len, BIO_MAX_SIZE);
	pages_todo = bytes2objects(bytes_todo, PAGE_SIZE);

#if USE_BIO_KMALLOC
	bio = bio_kmalloc(gfp_flags, pages_todo);
#else
#  if BIO_ALLOC_4_ARGS
	bio = bio_alloc(bdev, pages_todo, op, gfp_flags);
#  else
	bio = bio_alloc(gfp_flags, 1);
#  endif
#endif

	if (!bio) {
		ret = -ENOMEM;
		bsmsg_bdev(bdev, "ERROR: Can not allocate bio(%llu, %llu) with %d segments",
			   sector, len, pages_todo);
		goto out;
	}

	/* bdev of bio must be initialized before bio_add_page() */
	COMPAT_BIO_SET_BDEV(bio, bdev);
	COMPAT_BIO_SECTOR(bio) = sector;

	while (bytes_remain) {
		struct page *page;
		unsigned int bytes = PAGE_SIZE;

		if (bytes > bytes_remain)
			bytes = bytes_remain;

		/* NOTE: bio-request MUST be equal or greater than SECTOR_SIZE,
		 * so the minimum size of added vector is SECTOR_SIZE.
		 * THIS NEED TO BE RECHECKED!!!
		 * Commented out temporarily.
		 */
		vec_size = bytes;
		if (bytes_todo < SECTOR_SIZE)
			vec_size = SECTOR_SIZE;

#if HAVE_BOUNCE_GFP
		page = alloc_page(q->bounce_gfp | gfp_flags);
#else
		page = alloc_page(GFP_NOIO | gfp_flags);
#endif
		if (!page) {
			ret = -ENOMEM;
			bsmsg_bdev(bdev, "ERROR: Can not allocate page for bio(%llu, %llu) with %d segments",
				   sector, len, pages_todo);
			goto free_bio_and_pages;
		}

		if (op != BLKSNAP_READ) {
			if (bytes < vec_size)
				memset(page_address(page), 0, SECTOR_SIZE);
			memcpy(page_address(page), data_ptr, bytes);
		}

		/* if the request queue limit is reached, just stop adding
		 * pages and process bio. A submit_io() caller should process
		 * return value and call it again with proper size and pointer
		 * to next data portion.
		 */
#if USE_BIO_KMALLOC
		if (bio_add_pc_page(q, bio, page, vec_size, 0) < vec_size)
#else
		if (bio_add_page(bio, page, vec_size, 0) < vec_size)
#endif
			break;

		bytes_remain -= vec_size;
		data_ptr += vec_size;
	}

	if (bytes_remain > 0)
		bytes_todo -= bytes_remain;

	blksnap_submit_bio_wait(bio, op);

	if (op == BLKSNAP_READ) {
		char *buf_ptr = buf;
		struct bio_vec *bvec;

		if (bytes_todo < SECTOR_SIZE) {
			bvec = bio->bi_io_vec;
			memcpy(buf_ptr, page_address(bvec->bv_page), bytes_todo);
		} else {
			blksnap_bio_for_each_segment_all(bvec, bio) {
				memcpy(buf_ptr, page_address(bvec->bv_page), bvec->bv_len);
				buf_ptr += bvec->bv_len;
			}
		}
	}

	ret = bytes_todo;

free_bio_and_pages:
	bio_free_pages(bio);
	bio_put(bio);
out:
	return ret;
}

/* reads/writes to/from continuous area of disk
 * @bdev:	partition or disk if no parititions
 * @start_sect:	sector related to begin of partition
 * @len_bytes:	length in bytes
 * @data:	pointer to preallocated buffer for read/write operation
 * @op:		[READ|WRITE|WRITE_FILTER_DISABLED] operation
 */
int blksnap_disk_io(struct block_device *bdev,
		    sector_t start_sect,
		    uint64_t len_bytes,
		    unsigned char *data,
		    int op)
{
	int ret = 0;
	int64_t bytes_left = len_bytes;

	while (bytes_left > 0) {
		int bytes_done = submit_io(bdev,
					   start_sect,
					   data + (len_bytes - bytes_left),
					   bytes_left,
					   op);
		/* check bytes_done == 0 to prevent infinity loop */
		if (bytes_done <= 0) {
			bsmsg_bdev(bdev,
				   "ERROR: Can not read/write sector %llu",
				   start_sect);
			ret = (bytes_done) ? bytes_done : -EIO;
			break;
		}

		bytes_left -= bytes_done;

		if (bytes_left && (ret % SECTOR_SIZE)) {
			bsmsg_bdev(bdev, "ERROR: Unaligned bio detected to submit!");
			bsmsg_bdev(bdev, "ERROR(2): bytes_left=%d ret=%d",
				   bytes_left, ret);
			ret = -EIO;
			break;
		}

		start_sect += bytes_done >> SECTOR_SHIFT;
	}

	return ret;
}
