#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/blkdev.h>
#include <linux/vmalloc.h>
#include <linux/math64.h>
#include <linux/bitops.h>
#include <linux/rwsem.h>

#include "blksnap.h"
#include "blksnap-bio-queue-share.h"
#include "blksnap-hash.h"
#include "blksnap-device.h"
#include "blksnap-compat.h"
#include "blksnap-shared.h"
#include "blksnap-cl-store.h"
#include "blksnap-cl-thread.h"
#include "blksnap-data.h"
#include "blksnap-cow-tree.h"
#include "blksnap-bitmap.h"
#include "blksnap-block.h"

/* For changelog that has fewer than this limit, print the individual block
 * numbers to the kernel log. Otherwise, only display the blocks count.
 */
#define CHANGELOG_SHOW_LIMIT		32

typedef unsigned long bitmap_step_t;

static inline bool is_last_chunk(struct blksnap_volume *volume,
				 sector_t nchunk)
{
	return (nchunk >= (volume->bitmap.nchunks - 1));
}

/* @volume - blksnap volume structure
 * @iter   - loop iterator to start work from the last point
 * Return: error code or 0 if success
 */
__always_inline
static int get_next_changed_block(struct blksnap_volume *volume,
				  struct blksnap_bitmap_iter *iter)
{
next_block:
	/* instantly go to next block */
	iter->nblock++;

check_block:
	if (!iter->chunk->changed_blocks || iter->nblock >= BITMAP_CHUNK_BLOCKS)
		goto next_chunk;

	if (iter->chunk->blocks[iter->nblock]) {
		iter->block = &iter->chunk->blocks[iter->nblock];
		return 0;
	} else {
		if (!*( (bitmap_step_t *) (iter->chunk->blocks + iter->nblock))) {
			iter->nblock += sizeof(bitmap_step_t);
			goto check_block;
		} else {
			goto next_block;
		}
	}

next_chunk:
	/* lookup next non-empty chunk.
	 * ATTENTION: chunk IS still read-locked after this loop
	 */
	/* no chunks with data */
	if (is_last_chunk(volume, iter->nchunk)) {
		up_read(&iter->chunk->lock_alloc);
		return EOFBITMAP;
	}

	up_read(&iter->chunk->lock_alloc);
	iter->nblock = 0;
	iter->nchunk++;
	iter->chunk = &volume->bitmap.chunks[iter->nchunk];
	down_read(&iter->chunk->lock_alloc);

	goto check_block;
}

/* @volume - blksnap volume structure
 * @iter   - uninitialized loop iterator.
 *           WARNING! All previous data in @iter is erased!
 * @offset - start from @offset sector of volume
 * Return: error code or 0 if success
 */
static int get_first_changed_block(struct blksnap_volume *volume,
				   struct blksnap_bitmap_iter *iter,
				   sector_t offset)
{
	memset(iter, 0, sizeof(struct blksnap_bitmap_iter));

	if (!offset)
		iter->chunk = &volume->bitmap.chunks[0];
	else
		abs_addr_to_index(volume, offset,  iter);

	down_read(&iter->chunk->lock_alloc);

	if (!iter->chunk->changed_blocks)
		goto skip_check;

	iter->block = &iter->chunk->blocks[iter->nblock];
	if (*iter->block) {
		return 0;
	}
skip_check:

	return get_next_changed_block(volume, iter);
}

/* @nchunk - index of chunk in volume->bitmap.chunks[]
 * @nblock - block offset within volume->bitmap.chunks[i].blocks
 * Return: absolute (from device start) page-aligned sector offset
 */
static inline sector_t __index_to_abs_addr(struct blksnap_volume *volume,
					   sector_t nchunk, uint16_t nblock)
{
	return volume->start_block
	       + ((nchunk * BITMAP_CHUNK_BLOCKS + nblock)
		  << PAGE_SECTORS_SHIFT);
}

/* Start block of volume can be 0 in abs value if the volume is a whole device.
 * On-disk store stores abs values and uses 0-block as EOF.
 * For this case start blocks should always be in RAM changelog to force agent
 * to grab it on incremental snapshot because it can't be stored at on-disk.
 */
static inline int add_start_block(struct blksnap_volume *volume)
{
	return block_set_bit(volume, volume->start_block, NEXT_SNAP);
}

/* Forced freeing bitmap chunks */
static void free_bitmap_chunks(struct blksnap_volume *volume)
{
	int i;
	struct bitmap_chunk *chunk;

	for (i = 0; i < volume->bitmap.nchunks; i++) {
		chunk = &volume->bitmap.chunks[i];
		down_write(&chunk->lock_alloc);
		if (chunk->changed_blocks) {
			kfree(chunk->blocks);
			chunk->changed_blocks = 0;
			chunk->blocks = NULL;
		}
		WARN_ON(chunk->blocks);
		up_write(&chunk->lock_alloc);
	}
}

/* See note about bitmap and volume addresses relations in doc/addressing.txt */
sector_t last_bitmap_block_addr(sector_t start_sector, sector_t nsectors)
{
	sector_t addr;

	/* volume end offset from device start */
	addr = start_sector + nsectors;

	/* one-block step back  */
	if (addr % PAGE_SECTORS)
		/* last bitmap block is incomplete */
		addr = aligned_block_address(addr);
	else
		addr -= PAGE_SECTORS;

	return addr;
}

int blksnap_blk_bitmap_create(struct blksnap_volume *volume)
{
	int i;
	int ret = 0;
	sector_t alloc_size, bitmap_nblocks;
	u32 reminder;

	if (!volume || !volume->bdev) {
		WARN_VOLUME(volume, 1, "Can't create block map: volume structure is incomplete");
		return -EINVAL;
	}

	if (volume->bitmap.chunks) {
		WARN_VOLUME(volume, 1, "Block map is already allocated");
		return -EEXIST;
	}

	/* Is changed blocks counter big enough? */
	if ((1 << (8 * sizeof(volume->bitmap.chunks->changed_blocks)))
	    < BITMAP_CHUNK_BLOCKS) {
		WARN(1, "Bitmap chunk size variable has insufficient type.");
		return -EBADTYPE;
	}

	if (BITMAP_CHUNK_SIZE - BITMAP_CHUNK_BLOCKS + sizeof(block_bm_t)
	    < sizeof(unsigned long *)) {
		WARN(1, "Bitmap chunk size can't support atomic bitmap ops.");
		return -ERANGE;
	}

#if HAVE_BDEV_NR_SECTORS
	volume->nsectors = bdev_nr_sectors(volume->bdev);
#else
	volume->nsectors = volume->bdev->bd_part->nr_sects;
#endif

	volume->bitmap.last_block = last_bitmap_block_addr(volume->start_sector, volume->nsectors);

	bitmap_nblocks = ((volume->bitmap.last_block - volume->start_block)
			  >> PAGE_SECTORS_SHIFT) + 1;

	volume->bitmap.nchunks = div_u64_rem(bitmap_nblocks,
					     BITMAP_CHUNK_BLOCKS,
					     &reminder);
	volume->bitmap.nchunks += reminder ? 1 : 0;

	alloc_size = volume->bitmap.nchunks * sizeof(struct bitmap_chunk);

	volume->bitmap.vmalloc = false;

	volume->bitmap.chunks = kzalloc(alloc_size, GFP_KERNEL | __GFP_NOWARN);
	if (!volume->bitmap.chunks) {
		volume->bitmap.chunks = vzalloc(alloc_size);
		if (volume->bitmap.chunks) {
			volume->bitmap.vmalloc = true;
		} else {
			WARN_VOLUME(volume, 1,
				    "Failed to allocate bitmap both with kmalloc and vmalloc [%llu B]",
				    (u64) alloc_size);
			return -ENOMEM;
		}
	}

	for (i = 0; i < volume->bitmap.nchunks; i++)
		init_rwsem(&volume->bitmap.chunks[i].lock_alloc);

	ret = add_start_block(volume);
	if (ret)
		goto fail;

	return 0;
fail:
	blksnap_blk_bitmap_free(volume);
	return ret;
}

/* filter write must be disabled */
void blksnap_blk_bitmap_free(struct blksnap_volume *volume)
{
	free_bitmap_chunks(volume);

	if (volume->bitmap.vmalloc)
		vfree(volume->bitmap.chunks);
	else
		kfree(volume->bitmap.chunks);

	volume->bitmap.chunks = NULL;
}

int blksnap_blk_bitmap_reset(struct blksnap_volume *volume)
{
	int ret;

	free_bitmap_chunks(volume);
	ret = add_start_block(volume);

	return ret;
}

void blksnap_blk_bitmap_apply_start_snapshot(struct blksnap_volume *volume)
{
	int rc = 0;
	struct blksnap_bitmap_iter iter;

	rc = get_first_changed_block(volume, &iter, 0);
	while (!rc) {
		if (test_bit(NEXT_SNAP, iter.block_lu)) {
			clear_bit(NEXT_SNAP, iter.block_lu);
			set_bit(THIS_SNAP, iter.block_lu);
		}
		rc = get_next_changed_block(volume, &iter);
	}
}

/* Write all active non-committed blocks to disk and mark them committed.
 */
int blksnap_blk_bitmap_write_changelog(struct blksnap_volume *volume)
{
	int rc, ret = 0;
	sector_t addr;
	struct blksnap_bitmap_iter iter;

	if (unlikely(!blksnap_cl_store_valid(volume->clstore)))
		return -EINVAL;

	/* Reset statistic of write operation */
	blksnap_cl_store_stat_reset(volume->clstore);

	rc = get_first_changed_block(volume, &iter, 0);
	while (!rc) {
		/* changelog store stores absolute addresses in sectors */
		addr = __index_to_abs_addr(volume, iter.nchunk, iter.nblock);

		ret = __cl_store_commit_block(volume, addr, &iter);
		if (ret) {
			up_read(&iter.chunk->lock_alloc);
			break;
		}
		rc = get_next_changed_block(volume, &iter);
	}

	blksnap_cl_store_stat_show(volume->clstore);

	return ret;
}

/* Print count of changed blocks. Print status of each block if changed blocks
 * count is fewer than the corresponding limit.
 */
void blksnap_show_changelog(struct blksnap_volume *volume)
{
	int rc;
	bool has_data;
	sector_t count = 0, addr;
	struct blksnap_bitmap_iter iter;
	struct blksnap_cow_block *cow_node;

	rc = get_first_changed_block(volume, &iter, 0);
	while (!rc) {
		/* THIS_SNAP or NEXT_SNAP is set */
		count++;
		rc = get_next_changed_block(volume, &iter);
	}

	bsmsg_volume(volume, "There are %lu changed blocks", count);

	if (count > CHANGELOG_SHOW_LIMIT)
		return;

	down_read(&volume->tree_rw_lock);

	rc = get_first_changed_block(volume, &iter, 0);
	while (!rc) {
		addr = __index_to_abs_addr(volume, iter.nchunk, iter.nblock);

		if (test_bit(COW_DATA, iter.block_lu)) {
			cow_node = __cow_tree_search(&volume->cow_tree, addr);
			if (!cow_node) {
				WARN(1, "blk %llu has COW_DATA flag, but doesn't have corresponding CoW node [%x]",
				     (u64) addr, *iter.block);
				BUG();
			}
			has_data = blksnap_data_cow_has_data(cow_node);
		} else {
			has_data = false;
		}

		printk(KERN_INFO "%s%llu%s",
		       test_bit(THIS_SNAP, iter.block_lu) ? "!" : "",
		       (unsigned long long) addr,
		       has_data ? "- " : " ");

		rc = get_next_changed_block(volume, &iter);
	}

	up_read(&volume->tree_rw_lock);
}

/* The block's chunk must be locked on reading outside.
 */
static bool __block_is_in_changelog(struct blksnap_volume *volume,
				    const unsigned long *block)
{
	/* The normal case, this block is marked as being a member of the
	 * current in-progress snapshot.
	 */
	if (test_bit(THIS_SNAP, block) && volume->disk)
		return true;

	/* A changelog is only allowed without a snapshot if the device is not
	 * currently mounted.  This facilites restoring a volume to a previous
	 * point in time and starting with an incremental snapshot.
	 */
	if (test_bit(NEXT_SNAP, block) && !volume->disk)
		return true;

	return false;
}

/* Return: count of 4K-sized blocks.
 */
sector_t blksnap_get_changelog_size(struct blksnap_volume *volume)
{
	int rc;
	struct blksnap_bitmap_iter iter;
	sector_t total = 0;

	rc = get_first_changed_block(volume, &iter, 0);
	while (!rc) {
		if (__block_is_in_changelog(volume, iter.block_lu))
			total++;
		rc = get_next_changed_block(volume, &iter);
	}

	return total;
}

/* offset: start block offset, in sectors from device start.
 */
u32 blksnap_get_changelog_extents(struct blksnap_volume *volume,
				  struct blksnap_ioctl_disk_extent extents[],
				  u32 nextents,
				  sector_t offset)
{
	int rc;
	sector_t addr;	/* in sectors */
	struct blksnap_bitmap_iter iter;
	const sector_t volume_start = volume->start_block;
	sector_t prev_block = 0;
	sector_t curr_block;
	bool have_data = false;
	u32 extent_index = 0;

	rc = get_first_changed_block(volume, &iter, offset);
	while (!rc) {
		if (!__block_is_in_changelog(volume, iter.block_lu))
			goto next_block;

		/* absolute block addr, in sectors */
		addr = __index_to_abs_addr(volume, iter.nchunk, iter.nblock);
		/* relative block addr */
		curr_block = addr - volume_start;

		/* Is it the first block selected to be added in changelog? */
		if (!have_data) {
			prev_block = curr_block;
			extents[extent_index].nblocks = 1;
			extents[extent_index].start_sect = prev_block;
			have_data = true;
			goto next_block;
		}

		/* Check if the continuous extent doesn't break with current
		 * block.
		 */
		if (curr_block != prev_block + PAGE_SECTORS) {
			if (extent_index >= nextents - 1) {
				up_read(&iter.chunk->lock_alloc);
				break;
			}
			extent_index++;
			extents[extent_index].nblocks = 1;
			extents[extent_index].start_sect = curr_block;
		} else {
			extents[extent_index].nblocks++;
		}
		prev_block = curr_block;
next_block:
		rc = get_next_changed_block(volume, &iter);
	}

	/* nextents = extent_index + 1 (as index starts from 0)
	 * BUT if there is NO extents to return, nextents MUST be 0.
	 */
	return extent_index + ((have_data == true) ? 1 : 0);
}

#ifdef BLKSNAP_DEBUG_CLSTORE
void block_flags_sanity_check(sector_t addr, const struct blksnap_bitmap_iter *index)
{
	if (unlikely(!(test_bit(THIS_SNAP, index->block_lu)
		       || test_bit(NEXT_SNAP, index->block_lu)))) {
		bsmsg("BUG: %s: block [%lu] flags: %.8x", __func__, addr,
							  index->block);
		BUG();
	}
}
#else
void block_flags_sanity_check(sector_t addr, const struct blksnap_bitmap_iter *index)
{
}
#endif

/* External locking is required */
/* 1) if (cancel): mark all this-snap blocks as next-snap blocks
 *    else:        remove all this-snap blocks from changelog
 * 2) write all changed blocks to .blksnap/changelog
 * Bitmap traversal is very expensive operation, so cycle clause must be
 * be optimized.
 */
int blksnap_blk_bitmap_apply_stop_snapshot(struct blksnap_volume *volume, bool cancel)
{
	int ret = 0, rc;
	sector_t addr;
	struct blksnap_bitmap_iter iter;
	unsigned int remove = 0, remain = 0;

	if (unlikely(!blksnap_cl_store_valid(volume->clstore)))
		return -EINVAL;

	/* Reset statistic of write operation */
	blksnap_cl_store_stat_reset(volume->clstore);

	rc = get_first_changed_block(volume, &iter, 0);
	while (!rc) {
		/* absolute block addr, in sectors */
		addr = __index_to_abs_addr(volume, iter.nchunk, iter.nblock);

		block_flags_sanity_check(addr, &iter);

		/* assign this-snap block for the next snapshot or remove it */
		if (test_bit(NEXT_SNAP, iter.block_lu)) {
			clear_bit(THIS_SNAP, iter.block_lu);
			remain++;
		} else {
			if (test_bit(THIS_SNAP, iter.block_lu)) {
				if (cancel) {
					/* THIS_SNAP -> NEXT_SNAP */
					set_bit(NEXT_SNAP, iter.block_lu);
					clear_bit(THIS_SNAP, iter.block_lu);
					remain++;
				} else {
					/* THIS_SNAP -> 0 */
					__block_reset(volume, &iter);
					remove++;
					goto next_block;
				}
			} else {
				WARN_VOLUME(volume, 1, "block %llu (%x) quick check failed on stopping snapshot",
					    (u64) addr, *iter.block);
				ret = 1;
				up_read(&iter.chunk->lock_alloc);
				break;
			}
		}

		if (unlikely(test_bit(THIS_SNAP, iter.block_lu)
			     || !test_bit(NEXT_SNAP, iter.block_lu))) {
			WARN(1, "Block [%llu] has invalid flags [%x] after clearing bitmap on stopping snapshot",
			     (u64) addr, *iter.block);
			ret = 1;
			up_read(&iter.chunk->lock_alloc);
			break;
		}

		/* commit block to .blksnap/changelog */
		ret = __cl_store_commit_block(volume, addr, &iter);
		if (ret) {
			up_read(&iter.chunk->lock_alloc);
			break;
		}
next_block:
		rc = get_next_changed_block(volume, &iter);
	}

	if (!ret) {
		ret = add_start_block(volume);

		/* Show block statistic of write operation */
		blksnap_cl_store_stat_show(volume->clstore);
	}

	if (!ret)
		bsmsg_volume(volume, "Bitmap prune: removed %u blocks (%u remained)",
			     remove, remain);
	else
		bsmsg_volume(volume, "WARNING: Bitmap prune is stopped due errors");

	return ret;
}

//static void __debug_print_bitmap(struct blksnap_volume *volume)
//{
//	sector_t i;

//	if (!volume || !volume->bdev) {
//		WARN_VOLUME(volume, 1, "%s: volume structure is incomplete", __func__);
//		return;
//	}

//	if (!volume->bitmap.chunks) {
//		WARN_VOLUME(volume, 1, "%s: bitmap is not allocated", __func__);
//		return;
//	}

//	bsmsg("%s: bitmap.nchunks = %llu", __func__, volume->bitmap.nchunks);
//	bsmsg("%s: bitmap alloc type: %s", __func__, volume->bitmap.vmalloc
//						     ? "vmalloc" : "kmalloc");
//	for (i = 0; i < volume->bitmap.nchunks; i++) {
//		struct bitmap_chunk *chunk = &volume->bitmap.chunks[i];
//		down_read(&chunk->lock_alloc);
//		if (chunk->changed_blocks) {
//			WARN_ON(!chunk->blocks);
//			bsmsg("volume->bitmap.chunks[%llu].changed_blocks = %llu",
//			      i, chunk->changed_blocks);
//		}
//		up_read(&chunk->lock_alloc);
//	}
//}
