#include <linux/types.h>
#include <linux/bitops.h>
#include <linux/math64.h>
#include <linux/rwsem.h>
#include "blksnap-volume.h"
#include "blksnap-device.h"
#include "blksnap-bitmap.h"
#include "blksnap-block.h"

#define BUG_ON_INVALID_ADDR(volume, addr)				       \
	if (unlikely(addr < volume->start_block				       \
		     || addr > volume->bitmap.last_block)) {		       \
		WARN_VOLUME(volume, 1, "invalid sector address %llu in  %s()", \
			    (u64) addr, __func__);			       \
		BUG();							       \
	}

static int alloc_bitmap_chunk(struct bitmap_chunk *bitmap_chunk)
{
	block_bm_t *new;

	if (bitmap_chunk->blocks)
		return 0;

	/* kzalloc is able to cause preemption.
	 * It'd be fatal for locked spinlock.
	 */
	new = kzalloc(BITMAP_CHUNK_SIZE, GFP_KERNEL);
	if (!new) {
		WARN_ONCE(1, "Can't allocate bitmap chunk");
		return -ENOMEM;
	}

	down_write(&bitmap_chunk->lock_alloc);

	/* chunk could've been allocated while spinlock was unlocked */
	if (bitmap_chunk->blocks) {
		up_write(&bitmap_chunk->lock_alloc);
		kfree(new);
		return 0;
	}

	WARN_ON(bitmap_chunk->changed_blocks);

	bitmap_chunk->blocks = new;
	up_write(&bitmap_chunk->lock_alloc);

	return 0;
}

/* Free the current chunk if it hasn't any flags */
static void free_bitmap_chunk(struct blksnap_bitmap_iter *index)
{
	down_write(&index->chunk->lock_alloc);

	if (index->chunk->changed_blocks)
		goto unlock;

	kfree(index->chunk->blocks);
	index->chunk->blocks = NULL;
unlock:
	up_write(&index->chunk->lock_alloc);
}

/* Convert absolute (from device start) page-aligned sector offset
 * to bitmap index. index->block can be assigned only under locking.
 */
__always_inline
void abs_addr_to_index(struct blksnap_volume *volume, sector_t abs_sectors,
		       struct blksnap_bitmap_iter *index)
{
	sector_t relative_block = (abs_sectors - volume->start_block)
				  >> PAGE_SECTORS_SHIFT;

	/* ordinary '/', '%' operations with u64 operand don't work on i386 */
	index->nchunk = div_u64_rem(relative_block, BITMAP_CHUNK_BLOCKS,
				    &index->nblock);
	index->chunk = &volume->bitmap.chunks[index->nchunk];
}

/* @addr: page-aligned block offset in sectors, from device start
 * Return: true if bit is set.
 */
int block_test_bit(struct blksnap_volume *volume, sector_t addr, u8 bit)
{
	int ret;
	struct blksnap_bitmap_iter index;

	BUG_ON_INVALID_ADDR(volume, addr);

	abs_addr_to_index(volume, addr, &index);

	down_read(&index.chunk->lock_alloc);
	if (!index.chunk->changed_blocks)
		ret = 0;
	else
		ret = test_bit(bit,
			       (unsigned long *) &index.chunk->blocks[index.nblock]);
	up_read(&index.chunk->lock_alloc);

	return ret;
}

/* @addr: page-aligned block offset in sectors, from device start
 * Return: error code.
 */
int block_set_bit(struct blksnap_volume *volume, sector_t addr, u8 bit)
{
	int ret;
	struct blksnap_bitmap_iter index;

	BUG_ON_INVALID_ADDR(volume, addr);

	abs_addr_to_index(volume, addr, &index);

	down_read(&index.chunk->lock_alloc);
	while (!index.chunk->blocks) {
		up_read(&index.chunk->lock_alloc);
		ret = alloc_bitmap_chunk(index.chunk);
		if (ret)
			return ret;

		down_read(&index.chunk->lock_alloc);
	}

	index.block = &index.chunk->blocks[index.nblock];

	if (!test_and_set_bit(CHANGED, index.block_lu))
		index.chunk->changed_blocks++;

	set_bit(bit, index.block_lu);
	up_read(&index.chunk->lock_alloc);

	return 0;
}

/* Restrictions:
 * 1) index->chunk must be locked on reading
 * 2) index->chunk->blocks must be allocated
 * 3) index->chunk->block must be assigned
 * @addr: page-aligned block offset in sectors, from device start
 * Return: true if bit is set at index->chunk->block
 */
__always_inline
bool __block_test_and_set_bit(struct blksnap_volume *volume,
			      struct blksnap_bitmap_iter *index,
			      u8 bit)
{
	if (!test_and_set_bit(CHANGED, index->block_lu))
		index->chunk->changed_blocks++;

	return test_and_set_bit(bit, index->block_lu);
}

/* @addr: page-aligned block offset in sectors, from device start
 * Return: error code.
 */
int block_test_and_set_bit(struct blksnap_volume *volume,
			   sector_t addr, u8 bit, bool *res)
{
	int ret;
	struct blksnap_bitmap_iter index;

	BUG_ON_INVALID_ADDR(volume, addr);

	abs_addr_to_index(volume, addr, &index);

	down_read(&index.chunk->lock_alloc);
	while (!index.chunk->blocks) {
		up_read(&index.chunk->lock_alloc);
		ret = alloc_bitmap_chunk(index.chunk);
		if (ret)
			return ret;

		down_read(&index.chunk->lock_alloc);
	}

	index.block = &index.chunk->blocks[index.nblock];

	*res = __block_test_and_set_bit(volume, &index, bit);

	up_read(&index.chunk->lock_alloc);

	return 0;
}

/* @addr: page-aligned block offset in sectors, from device start
 * Return: true if block has at least one raised flag.
 */
bool block_is_active(struct blksnap_volume *volume, sector_t addr)
{
	struct blksnap_bitmap_iter index;
	bool ret;

	BUG_ON_INVALID_ADDR(volume, addr);

	abs_addr_to_index(volume, addr, &index);

	down_read(&index.chunk->lock_alloc);

	if (!index.chunk->changed_blocks)
		ret = false;
	else
		ret = !!index.chunk->blocks[index.nblock];

	up_read(&index.chunk->lock_alloc);

	return ret;
}

/* @addr: page-aligned block offset in sectors, from device start */
void block_clear_bit(struct blksnap_volume *volume, sector_t addr, u8 bit)
{
	struct blksnap_bitmap_iter index;

	BUG_ON_INVALID_ADDR(volume, addr);

	abs_addr_to_index(volume, addr, &index);

	down_read(&index.chunk->lock_alloc);

	/* chunk isn't allocated */
	if (!index.chunk->changed_blocks)
		goto unlock;

	index.block = &index.chunk->blocks[index.nblock];

	/* bit is already cleared */
	if (!(*index.block & (1 << bit)))
		goto unlock;

	/* block certainly bears the flag */
	clear_bit(bit, index.block_lu);
	if (!(*index.block & BLOCK_STATE_MASK)
	    && test_and_clear_bit(CHANGED, index.block_lu)) {
		index.chunk->changed_blocks--;
		if (!index.chunk->changed_blocks) {
			up_read(&index.chunk->lock_alloc);
			free_bitmap_chunk(&index);
			return;
		}
	}
unlock:
	up_read(&index.chunk->lock_alloc);
}

/* Restrictions:
 * 1) index->chunk must be locked on reading
 * 2) index->chunk->blocks must be allocated
 * 3) index->chunk->block must be assigned
 */
void __block_reset(struct blksnap_volume *volume, struct blksnap_bitmap_iter *index)
{
	if (!test_and_clear_bit(CHANGED, index->block_lu))
		return;

	*index->block = 0;

	index->chunk->changed_blocks--;

	if (!index->chunk->changed_blocks) {
		up_read(&index->chunk->lock_alloc);
		free_bitmap_chunk(index);
		down_read(&index->chunk->lock_alloc);
	}
}
