#include <linux/kernel.h>
#include <linux/err.h>
#include <linux/types.h>
#include <linux/blkdev.h>
#include <asm/string.h>

#include "blksnap.h"
#include "blksnap-bio-queue-share.h"
#include "blksnap-hash.h"
#include "blksnap-device.h"
#include "blksnap-compat.h"
#include "blksnap-shared.h"
#include "blksnap-cl-store.h"
#include "blksnap-cl-buffer.h"

#define INITIAL_PAGE_COUNT	2
#define CL_PAGE_BUFFER_BYTES	PAGE_SIZE
#define CL_PAGE_BUFFER_CAPACITY	(CL_PAGE_BUFFER_BYTES / sizeof(sector_t))

struct blksnap_cl_page {
	sector_t buf[CL_PAGE_BUFFER_CAPACITY];

	/* Buffer position */
	unsigned int position;

	/* Pointer to the next element */
	struct blksnap_cl_page *next, *prev;
};

struct blksnap_cl_buffer {
	/* List of page pointers */
	struct blksnap_cl_page *curr_page;

	/* Callback to send full page signal */
	flush_page_func flush_callback;

	/* Callback to write passed page */
	writer_func write_callback;

	/* Statistic information */
	struct blksnap_cl_stat statistics;

	/* Mutex to protect current page pointer */
	struct mutex page_lock;

	/* User data for flush and writer callback */
	void *flush_fn_data, *writer_fn_data;
};

static int cl_page_flush_call(struct blksnap_cl_buffer *buffer,
			      struct blksnap_cl_page *page)
{
	int res = 0;

	/* Call flush callback */
	res = buffer->flush_callback(page, buffer->flush_fn_data);
	if (res) {
		bsmsg("ERROR: %s got error %d in flush callback",
		      __func__, res);
		buffer->statistics.error_count++;
	}

	return res;
}

static inline bool cl_page_is_full(struct blksnap_cl_page *page)
{
	if (warn_null_pointer(page, __func__))
		return true;
	return page->position == CL_PAGE_BUFFER_CAPACITY;
}

static int cl_page_add(struct blksnap_cl_buffer *buffer)
{
	/* Use 3 variables to make inserting page more clear, no other
	 * reasons to do this
	 */
	struct blksnap_cl_page *page, *next_page, *prev_page;

	page = kzalloc(sizeof(*page), GFP_KERNEL);
	if (!page) {
		bsmsg("ERROR: %s: Failed to allocate page", __func__);
		return -ENOMEM;
	}

	/* Updating stat information */
	buffer->statistics.allocated_pages++;

	if (buffer->curr_page) {
		/* Add new page to list */
		next_page = buffer->curr_page->next;
		prev_page = buffer->curr_page;

		page->prev = prev_page;
		page->next = next_page;
		next_page->prev = page;
		prev_page->next = page;
	} else {
		/* Add first page */
		buffer->curr_page = page;
		buffer->curr_page->next = page;
		buffer->curr_page->prev = page;
	}

	return 0;
}

static int cl_page_remove(struct blksnap_cl_buffer *buffer,
			  struct blksnap_cl_page *page)
{
	int res = 0;

	/* Is it not last page? */
	if (page->next != page) {
		/* Remove page from buffer */
		page->prev->next = page->next;
		page->next->prev = page->prev;

		/* Before removing curr_page - change to the next element */
		if (page == buffer->curr_page)
			buffer->curr_page = page->next;
	} else {
		if (buffer->curr_page != page) {
			bsmsg("Wrong page address to remove");
			return -EINVAL;
		}

		/* Set page pointer in buffer to NULL. It was a last page */
		buffer->curr_page = NULL;
	}

	/* Call flush function if page is not empty and not full */
	if (page->position && !cl_page_is_full(page))
		res = blksnap_cl_page_flush(buffer, page);
	kfree(page);
	buffer->statistics.allocated_pages--;

	return res;
}

struct blksnap_cl_buffer *blksnap_cl_buffer_init(
					 flush_page_func flush_callback,
					 void *flush_page_data,
					 writer_func write_callback,
					 void *writer_data)
{
	struct blksnap_cl_buffer *buffer;
	int i, err = 0;

	/* Allocating buffer */
	buffer = kzalloc(sizeof(*buffer), GFP_KERNEL);
	if (!buffer) {
		bsmsg("ERROR: %s: Can not allocate buffer memory", __func__);
		return ERR_PTR(-ENOMEM);
	}

	/* Init page locking */
	mutex_init(&buffer->page_lock);

	buffer->flush_callback = flush_callback;
	buffer->write_callback = write_callback;

	/* Save user data for flush and writer cb */
	buffer->flush_fn_data = flush_page_data;
	buffer->writer_fn_data = writer_data;

	/* Lock buffer while adding pages */
	mutex_lock(&buffer->page_lock);

	/* Add initial pages */
	for (i = 0; i < INITIAL_PAGE_COUNT; i++) {
		err = cl_page_add(buffer);
		if (err)
			goto destroy_buffer;
	}

	mutex_unlock(&buffer->page_lock);

	return buffer;

destroy_buffer:
	mutex_unlock(&buffer->page_lock);
	if (blksnap_cl_buffer_destroy(buffer))
		bsmsg("ERROR: %s: Buffer destroy failed", __func__);

	return ERR_PTR(err);
}

int blksnap_cl_buffer_destroy(struct blksnap_cl_buffer *buffer)
{
	int res = 0;

	/* Lock buffer before removing pages */
	mutex_lock(&buffer->page_lock);

	/* Destroy all pages */
	while (buffer->curr_page && !res)
		res = cl_page_remove(buffer, buffer->curr_page->next);

	/* Unlock buffer */
	mutex_unlock(&buffer->page_lock);

	kfree(buffer);

	return res;
}

/* @addr - block offset, in sectors from device start */
int blksnap_cl_buffer_add_block(struct blksnap_cl_buffer *buffer,
				sector_t addr)
{
	int res = 0;

	/* Lock buffer before working with current_page */
	mutex_lock(&buffer->page_lock);

	if (unlikely(!buffer->curr_page)) {
		bsmsg("No page available in cycle buffer");
		mutex_unlock(&buffer->page_lock);
		return -EINVAL;
	}

	/* Add block to current position */
	buffer->curr_page->buf[buffer->curr_page->position] = addr;
	buffer->curr_page->position++;

	/* Update statistics */
	buffer->statistics.saved_block_count++;

	/* if current page is full */
	if (unlikely(cl_page_is_full(buffer->curr_page))) {
		struct blksnap_cl_page *page_to_flush = NULL;

		/* We need to add another page if next is also full */
		if (cl_page_is_full(buffer->curr_page->next)) {
			/* We got buffer overflow. Add new page */
			res = cl_page_add(buffer);
			if (res) {
				bsmsg("ERROR: %s: Unable to add new page [%d]",
				      __func__, res);
				buffer->statistics.error_count++;
			}

			/* Update statistics */
			buffer->statistics.overflow_count++;
		}
		page_to_flush = buffer->curr_page;

		/* Go to next page */
		buffer->curr_page = buffer->curr_page->next;

		res = cl_page_flush_call(buffer, page_to_flush);
		if (res) {
			bsmsg("ERROR: %s: Unable to flush buffer page [%d]",
			      __func__, res);
			buffer->statistics.error_count++;
		}
	}

	mutex_unlock(&buffer->page_lock);

	return res;
}

int blksnap_cl_get_stat(struct blksnap_cl_buffer *buffer,
			struct blksnap_cl_stat *stats)
{
	/* Just copy current stats to user defined memory */
	memcpy(stats, &buffer->statistics, sizeof(buffer->statistics));

	return 0;
}

static int clean_cl_page(struct blksnap_cl_page *page)
{
	/* Set position pointer to 0 */
	page->position = 0;

	/* Clean buffer */
	memset((unsigned char *) page->buf, 0, CL_PAGE_BUFFER_BYTES);

	return 0;
}

int blksnap_cl_page_flush(struct blksnap_cl_buffer *buffer,
		  struct blksnap_cl_page *page)
{
	int res = 0;

	/* Calling flush callback */
	res = buffer->write_callback((sector_t *) page->buf,
				     CL_PAGE_BUFFER_BYTES,
				     buffer->writer_fn_data);
	if (res) {
		bsmsg("ERROR: %s: Unable to write changes [%d]", __func__, res);
		buffer->statistics.error_count++;
		return res;
	}

	/* Update statistic information */
	buffer->statistics.pages_written++;

	cl_store_usage_add_flushed_blocks(buffer->writer_fn_data,
					  page->position * sizeof(sector_t));

	res = clean_cl_page(page);

	return res;
}

