/* {{{1 Preamble */
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/blkdev.h>
#include <linux/vmalloc.h>

#include "blksnap.h"
#include "blksnap-bio-queue-share.h"
#include "blksnap-hash.h"
#include "blksnap-device.h"
#include "blksnap-compat.h"
#include "blksnap-shared.h"
#include "blksnap-extent.h"
#include "blksnap-device-utils.h"
#include "blksnap-volume.h"
#include "blksnap-data.h"
#include "blksnap-cow-tree.h"
#include "blksnap-bio.h"
#include "blksnap-bio-submit.h"
#include "blksnap-bitmap.h"
#include "blksnap-block.h"

#include "blksnap-cl-buffer.h"
#include "blksnap-cl-thread.h"
#include "blksnap-cl-store.h"


/* amount of sector_t objects in given sectors */
#define SECTOR_BLOCKS(sectors) ((sectors << SECTOR_SHIFT) / sizeof(sector_t))

/* {{{1 Stat and show functions */

inline void cl_store_stat_inc_blocks(struct blksnap_cl_store *store)
{
	store->stat.blocks++;
}

inline void cl_store_stat_inc_added(struct blksnap_cl_store *store)
{
	store->stat.added++;
}

void cl_store_usage_add_flushed_blocks(struct blksnap_cl_store *store,
				       uint64_t bytes)
{
	/* if it's the last flush - exclude 0-block */
	if (bytes % PAGE_SIZE)
		bytes -= sizeof(sector_t);

	/* if it's the first flush after reading on-disk store - exclude
	 * uncommited blocks.
	 */
	store->usage = (store->usage % PAGE_SIZE) ?
		       (store->usage & ~(PAGE_SIZE - 1)) + bytes :
		       store->usage + bytes;
}

/* {{{1 Callback functions */
/* mark clstore as invalid and try to reset clstore */
void invalidate_cl_store(void *data)
{
	struct blksnap_cl_store *store = (struct blksnap_cl_store *) data;
	int res;

	if (unlikely(!blksnap_cl_store_valid(store)))
		return;

	res = cl_store_reset(store);
	if (res)
		bsmsg_bdev(store->bdev,
			   "ERROR: resetting cl store failed with error code: %d",
			   res);

	atomic_set(&store->status, CL_STORE_INVALID);
}

static int cl_store_flush(sector_t *page_buffer, unsigned int len,
			  void *data)
{
	struct blksnap_cl_store *store = (struct blksnap_cl_store *) data;
	int ret = -EINVAL;
	sector_t todo_sec;

	DCLS("%s(%p, %u, %p)", __func__, page_buffer, len, store);

	if (!store)
		return -EINVAL;

	if (cl_store_status(store) != CL_STORE_MAPPED &&
	    cl_store_status(store) != CL_STORE_SHUTDOWN)
		return -ECLSNOMAP;

	if (len % SECTOR_SIZE) {
		bsmsg_bdev(store->bdev,
			   "ERROR: length of buffer passed to %s() MUST be aligned to %d bytes",
			   __func__, SECTOR_SIZE);
		return -EINVAL;
	}

	todo_sec = len >> SECTOR_SHIFT;

	while (todo_sec) {
		unsigned int iter_sec;
		sector_t relative_start;

		/* Try to use a minimum value of:
		 * - sectors remain till the end of the current extent
		 * - sectors remain to write
		 */
		iter_sec = MIN(store->cur_pos.ext->nsectors -
				   store->cur_pos.pos,
			       todo_sec);

		/* iter_sec == 0 means the current extent is full so try to get
		 * next extent
		 */
		if (!iter_sec) {
			store->cur_pos.pos = 0;
			store->cur_pos.ext = store->cur_pos.ext->next;
			if (unlikely(!store->cur_pos.ext)) {
				/* Store is full */
				return -EIO;
			}
			/* Restart process with next extent */
			continue;
		}

		relative_start = volume_relative_address(
				store->bdev,
				store->cur_pos.ext->start);

		DCLS("%s(%p, %u, %p) blksnap_disk_io(%llu, %llu)",
		     __func__, page_buffer, len, store,
		     relative_start + store->cur_pos.pos,
		     iter_sec << SECTOR_SHIFT);

		ret = blksnap_disk_io(store->bdev,
			      relative_start + store->cur_pos.pos,
			      iter_sec << SECTOR_SHIFT,
			      (unsigned char *) page_buffer,
			      BLKSNAP_WRITE_FILTER_DISABLED);

		if (unlikely(ret))
			break;

		todo_sec -= iter_sec;
		page_buffer += iter_sec << SECTOR_SHIFT;
		store->cur_pos.pos += iter_sec;
	}

	return ret;
}

/* {{{1 Internal routines */

/* Count valid blocks in given buffer
 * Buffer is terminated with 0 value
 */
static int blocks_in_buffer(sector_t *buf, int max_items)
{
	sector_t *pos = buf;

	for (; (pos - buf < max_items) && *pos; pos++)
		;

	return pos - buf;
}

/* Import block numbers from buffer and insert blocks to the block tree
 * max_items is in sector_t items
 */
static int cl_store_import_blocks(struct block_device *bdev, sector_t *buf,
				  int max_items)
{
	sector_t sector, *pos = buf;
	struct blksnap_volume *volume;
	bool commit = true;
	int blocks;
	int err;

	volume = lookup_volume_by_bdev(bdev);
	if (!volume) {
		bsmsg_bdev(bdev, "ERROR: no blksnap volume for block device");
		return -EINVAL;
	}

	blocks = blocks_in_buffer(buf, max_items);
	if (blocks < max_items)
		commit = false;

	while (pos - buf < blocks) {
		sector = *pos++;
		err = block_set_bit(volume, sector, NEXT_SNAP);
		if (unlikely(err))
			return err;
		if (commit)
			/* No need to check ret value for the same addr twice */
			block_set_bit(volume, sector, COMMITTED);
	}

	/* Return number of blocks processed */
	return pos - buf;
}

static int cl_store_read(struct blksnap_cl_store *store)
{
	int blocks, ret;
	sector_t *buf;
	uint64_t todo_sec, cur_ext_pos;
	struct blksnap_extent *cur_extent;
	const unsigned int buf_size_sec = BIO_MAX_SIZE >> SECTOR_SHIFT;

	buf = vmalloc(buf_size_sec << SECTOR_SHIFT);
	if (!buf) {
		bsmsg("ERROR: Unable to allocate memory in %s", __func__);
		return -ENOMEM;
	}

	/* Initialize current position in store to 1st extent and 0 sector */
	store->cur_pos.ext = store->extents.head;
	store->cur_pos.pos = 0;

	cur_extent = store->extents.head;
	do {
		cur_ext_pos = 0;
		while (cur_extent->nsectors > cur_ext_pos) {
			todo_sec = MIN(cur_extent->nsectors - cur_ext_pos,
				       buf_size_sec);

			ret = blksnap_disk_io(
					store->bdev,
					volume_relative_address(
						store->bdev,
						cur_extent->start) +
					cur_ext_pos,
					todo_sec << SECTOR_SHIFT,
					(unsigned char *) buf,
					BLKSNAP_READ);
			if (ret)
				goto end;
			blocks = cl_store_import_blocks(store->bdev, buf,
					SECTOR_BLOCKS(todo_sec));
			if (blocks < 0) {
				/* error occured */
				ret = blocks;
				goto end;
			}

			store->usage += blocks * sizeof(sector_t);

			if (blocks < SECTOR_BLOCKS(todo_sec))
				goto end;

			cur_ext_pos += todo_sec;
		}
		cur_extent = cur_extent->next;
	} while (cur_extent);

	if (!cur_extent) {
		bsmsg_bdev(store->bdev, "ERROR: changelog store is full");
		ret = -ECLSFULL;
	}
end:
	/* Update current position in cl store only if no error occured */
	if (!ret) {
		store->cur_pos.ext = cur_extent;
		store->cur_pos.pos = cur_ext_pos;
	}
	vfree(buf);
	return ret;
}

int cl_store_reset(struct blksnap_cl_store *store)
{
	int ret;
	sector_t *buf;
	uint64_t todo_sec;

	if (warn_null_pointer(store, __func__) ||
	    warn_null_pointer(store->bdev, __func__) ||
	    warn_null_pointer(store->extents.head, __func__))
		return -EINVAL;

	/* Reset cl store, 0 tells the end of cl store data */
	buf = kzalloc(PAGE_SIZE, GFP_KERNEL);
	if (!buf) {
		bsmsg("ERROR: Unable to allocate memory in %s", __func__);
		return -ENOMEM;
	}

	/* store->extents always points to the first extent of cl store */
	todo_sec = MIN(store->extents.head->nsectors, PAGE_SECTORS);

	ret = blksnap_disk_io(
			store->bdev,
			volume_relative_address(
				store->bdev,
				store->extents.head->start),
			todo_sec << SECTOR_SHIFT,
			(unsigned char *) buf,
			BLKSNAP_WRITE_FILTER_DISABLED);

	/* Initialize current postition in store to 1st extent and 0 sector */
	store->cur_pos.ext = store->extents.head;
	store->cur_pos.pos = 0;
	store->usage = 0;

	kfree(buf);
	return ret;
}

static int blksnap_cl_store_thread_init(struct blksnap_cl_store *store)
{
	int ret = 0;

	store->opts.invalidate_cb = invalidate_cl_store;
	store->opts.invalidate_data = store;
	store->opts.writer_cb = cl_store_flush;
	store->opts.writer_cb_data = store;

	store->thread = blksnap_cl_thread_init(&store->opts, store->bdev);

	if (IS_ERR(store->thread)) {
		ret = PTR_ERR(store->thread);
		store->thread = NULL;
	}

	return ret;
}

/* {{{1 Exported routines */

int cl_store_map(struct blksnap_volume *volume,
		 struct blksnap_ioctl_map *map,
		 int operation)
{
	int ext_idx, ret = 0;
	struct blksnap_extent *new;
	struct blksnap_cl_store *store = volume->clstore;

	mutex_lock(&store->lock);

	if (atomic_read(&store->status) == CL_STORE_INVALID) {
		ret = -EINVAL;
		goto unlock;
	}

	if (is_first_map_ioctl(map)) {
		if (atomic_read(&store->status) != CL_STORE_NOSTORE) {
			ret = -EEXIST;
			goto unlock;
		}
		atomic_set(&store->status, CL_STORE_LOADING);
	}

	/* Creating cl extent list. Converting start offsets from ioctl to
	 * volume related offsets because process_page requires it.
	 */
	for (ext_idx = 0; ext_idx < map->nextents; ext_idx++) {
		new = __blksnap_extent_add(map->extents[ext_idx].start_sect,
					   map->extents[ext_idx].nsectors,
					   store->extents.head);
		if (IS_ERR(new)) {
			ret = PTR_ERR(new);
			goto free_ext;
		}
		store->extents.head = new;
		store->size += map->extents[ext_idx].nsectors << SECTOR_SHIFT;
	}

	if (is_last_map_ioctl(map)) {
		if ((operation & MAP_RESET) && (operation & MAP_READ)) {
			bsmsg_bdev(store->bdev,
				   "ERROR: empty operation set for mapping changelog store (%d)\n",
				   operation);
			ret = -EINVAL;
			goto free_ext;
		}

		/* Initialize current postition in store to
		 * the 1st extent and 0 sector
		 */
		store->cur_pos.ext = store->extents.head;
		store->cur_pos.pos = 0;

		/* Reset or read changelog store */
		if (operation & MAP_READ) {
			ret = cl_store_read(store);
		} else {
			if (operation & MAP_RESET) {
				ret = cl_store_reset(store);
			} else {
				bsmsg_bdev(store->bdev,
					   "ERROR: bad operation set for mapping changelog store (%d)\n",
					   operation);
				ret = -EINVAL;
			}
		}

		if (ret)
			goto free_ext;

		ret = blksnap_cl_store_thread_init(store);
		if (ret) {
			bsmsg_bdev(store->bdev,
				   "ERROR: Changelog store is NOT mapped [%d]",
				   ret);
			goto free_ext;
		}

		atomic_set(&store->status, CL_STORE_MAPPED);

		if (operation & MAP_WRITE)
			ret = blksnap_blk_bitmap_write_changelog(volume);

		blksnap_extent_show(store->extents.head, store->bdev, "chlog");
		bsmsg_bdev(store->bdev,
			   "Changelog store is mapped (disk: %llu/%llu) [%s %s %s]",
			   store->usage, store->size,
			   (operation & MAP_READ)  ? "rd" : "-",
			   (operation & MAP_RESET) ? "rst" : "-",
			   (operation & MAP_WRITE) ? "wr" : "-");
	}

unlock:
	mutex_unlock(&store->lock);
	return ret;
free_ext:
	atomic_set(&store->status, CL_STORE_INVALID);
	blksnap_extent_free(store->extents.head);
	store->extents.head = NULL;
	goto unlock;
}

struct blksnap_cl_store *cl_store_init(struct block_device *bdev)
{
	struct blksnap_cl_store *store;

	store = kzalloc(sizeof(struct blksnap_cl_store), GFP_KERNEL);
	if (!store) {
		bsmsg_bdev(bdev,
			   "ERROR: failed to allocate memory for changelog store");
		return ERR_PTR(-ENOMEM);
	}

	store->bdev = bdev;

	mutex_init(&store->extents.lock);
	mutex_init(&store->lock);

	return store;
}

void cl_store_destroy(struct blksnap_cl_store *store)
{
	if (!store)
		return;

	if (!blksnap_cl_store_valid(store))
		return;

	mutex_lock(&store->lock);

	atomic_set(&store->status, CL_STORE_SHUTDOWN);

	if (store->thread) {
		blksnap_cl_thread_add_block(store->thread, 0UL);
		blksnap_cl_thread_destroy(store->thread);
	}

	mutex_unlock(&store->lock);

	bsmsg_bdev(store->bdev, "Changelog store is commited (disk: %llu/%llu)",
		   store->usage, store->size);

	if (store->extents.head)
		blksnap_extent_free(store->extents.head);

	kfree(store);
}

int cl_store_status(struct blksnap_cl_store *store)
{
	if (!store)
		return CL_STORE_NOSTORE;

	return atomic_read(&store->status);
}

/* WARNING! validate chlog-store before calling this function!
 * Restrictions:
 * 1) chunk must be locked on reading
 * 2) i->chunk->blocks must be allocated
 * 3) i->chunk->block must be assigned
 * 4) blksnap_cl_store_stat_reset() must be called before the first commit
 * @addr - block offset, in sectors from device start
 */
int __cl_store_commit_block(struct blksnap_volume *volume, sector_t addr,
			    struct blksnap_bitmap_iter *index)
{
	struct blksnap_cl_store *cl_store = volume->clstore;

	/* 0L is the EOF-value of changelog file. It will be added on detach. */
	if (!addr)
		return 0;

	cl_store_stat_inc_blocks(cl_store);

	if (__block_test_and_set_bit(volume, index, COMMITTED))
		return 0;

	cl_store_stat_inc_added(cl_store);

	return blksnap_cl_thread_add_block(cl_store->thread, addr);
}

/* WARNING! validate chlog-store before calling this function!
 * @addr - block offset, in sectors from device start.
 */
int cl_store_commit_block(struct blksnap_volume *volume, sector_t addr,
			  bool stat)
{
	int ret;
	bool committed;

	struct blksnap_cl_store *cl_store = volume->clstore;
	/* 0L is the EOF-value of changelog file. It will be added on detach. */
	if (!addr)
		return 0;

	if (stat)
		cl_store_stat_inc_blocks(cl_store);

	ret = block_test_and_set_bit(volume, addr, COMMITTED, &committed);

	if (ret || committed)
		return ret;

	if (stat)
		cl_store_stat_inc_added(cl_store);

	return blksnap_cl_thread_add_block(cl_store->thread, addr);
}
