#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/spinlock.h>
#include <linux/err.h>
#include <linux/kthread.h>
#include <linux/blkdev.h>
#include <linux/slab.h>

#include "blksnap.h"
#include "blksnap-bio-queue-share.h"
#include "blksnap-hash.h"
#include "blksnap-device.h"
#include "blksnap-compat.h"
#include "blksnap-shared.h"
#include "blksnap-cl-buffer.h"
#include "blksnap-cl-thread.h"

#define CL_THREAD_NAME	"cl"

static struct kmem_cache *cmd_cache;

enum list_command {
	CL_ADD,
	CL_FLUSH
};

struct cl_command_entry {
	enum list_command command;
	void *data;
	sector_t block_addr;
	struct list_head list;
};

struct cl_thread {
	wait_queue_head_t pending_event;
	struct blksnap_cl_buffer *cl_buffer;
	struct task_struct *task;
	struct cl_thread_opts options;
	struct list_head command_list;
	spinlock_t lock;
};

int blksnap_cl_cmd_cache_create(void)
{
	if (unlikely(cmd_cache)) {
		WARN(1, "Detected try to re-create a cache, %s", __func__);
		return -EEXIST;
	}

	cmd_cache = KMEM_CACHE(cl_command_entry, 0);
	if (!cmd_cache) {
		bsmsg("ERROR: Can't create cache in %s", __func__);
		return -ENOMEM;
	}

	return 0;
}

void blksnap_cl_cmd_cache_destroy(void)
{
	if (unlikely(!cmd_cache)) {
		WARN(1, "Detected try to destroy an empty cache, %s", __func__);
		return;
	}

	kmem_cache_destroy(cmd_cache);
	cmd_cache = NULL;
}

static int command_list_empty(struct list_head *list)
{
	return list_empty(list);
}

/* Add command to list. Should be called under spinlock */
static void __push_entry(struct cl_thread *thread,
			 struct cl_command_entry *entry)
{
	list_add_tail(&entry->list, &thread->command_list);
	wake_up(&thread->pending_event);
}

/* Add command to list. Thread-safe */
static void push_entry(struct cl_thread *thread, struct cl_command_entry *entry)
{
	spin_lock_irq(&thread->lock);
	__push_entry(thread, entry);
	spin_unlock_irq(&thread->lock);
}

/* Add command to list. Irq-safe */
static void push_entry_irq(struct cl_thread *thread,
			   struct cl_command_entry *entry)
{
	unsigned long flags;

	spin_lock_irqsave(&thread->lock, flags);
	__push_entry(thread, entry);
	spin_unlock_irqrestore(&thread->lock, flags);
}

/* Get command from list. Should be called under spinlock */
static struct cl_command_entry *__pop_entry(struct cl_thread *thread)
{
	struct list_head *elem = thread->command_list.next;
	struct cl_command_entry *entry =
		list_entry(elem, struct cl_command_entry, list);

	list_del(&entry->list);

	return entry;
}

/* Get command from list. Thread-safe */
static struct cl_command_entry *pop_entry(struct cl_thread *thread)
{
	struct cl_command_entry *result;

	spin_lock_irq(&thread->lock);
	result = __pop_entry(thread);
	spin_unlock_irq(&thread->lock);

	return result;
}

static void invalidate_changelog(struct cl_thread *thread)
{
	if (thread->options.invalidate_cb)
		thread->options.invalidate_cb(thread->options.invalidate_data);
}

static int cl_thread_flush_cb(struct blksnap_cl_page *page, void *data)
{
	struct cl_thread *thread = (struct cl_thread *) data;
	struct cl_command_entry *entry;

	FALL_ON_NO_CACHE(cmd_cache, -1);

	entry = kmem_cache_alloc(cmd_cache, GFP_KERNEL);
	if (!entry) {
		bsmsg("ERROR: Unable to allocate memory in %s()", __func__);
		return -ENOMEM;
	}

	entry->command = CL_FLUSH;
	entry->data = (void *) page;
	push_entry(thread, entry);

	return 0;
}

static void process_command(struct cl_thread *thread)
{
	struct cl_command_entry *entry;
	int result = 0;

	FALL_ON_NO_CACHE(cmd_cache,);

	entry = pop_entry(thread);

	switch (entry->command) {
	case CL_ADD:
		result = blksnap_cl_buffer_add_block(thread->cl_buffer,
						     entry->block_addr);
		break;
	case CL_FLUSH:
		result = blksnap_cl_page_flush(thread->cl_buffer,
				(struct blksnap_cl_page *) entry->data);
		break;
	default:
		bsmsg("ERROR: Unknown command in %s()", __func__);
		BUG();
	}

	if (result)
		invalidate_changelog(thread);

	kmem_cache_free(cmd_cache, entry);
}

static void process_all_commands(struct cl_thread *thread)
{
	while (!command_list_empty(&thread->command_list))
		process_command(thread);
}

static int cl_thread_fn(void *data)
{
	struct cl_thread *thread = (struct cl_thread *) data;

	while (!kthread_should_stop() ||
	       !command_list_empty(&thread->command_list)) {

		wait_event_interruptible(thread->pending_event,
				!command_list_empty(&thread->command_list) ||
				kthread_should_stop());

		if (command_list_empty(&thread->command_list))
			continue;

		process_command(thread);
	}

	return 0;
}

/* Init cl processing thread */
struct cl_thread *blksnap_cl_thread_init(struct cl_thread_opts *options,
					 struct block_device *bdev)
{
	unsigned int result;
	struct cl_thread *thread;

	FALL_ON_NO_CACHE(cmd_cache, ERR_PTR(-1));

	thread = kmalloc(sizeof(*thread), GFP_KERNEL);
	if (!thread) {
		bsmsg("ERROR: Can not allocate memory in %s()", __func__);
		return ERR_PTR(-ENOMEM);
	}

	init_waitqueue_head(&thread->pending_event);

	memcpy(&thread->options, options, sizeof(*options));

	thread->cl_buffer = blksnap_cl_buffer_init(cl_thread_flush_cb,
					   (void *) thread,
					   thread->options.writer_cb,
					   thread->options.writer_cb_data);
	if (IS_ERR(thread->cl_buffer)) {
		bsmsg("ERROR: Unable to create cl buffer [%d]",
		      PTR_ERR(thread->cl_buffer));

		result = PTR_ERR(thread->cl_buffer);
		goto clean_cl_thread;
	}

	INIT_LIST_HEAD(&thread->command_list);
	spin_lock_init(&thread->lock);

	thread->task = kthread_run(cl_thread_fn, thread, "%s-%u:%u",
				   CL_THREAD_NAME,
				   MAJOR(bdev->bd_dev), MINOR(bdev->bd_dev));

	if (IS_ERR(thread->task)) {
		bsmsg("ERROR: Unable to run kthread %s", CL_THREAD_NAME);
		result = PTR_ERR(thread->task);
		goto clean_cl_buffer;
	}

	return thread;

clean_cl_buffer:
	blksnap_cl_buffer_destroy(thread->cl_buffer);
clean_cl_thread:
	kfree(thread);
	return ERR_PTR(result);
}

void blksnap_cl_thread_destroy(struct cl_thread *thread)
{
	kthread_stop(thread->task);
	process_all_commands(thread);

	/* Destroy cl buffer. This will flush all pages */
	blksnap_cl_buffer_destroy(thread->cl_buffer);

	kfree(thread);
}

/* @addr - block offset, in sectors from device start
 * addr 0 should not be placed anywhere but at the end of changelog store.
 */
int blksnap_cl_thread_add_block(struct cl_thread *thread, sector_t addr)
{
	struct cl_command_entry *entry;

	FALL_ON_NO_CACHE(cmd_cache, -1);

	entry = kmem_cache_alloc(cmd_cache, GFP_KERNEL);
	if (!entry) {
		bsmsg("ERROR: Unable to allocate memory in %s()", __func__);
		invalidate_changelog(thread);
		return -ENOMEM;
	}

	entry->command = CL_ADD;
	entry->block_addr = addr;
	push_entry_irq(thread, entry);

	return 0;
}

