#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/blkdev.h>

#include "blksnap.h"
#include "blksnap-bio-queue-share.h"
#include "blksnap-bio-queue.h"
#include "blksnap-hash.h"
#include "blksnap-device.h"
#include "blksnap-compat.h"
#include "blksnap-shared.h"
#include "blksnap-device-utils.h"
#include "blksnap-volume.h"
#include "blksnap-bitmap.h"
#include "blksnap-block.h"
#include "blksnap-data.h"
#include "blksnap-cow-tree.h"
#include "blksnap-bitmap.h"
#include "blksnap-rbtree-foreach.h"

struct blksnap_cow_block *cow_tree_search(struct blksnap_volume *volume,
					  sector_t addr)
{
	struct blksnap_cow_block *cow_blk;

	down_read(&volume->tree_rw_lock);
	cow_blk = __cow_tree_search(&volume->cow_tree, addr);
	up_read(&volume->tree_rw_lock);

	return cow_blk;
}

struct blksnap_cow_block *__cow_tree_search(struct rb_root *root, sector_t addr)
{
	struct rb_node *node = root->rb_node;

	while (node) {
		struct blksnap_cow_block *this =
			container_of(node, struct blksnap_cow_block, rb_node);

		if (addr < this->addr)
			node = node->rb_left;
		else if (addr > this->addr)
			node = node->rb_right;
		else
			return this;
	}

	return NULL;
}

/* should be called with down_write(tree_rw_lock) */
static struct blksnap_cow_block *__tree_insert(struct rb_root *root,
					       struct blksnap_cow_block *cow_blk)
{
	struct rb_node **node = &root->rb_node;
	struct rb_node *parent = NULL;

	WARN_ON(blksnap_data_cow_has_data(cow_blk));

	while (*node) {
		struct blksnap_cow_block *this =
			container_of(*node, struct blksnap_cow_block, rb_node);

		parent = *node;

		if (cow_blk->addr < this->addr) {
			node = &((*node)->rb_left);
		} else {
			if (cow_blk->addr > this->addr) {
				node = &((*node)->rb_right);
			} else {
				/* Block is already there, nothing to do */
				kfree(cow_blk);
				return this;
			}
		}
	}

	D(">>> %llu", blk->addr);

	rb_link_node(&cow_blk->rb_node, parent, node);
	rb_insert_color(&cow_blk->rb_node, root);

	return cow_blk;
}

/* @addr - page-aligned block offset, in sectors from device start */
static struct blksnap_cow_block *alloc_cow_block(sector_t addr)
{
	struct blksnap_cow_block *cow_blk;
	gfp_t gfp_mask = GFP_NOIO;

	/* kmem cache is inappropriate for cow tree */
	cow_blk = kzalloc(sizeof(*cow_blk), gfp_mask);
	if (!cow_blk)
		return NULL;

	cow_blk->addr = addr;

	/* init cow */
	cow_blk->committed = false;
	rwlock_init(&cow_blk->lock);

	return cow_blk;
}

/* @addr - page-aligned block offset, in sectors from device start
 * Allocate node and insert it in the tree or return node pointer
 * if it is already in the tree.
 */
struct blksnap_cow_block *cow_tree_insert(struct blksnap_volume *volume,
					  sector_t addr)
{
	struct blksnap_cow_block *cow_blk, *new;

	down_read(&volume->tree_rw_lock);

	/* Addr is already in the cow tree */
	if (block_test_bit(volume, addr, COW_DATA)) {
collision:
		cow_blk = __cow_tree_search(&volume->cow_tree, addr);
		up_read(&volume->tree_rw_lock);
		BUG_ON(!cow_blk);
		return cow_blk;
	}

	/* Addr isn't in the cow tree */
	new = alloc_cow_block(addr);
	if (!new) {
		bsmsg("ERROR: failed to allocate cow block");
		up_read(&volume->tree_rw_lock);
		return NULL;
	}

	up_read(&volume->tree_rw_lock);
	down_write(&volume->tree_rw_lock);

	/* The same CoW block sneaked into the tree during allocation */
	if (block_test_bit(volume, addr, COW_DATA)) {
		downgrade_write(&volume->tree_rw_lock);
		kfree(new);
		goto collision;
	}

	cow_blk = __tree_insert(&volume->cow_tree, new);

	/* addr is already in the tree, but without COW_DATA flag */
	BUG_ON(cow_blk != new);

	/* no need to check ret value, cus it was checked WHERE??? */
	block_set_bit(volume, addr, COW_DATA);

	up_write(&volume->tree_rw_lock);

	return cow_blk;
}

static bool cow_block_sanity_check_prune(struct blksnap_volume *volume,
					 struct blksnap_cow_block *cow_blk)
{
	/* If block is in flight, we can't remove it at all */
	if (WARN_ON(block_test_bit(volume, cow_blk->addr, IN_FLIGHT)))
		return false;

	if (WARN_ON(!block_test_bit(volume, cow_blk->addr, COW_DATA)))
		blksnap_data_cow_show(cow_blk);

	if (WARN_ON(blksnap_data_cow_has_data(cow_blk) &&
		!block_test_bit(volume, cow_blk->addr, NEXT_SNAP)))
		blksnap_data_cow_show(cow_blk);

	return true;
}


/* Return zero if no errors were found */
int cow_tree_empty(struct blksnap_volume *volume)
{
	int ret = 0;
	struct rb_root *root = &volume->cow_tree;
	struct rb_node *node;
	struct rb_node *prev = NULL;

	down_write(&volume->tree_rw_lock);
	while ((node = rb_first(root)) != NULL) {
		struct blksnap_cow_block *cow_blk =
			rb_entry(node, struct blksnap_cow_block, rb_node);

		if (node == prev) {
			bsmsg("%s: BUG rb_tree is broken. %p points to itself",
			      __func__, node);
			break;
		}

		if (cow_block_sanity_check_prune(volume, cow_blk)) {
			rb_erase(node, root);
			blksnap_data_cow_free(cow_blk);
			block_clear_bit(volume, cow_blk->addr, COW_DATA);
			kfree(cow_blk);
		} else {
			ret = 1;
		}

		prev = node;
	}
	up_write(&volume->tree_rw_lock);

	return ret;
}

/* For debug */
void cow_block_sanity_check(struct blksnap_volume *volume,
			    struct blksnap_cow_block *cow_blk, bool inserted)
{
	bool has_data = blksnap_data_cow_has_data(cow_blk);
	bool inflight = block_test_bit(volume, cow_blk->addr, IN_FLIGHT);
	bool nextsnap = block_test_bit(volume, cow_blk->addr, NEXT_SNAP);
	bool cow_data = block_test_bit(volume, cow_blk->addr, COW_DATA);

	if (!inserted && nextsnap && !inflight && !has_data && volume->disk) {
		bsmsg_volume(volume,
			     "WARNING: Block %llu failed general quick check%s",
			     (unsigned long long) cow_blk->addr,
			     !has_data ? " (this is probably okay)" : "");
		blksnap_data_cow_show(cow_blk);
	}

	if (has_data && inserted) {
		bsmsg_volume(volume,
			     "WARNING: Inserted block %llu has data",
			     (unsigned long long) cow_blk->addr);
		blksnap_data_cow_show(cow_blk);
	}

	if (has_data && inflight) {
		bsmsg_volume(volume,
			     "WARNING: Existing block %llu is in-flight with data",
			     (unsigned long long) cow_blk->addr);
		blksnap_data_cow_show(cow_blk);
	}

	if (!has_data && cow_data) {
		bsmsg_volume(volume,
			     "WARNING: Block %llu is flagged as cow but don't have cow data",
			     (unsigned long long) cow_blk->addr);
		blksnap_data_cow_show(cow_blk);
	}

	if (has_data && !cow_data) {
		bsmsg_volume(volume,
			     "WARNING: Block %llu has allocated cow but not flagged as cow",
			     (unsigned long long) cow_blk->addr);
		blksnap_data_cow_show(cow_blk);
	}

	if (has_data && !nextsnap) {
		bsmsg_volume(volume,
			     "WARNING: Block %llu has data but not in next snapshot",
			     (unsigned long long) cow_blk->addr);
		blksnap_data_cow_show(cow_blk);
	}
}
