/* {{{1 Preamble */
#include <linux/types.h>
#include <linux/gfp.h>
#include <linux/slab.h>
#include <linux/kernel.h>
#include <linux/blkdev.h>

#include "blksnap.h"
#include "blksnap-bio-queue-share.h"
#include "blksnap-hash.h"
#include "blksnap-device.h"
#include "blksnap-data.h"
#include "blksnap-compat.h"
#include "blksnap-shared.h"
#include "blksnap-extent.h"

/* A cow's page pointer is set to this value if it intended to be stored in
 * memory.  The logic is as follows:
 *
 *   * If a cow_blk's page pointer is NULL, it is reserved on disk and has not
 *     been commited yet.
 *   * If a cow_blk's page pointer is not NULL and is equal to the sentinel value,
 *     then it is reserved in memory and has not been committed yet.
 *   * If a cow_blk's page pointer is not NULL and not equal to the sentinel
 *     value, then it's a valid page pointer and contains cow data.
 *     * if committed = true, the cow_blk stores in memory
 *     * if committed = false, the cow_blk stores in disk but data has not been
 *     commited to disk yet.
 *
 * Additionally, if a cow_blk's page pointer is NULL, the 'committed' field will
 * tell you whether the data is on disk or not.
 *
 * Also worth noting, for a cow reserved on disk, the page pointer is valid
 * during the time it's being processed by the device driver.  Once the end_io
 * callback occurs (indicating the block was successfully written), the page
 * pointer is reset to NULL.
 */
#define CLAIM_SENTINEL	(0xdeadbeef)

/* {{{1 Setup/Teardown */
struct blksnap_data_store *blksnap_data_store_alloc(struct block_device *bdev)
{
	gfp_t gfp_mask = GFP_NOIO;
	struct blksnap_data_store *store;

	store = kzalloc(sizeof(*store), gfp_mask);
	if (!store)
		return ERR_PTR(-ENOMEM);

	store->bdev = bdev;
	mutex_init(&store->extents.lock);
	mutex_init(&store->reservation);

	return store;
}

void blksnap_data_store_free(struct blksnap_data_store *store)
{
	if (!store)
		return;

	blksnap_extent_free(store->extents.head);
	kfree(store);
}

/* {{{1 Store Operations */
bool blksnap_data_store_is_valid(struct blksnap_data_store *store)
{
	return (store && store->valid);
}

void blksnap_data_store_reset(struct blksnap_data_store *store)
{
	bsmsg_bdev(store->bdev,
		   "Data store utilization (memory: %llu/%llu, disk: %llu/%llu)",
		   store->npages_mem_now, store->npages_mem_max,
		   store->npages_disk_now, store->npages_disk_max);

	store->npages_mem_now = store->npages_disk_now = 0;
}

int blksnap_data_store_unmap(struct blksnap_data_store *store)
{
	/* Reserve store to prevent posting cow_ops during unmap */
	mutex_lock(&store->reservation);

	store->valid = false;

	blksnap_data_store_reset(store);
	blksnap_extent_free(store->extents.head);
	store->extents.head = NULL;

	/* Set maximum number of pages in memory and disk to 0. Because we
	 * unmaped data store */
	store->npages_mem_max = store->npages_disk_max = 0;

	/* Releasing mutex after setting npage_max to 0. */
	mutex_unlock(&store->reservation);

	return 0;
}

int blksnap_data_store_map(struct blksnap_data_store *store,
			   struct blksnap_ioctl_map *map)
{
	unsigned int i;
	struct blksnap_extent *new;

	if (is_first_map_ioctl(map)) {
		if (store->npages_mem_max > 0 || store->npages_disk_max > 0) {
			bsmsg_bdev(store->bdev,
				   "WARNING: Attempt to map already mapped data store");
			return -EEXIST;
		}
		store->npages_mem_max = map->npages;
	}

	for (i = 0; i < map->nextents; i++) {
		new = __blksnap_extent_add(map->extents[i].start_sect,
					   map->extents[i].nsectors,
					   store->extents.head);
		if (IS_ERR(new))
			return PTR_ERR(new);
		store->extents.head = new;

		store->npages_disk_max +=
			map->extents[i].nsectors >> PAGE_SECTORS_SHIFT;
	}

	if (is_last_map_ioctl(map)) {
		blksnap_extent_show(store->extents.head, store->bdev, "data");

		bsmsg_bdev(store->bdev,
			   "Data store is mapped (memory: %llu, disk: %llu)",
			   store->npages_mem_max, store->npages_disk_max);
		store->valid = true;
	}

	return 0;
}

/* {{{1 CoW Operations */
/* {{{2 Alloc CoW */
int blksnap_data_cow_alloc(struct blksnap_data_store *store,
			   struct blksnap_cow_block *cow_blk)
{
	int err = 0;

	mutex_lock(&store->reservation);

	/* Attempt to reserve in-memory slot first */
	if (store->npages_mem_now < store->npages_mem_max) {
		store->npages_mem_now++;
		cow_blk->page = (struct page *) CLAIM_SENTINEL;
		D("+L  M (0x%p)", cow_blk);
		goto out;
	}

	/* Fall back to on-disk slot if in-memory fails */
	if (store->npages_disk_now  < store->npages_disk_max) {

		if (blksnap_extent_find(store->extents.head,
					PAGE_SECTORS * store->npages_disk_now,
					PAGE_SECTORS,
					&cow_blk->extent_offset)) {

			store->npages_disk_now++;
			D("+L  D (0x%p, %llu)", cow_blk, cow_blk->extent_offset);
			goto out;
		}
	}

	/* No slot available, store is exhausted */
	err = -ENOSPC;
out:
	mutex_unlock(&store->reservation);
	return err;
}

/* {{{2 Has Data */
static inline bool __cow_blk_has_mem_data(struct blksnap_cow_block *cow_blk)
{
	return (cow_blk->page && cow_blk->page != (struct page *) CLAIM_SENTINEL && cow_blk->committed);
}

static inline bool __cow_blk_has_disk_data(struct blksnap_cow_block *cow_blk)
{
	return (cow_blk->committed && cow_blk->page == NULL);
}

static inline bool __cow_blk_has_data(struct blksnap_cow_block *cow_blk)
{
	return __cow_blk_has_mem_data(cow_blk) || __cow_blk_has_disk_data(cow_blk);
}

/*
 * Returns true if the cow_blk either has data right now, or will soon have data
 * because a write to disk is waiting on the end_io callback.  If a cow_blk is
 * either unallocated or has been allocated by never put'ed, then false will be
 * returned.
 *
 * If this function returns true, it is safe to call get() to fetch the data.
 */
bool blksnap_data_cow_has_data(struct blksnap_cow_block *cow_blk)
{
	bool has_data;
	unsigned long flags;

	read_lock_irqsave(&cow_blk->lock, flags);
	has_data = __cow_blk_has_data(cow_blk);
	read_unlock_irqrestore(&cow_blk->lock, flags);

	return has_data;
}

/* {{{2 Commit */
void __blksnap_data_cow_disk_commit(struct blksnap_cow_block *cow_blk)
{
	struct page *page;

	/* Clear the cached page, future calls to get() will be redirected to
	 * the disk. Nullify pointer at first to avoid collisions on
	 * rescheduling.
	 */
	page = cow_blk->page;
	cow_blk->page = NULL;
	put_page(page);

	/* Page is on disk now
	 */
	cow_blk->committed = true;
}

void blksnap_data_cow_disk_commit(struct blksnap_cow_block *cow_blk)
{
	unsigned long flags;

	write_lock_irqsave(&cow_blk->lock, flags);
	__blksnap_data_cow_disk_commit(cow_blk);
	write_unlock_irqrestore(&cow_blk->lock, flags);
}

/* {{{2 Free */
void __blksnap_data_cow_free(struct blksnap_cow_block *cow_blk)
{
	D("-L  %c (0x%p, %llu, 0x%p)", __cow_blk_has_mem_data(cow_blk) ? 'M' : 'D',
	  cow_blk, cow_blk->extent_offset, cow_blk->page);

	if (__cow_blk_has_mem_data(cow_blk)) {
		struct page *page;

		page = cow_blk->page;
		cow_blk->page = NULL;
		put_page(page);
	}
	/* as __cow_blk_has_mem_data() checks against committed flag, we have to
	 * clear it _AFTER_ check
	 */
	cow_blk->committed = false;
}

void blksnap_data_cow_free(struct blksnap_cow_block *cow_blk)
{
	unsigned long flags;

	write_lock_irqsave(&cow_blk->lock, flags);
	__blksnap_data_cow_free(cow_blk);
	write_unlock_irqrestore(&cow_blk->lock, flags);
}

/* {{{2 Show */
void blksnap_data_cow_show(struct blksnap_cow_block *cow_blk)
{
	unsigned long flags;

	read_lock_irqsave(&cow_blk->lock, flags);
	bsmsg(" L  (0x%p, %llu) committed: %s, page: 0x%p",
	      cow_blk, cow_blk->extent_offset,
	      (cow_blk->committed) ? "yes" : "no", cow_blk->page);
	read_unlock_irqrestore(&cow_blk->lock, flags);
}

/* {{{2 Put */
static sector_t store_address(struct blksnap_data_store *store,
			      struct blksnap_cow_block *cow_blk)
{
	return blksnap_extent_unmap(store->extents.head, cow_blk->extent_offset);
}

static bool cow_blk_put_disk(struct blksnap_data_store *store,
			   struct blksnap_cow_block *cow_blk, struct bio *bio)
{
	/* Clear committed flag so the cow_blk isn't free'd before the
	 * end_io callback executes. */
	cow_blk->committed = false;

	/* Rewrite this bio to store the cow_blk data on the disk.  The caller
	 * will enqueue this bio once we return. */
 	COMPAT_BIO_SET_OP(bio, WRITE, 0);
	/* FIXME: this is not required for now, as we're in queue of underlying
	 * device of bdev and this change does nothing. To add support of
	 * separate disks for *-store, it is needed to reinject CoW bio
	 * requests to the queue of disk where data store is placed.
	 */
	/* bdev of bio = bdev of store; */
	/* absolute sector */
	COMPAT_BIO_SECTOR(bio) = store_address(store, cow_blk);

	/* Cache the page here in case the user reads this block from the
	 * snapshot device before it hits the disk. */
	cow_blk->page = bio_page(bio);

	D(" L  D >>> %llu (0x%p, %llu)", COMPAT_BIO_SECTOR(bio), cow_blk, cow_blk->extent_offset);

	/* Return false, indicating that the caller should not free the bio, and
	 * rather enqueue it. */
	return false;
}

static bool cow_blk_put_mem(struct blksnap_cow_block *cow_blk, struct bio *bio)
{
	/* This page stops here, nothing left to do. */
	cow_blk->page = bio_page(bio);
	cow_blk->committed = true;

	D(" L  M >>> %llu (0x%p, 0x%p)", COMPAT_BIO_SECTOR(bio), cow_blk, cow_blk->page);
	return true;
}

/* The return value indicates whether the caller can free the bio passed in as
 * a parameter.  If the data is stored in memory, there is nothing more to do
 * and the caller can free the bio.  If the data should go to disk, the bio is
 * redirected to the appropriate location on disk and false is returned.  This
 * indicates to the caller that instead of freeing the bio, it should instead
 * enqueue the bio into the respective request queue.
 */
bool blksnap_data_cow_put(struct blksnap_data_store *store,
			  struct blksnap_cow_block *cow_blk,
			  struct bio *bio,
			  int *err)
{
	bool completed;
	unsigned long flags;

	if (warn_null_pointer(cow_blk, __func__)) {
		*err = -EINVAL;
		return true;
	}

	write_lock_irqsave(&cow_blk->lock, flags);

	if (__cow_blk_has_data(cow_blk)) {
		bsmsg_bdev(store->bdev,
			   "WARNING: Attempt to put already valid cow_blk");
		completed = true;
		*err = -EINVAL;
		goto out;
	}

	completed = (cow_blk->page ?
		     cow_blk_put_mem(cow_blk, bio) :
		     cow_blk_put_disk(store, cow_blk, bio));
	*err = 0;
out:
	write_unlock_irqrestore(&cow_blk->lock, flags);
	return completed;
}

/* {{{2 Get */
static bool cow_blk_get_disk(struct blksnap_data_store *store,
			   struct blksnap_cow_block *cow_blk, struct bio *bio)
{
	sector_t start, offset;

	/* Distance from the aligned block address */
	offset = COMPAT_BIO_SECTOR(bio) - bio_block_address(bio);

	/* absolute sector */
	start = store_address(store, cow_blk);
	D(" L  REMAP %llu -> %llu (+%llu)", COMPAT_BIO_SECTOR(bio), start, offset);

	COMPAT_BIO_SECTOR(bio) = start + offset;

	/* False indicates the I/O is not complete and the caller must submit
	 * the bio to the underlying disk after this call returns */
	return false;
}

static bool cow_blk_get_mem(struct blksnap_cow_block *cow_blk, struct bio *bio)
{
	sector_t start_sect, offset_sect;
	char *data_bio, *data_cow_blk;
	compat_bio_vec_t bv;
	compat_bvec_iter_t iter;
	unsigned int offset_src, offset_dst, copy_bytes;

	/* Calculate the aligned block address */
	start_sect = bio_block_address(bio);

	/* Distance from the aligned block address */
	offset_sect = COMPAT_BIO_SECTOR(bio) - start_sect;

	data_cow_blk = COMPAT_KMAP_ATOMIC(cow_blk->page);

	offset_src = offset_sect << SECTOR_SHIFT;
	bio_for_each_segment(bv, bio, iter) {
		offset_dst = COMPAT_BIO_ITER_OFFSET(bio, iter);
		copy_bytes = COMPAT_BIO_ITER_LEN(bio, iter);

		data_bio = COMPAT_KMAP_ATOMIC(
				COMPAT_BIO_ITER_PAGE(bio, iter)
			   );

		memcpy(data_bio + offset_dst,
		       data_cow_blk + offset_src,
		       copy_bytes);

		COMPAT_KUNMAP_ATOMIC(data_bio);
		offset_src += copy_bytes;
	} /* end of bio_for_each_segment() */

	COMPAT_KUNMAP_ATOMIC(data_cow_blk);

	D(" L  M <<< %llu (0x%p, 0x%p) [%u, +%llu]", start_sect, cow_blk, cow_blk->page,
	  COMPAT_BIO_SIZE(bio), offset_sect);

	/* True indicates I/O is complete and the caller can call bio_endio()
	 * without further processing */
	return true;
}

bool blksnap_data_cow_get(struct blksnap_data_store *store,
			    struct blksnap_cow_block *cow_blk,
			    struct bio *bio,
			    int *err)
{
	bool completed;
	unsigned long flags;

	if (warn_null_pointer(cow_blk, __func__)) {
		*err = -EINVAL;
		return true;
	}

	read_lock_irqsave(&cow_blk->lock, flags);

	if (!__cow_blk_has_data(cow_blk)) {
		bsmsg_bdev(store->bdev,
			   "WARNING: attempt to get invalid cow_blk");
		completed = true;
		*err = -EINVAL;
		goto out;
	}

	completed = (__cow_blk_has_mem_data(cow_blk) ?
		     cow_blk_get_mem(cow_blk, bio):
		     cow_blk_get_disk(store, cow_blk, bio));

	*err = 0;
out:
	read_unlock_irqrestore(&cow_blk->lock, flags);
	return completed;
}

