#include <linux/kernel.h>
#include <linux/blkdev.h>

#include "blksnap.h"
#include "blksnap-bio-queue-share.h"
#include "blksnap-hash.h"
#include "blksnap-device.h"
#include "blksnap-compat.h"
#include "blksnap-shared.h"
#include "blksnap-device-utils.h"
#include "blksnap-volume.h"

/* List of physical devices (those that have a request queue) */
LIST_HEAD(blksnap_devices);

/* {{{1 Device Utility Routines */
static bool device_match(struct block_device *bdev,
		  unsigned int major, unsigned int minor)
{
	return (bdev &&
		MAJOR(bdev->bd_dev) == major &&
		MINOR(bdev->bd_dev) == minor);
}

/* Check if device is a parent or child of already
 * attached devices
 */
bool blksnap_bdev_is_duplicate(struct block_device *bdev)
{
	struct list_head *p, *q;

	list_for_each_safe(p, q, &blksnap_volumes) {
		struct blksnap_volume *volume =
			list_entry(p, struct blksnap_volume, classmates);
#if HAVE_BDEV_WHOLE
		if (bdev_whole(volume->bdev) == bdev
		    || volume->bdev == bdev_whole(bdev))
#else
		if (volume->bdev->bd_contains == bdev
		    || volume->bdev == bdev->bd_contains)
#endif
			return true;
	}
	return false;
}

static struct blksnap_device *lookup_device_by_dev(unsigned int major,
						   unsigned int minor)
{
	struct list_head *p, *q;

	list_for_each_safe(p, q, &blksnap_devices) {
		struct blksnap_device *bd =
			list_entry(p, struct blksnap_device, siblings);

		if (device_match(bd->bdev, major, minor))
			return bd;
	}

	list_for_each_safe(p, q, &blksnap_volumes) {
		struct blksnap_volume *volume =
			list_entry(p, struct blksnap_volume, classmates);

		if (volume->device && device_match(volume->bdev, major, minor))
			return volume->device;
	}

	return NULL;
}

struct blksnap_device *blksnap_lookup_device_by_bdev(struct block_device *bdev)
{
	struct blksnap_device *bd = lookup_device_by_dev(MAJOR(bdev->bd_dev),
							 MINOR(bdev->bd_dev));

#if HAVE_BDEV_WHOLE
	if (!bd && bdev_whole(bdev) && (bdev_whole(bdev) != bdev)) {
#else
	if (!bd && bdev->bd_contains && (bdev->bd_contains != bdev)) {
#endif
		/* Kernel version 2.6.32 (and possibly others) does not remap
		 * the bio's block device to the parent block device. So this
		 * additional lookup is necessary for partitions that share the
		 * same parent block device but are not attached.
		 */
#if HAVE_BDEV_WHOLE
		bd = lookup_device_by_dev(MAJOR(bdev_whole(bdev)->bd_dev),
					  MINOR(bdev_whole(bdev)->bd_dev));
#else
		bd = lookup_device_by_dev(MAJOR(bdev->bd_contains->bd_dev),
					  MINOR(bdev->bd_contains->bd_dev));
#endif
	}

	return bd;
}

struct blksnap_device *blksnap_lookup_device_by_bio(struct bio *bio)
{
#if BIO_HAS_BI_DISK
	struct blksnap_device *bd;
#if BDGET_DISK_EXPORTED
	struct block_device *bdev = bdget_disk(bio->bi_disk, bio->bi_partno);
#else
	struct block_device* (*blksnap_bdget_disk)(struct gendisk *, int) = (struct block_device* (*)(struct gendisk *, int)) (BDGET_DISK_ADDR + COMPAT_SYS_CALL_ADDR_DISP);
	struct block_device *bdev = blksnap_bdget_disk(bio->bi_disk, bio->bi_partno);
#endif

	if (!bdev)
		return NULL;

	bd = blksnap_lookup_device_by_bdev(bdev);

	/* Don't forget to release bdev device grabbed with bdget inside
	 * bdget_disk()!
	 */
	bdput(bdev);

	return bd;
#else
	return blksnap_lookup_device_by_bdev(bio->bi_bdev);
#endif
}

