#ifndef _BLKSNAP_DEVICE_UTILS_H
#define _BLKSNAP_DEVICE_UTILS_H

#include <linux/fs.h>

#include "blksnap-shared.h"
#include "blksnap-device.h"
#include "blksnap-config.h"

static char missing_disk[] = "DISK POINTER MISSING!";

/* List of physical devices (those that have a request queue) */
extern struct list_head blksnap_devices;

static inline const char *blksnap_bdev_get_name(struct block_device *bdev, char *buf)
{
#if HAVE_BDEVNAME
	char tmp[BDEVNAME_SIZE];
#endif
	if (!bdev || !bdev->bd_disk)
		return missing_disk;

	if (!bdev->bd_disk->disk_name)
		return UNKNOWN_NAME;

#if HAVE_BDEVNAME
	snprintf(buf, BDEVNAME_SIZE, "/dev/%s", bdevname(bdev, tmp));
#else
	snprintf(buf, BDEVNAME_SIZE, "/dev/%pg", bdev);
#endif
	return buf;
}

static inline const char *blksnap_device_name(struct blksnap_device *bdp, char *buf)
{
	return bdp ? blksnap_bdev_get_name(bdp->bdev, buf) : missing_disk;
}

static inline bool blksnap_is_device_mounted_rw(struct block_device *bdev)
{
	return (bdev->bd_super != NULL
#if HAVE_SB_RDONLY_FLAG
		&& !(bdev->bd_super->s_flags & SB_RDONLY));
#else
		&& !(bdev->bd_super->s_flags & MS_RDONLY));
#endif
}

static inline void blksnap_hold_bdev(struct block_device *bdev)
{
#if GENDISK_HAS_OPEN_PARTITIONS
	mutex_lock(&bdev->bd_disk->open_mutex);
	bdev->bd_disk->open_partitions++;
	mutex_unlock(&bdev->bd_disk->open_mutex);
#else
	mutex_lock_nested(&bdev->bd_mutex, 1);
	bdev->bd_part_count++;
	mutex_unlock(&bdev->bd_mutex);
#endif
}

static inline void blksnap_release_bdev(struct block_device *bdev)
{
#if GENDISK_HAS_OPEN_PARTITIONS
	mutex_lock(&bdev->bd_disk->open_mutex);
	bdev->bd_disk->open_partitions--;
	mutex_unlock(&bdev->bd_disk->open_mutex);
#else
	mutex_lock_nested(&bdev->bd_mutex, 1);
	bdev->bd_part_count--;
	mutex_unlock(&bdev->bd_mutex);
#endif
}

/* Wrapper function around the bd_part block_device member, to avoid the
 * incidents when bd_part is unitialized when the volume is the whole device */
static inline sector_t blksnap_get_nr_sects(struct block_device *bdev)
{
#if HAVE_BDEV_NR_SECTORS
	return bdev_whole(bdev) == bdev ? get_capacity(bdev->bd_disk)
		:  bdev_nr_sectors(bdev);
#else
	return bdev->bd_contains == bdev ? get_capacity(bdev->bd_disk)
		: bdev->bd_part->nr_sects;
#endif
}

bool blksnap_bdev_is_duplicate(struct block_device *bdev);
struct blksnap_device *blksnap_lookup_device_by_bdev(struct block_device *bdev);
struct blksnap_device *blksnap_lookup_device_by_bio(struct bio *bio);

#endif /* _BLKSNAP_DEVICE_UTILS_H */

