#ifndef _BLKSNAP_DEVICE_H
#define _BLKSNAP_DEVICE_H

#include <linux/blkdev.h>

#include "blksnap.h"
#include "blksnap-compat.h"
#include "blksnap-hash.h"

#include "blksnap-bio-queue-share.h"

/* Version of metadata file format: 1 - 255
 */
#define BLKSNAP_METADATA_VERSION 0x01
/* Magic number of metadata file:
 * PATTERN | BLKSNAP_METADATA_VERSION
 */
#define BLKSNAP_METADATA_MAGIC (0x00C87000C8700000 | BLKSNAP_METADATA_VERSION)

struct blksnap_device;
struct blksnap_volume;

/* Creating hash as a struct member for extensibility */
struct blksnap_meta_on_disk {
	u64 magic; /* Magic number to validate metadata file type */
	/* Hash of strong variable blocks of volume */
	u32 hash[BLKSNAP_HASH_DIGEST_WORDS];
};

struct blksnap_extents {
	/* head is the closest extent to a device start */
	struct blksnap_extent *head;
	struct mutex lock;
};

/* Metadata object with extent maps to get hash and to store metadata */
struct blksnap_metadata {
	/* Extents to get checksum for metadata logic */
	struct blksnap_extent *hashblocks;

	/* Extents of the metadata storage. Needed to write latest metadata to
	 * disk after device unmount */
	struct blksnap_extents extents;

	/* Metadata object to store and restore to/from disk */
	struct blksnap_meta_on_disk data;
};

typedef uint8_t block_bm_t;

/* Actual memory size.
 * Must be long-aligned for long-stepped bitmap traversing.
 */
#define BITMAP_CHUNK_SIZE	PAGE_SIZE

/* How many blocks chunk can contain. Block pointer must support
 * dereferencing as (unsigned long) for bitmap atomic operations.
 */
#define BITMAP_CHUNK_BLOCKS	(u32) (BITMAP_CHUNK_SIZE		\
				       - (sizeof(unsigned long)		\
				          - sizeof(block_bm_t)))	\

/* Bitmap is distributed at chunks. Each chunk represents contigious disk area
 * containing BITMAP_CHUNK_SIZE page-sized blocks (16 Mb currently).
 * @changed_blocks: counter of active blocks within disk area
 * @blocks: byte array, which is allocated and freed dynamically
 * @lock_alloc: RW spinlock. Prevents from dereferencing chunk->blocks pointer
 *		to unallocated memory. Flags values themselves are protected
 *		due to atomic bitwise operations.
 */
struct bitmap_chunk {
	uint16_t changed_blocks;
	block_bm_t *blocks;
	struct rw_semaphore lock_alloc;
};

/* Bitmap of volume blocks.
 * Each byte represents page-sized block and stores block flags
 * (see blksnap-block.h).
 * @chunks - allocates on attach, freed on detach
 * @nchunks - constant value, depending on volume size
 * @last_block - the last block offset from device start, in sectors
 *		 (see note in doc/addressing.txt)
 * @vmalloc - stores the way chunks were allocated
 */
struct blksnap_bitmap {
	struct bitmap_chunk *chunks;
	sector_t nchunks;
	sector_t last_block;
	bool vmalloc;
};

/* Iterator for bitmap traversing (also can be used as index)
 * @chunk:    pointer to a current chunk
 * @block:    (block_t *) pointer to a current bitmap block
 * @block_lu: (unsigned long *) pointer to a current bitmap block
 *	      blocks area.
 * @nchunk:   index of a current chunk in bitmap->chunks[] array
 * @nblock:   index of a current block within current chunk (chunk->blocks[])
 * @byte_idx: for internal use during traversing. An index within sizeof(long)
 */
struct blksnap_bitmap_iter {
	struct bitmap_chunk *chunk;
	union {
		block_bm_t *block;
		unsigned long *block_lu;  /* for atomic bitmap operations */
	};
	sector_t nchunk;
	int nblock;
	int byte_idx;
};

/* A wrapper around the block device structure, contains all additional metadata
 * to manage snapshots. */
struct blksnap_device {

	/* The block device structure used by the kernel to represent this
	 * device. */
	struct block_device *bdev;

#if HAVE_MAKE_REQUEST_FN
	/* The original request function, used to submit I/O once our layer has
	 * completed it's function. */
	make_request_fn *make_request;
#else
	/* The original structure of block device operations */
	const struct block_device_operations *original_bd_ops;

	/* Our own structure of block device operations */
	const struct block_device_operations *blksnap_bd_ops;
#endif

	/* The number of volume siblings under snapshot */
	unsigned int snapshot_count;

	/* Back reference to the list of all physical devices */
	struct list_head siblings;

	/* List of child volumes */
	struct list_head volumes;

	/* Used to protect access to snapshot count */
	struct mutex lock;
};

/* Common volume state bit flags */
/* if true, this volume is dirty
 */
#define VOLUME_DIRTY		0
/* if true, underlying disk had been changed while was not being protected
 */
#define VOLUME_CHANGED		1
/* If true, snapshot is running on this volume
 */
#define DIRTY_FLAG_PROVEN	2
/* if true, chlog and metadata are mapped OR snapshot has been done and
 * decision about dirty status is accepted
 */
#define VOLUME_SNAP_ON		3
/* If true, incremental snapshot is getting from this volume,
 * otherwise - base snapshot
 */
#define VOLUME_SNAP_INCR	4
/* Volume is grabbing a map at that moment, have to process incoming
 * bio write-request in other way
 */
#define VOLUME_GRAB_MAP		5

/* Filesystem properties bit flags  */
#define VOLUME_FS_XFS_VERSION_5	0


/* The volume representation. It is not always that case that the
 * volume is the actual device used for attachment. For example, sda1
 * is the first partition on a particular disk. But this block device doesn't
 * have a request queue, so it cannot be attached to directly. The underlying
 * device (sda) must be attached to. This is needed internally to access the
 * appropriate data stuctures when I/O requests are made. They are kept
 * seperate from the device structure so we can refer to them on user
 * initiation. */
struct blksnap_volume {

	/* Unique identifier. This is used for the minor number when creating
	 * a block device for a snapshot request and also for the drive letter
	 * in the device name. */
	int id;

	/* User-specified name for this block device.  This is probably the path
	 * to the block device or a shorted version of it, but it can be
	 * anything you want as long as it's unique. */
	char name[BLKSNAP_NAME_SIZE];

	/* See flags above */
	unsigned long flags;

	/* Filesystem properties flags */
	unsigned long fs_flags;

	/* Extents of a file is currently being grabbed with grab_map ioctl */
	struct blksnap_extents grabbing_extents;

	/* Siblings: other volumes that share the same parent */
	struct list_head siblings;

	/* Classmates: the set of all volumes */
	struct list_head classmates;

	/* Used for device lookups and by the on-disk changelog store */
	struct block_device *bdev;

	/* The underlying physical device */
	struct blksnap_device *device;

	/* On-disk storage - changelog store */
	struct blksnap_cl_store *clstore;

	/* On-disk storage - COW block data */
	struct blksnap_data_store *data;

	/* Block device exposed on snapshot request. Snapshot data can be read
	 * from this device and will not change until snapshot is complete. */
	struct gendisk *disk;

	struct blksnap_bitmap bitmap;

	/* Size of volume in sectors */
	sector_t nsectors;

	/* CoW tree stores addresses and data of blocks, which were modified
	 * during snapshot. It is needed to represent a consistent image of
	 * protected volume in a point of snapshot start. CoW tree lives only
	 * during snapshot. */
	struct rb_root cow_tree;

	/* Used to protect access to the cow tree */
	struct rw_semaphore tree_rw_lock;

	/* Used to protect changing/checking cow-flags */
	struct mutex cow_lock;

	/* A link to another volume. This is used for the restore
	 * operation to copy data from one device to another. The link structure
	 * describes the pairing and is used to redirect I/O to the appropriate
	 * location. */
	struct {
		struct blksnap_volume *to;
		enum {
			LINK_NONE,
			LINK_TO,
			LINK_FROM,
		} direction;
	} link;

	/* Number of open handles in userspace. Destroying the snapshot device
	 * while userspace has open handles results in I/O errors. */
	atomic_t snapshot_holders;

	/* Any errors that occur during snapshot are recorded here. If errors
	 * are present, the snapshot device will fail incoming read requests. */
	atomic_t snapshot_errors;

	/* The start volume relative to the start of physical device.
	 * It is easy to calculate this value on flight, but it is used
	 * very often so storing it improves overall performance
	 */
	sector_t start_block;

	/* Actual (unaligned) volume start sector */
	sector_t start_sector;

	/* A pointer to the superblock is necessary to freeze and thaw the
	 * filesystem when a snapshot operation requires it. */
	struct super_block *frozen_sb;

	/* FileSystem type */
	enum blksnap_fs_type fstype;

	/* Logical metadata storage */
	struct blksnap_metadata metadata;

	/* Flush bio queue to process original write bios */
	struct blksnap_bio_queue write_queue;
};

#endif /* _BLKSNAP_DEVICE_H */
