#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/blkdev.h>

#include "blksnap.h"
#include "blksnap-bio-queue-share.h"
#include "blksnap-hash.h"
#include "blksnap-device.h"
#include "blksnap-compat.h"
#include "blksnap-shared.h"
#include "blksnap-extent.h"

#define FOREACH_EXT(xts, xt) for ((xt) = (xts); (xt) && ((xt)->nsectors); (xt) = (xt)->next)

static struct kmem_cache *extent_cache = NULL;

int blksnap_extent_cache_create(void)
{
	if (unlikely(extent_cache)) {
		WARN(1, "Detected try to re-create a cache, %s", __func__);
		return -EEXIST;
	}

	extent_cache = KMEM_CACHE(blksnap_extent, 0);
	if (!extent_cache) {
		bsmsg("ERROR: Can't create cache in %s", __func__);
		return -ENOMEM;
	}

	return 0;
}

void blksnap_extent_cache_destroy(void)
{
	if (unlikely(!extent_cache)) {
		WARN(1, "Detected try to destroy an empty cache, %s", __func__);
		return;
	}

	kmem_cache_destroy(extent_cache);
	extent_cache = NULL;
}

static struct blksnap_extent *extent_alloc(void)
{
	gfp_t gfp_mask = GFP_KERNEL;
	struct blksnap_extent *ext;

	FALL_ON_NO_CACHE(extent_cache, NULL);

	ext = kmem_cache_alloc(extent_cache, gfp_mask);

	return ext;
}

static void extent_free(struct blksnap_extent *ext)
{
	BUG_ON(!ext);

	FALL_ON_NO_CACHE(extent_cache, );

	kmem_cache_free(extent_cache, ext);
}

static struct blksnap_extent *extent_new(sector_t start, size_t nsectors,
					 struct blksnap_extent *next)
{
	struct blksnap_extent *ext = extent_alloc();

	if (!ext)
		return ERR_PTR(-ENOMEM);

	ext->nsectors = nsectors;
	ext->start = start;
	ext->next = next;

	return ext;
}

static bool blk_addr_ext_merge(struct blksnap_extent *ext,
			       sector_t start, u32 nsectors)
{
	/* If the new extent is on the right or on the left of ext node, we
	 * don't need to merge */
	return !(ext->start + ext->nsectors < start ||
		 start + nsectors < ext->start);
}

static bool extent_preceeds(sector_t start, size_t nsectors,
			    struct blksnap_extent *ext)
{
	BUG_ON(!ext);

	return ((start + nsectors) < ext->start);
}

static void extent_free_range(struct blksnap_extent *start,
			      struct blksnap_extent *end)
{
	if (!start)
		return;

	if (start == end) {
		extent_free(start);
		return;
	}

	while (start != end) {
		struct blksnap_extent *ext = start;

		start = start->next;
		extent_free(ext);
	}
}

/* Add new extent to list of extents
 * @start, @nsectors:	address and length of extent to be added to list
 * @head:		minor extent of a list
 * Return value: pointer to the actual list' head.
 * WARNING: not thread-safe
 */
struct blksnap_extent *__blksnap_extent_add(sector_t start, size_t nsectors,
					    struct blksnap_extent *head)
{
	struct blksnap_extent *tmp, *prev, *next, *new;

	if ((!start && head) || !nsectors) {
		bsmsg("WARNING: try to add invalid extent (0x%p, %llu, %llu)",
		      head, start, nsectors);
		return head;
	}

	D(" E add: %llu +%u (0x%p)", start, nsectors, head);

	if (!head || extent_preceeds(start, nsectors, head)) {
		/* {start, nsectors} extent is on the left from the head */
		new = extent_new(start, nsectors, head);
		if (IS_ERR(new))
			goto no_memory;
		return new;
	}

	/* tmp - the first extent that ends on the right of 'start' */
	for (tmp = head, prev = NULL;
	     tmp && (start > tmp->start + tmp->nsectors);
	     prev = tmp, tmp = tmp->next);

	for (next = tmp;
	     next && next->next && (start + nsectors >= next->next->start);
	     prev = next, next = next->next);

	if (tmp && blk_addr_ext_merge(tmp, start, nsectors)) {
		struct blksnap_extent *foo = tmp->next;
		u64 t_len;

		/* Check for tmp->nsectors overflow. If we assume that the
		 * two values are non-negative a and b, then if (a > max - y)
		 * then we must abort this operation. */
		tmp->start = MIN(start, tmp->start);
		t_len = MAX(start + nsectors, next->start + next->nsectors)
		        - tmp->start;

		if (tmp != next)
			extent_free_range(foo, next->next);

		if (SIZE_MAX < t_len) {
			tmp->nsectors = SIZE_MAX;
			new = extent_new(tmp->nsectors + SIZE_MAX,
					 t_len - SIZE_MAX,
					 next->next);
			if (IS_ERR(new))
				goto no_memory;
			tmp->next = new;
		} else {
			tmp->nsectors = t_len;
			tmp->next = next->next;
		}
	} else {
		/* {start, nsectors} extent is on the left of tmp */
		new = extent_new(start, nsectors, tmp);
		if (IS_ERR(new))
			goto no_memory;
		prev->next = new;
	}

	return head;

no_memory:
	bsmsg("FATAL: Can't allocate memory for new extent");
	/* We can't do anything with this error in this function,
	 * so just return it to our parent
	 */
	return ERR_PTR(-ENOMEM);
}

sector_t blksnap_extent_unmap(struct blksnap_extent *head, sector_t offset)
{
	sector_t relative_start = 0; /* physical extent volume related start */
	struct blksnap_extent *ext;

	FOREACH_EXT(head, ext) {
		if (offset >= relative_start &&
		    offset < relative_start + ext->nsectors)
			return ext->start + (offset - relative_start);
		relative_start += ext->nsectors;
	}

	return 0;
}

int blksnap_extent_find(struct blksnap_extent *head, sector_t start,
			size_t nsectors, sector_t *addr)
{
	sector_t relative_start = 0; /* physical extent volume related start */
	struct blksnap_extent *xt, *x;

	D("(%s) start %llu, len %u", __func__, start, nsectors);

	BUG_ON(!head);

	FOREACH_EXT(head, xt) {
		D("(%s) looking at %llu, in (%llu,%u)", __func__,
		  relative_start, xt->start, xt->nsectors);
		if (start >= relative_start &&
		    start < relative_start + xt->nsectors) {
			if (nsectors <= xt->nsectors - (start - relative_start)) {
				*addr = start;
				return 1;
			} else {
				relative_start += xt->nsectors;
				break;
			}
		}
		relative_start += xt->nsectors;
	}

	if (!xt) {
		bsmsg("WARNING: Size of extent map %p is smaller than should",
		      head);
		return 0;
	}

	/* Should be in the begining of the first 'large enough' extent */
	FOREACH_EXT(xt->next, x) {
		if (x->nsectors >= nsectors) {
			*addr = relative_start;
			return 1;
		}
		relative_start += x->nsectors;
	}

	/* Out of range */
	return 0;
}

void blksnap_extent_free(struct blksnap_extent *head)
{
	extent_free_range(head, NULL);
}

#if defined(BLKSNAP_DEBUG_EXTENTS)
void blksnap_extent_show(const struct blksnap_extent *head,
			 const struct block_device *bdev,
			 const char *type)
{
	bsmsg_bdev(bdev, "%s extents:", type);

	for (; head; head = head->next)
		bsmsg(" E (%llu +%u) 0x%p -> 0x%p", head->start, head->nsectors,
		      head, head->next);
}
#else /* !BLKSNAP_DEBUG_EXTENTS */
void blksnap_extent_show(const struct blksnap_extent *head,
			 const struct block_device *bdev,
			 const char *type)
{
	return;
}
#endif /* !BLKSNAP_DEBUG_EXTENTS */
