/*
 * As blk-snapshot is a block layer solution to COW snapshots, additional
 * "help" is inlisted from userspace to obtain on-disk space to store
 * persistent information. A file is created and allocated in userspace and
 * those extents (the blocks that make up the file) are passed into the modules
 * through the ioctl interface. This "file" is never modified from userspace
 * and serves as a reserved on-disk area to store information. This information
 * may be the on-disk representation of the tree (changelog store) or it may
 * contain original disk data that was modified during a snapshot (data).
 *
 * The changelog store is persisted to disk so that an incremental snapshot can
 * be taken after the machine powers on. The on-disk tree is loaded into memory
 * and the module has reclaimed its original state, allowing us to pick up
 * where we left off in the event of a reboot or other power event.
 *
 * COW data is written to disk to prevent running out of memory during
 * a snapshot.  This data is discared after each snapshot.
 *
 * This interface serves as an abstraction for the userspace files used by the
 * metadata, changelog store and data implementations. At this level, a file is
 * just a set of block extents (typically not contiguous).
 *
 * Ignore the name, this is an implementation of an interval list.  Each item
 * in the list contains an address and a length.  As items are added to the
 * list, the list maintains an ascending order with no overlaps between items.
 */

#ifndef _BLKSNAP_EXTENT_H
#define _BLKSNAP_EXTENT_H

#include "blksnap-shared.h"

struct blksnap_extent {
	/* sectors from device start */
	sector_t start;
	/* sectors */
	size_t nsectors;
	struct blksnap_extent *next;
};

/* setup/teardown */
int blksnap_extent_cache_create(void);
void blksnap_extent_cache_destroy(void);

/* calculates summary of all extents length in bytes */
static inline uint64_t blksnap_extents_size(struct blksnap_extent *first)
{
	struct blksnap_extent *extent;
	uint64_t extents_size = 0;

	for (extent = first; extent; extent = extent->next)
		extents_size += extent->nsectors << SECTOR_SHIFT;

	return extents_size;
}

static inline uint32_t blksnap_extents_count(struct blksnap_extent *first)
{
	struct blksnap_extent *extent;
	uint32_t extents_count = 0;

	for (extent = first; extent; extent = extent->next, extents_count++);

	return extents_count;
}

/* add a region to an extent list */
struct blksnap_extent *__blksnap_extent_add(sector_t start, size_t nsectors,
					    struct blksnap_extent *head);

static inline int blksnap_extent_add(sector_t start, size_t nsectors,
				     struct blksnap_extents *extents)
{
	struct blksnap_extent *ext;

	ext = __blksnap_extent_add(start, nsectors, extents->head);

	if (ext)
		extents->head = ext;

	return ext ? 0 : -1;
}

/* TODO */
sector_t blksnap_extent_unmap(struct blksnap_extent *head, sector_t offset);

/* TODO */
int blksnap_extent_find(struct blksnap_extent *head,
			sector_t start, size_t nsectors, sector_t *addr);

/* release all memory associated with an extent list */
void blksnap_extent_free(struct blksnap_extent *head);

/* dump the extent list to the kernel log */
void blksnap_extent_show(const struct blksnap_extent *head,
			 const struct block_device *bdev,
			 const char *type);

#endif /* _BLKSNAP_EXTENT_H */
