#include <asm-generic/errno-base.h>
#include <linux/kernel.h>
#include <linux/vmalloc.h>
#include <linux/blkdev.h>
#include <linux/string.h>

#include "blksnap.h"
#include "blksnap-bio-queue-share.h"
#include "blksnap-hash.h"
#include "blksnap-device.h"
#include "blksnap-compat.h"
#include "blksnap-shared.h"

uint64_t blksnap_get_data_hash_size(uint64_t data_size)
{
	return round_up(data_size + sizeof(__be64) + 1, BLKSNAP_HASH_MSG_BYTES);
}

/* Calculates hash from given data. Use pseudocode from link
 * https://en.wikipedia.org/wiki/SHA-1.
 * The main transformation do with sha_transform() helper from cryptohash.h.
 *
 * Return value:
 *     0 - on success, error code otherwise
 */
int blksnap_calculate_hash(u8 *data_hash, uint64_t data_size, u32 *out)
{
	const uint32_t bits_offset = BLKSNAP_HASH_MSG_BYTES - sizeof(__be64);
	uint32_t digest[BLKSNAP_HASH_DIGEST_WORDS];
        uint32_t ws[BLKSNAP_HASH_WS_WORDS];
	uint64_t msg_size = blksnap_get_data_hash_size(data_size);
	uint64_t aligned_size, blocks;
	uint8_t *msg = data_hash, *msg_cur;
	__be64 *length_ptr;
	__be32 *digest_be;
	int i;

	/* Initialize variables: */
	blksnap_sha_init(digest);

	/* Set workspace to work properly in case if len < BLKSNAP_HASH_WS_BYTES */
	memset(ws, 0, BLKSNAP_HASH_WS_BYTES);

	/* reset all tail of message to fit step "append 0 ≤ k < 512 bits '0'"
	 */
	memset(&msg[data_size], 0, msg_size - data_size);

	/* Pre-processing: append the bit '1' to the message e.g. by adding
	 * 0x80 if message length is a multiple of 8 bits. Increase size of
	 * message data.
	 */
	msg[data_size++] = 0x80;

	/* append 0 ≤ k < 512 bits '0', such that the resulting message length
	 * in bits is congruent to −64 ≡ 448 (mod 512).
	 * Actually calculate numbers only. We have already resetted tail of
	 * message.
	 */
	aligned_size = round_up(data_size, BLKSNAP_HASH_MSG_BYTES);
	blocks = aligned_size / BLKSNAP_HASH_MSG_BYTES;
	msg_cur = msg;

	/* append ml, the original message length, as a 64-bit big-endian
	 * integer. Thus, the total length is a multiple of 512 bits.
	 */
	length_ptr = (__be64 *)(msg_cur + aligned_size - sizeof(__be64));
	if (aligned_size - data_size < sizeof(__be64)) {
		length_ptr = (__be64 *)(msg_cur + aligned_size + bits_offset);
		blocks++;
	}
	/* ml = message length in bits (always a multiple of the number of bits
	 * in a character).
	 */
	*length_ptr = cpu_to_be64((data_size - 1) << 3);

	while (blocks--) {
		blksnap_sha_transform(digest, msg_cur, ws);
		msg_cur += BLKSNAP_HASH_MSG_BYTES;
	}

	/* Produce the final hash value (big-endian) as a 160-bit number:
	 */
	digest_be = (__force __be32 *) digest;
	for (i = 0; i < BLKSNAP_HASH_DIGEST_WORDS; i++)
		digest_be[i] = cpu_to_be32(digest[i]);

	memcpy(out, digest_be, sizeof(digest));

	return 0;
}

void blksnap_hasher_init(struct blksnap_hasher *hasher, uint64_t total_data_size)
{
	const uint32_t bits_offset = BLKSNAP_HASH_MSG_BYTES - sizeof(__be64);
	__be64 *length_ptr;

	/* Initialize variables: */
	blksnap_sha_init(hasher->digest);

	/* Set workspace to work properly in case if len < BLKSNAP_HASH_WS_BYTES */
	memset(hasher->ws, 0, BLKSNAP_HASH_WS_BYTES);

	/* reset all tail of message to fit step "append 0 ≤ k < 512 bits '0'"
	 */
	memset(hasher->tail_data, 0, BLKSNAP_HASH_MSG_BYTES);

	/* Pre-processing: append the bit '1' to the message e.g. by adding
	 * 0x80 if message length is a multiple of 8 bits. Increase size of
	 * message data.
	 */
	hasher->tail_data[0] = 0x80;

	length_ptr = (__be64 *)(hasher->tail_data + bits_offset);

	/* ml = message length in bits (always a multiple of the number of bits
	 * in a character).
	 */
	*length_ptr = cpu_to_be64(total_data_size << 3);
}

void blksnap_hasher_update(struct blksnap_hasher *hasher, uint8_t *data, uint64_t data_size)
{
	uint8_t *data_pos = data;
	uint64_t data_blocks = data_size / BLKSNAP_HASH_MSG_BYTES;

	while (data_blocks--) {
		blksnap_sha_transform(hasher->digest, data_pos, hasher->ws);
		data_pos += BLKSNAP_HASH_MSG_BYTES;
	}
}

void blksnap_hasher_finalize(struct blksnap_hasher *hasher, u32 *out)
{
	__be32 *digest_be;
	int i;

	/* Process the tail data
	 */
	blksnap_sha_transform(hasher->digest, hasher->tail_data, hasher->ws);

	/* Produce the final hash value (big-endian) as a 160-bit number:
	 */
	digest_be = (__force __be32 *) hasher->digest;
	for (i = 0; i < BLKSNAP_HASH_DIGEST_WORDS; i++)
		digest_be[i] = cpu_to_be32(hasher->digest[i]);

	memcpy(out, digest_be, sizeof(hasher->digest));
}
