/* {{{1 Preamble */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/blkdev.h>
#include <linux/delay.h>
#include <linux/reboot.h>
#include <linux/slab.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/bio.h>
#include <linux/mm_types.h>

#include "blksnap.h"
#include "blksnap-bio-queue-share.h"
#include "blksnap-bio-queue.h"
#include "blksnap-hash.h"
#include "blksnap-device.h"
#include "blksnap-fs-types.h"
#include "blksnap-compat.h"
#include "blksnap-shared.h"
#include "blksnap-main.h"
#include "blksnap-data.h"
#include "blksnap-cow-tree.h"
#include "blksnap-extent.h"
#include "blksnap-cl-store.h"
#include "blksnap-write-queue.h"
#include "blksnap-snapshot.h"
#include "blksnap-device-utils.h"
#include "blksnap-volume.h"
#include "blksnap-bio-submit.h"
#include "blksnap-metadata.h"
#include "blksnap-bitmap.h"
#include "version.h"

#ifndef BLKSNAP_LICENSE
#  define BLKSNAP_LICENSE	"GPL"
#endif

#define AUTHOR			"Quest Software Inc."
#define LICENSE			BLKSNAP_LICENSE

/* The mode passed to blkdev_get() and blkdev_put() to attach and detach from
 * block devices respectively. */
#define BDEV_MODE		(FMODE_READ)

/* Number of attempts to acquire the ioctl lock before giving up. */
#define IOCTL_LOCK_TRIES	15

/* Number of milliseconds to wait before attempting to acquire the ioctl lock
 * after a failed attempt. */
#define IOCTL_LOCK_SLEEP_MS	200

/* Number of milliseconds to wait before attempting to detach a busy volume
 * again. */
#define DETACH_VOLUME_SLEEP_MS	100

/* {{{1 Static Globals */

/* The module can be disabled completely on the command line.
 *
 * It would seem obvious to use a bool here, but kernels older than 2.6.28 do
 * now allow bools as module parameters.
 */
static int disabled = 0;

module_param(disabled, int, 0444);
MODULE_PARM_DESC(disabled, "Disable this module completely");

/* Commandline argument used to specify which devices should be attached. The
 * format is "devices=<device>, ...". A device can be specified in one of two
 * ways. The short format can be a block device residing in /dev, e.g. "sda".
 * The "/dev/" prefix will be added to such device names. A full path can also
 * be specified ("/dev/sdc") and the path will be used unmodified.
 */
static char *devices = NULL;

module_param(devices, charp, 0444);
MODULE_PARM_DESC(devices, "A comma separated list of block devices");

/* Provide mutual exclusion for ioctl commands */
static DEFINE_MUTEX(blksnap_ioctl_lock);

/* Next unused device id. */
static atomic_t blksnap_next_id;

/* Store the kernel version during compliation.  This is compared to the running
 * version at module load time to catch a mismatch. */
static char *compile_version = UTS_RELEASE;

/* Indicate if the blk-snapshot has been destroyed already to prevent duplicate
 * blksnap_free calls */
static atomic_t destroyed;

/* }}}1 */

/* {{{1 Public Interface */

/* {{{1 I/O Routing */

static int bio_is_write(struct bio *bio)
{
	return ((bio_data_dir(bio) == WRITE) && bio_has_data(bio));
}

bool page_is_marked(struct page *page)
{
	bool ret = false;

	uint64_t *buf = COMPAT_KMAP_ATOMIC(page);

	if (BLKSNAP_MAP_SIGN == *buf)
		ret = true;

	COMPAT_KUNMAP_ATOMIC(buf);

	return ret;
}

int add_marked_sectors(struct bio *bio, struct blksnap_volume *volume)
{
	sector_t curr;
	unsigned int mapped = 0;
	compat_bvec_iter_t iter;
	compat_bio_vec_t bvec;

	curr = COMPAT_BIO_SECTOR(bio);

	bio_for_each_segment(bvec, bio, iter) {
		/* Break cycle if volume was detached or grabbing - cancelled */
		int ret = 1;
		unsigned int nsectors =
			sector_count(COMPAT_BIO_ITER_LEN(bio, iter));

		if (!page_is_marked(COMPAT_BIO_ITER_PAGE(bio, iter)))
			continue;

		mutex_lock(&volume->grabbing_extents.lock);
		/* On detach or mapping cancel this conditional must protect
		 * from dereference of invalid pointer or memory leak
		 */
		if (likely(volume && volume_is_grabbing_map(volume))) {
			ret = blksnap_extent_add(curr, nsectors,
						 &volume->grabbing_extents);
		}
		mutex_unlock(&volume->grabbing_extents.lock);

		/* If adding failed - not write signed blocks to volume */
		mapped++;

		if (ret) {
			bsmsg_volume(volume,
				   "ERROR: Can't add extent [%lu:%u]",
				   curr, SECTOR_SIZE);
			break;
		}

		curr += nsectors;
	}

	return mapped;
}

/* custom make_request function - main filter to process bio-requests by module
 * @q: request queue for whole disk device
 * @bio: bio request already arranged to whole disk device.
 * It means that:
 * - bi_bdev/bi_disk points to whole disk;
 * - bi_sector is an absolute sector from begin of the whole disk device.
 * AGAIN. WARNING! bi_sector is ABSOLUTE number from start of whole disk.
 */
#if HAVE_MAKE_REQUEST_FN
static MRF_TYPE blksnap_make_request(struct request_queue *q, struct bio *bio)
{
	struct blksnap_device *bd = blksnap_lookup_device_by_bio(bio);
	make_request_fn *make_request;
	struct request_queue *queue;
	/* init ret variable with appropriate type and value */
	MRF_RET_VAR_INIT;

	BUG_ON(!bd);

	/* Store make_request() function pointer and queue ptr to avoid
	 * crash if bd will be destroyed during record_write_bio() runs
	 */
	make_request = bd->make_request;
	queue = bdev_get_queue(bd->bdev);

	/* If bio is NOT write OR blksnap's generated bio,
	 * skip local processing and pass this bio directly to
	 * make_request() function of underlaying device.
	 */
	if (bio_is_write(bio) && !blksnap_owns_bio(bio)) {
		/* Lookup volume to check advanced flags */
		struct blksnap_volume *volume =
			lookup_volume_by_addr(bd, COMPAT_BIO_SECTOR(bio));

		/* The partition of disk is not protected. Skip processing bio.
		 */
		if (unlikely(!volume)) {
			D(" X  %llu 0x%p", COMPAT_BIO_SECTOR(bio), bio);
			goto make_request;
		}

		if (volume_is_grabbing_map(volume)) {
			int mapped = 0;

			mapped = add_marked_sectors(bio, volume);

			/* if we've got a mapped write bio requests from bsctl,
			 * we MUST do not really write data to our store files
			 */
			if (mapped) {
				blksnap_bio_endio(bio, 0);
				goto out;
			}
		}

		D("@@@ %llu (%llu)  W  %u [%s]", bio_block_address(bio),
		  COMPAT_BIO_SECTOR(bio), COMPAT_BIO_SIZE(bio), current->comm);

		/* Snapshot is in progress. Enqueue bio to write_bio kthread
		 * and finish.
		 */
		if (volume_has_snapshot(volume)) {
			blksnap_write_bio_enqueue(volume, bio);
			goto out;
		}

		/* Nothing to do with dirty proved volume
		 * until take base snapshop.
		 */
		if (test_bit(VOLUME_DIRTY, &volume->flags) &&
		    test_bit(DIRTY_FLAG_PROVEN, &volume->flags))
			goto make_request;

		record_write_bio(volume, bio);
	}

	/* Sanity check, bi_next is used in the layers below us */
	WARN_ON(bio->bi_next);

make_request:
	MRF_LET_RET_VAR blksnap_make_request_compat(make_request, queue, bio);

out:
	return MRF_RET_VAR;
}
#else
static MRF_TYPE blksnap_submit_bio(struct bio *bio)
{
	MRF_RET_VAR_INIT;
	struct blksnap_device *bd = blksnap_lookup_device_by_bio(bio);

	BUG_ON(!bd);

	/* If bio is NOT write OR blksnap's generated bio,
	 * skip local processing and pass this bio directly to
	 * make_request() function of underlaying device.
	 */
	if (bio_is_write(bio) && !blksnap_owns_bio(bio)) {
		/* Lookup volume to check advanced flags */
		struct blksnap_volume *volume =
			lookup_volume_by_addr(bd, COMPAT_BIO_SECTOR(bio));

		/* The partition of disk is not protected. Skip processing bio.
		 */
		if (unlikely(!volume)) {
			D(" X  %llu 0x%p", COMPAT_BIO_SECTOR(bio), bio);
			goto submit_bio;
		}

		if (volume_is_grabbing_map(volume)) {
			int mapped = 0;

			mapped = add_marked_sectors(bio, volume);

			/* if we've got a mapped write bio requests from bsctl,
			 * we MUST do not really write data to our store files
			 */
			if (mapped) {
				blksnap_bio_endio(bio, 0);
				goto out;
			}
		}

		D("@@@ %llu (%llu)  W  %u [%s]", bio_block_address(bio),
		  COMPAT_BIO_SECTOR(bio), COMPAT_BIO_SIZE(bio), current->comm);

		/* Snapshot is in progress. Enqueue bio to write_bio kthread
		 * and finish.
		 */
		if (volume_has_snapshot(volume)) {
			blksnap_write_bio_enqueue(volume, bio);
			goto out;
		}

		/* Nothing to do with dirty proved volume
		 * until take base snapshop.
		 */
		if (test_bit(VOLUME_DIRTY, &volume->flags) &&
		    test_bit(DIRTY_FLAG_PROVEN, &volume->flags))
			goto submit_bio;

		record_write_bio(volume, bio);
	}

	/* Sanity check, bi_next is used in the layers below us */
	WARN_ON(bio->bi_next);

submit_bio:
	if (bd->original_bd_ops != NULL && bd->original_bd_ops->submit_bio != NULL) {
		MRF_LET_RET_VAR bd->original_bd_ops->submit_bio(bio);
	} else {
		MRF_LET_RET_VAR blksnap_submit_bio_compat(bio);
	}

out:
	return MRF_RET_VAR;
}

#endif

/* {{{1 Ioctl Interface (Commands) */
/* {{{2 Attach Volume */
/* {{{3 Attach Logic */

static void override_request_queue(struct blksnap_device *disk)
{
#if HAVE_MAKE_REQUEST_FN
	struct request_queue *q = bdev_get_queue(disk->bdev);
	disk->make_request = xchg(&q->make_request_fn, blksnap_make_request);
#else
	struct block_device_operations *bd_ops = (struct block_device_operations*)disk->blksnap_bd_ops;
	memcpy(bd_ops, disk->original_bd_ops, sizeof(struct block_device_operations));
	bd_ops->submit_bio = blksnap_submit_bio;

	(void) xchg(&disk->bdev->bd_disk->fops, bd_ops);
#endif
}

static struct blksnap_device *alloc_device(struct block_device *bdev)
{
	struct blksnap_device *device;

	if (!bdev || !bdev->bd_disk)
		return NULL;

	device = kzalloc(sizeof(*device), GFP_KERNEL);
	if (!device)
		return ERR_PTR(-ENOMEM);

#if !HAVE_MAKE_REQUEST_FN
	device->original_bd_ops = bdev->bd_disk->fops;
	device->blksnap_bd_ops = kzalloc(sizeof(struct block_device_operations), GFP_KERNEL);
	if (!device->blksnap_bd_ops)
		return ERR_PTR(-ENOMEM);
#endif

	mutex_init(&device->lock);

	INIT_LIST_HEAD(&device->volumes);

	device->bdev = bdev;

	list_add_tail(&device->siblings, &blksnap_devices);
	override_request_queue(device);
	bsmsg_bdev(device->bdev, "Disk is attached");

	return device;
}

#define UNINITIALIZED_VOLUME_ID	(-1)

static struct blksnap_volume *alloc_volume(struct block_device *bdev,
					   const char *name)
{
	int bitmap_create_res;
	gfp_t gfp_mask = GFP_KERNEL;
	struct blksnap_volume *volume;

	volume = kzalloc(sizeof(*volume), gfp_mask);
	if (!volume)
		return ERR_PTR(-ENOMEM);

	volume->bdev = bdev;
	volume->cow_tree = RB_ROOT;
	volume->id = UNINITIALIZED_VOLUME_ID;

	volume->start_sector = get_start_sect(volume->bdev);
	volume->start_block = aligned_block_address(volume->start_sector);

	bitmap_create_res = blksnap_blk_bitmap_create(volume);
	if (bitmap_create_res)
		goto bitmap_fail;

	init_rwsem(&volume->tree_rw_lock);
	mutex_init(&volume->grabbing_extents.lock);
	mutex_init(&volume->cow_lock);

	atomic_set(&volume->snapshot_holders, 0);

	snapshot_error_reset(volume);

	strncpy(volume->name, name, BLKSNAP_NAME_SIZE);

	/* Volume is dirty when it is just attached */
	set_bit(VOLUME_DIRTY, &volume->flags);

	/* Assuming that volume was not changed while not being attached */
	clear_bit(VOLUME_CHANGED, &volume->flags);

	return volume;

bitmap_fail:
	kfree(volume);
	return ERR_PTR(bitmap_create_res);
}

static void detach_device(struct blksnap_device *disk);

static int blksnap_ioctl_attach(struct blksnap_ioctl_attach *attach)
{
	int res = 0;
	int ext_idx;
	struct block_device *bdev;
	struct blksnap_device *disk;
	struct blksnap_volume *volume = NULL;
	char *path = attach->name;
	struct blksnap_extent *new;

	bdev = blkdev_get_by_path(path, BDEV_MODE, NULL);
	if (IS_ERR(bdev))
		return PTR_ERR(bdev);

	volume = lookup_volume_by_bdev(bdev);
	if (volume) {
		COMPAT_BLKDEV_PUT(bdev, BDEV_MODE);
		return -EEXIST;
	}

	if (blksnap_bdev_is_duplicate(bdev)) {
		COMPAT_BLKDEV_PUT(bdev, BDEV_MODE);
		return -EINVAL;
	}

	volume = alloc_volume(bdev, path);
	if (IS_ERR(volume)) {
		COMPAT_BLKDEV_PUT(bdev, BDEV_MODE);
		bsmsg_bdev(bdev, "ERROR: failed to allocate volume");
		return PTR_ERR(volume);
	}

	/* Increment counters held to prevent repartition this device */
	blksnap_hold_bdev(bdev);

	volume->data = blksnap_data_store_alloc(volume->bdev);
	if (IS_ERR(volume->data)) {
		bsmsg_volume(volume, "ERROR: Failed to create COW store");
		res = PTR_ERR(volume->data);
		goto err;
	}

	/* Allocate cl_store */
	volume->clstore = cl_store_init(volume->bdev);
	if (IS_ERR(volume->clstore)) {
		bsmsg_volume(volume, "ERROR: Failed to create CL store");
		res = PTR_ERR(volume->clstore);
		goto err_clstore;
	}

#if HAVE_BDEV_WHOLE
	disk = blksnap_lookup_device_by_bdev(bdev_whole(bdev));
#else
	disk = blksnap_lookup_device_by_bdev(bdev->bd_contains);
#endif
	if (!disk)
#if HAVE_BDEV_WHOLE
		disk = alloc_device(bdev_whole(bdev));
#else
		disk = alloc_device(bdev->bd_contains);
#endif

	if (!disk || IS_ERR(disk)) {
		bsmsg_volume(volume, "ERROR: Failed to attach disk");
		res = -EFAULT;
		goto err_disk;
	}

	volume->device = disk;
	volume->id = atomic_add_return(1, &blksnap_next_id) - 1;

	volume->fstype = attach->fstype;

	bsmsg_volume(volume, "FileSystem is: %s",
		     fstype_to_string(volume->fstype));

	mutex_init(&volume->metadata.extents.lock);

	for (ext_idx = 0; ext_idx < attach->nextents; ext_idx++) {
		new = __blksnap_extent_add(
					attach->hashblocks[ext_idx].start_sect,
					attach->hashblocks[ext_idx].nsectors,
					volume->metadata.hashblocks);
		if (IS_ERR(new)) {
			bsmsg_volume(volume, "ERROR: failed to add hasblocks extent");
			goto err_ext;
		}
		volume->metadata.hashblocks = new;
		bsmsg_volume(volume, "Add hashblocks extent #%d [%lu, %lu]",
			     ext_idx,
			     attach->hashblocks[ext_idx].start_sect,
			     attach->hashblocks[ext_idx].nsectors);
	}

	get_blocks_hash(bdev, &volume->metadata);

	/* Set VOLUME_CHANGED flag in case if volume is already mounted */
	if ((attach->flags & BLKSNAP_ATTACH_MOUNT_RW) ||
	    blksnap_is_device_mounted_rw(bdev)) {
		set_bit(VOLUME_CHANGED, &volume->flags);
		bsmsg_volume(volume, "WARNING: volume is mounted in RW mode");
	}

	if (attach->flags & BLKSNAP_ATTACH_XFS_VERSION_5)
		set_bit(VOLUME_FS_XFS_VERSION_5, &volume->fs_flags);

	list_add_tail(&volume->siblings,   &disk->volumes);
	list_add_tail(&volume->classmates, &blksnap_volumes);

	bsmsg_volume(volume, "Volume is attached");
out:
	return res;
err_ext:
	/* free extents that has been already allocated */
	blksnap_extent_free(volume->metadata.hashblocks);
	if (volume->device && list_empty(&volume->device->volumes))
		detach_device(volume->device);
err_disk:
	if (volume->clstore && !IS_ERR(volume->clstore))
		kfree(volume->clstore);
err_clstore:
	if (volume->data && !IS_ERR(volume->data))
		kfree(volume->data);
err:
	blksnap_release_bdev(bdev);
	COMPAT_BLKDEV_PUT(bdev, BDEV_MODE);
	blksnap_blk_bitmap_free(volume);
	kfree(volume);
	goto out;
}
/* }}}3 */

/* {{{2 Detach Volume */
/* {{{3 Detach Logic */

static void restore_request_queue(struct blksnap_device *disk)
{
#if HAVE_MAKE_REQUEST_FN
	struct request_queue *q = bdev_get_queue(disk->bdev);
	(void) xchg(&q->make_request_fn, disk->make_request);
#else
	(void) xchg(&disk->bdev->bd_disk->fops, disk->original_bd_ops);
#endif
}

static void detach_device(struct blksnap_device *disk)
{
	restore_request_queue(disk);
	bsmsg_bdev(disk->bdev, "Disk is detached");
	list_del(&disk->siblings);
#if !HAVE_MAKE_REQUEST_FN
	disk->original_bd_ops = NULL;
	if (disk->blksnap_bd_ops && !IS_ERR(disk->blksnap_bd_ops))
		kfree(disk->blksnap_bd_ops);
#endif
	kfree(disk);
}

/* detach_volume_once doesn't check the snapshot status. Make sure snap was not
 * started for volume before calling it */
static int detach_volume_once(struct blksnap_volume *volume)
{
	int meta_mode = BLKSNAP_META_STORE;

	if (cl_store_status(volume->clstore) == CL_STORE_LOADING)
		return -EBUSY;

	/* If cl-store is NOT valid, we HAVE to get BI after next attach */
	if (blksnap_cl_store_valid(volume->clstore)
	    && !test_bit(VOLUME_DIRTY, &volume->flags)) {
		if (blksnap_blk_bitmap_write_changelog(volume))
			meta_mode = BLKSNAP_META_RESET;
		else
			meta_mode = BLKSNAP_META_UPDATE;
	} else {
		meta_mode = BLKSNAP_META_RESET;
	}

	update_metadata(volume, meta_mode);

	list_del(&volume->classmates);
	list_del(&volume->siblings);

	blksnap_data_store_free(volume->data);

	/* restore request queue before destroying bitmap  */
	if (volume->device && list_empty(&volume->device->volumes))
		detach_device(volume->device);

	blksnap_blk_bitmap_free(volume);

	/* Free memory of metadata and clstore, destroy clstore kthread */
	blksnap_extent_free(volume->metadata.extents.head);
	blksnap_extent_free(volume->metadata.hashblocks);
	cl_store_destroy(volume->clstore);
	volume->clstore = NULL;

	/* Cancel extents grabbing and remove extent list */
	clear_bit(VOLUME_GRAB_MAP, &volume->flags);
	mutex_lock(&volume->grabbing_extents.lock);
	blksnap_extent_free(volume->grabbing_extents.head);
	mutex_unlock(&volume->grabbing_extents.lock);

	bsmsg_volume(volume, "Volume is detached");


	/* Decrement counters held to prevent repartition this device */
	blksnap_release_bdev(volume->bdev);

	/* This command MUST to be called AFTER detach_device() as on some
	 * kernel version after blkdev_put() bdev->bd_disk could be NULL
	 */
	COMPAT_BLKDEV_PUT(volume->bdev, BDEV_MODE);

	kfree(volume);
	return 0;
}

static int detach_volume(struct blksnap_volume *volume)
{
	int ret = 0;

	do {
		if (ret != 0)
			msleep(DETACH_VOLUME_SLEEP_MS);

		/* In case if we're having snap running, lets stop it */
		if (volume->disk)
			stop_snapshot(volume, true);

		ret = detach_volume_once(volume);
	} while (ret == -EBUSY);

	return ret;
}

/* }}}3 */

static int blksnap_ioctl_detach(struct blksnap_ioctl_detach *detach)
{
	struct blksnap_volume *volume;

	volume = lookup_volume_by_path(detach->name);
	if (!volume)
		return -ENODEV;

	if (volume->disk)
		return -EHASSNAP;

	return detach_volume_once(volume);
}

/* {{{2 List Attached Volumes */

static void volume_to_ioctl_device(struct blksnap_volume *volume,
				   struct blksnap_ioctl_device *bid)
{
	if (test_bit(VOLUME_DIRTY, &volume->flags))
		bid->flags |= BLKSNAP_DEVICE_DIRTY;

	if (blksnap_cl_store_valid(volume->clstore))
		bid->flags |= BLKSNAP_DEVICE_CHLOG;

	if (volume->metadata.extents.head)
		bid->flags |= BLKSNAP_DEVICE_METADATA;

	if (blksnap_data_store_is_valid(volume->data))
		bid->flags |= BLKSNAP_DEVICE_DATA;

	if (snapshot_is_incremental(volume))
		bid->flags |= BLKSNAP_DEVICE_INCR;

	memcpy(bid->name, volume_name(volume), BLKSNAP_NAME_SIZE);

	if (volume->device && volume->device->bdev
	    && volume->device->bdev->bd_disk) {
		char buf[BDEVNAME_SIZE];
	        const char *buf_ptr;
		size_t copy_size;

		/* device_name() can return reference to other object than buf[]
		 */
		buf_ptr = blksnap_device_name(volume->device, buf);

		/* NOTE: memcpy() in kernels >= 4.13 check if size parameter is not large
		 * than size of 2nd object. Following manipulations with copy_size are
		 * required to avoid getting an error:
		 * "detected read beyond size of object passed as 2nd parameter"
		 */
		copy_size = min3((size_t) (strlen(buf_ptr) + 1),
				 (size_t) BDEVNAME_SIZE,
				 (size_t) BLKSNAP_NAME_SIZE);

		memcpy(bid->device_name, buf_ptr, copy_size);
	}

	if (volume->disk)
		memcpy(bid->snapshot_name, blksnap_snapshot_name(volume),
		       BLKSNAP_NAME_SIZE);

	if (volume->link.to)
		memcpy(bid->link_name, volume_name(volume->link.to),
		       BLKSNAP_NAME_SIZE);
}

static int blksnap_ioctl_info(struct blksnap_ioctl_device *device)
{
	struct blksnap_volume *volume;

	volume = lookup_volume_by_path(device->name);
	if (!volume)
		return -ENODEV;

	volume_to_ioctl_device(volume, device);

	return 0;
}

static int blksnap_ioctl_list(struct blksnap_ioctl_list *list)
{
	int ret = 0;
	unsigned int count = 0;
	struct list_head *p, *q;

	list_for_each_safe(p, q, &blksnap_volumes) {
		count++;
	}

	/* if it's a request of devices number or noone device is attached */
	if (list->ndevices == BLKSNAP_DL_REQ || count == 0) {
		list->ndevices = count;
		return ret;
	}

	/* if number of devices are changed since 1st requiest, try again */
	if (list->ndevices != count) {
		return -EAGAIN;
	}

	count = 0;

	list_for_each_safe(p, q, &blksnap_volumes) {
		struct blksnap_volume *volume = list_entry(p,
							  struct blksnap_volume,
							  classmates);
		volume_to_ioctl_device(volume, &list->devices[count]);
		count++;
	}
	list->ndevices = count;

	return ret;
}

/* {{{2 Start Snapshot ioctl */

static int volume_sanity_check(struct blksnap_volume *volume)
{
	sector_t start_sector_cur;
	sector_t start_block_cur;
	sector_t last_block_cur;

	if (!volume || !volume->device)
		return -ENODEV;

	/* XXX: This should never happen. */
	if (!volume->bdev->bd_disk)
		return -ENODEV;

	if (volume->disk)
		return -EHASSNAP;

	start_sector_cur = get_start_sect(volume->bdev);
	start_block_cur = aligned_block_address(start_sector_cur);
#if HAVE_BDEV_NR_SECTORS
	last_block_cur = last_bitmap_block_addr(start_sector_cur, bdev_nr_sectors(volume->bdev));
#else
	last_block_cur = last_bitmap_block_addr(start_sector_cur, volume->bdev->bd_part->nr_sects);
#endif
	if (volume->start_block != start_block_cur
	    || volume->bitmap.last_block != last_block_cur)
	{
		bsmsg_volume(volume, "ERROR: failed to check volume size: start block=%lu, last block=%lu, start block current=%lu, last block curent=%lu",  volume->start_block, volume->bitmap.last_block, start_block_cur, last_block_cur);
		return -ESIZEVOLUME;
	}

	return 0;
}

static int blksnap_ioctl_start(struct blksnap_ioctl_start *start)
{
	int err;
	struct blksnap_volume *volume;
	bool skip_freeze;

	volume = lookup_volume_by_path(start->name);

	err = volume_sanity_check(volume);
	if (err)
		return err;

	if (cl_store_status(volume->clstore) == CL_STORE_LOADING)
		return -EBUSY;

	if (!blksnap_data_store_is_valid(volume->data))
		return -EDTSNOMAP;

	skip_freeze = test_bit(VOLUME_FS_XFS_VERSION_5, &volume->fs_flags);

	if (!skip_freeze) {
		if (IS_ERR(freeze_volume(volume))) {
			bsmsg_volume(volume, "ERROR: failed to freeze filesystem");
			return -ENXIO;
		}
	}

	bsmsg_volume(volume, "Starting snapshot...");

	/* Update metadata file for snapshot to not being dirty after restore */
	update_metadata(volume, BLKSNAP_META_UPDATE);

	err = start_snapshot(volume, start);

	/* Reset metadata file on original volume so reset/power failure during
	 * snapshot doesn't put volume into a clean mode after next attach */
	update_metadata(volume, BLKSNAP_META_RESET);

	if (!skip_freeze) {
		if (thaw_volume(volume)) {
			bsmsg_volume(volume, "ERROR: failed to thaw filesystem");
			err = -ENXIO;
		}
	}

	if (err)
		return err;

	/* Tell userspace the name of the snapshot block device */
	memcpy(start->snapshot_name, blksnap_snapshot_name(volume), BLKSNAP_NAME_SIZE);

	return 0;
}

/* {{{2 Stop Snapshot ioctl */

static int blksnap_ioctl_stop(struct blksnap_ioctl_stop *stop)
{
	struct blksnap_volume *volume;

	volume = lookup_volume_by_path(stop->name);
	if (!volume || !volume->device)
		return -ENODEV;

	if (!volume->disk)
		return -ENOSNAP;

	if (atomic_read(&volume->snapshot_holders) > 0)
		return -EBUSY;

	bsmsg_volume(volume, "Stopping snapshot...");

	return stop_snapshot(volume, stop->merge != 0);
}

/* }}}2 */

/* {{{2 Show Volume Information */

static int blksnap_ioctl_show(struct blksnap_ioctl_show *show)
{
	struct blksnap_volume *volume;

	volume = lookup_volume_by_path(show->name);
	if (!volume || !volume->device)
		return -ENODEV;

	if (!volume->disk)
		return -ENOSNAP;

	memcpy(show->snapshot_name, blksnap_snapshot_name(volume), BLKSNAP_NAME_SIZE);

	if (volume->link.to)
		memcpy(show->link_name, volume_name(volume->link.to),
		       BLKSNAP_NAME_SIZE);

	return 0;
}

/* {{{2 Changelog */

static void do_ioctl_changelog(struct blksnap_volume *volume,
			       struct blksnap_ioctl_changelog *changelog)
{
	sector_t offset = 0;

	if (changelog->offset)
		offset = volume->start_block + changelog->offset + 1;

	changelog->nextents =
		blksnap_get_changelog_extents(volume,
					      changelog->extents,
					      changelog->nextents,
					      offset);
}

static int blksnap_ioctl_changelog(struct blksnap_ioctl_changelog *changelog)
{
	struct blksnap_volume *volume;

	volume = lookup_volume_by_path(changelog->name);
	if (!volume)
		return -ENODEV;

	if (test_bit(VOLUME_DIRTY, &volume->flags))
		return -EDIRTY;

	if (changelog->nextents == 0) {
		if (changelog->offset != 0)
			return -EFAULT;

		/* nblocks set to the number of nodes in the tree each
		 * node of tree refers to one disk block exactly so
		 * (the size of tree == number of blocks for current changelog)
		 */
		changelog->nblocks = blksnap_get_changelog_size(volume);
		return 0;
	}

	do_ioctl_changelog(volume, changelog);
	return 0;
}

/* {{{2 Map Store */
static int blksnap_ioctl_map_chlog(struct blksnap_volume *volume,
				   struct blksnap_ioctl_map *map)
{
	int ret = 0;
	int ops = MAP_RESET | MAP_WRITE;

	if (is_first_map_ioctl(map) &&
	    cl_store_status(volume->clstore) == CL_STORE_LOADING)
		return -EBUSY;

	/* If volume is NOT dirty when cl store is mapping, it means that
	 * base snapshot has been performed or metadata was updated, so
	 * we SHOULDN'T MERGE this store, just RESET and STORE.
	 * Also, we shouldn't merge it if snapshot is in progress, because
	 * it means that changelog in store is obsolete.
	 */
	if (test_bit(VOLUME_DIRTY, &volume->flags)
	    && !test_bit(VOLUME_CHANGED, &volume->flags)
	    && !volume_has_snapshot(volume))
		ops = MAP_READ | MAP_WRITE;

	ret = cl_store_map(volume, map, ops);

	/* If metadata is mapped at that moment and disk is not changed since
	 * last attach, consider disk is clean and incremental snapshot
	 * can be performed
	 */
	if (is_last_map_ioctl(map)) {
		if (!ret && volume->metadata.extents.head &&
		    !test_bit(VOLUME_CHANGED, &volume->flags))
			clear_bit(VOLUME_DIRTY, &volume->flags);

		/* if metadata is mapped, process bitmap and prove DIRTY flag */
		if (volume->metadata.extents.head) {
			set_bit(DIRTY_FLAG_PROVEN, &volume->flags);

			/* If disk was changed being unprotected, changelog goes
			 * to hell. If snapshot is in progress, changelog wasn't
			 * read from file, so RAM-changelog is valid. */
			if (test_bit(VOLUME_DIRTY, &volume->flags)
			    && !volume_has_snapshot(volume))
				ret = blksnap_blk_bitmap_reset(volume);
		}
	}

	return ret;
}

static int blksnap_ioctl_map_data(struct blksnap_volume *volume,
				  struct blksnap_ioctl_map *map)
{
	if (volume->disk) {
		return -EHASSNAP;
	}

	return blksnap_data_store_map(volume->data, map);
}

static int blksnap_ioctl_unmap_data(struct blksnap_volume *volume)
{
	if (!volume) {
		return -EINVAL;
	}

	if (volume->disk) {
		return -EHASSNAP;
	}

	return blksnap_data_store_unmap(volume->data);
}

static int blksnap_ioctl_map_metadata(struct blksnap_volume *volume,
				      struct blksnap_ioctl_map *map)
{
	struct blksnap_meta_on_disk blk_meta;
	struct blksnap_extent *new;
	int result = 0;
	int ext_idx;

	/* Return if we have already mapped metadata extents */
	if (volume->metadata.extents.head)
		return -EEXIST;

	/* We do not support files that requires more than one ioctl calls to
	 * process all extents */
	if (!is_first_map_ioctl(map) || !is_last_map_ioctl(map)) {
		bsmsg_volume(volume, "ERROR: Metadata file is too big");
		return -EINVAL;
	}

	/* Add fiemap to blksnap volume */
	for (ext_idx = 0; ext_idx < map->nextents; ext_idx++) {
		new = __blksnap_extent_add(map->extents[ext_idx].start_sect,
					   map->extents[ext_idx].nsectors,
					   volume->metadata.extents.head);
		if (IS_ERR(new)) {
			blksnap_extent_free(volume->metadata.extents.head);
			volume->metadata.extents.head = NULL;
			return PTR_ERR(new);
		}
		volume->metadata.extents.head = new;
	}

	blksnap_extent_show(volume->metadata.extents.head, volume->bdev,
			    "metadata");

	/* if volume is not dirty at that point, so we sure as minimum
	 * one snapshot was finished successfully, so we can clear
	 * VOLUME_CHANGED flag
	 */
	if (!test_bit(VOLUME_DIRTY, &volume->flags)) {
		clear_bit(VOLUME_CHANGED, &volume->flags);
		/* Input: Dirty=0 */
		return 0;
	}

	/* At that point if VOLUME_CHANGED so the only way to have it is
	 * mounted in rw-mode volume before attach, so even flag
	 * UPDATE_METADATA should be ignored.
	 */
	if (test_bit(VOLUME_CHANGED, &volume->flags)) {
		bsmsg_volume(volume, "Volume was changed, set it as dirty");
		/* Input: Dirty=1, Changed=1 */
		goto changed;
	}

	/* Update metadata doesn't compare sb-hash with previous stored one.
	 * it uses after roll-back operation and always mark volume not dirty
	 * if it was attached before mount in rw-mode
	 */
	if (map->flags & BLKSNAP_STORE_UPDATE_METADATA) {
		/* Forced update - ignore dirty and mount state */
		if (update_metadata(volume, BLKSNAP_META_UPDATE)) {
			bsmsg_volume(volume, "ERROR: Failed to update metadata");
			return -EIO;
		}

		clear_bit(VOLUME_DIRTY, &volume->flags);
		set_bit(DIRTY_FLAG_PROVEN, &volume->flags);
		/* Input: Dirty=1, Changed=0, Update=1 */
		return 0;
	}

	/* Reading sha1sum from metadata */
	result = process_metadata(volume->bdev, &volume->metadata,
				  &blk_meta, BLKSNAP_READ);

	if (result) {
		bsmsg_volume(volume,
			     "ERROR: Unable to read metadata file [%d]",
			     result);
		set_bit(VOLUME_CHANGED, &volume->flags);
		/* Input: Dirty=1, Changed=Error=1 */
		goto changed;
	}

	D("%s: Got hash from metadata: %x %x %x %x", volume_name(volume),
	  blk_meta.hash[0], blk_meta.hash[1],
	  blk_meta.hash[2], blk_meta.hash[3]);

	if (hash_sums_equals(blk_meta.hash, volume->metadata.data.hash)) {
		bsmsg_volume(volume, "Volume was not changed");
		if (blksnap_cl_store_valid(volume->clstore)) {
			clear_bit(VOLUME_DIRTY, &volume->flags);
			set_bit(DIRTY_FLAG_PROVEN, &volume->flags);
		}
		/* Input: Dirty=1, Changed=0, Hash=ok */
		return 0;
	}

	bsmsg_volume(volume, "Volume was changed, set it as dirty");
	set_bit(VOLUME_CHANGED, &volume->flags);

changed:
	/* Got here just if Dirty=1, Changed=1 */
	set_bit(VOLUME_DIRTY, &volume->flags);
	set_bit(DIRTY_FLAG_PROVEN, &volume->flags);

	if (blksnap_cl_store_valid(volume->clstore)) {
		/* If volume is dirty at this moment, we can safely purge
		 * chlog-store and bitmap as the next snapshot will be BI.
		 * But if snapshot is in progress, changelog contains blocks
		 * changed since snapshot start and can't be reset. In this
		 * case changelog after snapshot ending might contain some more
		 * blocks, but so what...
		 */
		if (!volume_has_snapshot(volume)) {
			cl_store_reset(volume->clstore);
			result = blksnap_blk_bitmap_reset(volume);
		}
	}

	return result;
}

static int blksnap_ioctl_map(struct blksnap_ioctl_map *map)
{
	int ret;
	struct blksnap_volume *volume;

	volume = lookup_volume_by_path(map->name);

	ret = volume_sanity_check(volume);
	if (ret)
		return ret;

	switch (map->type) {
	case BLKSNAP_MAP_TYPE_CHLOG:
		ret = blksnap_ioctl_map_chlog(volume, map);
		break;
	case BLKSNAP_MAP_TYPE_DATA:
		ret = blksnap_ioctl_map_data(volume, map);
		break;
	case BLKSNAP_MAP_TYPE_METADATA:
		ret = blksnap_ioctl_map_metadata(volume, map);
		break;
	default:
		ret = -EINVAL;
		break;
	}

	return ret;
}

/* BLKSNAP_GRAB_MAP_START:
 *   Return error if grabbing is already started or extents storage isn't empty
 *   Turn on grabbing mode on a volume
 * BLKSNAP_GRAB_MAP_END:
 *   Return error if grabbing wasn't started or extents storage is empty
 *   Turn off grabbing mode on a volume
 *   Return number of grabbed extents to userspace
 * BLKSNAP_GRAB_MAP_FETCH_EXT:
 *   Return error if grabbing mode is on or extents storage is empty
 *   Copy extents from temporary storage to userspace and remove them
 * BLKSNAP_GRAB_MAP_CANCEL:
 *   Return error if grabbing mode is off
 *   Turn off grabbing mode for a volume
 *   Remove grabbed extents from temporary storage in volume structure
 */
static int blksnap_ioctl_grab_map(struct blksnap_ioctl_grab_map *grab_map)
{
	int ret;
	struct blksnap_volume *volume;

	volume = lookup_volume_by_path(grab_map->name);

	ret = volume_sanity_check(volume);
	if (ret)
		return ret;

	switch (grab_map->cmd) {
	case BLKSNAP_GRAB_MAP_START:
		if (volume_is_grabbing_map(volume))
			return -EBUSY;

		if (volume->grabbing_extents.head) {
			bsmsg_volume(volume,
				     "ERROR: some extents are already grabbed");
			return -EINVAL;
		}

		set_bit(VOLUME_GRAB_MAP, &volume->flags);
		bsmsg_volume(volume, "File map grabbing has started");

		return 0;
	case BLKSNAP_GRAB_MAP_END:
		if (!volume_is_grabbing_map(volume)) {
			bsmsg_volume(volume,
				"ERROR: attempt to finish grabbing not being started");
			return -EINVAL;
		}

		clear_bit(VOLUME_GRAB_MAP, &volume->flags);

		if (!volume->grabbing_extents.head) {
			bsmsg_volume(volume,
				"ERROR: there are no grabbed extents on grabbing end");
			return -EIO;
		}

		grab_map->nextents =
			blksnap_extents_count(volume->grabbing_extents.head);
		bsmsg_volume(volume, "File extents map grabbing ended");

		return 0;
	case BLKSNAP_GRAB_MAP_FETCH_EXT: {
		/* TODO: In case of plenty extents - multi-ioctl?
		 *	 Investigation is needed. Page sized limit or 4Mb limit?
		 */
		int i = 0;
		struct blksnap_extent *extent;

		if (volume_is_grabbing_map(volume)) {
			bsmsg_volume(volume,
				"ERROR: %s: grabbed extents can't be fetched because grabbing wasn't finished", __func__);
			return -EINVAL;
		}
		if (!volume->grabbing_extents.head) {
			bsmsg_volume(volume,
				"ERROR: %s: there are no extents to fetch", __func__);
			return -ESPIPE;
		}
		/* number of requested extents must be valid */
		if (grab_map->nextents !=
		    blksnap_extents_count(volume->grabbing_extents.head)) {
			bsmsg_volume(volume,
				     "ERROR: %s: number of requested extents isn't match number of grabbed ones", __func__);
			return -EINVAL;
		}

		for (extent = volume->grabbing_extents.head; extent;
		     extent = extent->next, i++) {
			grab_map->extents[i].start_sect = extent->start;
			grab_map->extents[i].nsectors = extent->nsectors;
		}

		mutex_lock(&volume->grabbing_extents.lock);
		blksnap_extent_free(volume->grabbing_extents.head);
		volume->grabbing_extents.head = 0;
		mutex_unlock(&volume->grabbing_extents.lock);

		return 0;
	}
	case BLKSNAP_GRAB_MAP_CANCEL:
		if (!volume_is_grabbing_map(volume)) {
			bsmsg_volume(volume, "ERROR: grabbing can't be cancelled because hasn't started");
			return -EINVAL;
		}

		bsmsg_volume(volume, "File extents map grabbing cancelled");

		clear_bit(VOLUME_GRAB_MAP, &volume->flags);

		mutex_lock(&volume->grabbing_extents.lock);
		blksnap_extent_free(volume->grabbing_extents.head);
		volume->grabbing_extents.head = 0;
		mutex_unlock(&volume->grabbing_extents.lock);

		return 0;
	default:
		bsmsg_volume(volume, "ERROR: Unknown command in %s()", __func__);
		BUG();
	}

}

static int blksnap_ioctl_get_map(struct blksnap_ioctl_map *map)
{
	int i = 0;
	int ret = 0;
	int store_nextents;
	struct blksnap_volume *volume;
	struct blksnap_extent *store_extents, *extent;

	volume = lookup_volume_by_path(map->name);

	ret = volume_sanity_check(volume);
	if (ret)
		return ret;

	switch (map->type) {
	case BLKSNAP_MAP_TYPE_CHLOG:
		if (!blksnap_cl_store_valid(volume->clstore))
			return -ECLSNOMAP;
		store_extents = volume->clstore->extents.head;
		store_nextents =
			blksnap_extents_count(volume->clstore->extents.head);
		break;
	case BLKSNAP_MAP_TYPE_DATA:
		if (!blksnap_data_store_is_valid(volume->data))
			return -EDTSNOMAP;
		store_extents = volume->data->extents.head;
		store_nextents =
			blksnap_extents_count(volume->data->extents.head);
		break;
	case BLKSNAP_MAP_TYPE_METADATA:
		if (!volume->metadata.extents.head)
			return -EMDSNOMAP;
		store_extents = volume->metadata.extents.head;
		store_nextents =
			blksnap_extents_count(volume->metadata.extents.head);
		break;
	default:
		return -EINVAL;
	}

	if (map->nextents == 0) {
		map->nextents = store_nextents;
		return ret;
	}
	if (map->nextents != store_nextents)
		return -EAGAIN;
	for (extent = store_extents; extent; extent = extent->next, i++) {
		map->extents[i].start_sect = extent->start;
		map->extents[i].nsectors = extent->nsectors;
	}
	return ret;
}

static int blksnap_ioctl_unmap(struct blksnap_ioctl_unmap *unmap)
{
	int ret;
	struct blksnap_volume *volume;

	volume = lookup_volume_by_path(unmap->name);

	ret = volume_sanity_check(volume);
	if (ret)
		return ret;

	switch (unmap->type) {
	case BLKSNAP_UNMAP_TYPE_DATA:
		ret = blksnap_ioctl_unmap_data(volume);
		break;
	default:
		ret = -EINVAL;
		break;
	}

	return ret;
}

/* {{{2 Dump Tree Contents */

static int blksnap_ioctl_dump(struct blksnap_ioctl_dump *dump)
{
	struct blksnap_volume *volume;

	volume = lookup_volume_by_path(dump->name);
	if (!volume)
		return -ENODEV;

	blksnap_show_changelog(volume);

	return 0;
}

/* {{{2 Restore Volume */

static void volume_link(struct blksnap_volume *source,
			struct blksnap_volume *target)
{
	BUG_ON(!source);
	BUG_ON(!target);

	source->link.to = target;
	source->link.direction = LINK_FROM;

	target->link.to = source;
	target->link.direction = LINK_TO;
}

static void volume_unlink(struct blksnap_volume *source,
			  struct blksnap_volume *target)
{
	BUG_ON(!source);
	BUG_ON(!target);

	source->link.to = NULL;
	source->link.direction = LINK_NONE;

	target->link.to = NULL;
	target->link.direction = LINK_NONE;
}

static int do_ioctl_restore(struct blksnap_volume *source,
			    struct blksnap_volume *target)
{
	if (!source)
		return -ENOLINK;

	if (!target)
		return -ENODEV;

	/* Restore volumes must be identical in size */
	if (volume_capacity(source) != volume_capacity(target))
		return -EFAULT;

	volume_link(source, target);
	return 0;
}

static int blksnap_ioctl_restore(struct blksnap_ioctl_restore *restore)
{
	int err;
	struct blksnap_volume *source, *target;

	if (!restore)
		return -EINVAL;

	D("received restore command from %s to %s",
	  restore->source_name, restore->target_name);

	source = lookup_volume_by_path(restore->source_name);

	if (!source)
		return -ENOLINK;

	err = volume_sanity_check(source);
	if (err)
		return err;

	/* TODO: target must not be mounted */

	target = lookup_volume_by_path(restore->target_name);

	err = volume_sanity_check(target);
	if (err)
		return err;

	return do_ioctl_restore(source, target);
}

/* {{{2 Reset Volume */

static int blksnap_ioctl_reset(struct blksnap_ioctl_reset *reset)
{
	struct blksnap_volume *volume;

	volume = lookup_volume_by_path(reset->name);
	if (!volume)
		return -ENODEV;

	if (volume->link.to == NULL)
		return -ENOLINK;

	volume_unlink(volume, volume->link.to);
	return 0;
}

/* {{{2 Check Attached Device */

static int blksnap_ioctl_check(struct blksnap_ioctl_check *check)
{
	struct blksnap_volume *volume;

	volume = lookup_volume_by_path(check->name);
	if (!volume)
		return -ENODEV;

	return volume_sanity_check(volume);
}

/* {{{2 Ioctl Routing */

static int __blksnap_ioctl(unsigned int cmd, struct blksnap_ioctl __user *user)
{
	int err;
	size_t len = sizeof(struct blksnap_ioctl);
	struct blksnap_ioctl *bsi, *tmp;

	bsi = kmalloc(len, GFP_KERNEL);
	if (!bsi)
		return -ENOMEM;

	if (copy_from_user(bsi, user, len)) {
		err = -EFAULT;
		goto out;
	}

	if (bsi->version != BLKSNAP_IOCTL_VERSION) {
		err = -EVERSION;
		goto out;
	}

	if (len < bsi->size) {
		len = bsi->size;

		tmp = krealloc(bsi, len, GFP_KERNEL);
		if (!tmp) {
			err = -ENOMEM;
			goto out;
		}
		bsi = tmp;

		if (copy_from_user(bsi, user, len)) {
			err = -EFAULT;
			goto out;
		}
	}

	switch (cmd) {
	case BLKSNAP_IOC_ATTACH:
		err = blksnap_ioctl_attach(&bsi->attach);
		break;
	case BLKSNAP_IOC_DETACH:
		err = blksnap_ioctl_detach(&bsi->detach);
		break;
	case BLKSNAP_IOC_INFO:
		err = blksnap_ioctl_info(&bsi->device);
		break;
	case BLKSNAP_IOC_LIST:
		err = blksnap_ioctl_list(&bsi->list);
		break;
	case BLKSNAP_IOC_START:
		err = blksnap_ioctl_start(&bsi->start);
		break;
	case BLKSNAP_IOC_STOP:
		err = blksnap_ioctl_stop(&bsi->stop);
		break;
	case BLKSNAP_IOC_SHOW:
		err = blksnap_ioctl_show(&bsi->show);
		break;
	case BLKSNAP_IOC_CHANGELOG:
		err = blksnap_ioctl_changelog(&bsi->changelog);
		break;
	case BLKSNAP_IOC_MAP:
		err = blksnap_ioctl_map(&bsi->map);
		break;
	case BLKSNAP_IOC_GRAB_MAP:
		err = blksnap_ioctl_grab_map(&bsi->grab);
		break;
	case BLKSNAP_IOC_GET_MAP:
		err = blksnap_ioctl_get_map(&bsi->map);
		break;
	case BLKSNAP_IOC_UNMAP:
		err = blksnap_ioctl_unmap(&bsi->unmap);
		break;
	case BLKSNAP_IOC_DUMP:
		err = blksnap_ioctl_dump(&bsi->dump);
		break;
	case BLKSNAP_IOC_RESTORE:
		err = blksnap_ioctl_restore(&bsi->restore);
		break;
	case BLKSNAP_IOC_RESET:
		err = blksnap_ioctl_reset(&bsi->reset);
		break;
	case BLKSNAP_IOC_CHECK:
		err = blksnap_ioctl_check(&bsi->check);
		break;
	default:
		err = -ENOTTY;
		break;
	}

	if (err)
		goto out;

	if (copy_to_user(user, bsi, len))
		err = -EFAULT;

out:
	kfree(bsi);
	return err;
}

static bool ioctl_trylock(void)
{
	unsigned int tries;

	for (tries = IOCTL_LOCK_TRIES; tries > 0; tries--) {
		if (mutex_trylock(&blksnap_ioctl_lock))
			return true;
		msleep(IOCTL_LOCK_SLEEP_MS);
	}

	return false;
}

static void ioctl_unlock(void)
{
	mutex_unlock(&blksnap_ioctl_lock);
}

static void ioctl_lock(void)
{
	mutex_lock(&blksnap_ioctl_lock);
}

static long blksnap_ioctl(struct file *file, unsigned int cmd,
			  unsigned long arg)
{
	int ret;

#ifndef BLKSNAP_NO_CAP_CHECK
	/* Elevated privileges are required to manipulate the kernel component
	 * unless stated otherwise. */
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;
#endif /* BLKSNAP_NO_CAP_CHECK */

	if (_IOC_TYPE(cmd) != BLKSNAP_IOCTL_MAGIC ||
	    _IOC_NR(cmd) > BLKSNAP_IOCTL_CMD_LAST)
		return -ENOTTY;

	if (!ioctl_trylock())
		return -EBUSY;

	ret = __blksnap_ioctl(cmd, (struct blksnap_ioctl __user *) arg);

	ioctl_unlock();
	return (long) ret;
}

/* {{{1 Misc Device */

static const struct file_operations blksnap_misc_fops = {
	.unlocked_ioctl = blksnap_ioctl,
	.compat_ioctl = blksnap_ioctl,
	.owner = THIS_MODULE,
	.llseek = noop_llseek,
};

static struct miscdevice blksnap_misc = {
	.minor = MISC_DYNAMIC_MINOR,
	.name = BLKSNAP_NAME,
	.fops = &blksnap_misc_fops,
};

/* {{{1 Setup/Teardown Routines */

static void detach_all_volumes(void)
{
	struct list_head *p, *q;

	list_for_each_safe(p, q, &blksnap_volumes) {
		struct blksnap_volume *volume =
			list_entry(p, struct blksnap_volume, classmates);
		detach_volume(volume);
	}
}

/* Define of reboot notifier callback */
static int reboot_notify(struct notifier_block *,
			 unsigned long, void *);

static struct notifier_block blksnap_notifier = {
	.notifier_call = reboot_notify,
	.next = NULL
};

static void __blksnap_free(void)
{
	/* Prevent to call this function twice */
	if (atomic_inc_return(&destroyed) > 1)
		return;

	detach_all_volumes();
	misc_deregister(&blksnap_misc);
	unregister_blkdev(blksnap_major, BLKSNAP_NAME);

	blksnap_extent_cache_destroy();
	blksnap_cl_cmd_cache_destroy();

}

static void teardown(void)
{
	/* Lock mutex to avoid call any ioctl during teardown */
	ioctl_lock();

	__blksnap_free();

	ioctl_unlock();

	/* Unregistering from reboot notifier */
	unregister_reboot_notifier(&blksnap_notifier);
}

/* Reboot notifier callback implemetation. This fuction is is calling during
 * reboot and shutdown syscalls.
 */
static int reboot_notify(struct notifier_block *this, unsigned long event,
			 void * buf)
{
	/* For now we have to detach all disks to store fresh metadata
	 */
	__blksnap_free();

	return NOTIFY_DONE;
}

/* For kernel versions that differ only slightly, it is possible to load
 * a module compiled for one version of the kernel into a non-matching version.
 * If this occurs, the module will load successfully but may exhibit strange
 * behaviour.
 */
static int check_kernel_version(void)
{
	char *runtime_version = utsname()->release;

	if (strcmp(compile_version, runtime_version)) {
		bsmsg("ERROR: Kernel version mismatch (%s != %s)",
		      compile_version, runtime_version);
		return -EFAULT;
	}

	return 0;
}

static bool is_file_exist(const char *name)
{
	struct file *fp;

	fp = filp_open(name, O_RDONLY, 0);

	if (fp == ERR_PTR(-ENOENT) )
		return false;

	if (!IS_ERR(fp))
		filp_close(fp, NULL);

	return true;
}

static int __init blksnap_init(void)
{
	int ret = 0;

	/* If the module should be disabled, load silently and do nothing. */
	if (disabled) {
		bsmsg("disabled from command line");
		return 0;
	}

	bsmsg("sizeof(struct rw_semaphore) = %lu", sizeof(struct rw_semaphore));
	bsmsg("sizeof(struct bitmap_chunk) = %lu", sizeof(struct bitmap_chunk));
	ret = check_kernel_version();
	if (ret)
		return ret;

	atomic_set(&blksnap_next_id, 0);
	atomic_set(&destroyed, 0);

	/* Lock mutex to avoid call any ioctl during init */
	ioctl_lock();

	if (is_file_exist("/dev/" BLKSNAP_NAME)) {
		ret = -EEXIST;
		/* nothing to teardown, so just go to out */
		goto out;
	}

	ret = misc_register(&blksnap_misc);
	if (ret) {
		/* Cannot call teardown() here, misc_deregister() will fault if
		 * blksnap_misc wasn't successfully registered. */
		goto out;
	}

	ret = blksnap_major = register_blkdev(0, BLKSNAP_NAME);
	if (blksnap_major < 0)
		goto err;

	/* Registering callback for reboot\shutdown syscalls */
	register_reboot_notifier(&blksnap_notifier);

	ret = blksnap_extent_cache_create();
	if (ret)
		goto err;

	ret = blksnap_cl_cmd_cache_create();
	if (ret)
		goto err;

	bsmsg("loaded, version %s (API version %u)",
	      BLKSNAP_VERSION, BLKSNAP_IOCTL_VERSION);
	bsmsg("block object size: %zu", sizeof(struct blksnap_cow_block));
	bsmsg("registered block device at major %d", blksnap_major);

	goto out;

err:
	teardown();
out:
	ioctl_unlock();
	return ret;
}

static void __exit blksnap_exit(void)
{
	if (disabled)
		return;

	teardown();
	bsmsg("unloaded");
}

module_init(blksnap_init);
module_exit(blksnap_exit);

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION("Rapid Recovery block device snapshot driver");
MODULE_VERSION(BLKSNAP_VERSION);
MODULE_LICENSE(LICENSE);
