#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/vmalloc.h>
#include <linux/blkdev.h>

#include "blksnap.h"
#include "blksnap-bio-queue-share.h"
#include "blksnap-hash.h"
#include "blksnap-device.h"
#include "blksnap-compat.h"
#include "blksnap-shared.h"
#include "blksnap-device-utils.h"
#include "blksnap-volume.h"
#include "blksnap-bio-submit.h"
#include "blksnap-bio.h"
#include "blksnap-extent.h"
#include "blksnap-metadata.h"

bool hash_sums_equals(__u32 left[], __u32 right[])
{
	return ( left[0] == right[0]) &&
		(left[1] == right[1]) &&
		(left[2] == right[2]) &&
		(left[3] == right[3]);
}

int get_blocks_hash(struct block_device *bdev,
		    struct blksnap_metadata *metadata)
{
	int ret = 0;
	unsigned char *data_buffer;
	uint64_t data_buffer_size = BIO_MAX_SIZE;
	uint64_t data_total_size = 0;
	uint64_t data_read_size = 0;
	struct blksnap_extent *extent;
	sector_t start_sect;
	size_t len_sect = 0;
	struct blksnap_hasher hasher;

	D("Using new blksnap hasher.");
	data_total_size = blksnap_extents_size(metadata->hashblocks);
	blksnap_hasher_init(&hasher, data_total_size);

	/* Create data buffer */
	data_buffer = vmalloc(data_buffer_size);
	if (!data_buffer) {
		bsmsg_bdev(bdev,
			   "ERROR: %s: Failed to allocate read buffer for blocks to calc hash",
			   __func__);
		return -ENOMEM;
	}

	for (extent = metadata->hashblocks; extent; extent = extent->next) {
		start_sect = volume_relative_address(bdev, extent->start);
		len_sect = extent->nsectors;

		while (len_sect) {
			data_read_size = ((len_sect << SECTOR_SHIFT) > data_buffer_size) ? data_buffer_size : (len_sect << SECTOR_SHIFT);
			ret = blksnap_disk_io(bdev, start_sect, data_read_size,
					      data_buffer, BLKSNAP_READ);
			if (ret) {
				bsmsg_bdev(bdev, "ERROR: Unable to read disk for metadata hash [%d]", ret);
				goto free_buffer;
			}

			blksnap_hasher_update(&hasher, data_buffer, data_read_size);

			start_sect += data_read_size >> SECTOR_SHIFT;
			len_sect -= data_read_size >> SECTOR_SHIFT;
		}
	}

	blksnap_hasher_finalize(&hasher, metadata->data.hash);

	D("Calculated hash sum: %x %x %x %x",
	  metadata->data.hash[0], metadata->data.hash[1],
	  metadata->data.hash[2], metadata->data.hash[3]);

free_buffer:
	vfree(data_buffer);

	return ret;
}

int process_metadata(struct block_device *bdev,
		     struct blksnap_metadata *metadata,
		     struct blksnap_meta_on_disk *data,
		     int op)
{
	int ret = 0;
	struct blksnap_extent *extent;
	unsigned char *data_position;
	size_t extent_size, req_size, remain_size;
	sector_t start_sector;
	uint32_t data_size = sizeof(struct blksnap_meta_on_disk);
	uint64_t extents_size = 0;

	extents_size = blksnap_extents_size(metadata->extents.head);

	/* size of struct blksnap_meta_on_disk is larger than size of metadata
	 * file? */
	if (WARN_ON(data_size > extents_size))
		return -E2BIG;

	data_position = (unsigned char *) data;
	remain_size = data_size;

	for (extent = metadata->extents.head; extent; extent = extent->next) {
		start_sector = volume_relative_address(bdev, extent->start);
		extent_size = extent->nsectors << SECTOR_SHIFT;

		req_size = MIN(extent_size, remain_size);
		ret = blksnap_disk_io(bdev, start_sector, req_size,
				      data_position, op);
		if (ret) {
			bsmsg_bdev(bdev, "ERROR: Unable to process metadata extent [%d]", ret);
			goto out;
		}

		remain_size -= req_size;
		data_position += req_size;

		/* Data position should be in sectors */
		if (!remain_size)
			break;
	}

out:
	return ret;
}

/* Update metadata store with hash of current hash blocks
 * If force is true - ignore dirty flag and mount-state of volume
 */
int update_metadata(struct blksnap_volume *volume, int mode)
{
	struct blksnap_metadata *metadata;
	int ret = 0;

	metadata = &volume->metadata;

	if (unlikely(!metadata->extents.head)) {
		bsmsg_volume(volume, "WARNING: %s: Metadata is not mapped",
			     __func__);
		return -ENOENT;
	}

	if (mode == BLKSNAP_META_RESET) {
		bsmsg_volume(volume, "Reset metadata");
		memset(&metadata->data.hash, 0x0, BLKSNAP_HASH_DIGEST_BYTES);
	} else if (mode == BLKSNAP_META_UPDATE) {
		bsmsg_volume(volume, "Update metadata");
		ret = get_blocks_hash(volume->bdev, metadata);
		if (ret) {
			bsmsg_volume(volume, "ERROR: Failed to get blocks hash, Reset metadata [%d]", ret);
			memset(&metadata->data.hash, 0x0, BLKSNAP_HASH_DIGEST_BYTES);
		}
	} else {
		bsmsg_volume(volume, "Store metadata");
	}

	ret = process_metadata(volume->bdev, metadata, &metadata->data,
			       BLKSNAP_WRITE);
	if (ret)
		bsmsg_volume(volume, "ERROR: Unable to store metadata");

	return ret;
}

