#include <linux/kernel.h>
#include <linux/blkdev.h>

#include "blksnap.h"
#include "blksnap-bio-queue-share.h"
#include "blksnap-hash.h"
#include "blksnap-device.h"
#include "blksnap-compat.h"
#include "blksnap-shared.h"
#include "blksnap-device-utils.h"
#include "blksnap-volume.h"

enum bsmsg_type {
	BSMSG_COMMON,
	BSMSG_BDEV,
	BSMSG_VOLUME
};

#if !HAVE_VA_FORMAT

static void _bsmsg(enum bsmsg_type type, const void *spec_arg, const char *fmt,
		   va_list *args)
{
	struct block_device *bdev;
	const struct blksnap_volume *volume;
	char buf[BDEVNAME_SIZE];

	switch (type) {
	case BSMSG_COMMON:
		printk("%s: ", BLKSNAP_LOG_PREFIX);
		break;
	case BSMSG_BDEV:
		bdev = (struct block_device *) spec_arg;
		printk("%s: %s [%u:%u]: ", BLKSNAP_LOG_PREFIX,
		       blksnap_bdev_get_name(bdev, buf),
		       bdev ? MAJOR(bdev->bd_dev) : 0,
		       bdev ? MINOR(bdev->bd_dev) : 0);
		break;
	case BSMSG_VOLUME:
		volume = spec_arg;
		printk("%s: %s [%u:%u]: ", BLKSNAP_LOG_PREFIX,
		       volume_name(volume),
		       volume ? (volume->bdev ? MAJOR(volume->bdev->bd_dev) : 0)
			      : 0,
		       volume ? (volume->bdev ? MINOR(volume->bdev->bd_dev) : 0)
			      : 0);
		break;
	}

	vprintk(fmt, *args);
	printk("\n");
}

#else /* HAVE_VA_FORMAT */

static void _bsmsg(enum bsmsg_type type, void *spec_arg, const char *fmt,
		   va_list *args)
{
	struct va_format vaf;
	struct block_device *bdev;
	const struct blksnap_volume *volume;
	char buf[BDEVNAME_SIZE];

	vaf.fmt = fmt;
	vaf.va = args;
	switch (type) {
	case BSMSG_COMMON:
		printk("%s: %pV\n", BLKSNAP_LOG_PREFIX, &vaf);
		break;
	case BSMSG_BDEV:
		bdev = spec_arg;
		printk("%s: %s [%u:%u]: %pV\n", BLKSNAP_LOG_PREFIX,
		       blksnap_bdev_get_name(bdev, buf),
		       bdev ? MAJOR(bdev->bd_dev) : 0,
		       bdev ? MINOR(bdev->bd_dev) : 0, &vaf);
		break;
	case BSMSG_VOLUME:
		volume = spec_arg;
		printk("%s: %s [%u:%u]: %pV\n", BLKSNAP_LOG_PREFIX,
		       volume_name(volume),
		       volume ? (volume->bdev ? MAJOR(volume->bdev->bd_dev) : 0)
			      : 0,
		       volume ? (volume->bdev ? MINOR(volume->bdev->bd_dev) : 0)
			      : 0, &vaf);
		break;
	}
}

#endif /* HAVE_VA_FORMAT */

void bsmsg(const char *fmt, ...)
{
	va_list args;

	if (warn_null_pointer(fmt, __func__)) return;

	va_start(args, fmt);
	_bsmsg(BSMSG_COMMON, NULL, fmt, &args);
	va_end(args);
}

void bsmsg_bdev(const struct block_device *bdev, const char *fmt, ...)
{
	va_list args;

	if (warn_null_pointer(fmt, __func__)) return;

	va_start(args, fmt);
	_bsmsg(BSMSG_BDEV, (void *) bdev, fmt, &args);
	va_end(args);
}

void bsmsg_volume(const struct blksnap_volume *volume, const char *fmt, ...)
{
	va_list args;

	if (warn_null_pointer(fmt, __func__)) return;

	va_start(args, fmt);
	_bsmsg(BSMSG_VOLUME, (void *) volume, fmt, &args);
	va_end(args);
}

