#ifndef _BLKSNAP_SHARED_H
#define _BLKSNAP_SHARED_H

#include <linux/blkdev.h>
#include <blksnap.h>

#include "blksnap-compat.h"
#include "blksnap-device.h"

#define SECTOR_SHIFT		9
#define PAGE_SHIFT		12
#define PAGE_SECTORS_SHIFT	(PAGE_SHIFT - SECTOR_SHIFT)
#define PAGE_SECTORS		(1 << PAGE_SECTORS_SHIFT)
#define SECTOR_SIZE	 	(1 << SECTOR_SHIFT)

#define UNKNOWN_NAME		"unknown"

#ifndef BLKSNAP_LOG_PREFIX
#  define BLKSNAP_LOG_PREFIX	BLKSNAP_NAME
#endif

#ifdef BLKSNAP_DEBUG_ALL
#define BLKSNAP_DEBUG_COMMON
#define BLKSNAP_DEBUG_CLSTORE
#define BLKSNAP_DEBUG_BIO
#define BLKSNAP_DEBUG_EXTENTS
#endif

#ifdef BLKSNAP_DEBUG_COMMON
#  define D(fmt, ...)		bsmsg(fmt, ## __VA_ARGS__)
#else /* ! BLKSNAP_DEBUG_COMMON */
#  define D(fmt, ...)		do { } while(0)
#endif /* ! BLKSNAP_DEBUG_COMMON */

#ifdef BLKSNAP_DEBUG_CLSTORE
#  define DCLS(fmt, ...)	bsmsg(fmt, ## __VA_ARGS__)
#else /* ! BLKSNAP_DEBUG_CLSTORE */
#  define DCLS(fmt, ...)	do { } while(0)
#endif

#ifdef BLKSNAP_DEBUG_BIO
#  define DBIO(fmt, ...)	bsmsg(fmt, ## __VA_ARGS__)
#else /* ! BLKSNAP_DEBUG_BIO */
#  define DBIO(fmt, ...)	do { } while(0)
#endif

#ifndef MIN
#define MIN(a, b) (((a) < (b)) ? (a) : (b))
#endif

#ifndef MAX
#define MAX(a, b) (((a) > (b)) ? (a) : (b))
#endif

#ifndef WARN_VOLUME
#define WARN_VOLUME(volume, cond, format, ...)	\
	WARN(cond, "[%d:%d]: " format,		\
	     MAJOR(volume->bdev->bd_dev),	\
	     MINOR(volume->bdev->bd_dev),	\
	     ## __VA_ARGS__)
#endif

#ifndef FALL_ON_NO_CACHE
# define FALL_ON_NO_CACHE(cache_name, err_code)				       \
	if (unlikely(!cache_name)) {					       \
		WARN(1, #cache_name " hasn't been allocated, %s()", __func__); \
		return err_code;					       \
	}
#endif

static inline sector_t aligned_block_address(sector_t addr)
{
	return addr & ~(PAGE_SECTORS - 1);
}

/* Calculate the 4K-aligned sector address given a bio. */
static inline sector_t bio_block_address(struct bio *bio)
{
	return aligned_block_address(COMPAT_BIO_SECTOR(bio));
}

/* is this the first map ioctl request */
static inline bool is_first_map_ioctl(struct blksnap_ioctl_map *map)
{
	return (map->flags & BLKSNAP_STORE_FIRST);
}

/* is this the last map ioctl request */
static inline bool is_last_map_ioctl(struct blksnap_ioctl_map *map)
{
	return (map->flags & BLKSNAP_STORE_LAST);
}

static inline uint64_t bytes2objects(uint64_t bytes, uint64_t obj_size)
{
	return (bytes / obj_size) + ((bytes % obj_size) ? 1 : 0);
}

/* WARNING! Mismatching between format and arguments in bsmsg* functins leads to
 * fatal accidents in RUNTIME.
 */
void bsmsg(const char *fmt, ...);

void bsmsg_bdev(const struct block_device *bdev, const char *fmt, ...);

void bsmsg_volume(const struct blksnap_volume *volume, const char *fmt, ...);

static inline int warn_null_pointer(const void *arg, const char *func) {
	if (unlikely(!arg)) {
		bsmsg("WARNING: %s: Input parameter isn't allocated", func);
		return 1;
	} else {
		return 0;
	}
}
/* {{{1 Structures */

/* Some device drivers use and modify fields within a bio before calling the
 * end_io callback. So we cannot rely on bi_sector to be the actual sector the
 * I/O request was performed on. There is really no way to infer this
 * information during copy-on-write, so when chaining bios together, this
 * wrapper allows us to preserve important information necessary for use within
 * the callback.
 *
 * See write_bio_do_cow() and bio_end_copy()
 *
 * TODO: If we allocate bios from a bio_set, additional space for each bio
 * allocation can be configured such that an additional allocation for
 * blksnap_bio_private may not be necessary.  This needs investigation.
 */
struct blksnap_bio_private {

	/* Some device drivers will change a bio's fields before calling the
	 * end_io callback, thus it cannot be relied on. */
	sector_t bi_sector;

	/* Some device drivers will change a bio's fields before calling the
	 * end_io callback, thus it cannot be relied on. */
	unsigned int bi_size;

	/* Store original index of bi_io_vect to use it in bi_end_io callback */
	unsigned short		bi_idx;

	/* The next bio, used for chaining. This is the primary mechanism used
	 * to do COW (write is chained to the read) and changelog store updates
	 */
	struct bio *next;

	/* Used for tree operations during COW */
	struct blksnap_volume *volume;
};

/* Some parameters of bio structure are broken in bi_end_io() stage,
 * this function restore correct parameters of bio request to be used
 * in our custom bi_end_io() and to reuse original changed bio request
 * to reinject it to the queue
 */
static inline void restore_bio_params(struct bio *bio, struct blksnap_bio_private *bp)
{
	BUG_ON(!bio || !bp);

	D("%s: bio(%p) sector=%lu idx=%d size=%u flags=%lu", __func__,
	  bio, COMPAT_BIO_SECTOR(bio), COMPAT_BIO_IDX(bio), COMPAT_BIO_SIZE(bio), bio->bi_flags);
	D("%s: bp(%p) sector=%lu idx=%d size=%u", __func__,
	  bp, bp->bi_sector, bp->bi_idx, bp->bi_size);

	COMPAT_BIO_IDX(bio) = bp->bi_idx;
	COMPAT_BIO_SIZE(bio) = bp->bi_size;
	COMPAT_BIO_SECTOR(bio) = bp->bi_sector;

	/* we shoud restore bi_remaining as it was decremented in bio_endio() */
	COMPAT_INC_BIO_REMAINING(bio);
}

/* Some parameters of bio structure are broken in bi_end_io() stage,
 * this function stores correct parameters of bio request to be used
 * in our custom bi_end_io() and to reuse original changed bio request
 * to reinject it to the queue
 */
static inline void store_bio_params(struct bio *bio, struct blksnap_bio_private *bp)
{
	BUG_ON(!bio || !bp);

	D("%s: bio(%p) sector=%lu idx=%d size=%u flags=%lu", __func__,
	  bio, COMPAT_BIO_SECTOR(bio), COMPAT_BIO_IDX(bio), COMPAT_BIO_SIZE(bio), bio->bi_flags);

	bp->bi_idx = COMPAT_BIO_IDX(bio);
	bp->bi_size = COMPAT_BIO_SIZE(bio);
	bp->bi_sector = COMPAT_BIO_SECTOR(bio);
}

/* Calculate the number of 4K boundaries a bio crosses. This is used to
 * determine how many blocks need to be retrieved from the disk. */
static inline unsigned int bio_block_count(struct bio *bio, sector_t block_addr)
{
	/* TODO: The filter routines should be doing large I/Os instead of
	 * iterating over each block. That change would allow this to go away.
	 */
	return (
		 (
		   (
		     COMPAT_BIO_SECTOR(bio)
		     + (
			 (COMPAT_BIO_SIZE(bio) - 1) >> SECTOR_SHIFT
		       )
		     + PAGE_SECTORS
		   ) & ~(PAGE_SECTORS - 1)
		 ) - block_addr
	       ) >> PAGE_SECTORS_SHIFT;
}

static inline unsigned int sector_count(unsigned int len)
{
	return (len >> SECTOR_SHIFT) +
		((len % SECTOR_SIZE) ? 1 : 0);
}
#endif /* _BLKSNAP_SHARED_H */

