#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/blkdev.h>

#include "blksnap.h"
#include "blksnap-bio-queue-share.h"
#include "blksnap-bio-queue.h"
#include "blksnap-hash.h"
#include "blksnap-device.h"
#include "blksnap-compat.h"
#include "blksnap-shared.h"
#include "blksnap-snapshot.h"
#include "blksnap-device-utils.h"
#include "blksnap-volume.h"
#include "blksnap-data.h"
#include "blksnap-cow-tree.h"
#include "blksnap-write-queue.h"
#include "blksnap-bitmap.h"
#include "blksnap-block.h"
#include "blksnap-cl-store.h"
#include "blksnap-bio-submit.h"

#define SNAPSHOT_PREFIX		"sbd"

#define COW_INFLIGHT_WAIT		(1 * HZ) /* 0.1 sec in jiffies */

/* Major number used for snapshot block devices. We ask the kernel to assign an
 * unused one and store it here. */
int blksnap_major;

/* {{{1 Snapshot Device Operations */

/* Please see blksnap-compat.h for RELEASE_* references
 */
static RELEASE_TYPE snapshot_release(RELEASE_PARAMS)
{
	struct blksnap_volume *volume = RELEASE_PRIVATE_DATA;

	/* We should always have a valid volume here, but just in case. */
	if (!WARN_ON(!volume))
		atomic_dec(&volume->snapshot_holders);

	return RELEASE_RET;
}

static int snapshot_open(OPEN_PARAMS)
{
	struct blksnap_volume *volume = OPEN_PRIVATE_DATA;
	OPEN_MODE_INIT;

	if (WARN_ON(!volume))
		return -ENXIO;

	if (mode & FMODE_WRITE)
		return -EPERM;

	atomic_inc(&volume->snapshot_holders);
	return 0;
}

static inline bool bio_must_split(struct bio *bio)
{
	return (bio_block_count(bio, bio_block_address(bio)) > 1);
}

static void snapshot_read_request(struct request_queue *q,
				  struct bio *bio);

static void snapshot_read_split(struct request_queue *q, struct bio *parent);

#if !HAVE_MAKE_REQUEST_FN
static MRF_TYPE snapshot_submit_bio(struct bio *bio)
{
	struct request_queue *q = COMPAT_BIO_GET_QUEUE(bio);
	struct blksnap_volume *volume = q->queuedata;

	if (!volume_has_snapshot(volume)) {
		blksnap_bio_endio(bio, -ENODEV);
		goto out;
	}

	if (snapshot_has_errors(volume)) {
		blksnap_bio_endio(bio, -EIO);
		goto out;
	}

	/*
	 * TODO: It is now possible for the user to unmount a volume during
	 * a snapshot.  We should consider failing this request here, or at
	 * least detect if the volume is still present in the system before
	 * proceeding.
	 *
	 * In the current state, you could start a snapshot, unmount the device,
	 * remove it from the system, and read a block from the snapshot device.
	 * This will cause terrible errors.
	 */

	/* Snapshot devices are read-only */
	if (bio_data_dir(bio) != READ) {
		bsmsg("attempted write at %llu on read-only snapshot device",
		      COMPAT_BIO_SECTOR(bio));
		blksnap_bio_endio(bio, -EROFS);
		goto out;
	}

	/* Remap the sector relative to the underlying device */
	/* NOTE: Can't use volume->start_block because it is page-aligned */
	COMPAT_BIO_SECTOR(bio) += volume->start_sector;

	/* Remap the device to the underlying device */
	COMPAT_BIO_SET_BDEV(bio, volume->device->bdev);

	if (bio_must_split(bio))
		snapshot_read_split(q, bio);
	else
		snapshot_read_request(q, bio);

out:
	return MRF_DEF_RET;
}
#endif

struct block_device_operations snapshot_ops = {
	.owner 		= THIS_MODULE,
	.open  		= snapshot_open,
	.release	= snapshot_release,
#if !HAVE_MAKE_REQUEST_FN
	.submit_bio	= snapshot_submit_bio,
#endif
};

/* {{{1 I/O Copy Filter */
/* @addr - page-aligned block offset, in sectors from device start */
static void copy_write(struct blksnap_volume *volume,
			      sector_t addr,
			      struct blksnap_cow_block *cow_blk)
{
	struct block_device *bdev = volume->device->bdev;
	struct bio *bio;
	struct page *page;
	struct blksnap_bio_private bp;
	gfp_t gfp_mask = (in_atomic()) ? GFP_ATOMIC : GFP_NOIO;
	int err;
	bool completed;

#if BIO_ALLOC_4_ARGS
	bio = bio_alloc(bdev, BIO_MAX_VECS, REQ_OP_READ, gfp_mask);
#else
	bio = bio_alloc(gfp_mask, 1);
#endif
	if (!bio) {
		bsmsg_volume(volume,
			     "%s: ERROR: Unable to allocate bio(%llu)",
			     __func__, addr);
		goto fail_snap;
	}

	COMPAT_BIO_SET_BDEV(bio, bdev);
	COMPAT_BIO_SECTOR(bio) = addr;
	COMPAT_BIO_SET_OP(bio, READ, READ_SYNC);

	page = alloc_page(gfp_mask);
	if (!page) {
		bsmsg_volume(volume,
			     "%s: ERROR: Unable to allocate bio(%llu) page",
			     __func__, addr);
		goto fail_bio;
	}

	if (bio_add_page(bio, page, PAGE_SIZE, 0) != PAGE_SIZE) {
		bsmsg_volume(volume,
			     "%s: ERROR: Unable to add page into bio(%llu)",
			     __func__, addr);
		goto fail_page;
	}

	mutex_lock(&volume->cow_lock);
	/* no need to check ret value for addr cus it was checked before */
	block_set_bit(volume, addr, IN_FLIGHT);
	mutex_unlock(&volume->cow_lock);

	store_bio_params(bio, &bp);
	if (bio->bi_next) {
		bsmsg_volume(volume,
			     "%s: ALERT #1: bio(0x%p, %llu:%u) has bi_next=0x%p",
			     __func__, bio, COMPAT_BIO_SECTOR(bio),
			     COMPAT_BIO_SIZE(bio), bio->bi_next);
		bio->bi_next = NULL;
	}
	err = blksnap_submit_bio_wait(bio, BLKSNAP_READ);
	restore_bio_params(bio, &bp);

	if (err) {
		bsmsg_volume(volume, "%s: ERROR Reading Original CoW BIO %llu (0x%p)",
			     __func__, COMPAT_BIO_SECTOR(bio), bio);
		mutex_lock(&volume->cow_lock);
		block_clear_bit(volume, addr, IN_FLIGHT);
		mutex_unlock(&volume->cow_lock);
		goto fail_page;
	}

	completed = blksnap_data_cow_put(volume->data, cow_blk, bio, &err);

	mutex_lock(&volume->cow_lock);
	block_clear_bit(volume, addr, IN_FLIGHT);
	mutex_unlock(&volume->cow_lock);

	if (err) {
		/* no need to set snapshot_error() as it has been set before
		 * inside blksnap_data_cow_put()
		 */
		bsmsg_volume(volume, "%s: ERROR Putting CoW BIO %llu (0x%p)",
			     __func__, COMPAT_BIO_SECTOR(bio), bio);
		/* An error occurred storing the page in the cow.  This means
		 * the snapshot is no longer valid.
		 */
		goto fail_page;
	}

	if (completed)
		/* The cow was successfully stored in memory and we can just
		 * free the bio and carry on with our lives.
		 */
		goto ok_bio;

	/* The cow requires a write to disk and the bio we passed in
	 * has been modified appropriately.  So we submit this bio
	 * again to store the COW data in the proper location.
	 */
	if (bio->bi_next) {
		bsmsg_volume(volume,
			     "%s: ALERT #2: bio(0x%p, %llu:%u) has bi_next=0x%p",
			     __func__, bio, COMPAT_BIO_SECTOR(bio),
			     COMPAT_BIO_SIZE(bio), bio->bi_next);
		bio->bi_next = NULL;
	}
	/* Submit bio to write original data to data store */
	err = blksnap_submit_bio_wait(bio, BLKSNAP_WRITE_FILTER_DISABLED);
	if (err) {
		bsmsg_volume(volume, "%s: ERROR Writing CoW Store BIO %llu (0x%p)",
			     __func__, COMPAT_BIO_SECTOR(bio), bio);
		goto fail_page;
	}

	/* Update cow_blk object */
	blksnap_data_cow_disk_commit(cow_blk);

	/* A bio was submitted with it's own bi_end_io, so it's required to
	 * force to release a bio object.
	 */
ok_bio:
	bio_put(bio);
	return; /* OK */

fail_page:
	put_page(page);
fail_bio:
	bio_put(bio);
fail_snap:
	snapshot_error(volume);
}

/* @addr - page-aligned block offset, in sectors from device start */
static inline void filter_write_copy(struct blksnap_volume *volume,
				     struct bio *bio,
				     sector_t addr)
{
	struct blksnap_cow_block *cow_blk;

	cow_blk = cow_tree_insert(volume, addr);
	if (unlikely(!cow_blk)) {
		bsmsg_volume(volume,
			     "WARNING: Failed to insert block %llu into cow tree",
			     (unsigned long long) addr);
		snapshot_error(volume);
		return;
	}

	/* If block has data already, get another one */
	if (blksnap_data_cow_has_data(cow_blk))
		return;

	if (blksnap_data_cow_alloc(volume->data, cow_blk)) {
		if (!snapshot_has_errors(volume))
			bsmsg_volume(volume,
				     "(%s): WARNING: Data store depleted",
				     blksnap_snapshot_name(volume));

		D("WARNING: Skipping COW for %llu, no space for CoW object", addr);

		snapshot_error(volume);
		return;
	}

	copy_write(volume, addr, cow_blk);

	return;
}

/* {{{2 I/O Copy iterator for bio
 */
void copy_write_bio(struct blksnap_volume *volume, struct bio *bio)
{
#if HAVE_MAKE_REQUEST_FN
	make_request_fn *make_request = NULL;
	struct request_queue *queue;
#endif
	sector_t start, curr;
	unsigned int i, count;

	start = bio_block_address(bio);
	count = bio_block_count(bio, start);

	for (i = 0, curr = start; i < count; i++, curr += PAGE_SECTORS) {
		D(" C  %llu (%u/%u) 0x%p (%s)", curr, i + 1, count, bio,
		  volume_name(volume));

		if (block_set_bit(volume, curr, NEXT_SNAP)) {
			set_bit(VOLUME_DIRTY, &volume->flags);
			snapshot_error(volume);
		}

		if (snapshot_has_errors(volume))
			continue;

		filter_write_copy(volume, bio, curr);
	}

#if HAVE_MAKE_REQUEST_FN
	if (volume->device != NULL && volume->device->make_request != NULL) {
		make_request = volume->device->make_request;
	}

	queue = COMPAT_BIO_GET_QUEUE(bio);
	blksnap_make_request_compat(make_request, queue, bio);
#else
	if (volume->device != NULL && volume->device->original_bd_ops != NULL && volume->device->original_bd_ops->submit_bio != NULL) {
		volume->device->original_bd_ops->submit_bio(bio);
	} else {
		blksnap_submit_bio_compat(bio);
	}
#endif
}

/* {{{1 I/O Record Filter
 * @addr - page-aligned block offset, in sectors from device start */
int filter_write_record(struct blksnap_volume *volume,
				sector_t addr)
{
	if (block_set_bit(volume, addr, NEXT_SNAP)) {
		/* if there is an issue to set block bit in bitmap,
		 * there is no way to make volume not dirty until
		 * base snapshot completed.
		 */
		set_bit(VOLUME_DIRTY | DIRTY_FLAG_PROVEN,
				&volume->flags);
		return -EINVAL;
	}

	/* Insert new/uncommited block to clstore */
	if (!test_bit(VOLUME_DIRTY, &volume->flags)
	    && blksnap_cl_store_valid(volume->clstore)) {
		if (unlikely(cl_store_commit_block(volume, addr, false)))
			bsmsg_volume(volume, "ERROR: Unable to put block %llu",
				     addr);
	}
	return 0;
}

/* {{{2 I/O Record iterator for bio
 */
void record_write_bio(struct blksnap_volume *volume, struct bio *bio)
{
	sector_t start, curr;
	unsigned int i, count;

	start = bio_block_address(bio);
	count = bio_block_count(bio, start);

	for (i = 0, curr = start; i < count; i++, curr += PAGE_SECTORS) {
		D(" R  %llu (%u/%u) 0x%p (%s)", curr, i + 1, count, bio,
		  volume_name(volume));
		if (filter_write_record(volume, curr)) {
			bsmsg_volume(volume,
				     "ERROR: Can't set block %llu as changed",
				     curr);
			return;
		}
	}
}
 
/* {{{1 I/O Filter */

struct bio *filter_write(struct blksnap_device *bd,
			 struct blksnap_volume *volume,
			 struct bio *bio)
{
	sector_t start, curr;
	unsigned int i, count;

	/* Nothing to do with dirty volume until taking base snap */
	if (test_bit(VOLUME_DIRTY, &volume->flags) &&
	    test_bit(DIRTY_FLAG_PROVEN, &volume->flags) &&
	    !volume_has_snapshot(volume)) {
		return bio;
	}

	start = bio_block_address(bio);
	count = bio_block_count(bio, start);

	for (i = 0, curr = start; i < count; i++, curr += PAGE_SECTORS) {
		D(" C  %llu (%u/%u) 0x%p (%s)", curr, i + 1, count, bio,
		  volume_name(volume));

		if (block_set_bit(volume, curr, NEXT_SNAP)) {
			set_bit(VOLUME_DIRTY, &volume->flags);
			snapshot_error(volume);
		}

		if (volume_has_snapshot(volume) && snapshot_has_errors(volume))
			continue;

		/* No need to make CoW-operation if
		 * - snapshot is not in progress, or
		 * - incremental snapshot is in progress, but this block wasn't
		 *   changed before starting snapshot */
		if (!volume_has_snapshot(volume)
		    || (snapshot_is_incremental(volume)
		        && !block_test_bit(volume, curr, THIS_SNAP)))
			filter_write_record(volume, curr);
		else
			filter_write_copy(volume, bio, curr);
	}

	return bio;
}

/* }}}1 */

/* {{{1 Snapshot Read Logic */
/* Returns a number of sectors to split bio request to:
 * - bio request with size <= PAGE_SIZE and bi_sector aligned to PAGE_SIZE
 * - rest of bio */
/* Due to big bio requests on kernels >= 4.3, a logic of this function was
 * inverted. Old version "splited" bio from "the begin", current version "splits"
 * bio from "the end" to avoid recursion in snapshot_read_split().
 */
/* TODO: need to be revised to simplify and speed up */
static inline int bio_first_sectors(struct bio *bio)
{
	sector_t bio_first_sector = COMPAT_BIO_SECTOR(bio);
	sector_t bio_last_sector =
			bio_first_sector
			+ (COMPAT_BIO_SIZE(bio) >> SECTOR_SHIFT);
	sector_t bio_last_aligned =
			(bio_last_sector >> PAGE_SECTORS_SHIFT)
			<< PAGE_SECTORS_SHIFT;

	if (bio_last_sector == bio_last_aligned)
		bio_last_aligned -= PAGE_SECTORS;

	return bio_last_aligned - bio_first_sector;
}

static void queue_submit_bio(struct request_queue *q, struct bio *bio)
{
	/* Sanity check, bi_next is used in the layers below us */
	WARN_ON(bio->bi_next);

#if HAVE_MAKE_REQUEST_FN
	generic_make_request(bio);
#else
	submit_bio_noacct(bio);
#endif
}

static void snapshot_read_split(struct request_queue *q, struct bio *parent)
{
	int first_sectors = bio_first_sectors(parent);
#if BIO_SPLIT_4_ARGS
	/* WARNING: new kernels >= 4.3 don't use queue segments limit so we
	 * need to handle big bio requests that contains up to BIO_MAX_PAGES
	 * segments
	 */
	struct bio *bio;
#else
	struct bio_pair *bp;
	struct blksnap_volume *volume = q->queuedata;

	/* FIXME: old bio_split() does't support bio requests with multiply
	 * segments, so it's just prevent to rise BUG() inside bio_split().
	 * In future we have to rewrite snapshot_read_request() to support big
	 * bio request without using bio_split()
	 */
	if (bio_segments(parent) > 1) {
		bsmsg_volume(volume,
			     "WARNING: BIO request (%p, %u, %lu) is too big to split",
			     parent, COMPAT_BIO_SIZE(parent), COMPAT_BIO_SECTOR(parent));
		blksnap_bio_endio(parent, -EIO);
		snapshot_error(volume);
		return;
	}
#endif

#if BIO_SPLIT_4_ARGS
	bio = COMPAT_BIO_SPLIT(parent, first_sectors);
	bio_chain(bio, parent);

	WARN_ON(bio_must_split(parent));

	snapshot_read_request(q, parent);

	/* iterate bio and split it with single-page block-aligned set of bios
	 */
	while (bio_must_split(bio)) {
		struct bio *tmp_bio;

		first_sectors = bio_first_sectors(bio);
		tmp_bio = COMPAT_BIO_SPLIT(bio, first_sectors);
		bio_chain(tmp_bio, bio);

		WARN_ON(bio_must_split(bio));

		snapshot_read_request(q, bio);

		bio = tmp_bio;
	}

	snapshot_read_request(q, bio);

#else
	bp = COMPAT_BIO_SPLIT(parent, first_sectors);

	snapshot_read_request(q, &bp->bio1);
	snapshot_read_request(q, &bp->bio2);
	bio_pair_release(bp);
#endif
}

static void snapshot_read_request(struct request_queue *q,
				  struct bio *bio)
{
	struct blksnap_volume *volume = q->queuedata;
	sector_t start;
	struct blksnap_cow_block *cow_blk;
	int err;
	bool endio;

	/* Calculate the aligned block address */
	start = bio_block_address(bio);

	/* If block is not found, we submit the request to the disk.
	 * Without "locking" this block, it's possible for filter_write_copy()
	 * (COW) to race with this request. If COW wins, we read modified rather
	 * than original data. Perhaps a place-holder block to temporary
	 * indicate this condition.
	 */
	mutex_lock(&volume->cow_lock);
	if (!block_is_active(volume, start)) {
		/* Block was not found */
		queue_submit_bio(q, bio);
		mutex_unlock(&volume->cow_lock);
		return;
	}

	/* Block was found, but without COW data */
	if (!block_test_bit(volume, start, COW_DATA)) {
		/* Nothing to do, just redirect a bio to underlying device */
		queue_submit_bio(q, bio);
		mutex_unlock(&volume->cow_lock);
		return;
	}
	mutex_unlock(&volume->cow_lock);

	cow_blk = cow_tree_search(volume, start);
	BUG_ON(!cow_blk);

	/* Block was found and COW_DATA flag is set, so this block should wait
	 * until data is committed.
	 */
	while (!blksnap_data_cow_has_data(cow_blk)) {
		if (snapshot_has_errors(volume)) {
			blksnap_bio_endio(bio, -EIO);
			return;
		}
		/* If current block is in-flight, just sleep and wait for end of
		 * cow-operation
		 */
		while (block_test_bit(volume, start, IN_FLIGHT)) {
			if (snapshot_has_errors(volume)) {
				blksnap_bio_endio(bio, -EIO);
				return;
			}
			io_schedule();
		}
	}

	endio = blksnap_data_cow_get(volume->data, cow_blk, bio, &err);

	if (endio) {
		/* In this case, __blksnap_data_cow_get() has populated
		 * the bio page with COW data from memory.  So there is
		 * nothing more to do beyond completing the bio.
		 */
		blksnap_bio_endio(bio, err ? -EIO : 0);

	} else {
		/* In this case, the call to __blksnap_data_cow_get()
		 * has modified this bio to reference the on-disk store
		 * block instead of the original block.  Thus,
		 * submitting this bio will cause the reader to receive
		 * the COW block as expected.
		 */
		queue_submit_bio(q, bio);
	}
}

#if HAVE_MAKE_REQUEST_FN
static MRF_TYPE snapshot_make_request(struct request_queue *q, struct bio *bio)
{
	struct blksnap_volume *volume = q->queuedata;

	if (!volume_has_snapshot(volume)) {
		blksnap_bio_endio(bio, -ENODEV);
		goto out;
	}

	if (snapshot_has_errors(volume)) {
		blksnap_bio_endio(bio, -EIO);
		goto out;
	}

	/*
	 * TODO: It is now possible for the user to unmount a volume during
	 * a snapshot.  We should consider failing this request here, or at
	 * least detect if the volume is still present in the system before
	 * proceeding.
	 *
	 * In the current state, you could start a snapshot, unmount the device,
	 * remove it from the system, and read a block from the snapshot device.
	 * This will cause terrible errors.
	 */

	/* Snapshot devices are read-only */
	if (bio_data_dir(bio) != READ) {
		bsmsg("attempted write at %llu on read-only snapshot device",
		      COMPAT_BIO_SECTOR(bio));
		blksnap_bio_endio(bio, -EROFS);
		goto out;
	}

	/* Remap the sector relative to the underlying device */
	/* NOTE: Can't use volume->start_block because it is page-aligned */
	COMPAT_BIO_SECTOR(bio) += volume->start_sector;

	/* Remap the device to the underlying device */
	COMPAT_BIO_SET_BDEV(bio, volume->device->bdev);

	if (bio_must_split(bio))
		snapshot_read_split(q, bio);
	else
		snapshot_read_request(q, bio);

out:
	return MRF_DEF_RET;
}
#endif

/* {{{1 Start Snapshot logic */

static int create_snapshot_device(struct blksnap_volume *volume)
{
	int ret = 0;
#if HAVE_ALLOC_DISK
	struct gendisk *disk = alloc_disk(1);
#else
	struct gendisk *disk = blk_alloc_disk(NUMA_NO_NODE);
	disk->minors = 1;
#endif

	if (WARN_ON(!disk))
		return -ENOMEM;

#if HAVE_MAKE_REQUEST_FN
	disk->queue = blksnap_alloc_queue(snapshot_make_request);
#else
#  if HAVE_ALLOC_DISK
	disk->queue = blksnap_alloc_queue();
#  endif
#endif
	if (WARN_ON(!disk->queue)) {
		ret = -ENODEV;
		goto err;
	}

	/* Set up default settings and copy settings from original bdev */
	blk_set_stacking_limits(&disk->queue->limits);
#if HAVE_BDEV_STACK_LIMITS
	bdev_stack_limits(&disk->queue->limits, volume->bdev, 0);
#endif

	/* It is needed to limit maximum request size to avoid
	 * passing to snapshot_make_request multi-vector/multi-segment bios.
	 * Don't forget to REMOVE this limit after adding support of
	 * multi-segment bio requests!
	 */
	blk_queue_max_hw_sectors(disk->queue, PAGE_SECTORS);
	/* FIXME: now snapshot_make_request() doesn't support a multi-segments
	 * bio requests. After adding support of multi-segmenting, don't forget
	 * to REMOVE these limits!
	 */
	blk_queue_max_segments(disk->queue, 1);
	blksnap_queue_set_flags(disk->queue);

	disk->queue->queuedata = volume;

	disk->fops = &snapshot_ops;
	disk->major = blksnap_major;
	disk->first_minor = volume->id;
	disk->private_data = volume;

	if (snprintf(disk->disk_name, DISK_NAME_LEN, "%s%d", SNAPSHOT_PREFIX,
		     volume->id) >= DISK_NAME_LEN) {
		ret = -EFAULT;
		goto err;
	}

	volume->disk = disk;

	/* set snapshot disk r/o on the lowest layer */
	set_disk_ro(disk, 1);

	/* set snapshot disk size from underlying's disk size */
	set_capacity(disk, volume_capacity(volume));

	/* don't add_disk() here because the VOLUME_SNAP_ON flag
	 * is not set yet.
	 */
out:
	return ret;

err:
	put_disk(disk);
	goto out;
}

static int start_snapshot_device(struct blksnap_volume *volume)
{
	if (WARN_ON(!volume->disk))
		return -ENODEV;

#if ADD_DISK_RET_INT
	return add_disk(volume->disk);
#else
	add_disk(volume->disk);
	return 0;
#endif
}

int start_snapshot(struct blksnap_volume *volume,
		   struct blksnap_ioctl_start *start)
{
	int rc = 0;

	if (WARN_ON(create_snapshot_device(volume)))
		return -EFAULT;

	/* mark snapshot incremental */
	if (start->flags & BLKSNAP_START_INCR) {
		if (test_bit(VOLUME_DIRTY, &volume->flags)) {
			bsmsg_volume(volume,
				     "ERROR: Try to start an incremental snapshot on a dirty volume");
			rc = -EDIRTY;
			goto free_disk;
		} else {
			set_bit(VOLUME_SNAP_INCR, &volume->flags);
		}
	}

	/* it's required to reset volume errors at start of snapshot
	 * because at stop of snapshot there could be postponed bio requests
	 * that could cause volume error after resetting it
	 */
	snapshot_error_reset(volume);

	/* Mark the blocks that this snapshot contains. A separate ioctl is
	 * used to request these (changelog).
	 * Changelog is disabled if device is dirty */
	if (!test_bit(VOLUME_DIRTY, &volume->flags))
		blksnap_blk_bitmap_apply_start_snapshot(volume);

	rc = blksnap_write_bio_kthread_init(volume);
	if (rc) {
		bsmsg_volume(volume,
			     "ERROR: Can't start Write BIO kthread");
		return rc;
	}

	set_bit(VOLUME_SNAP_ON, &volume->flags);
	/* add_disk() after setting VOLUME_SNAP_ON bit to avoid
	 * "Buffer I/O errors" in dmesg
	 */
	rc = start_snapshot_device(volume);
	if (rc) {
		bsmsg_volume(volume,
			     "ERROR: Can't start snapshot device");
		return rc;
	}

	bsmsg_volume(volume, "Snapshot available at %s", blksnap_snapshot_name(volume));

out:
	return rc;
free_disk:
	put_disk(volume->disk);
	goto out;
}

/* }}}1 */

/* {{{1 Stop Snapshot Logic */

static void destroy_snapshot_device(struct blksnap_volume *volume)
{

	if (!volume->disk)
		return;

	del_gendisk(volume->disk);
	put_disk(volume->disk);
#if HAVE_ALLOC_DISK
#  if HAVE_BLK_CLEANUP_QUEUE
	blk_cleanup_queue(volume->disk->queue);
#  else
	blk_put_queue(volume->disk->queue);
#  endif
#endif
	volume->disk = NULL;
} 

int stop_snapshot(struct blksnap_volume *volume, bool cancel)
{
	int ret = 0, ret_tmp;
	bool frozen;
	bool force_cancel = false;
	bool skip_freeze;

	clear_bit(VOLUME_SNAP_ON, &volume->flags);

	blksnap_write_bio_kthread_destroy(volume);

	destroy_snapshot_device(volume);

	blksnap_data_store_reset(volume->data);

	skip_freeze = test_bit(VOLUME_FS_XFS_VERSION_5, &volume->fs_flags);

	if (!skip_freeze) {
		if (IS_ERR(freeze_volume(volume))) {
			bsmsg_volume(volume, "ERROR: failed to freeze filesystem");
			ret = -ENXIO;
			frozen = false;
		} else {
			frozen = true;
		}
	}

 	if (ret || snapshot_has_errors(volume)) {
		snapshot_error_reset(volume);
		if (!cancel)
			cancel = force_cancel = true;
	}

	if (blksnap_cl_store_valid(volume->clstore) && !cancel)
		cl_store_reset(volume->clstore);

	/* Deal with this-snap blocks, write changelog to disk */
	ret_tmp = blksnap_blk_bitmap_apply_stop_snapshot(volume, cancel);
	if (ret_tmp) {
		/* changelog is corrupted, set volume as dirty */
		ret = ret_tmp;
		set_bit(VOLUME_DIRTY, &volume->flags);
		set_bit(DIRTY_FLAG_PROVEN, &volume->flags);
		if (!cancel)
			cancel = force_cancel = true;
	}

	if (!cancel && !ret) {
		/* On completion of a snapshot, this device is now considered
		 * clean */
		clear_bit(VOLUME_CHANGED, &volume->flags);
		clear_bit(VOLUME_DIRTY, &volume->flags);
		set_bit(DIRTY_FLAG_PROVEN, &volume->flags);
	}

	/* clear snapshot incremental flag at the end of snapshot */
	clear_bit(VOLUME_SNAP_INCR, &volume->flags);

	if (!skip_freeze) {
		if (frozen && thaw_volume(volume)) {
			bsmsg_volume(volume, "ERROR: failed to thaw filesystem");
			ret = -ENXIO;
		}
	}

	if (cow_tree_empty(volume)) {
		bsmsg_volume(volume, "ERROR: CoW tree has had errors on %s snapshot",
			     cancel ? "cancelling" : "completing");
		ret = -ENXIO;
	}

	if (!force_cancel) {
		bsmsg_volume(volume, "Snapshot ended (mode: %s)%s",
			     cancel ? "cancel" : "stop",
			     ret ? ", error(s) occurred" : "");
	} else {
		bsmsg_volume(volume, "WARNING: Snapshot was cancelled due to errors. Volume is%s dirty.",
			     test_bit(VOLUME_DIRTY, &volume->flags) ? "": " not");
		ret = 0;  /* agent wishes no errors in this case */
	}

	return ret;
}

/* }}}1 */

