#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/blkdev.h>

#include "blksnap.h"
#include "blksnap-bio-queue-share.h"
#include "blksnap-hash.h"
#include "blksnap-device.h"
#include "blksnap-compat.h"
#include "blksnap-shared.h"
#include "blksnap-device-utils.h"
#include "blksnap-volume.h"

/* List of volumes. These are block devices that the use maniplates through the
 * ioctl interface. */
LIST_HEAD(blksnap_volumes);

/* {{{1 Volume Utility Routines */
struct blksnap_volume *lookup_volume_by_bdev(struct block_device *bdev)
{
	struct list_head *p, *q;

	if (warn_null_pointer(bdev, __func__))
		return NULL;

	list_for_each_safe(p, q, &blksnap_volumes) {
		struct blksnap_volume *volume =
			list_entry(p, struct blksnap_volume, classmates);

		if (volume->bdev->bd_dev == bdev->bd_dev)
			return volume;
	}

	return NULL;
}

struct blksnap_volume *lookup_volume_by_path(const char *path)
{
	struct block_device *bdev;

	bdev = blksnap_lookup_bdev_compat(path);
	if (IS_ERR(bdev))
		return NULL;

	return lookup_volume_by_bdev(bdev);
}

bool volume_contains(struct blksnap_volume *volume, sector_t addr)
{
	if (WARN_ONCE(!volume->bdev, "volume went missing"))
		return false;

	return (get_start_sect(volume->bdev) <= addr &&
		(get_start_sect(volume->bdev) + blksnap_get_nr_sects(volume->bdev) >
		 addr)) ? true : false;
}

struct blksnap_volume *lookup_volume_by_addr(struct blksnap_device *bd,
					     sector_t addr)
{
	struct list_head *p, *q;

	list_for_each_safe(p, q, &bd->volumes) {
		struct blksnap_volume *volume =
			list_entry(p, struct blksnap_volume, siblings);

		if (volume_contains(volume, addr))
			return volume;
	}

	return NULL;
}

#if THAW_BDEV_2_ARGS
struct super_block *freeze_volume(struct blksnap_volume *volume)
{
	/* Freeze and thaw is supported on all filesystems that provide an
	 * implementation for freeze_fs(). Not all of them do, however. Ext2 is
	 * one example. For those filesystems, there is no way to guarantee
	 * a filesystem-consistent snapshot.
	 */
	volume->frozen_sb = freeze_bdev(volume->bdev);

	return volume->frozen_sb;
}
#else
struct blksnap_volume *freeze_volume(struct blksnap_volume *volume)
{
	/* Freeze and thaw is supported on all filesystems that provide an
	 * implementation for freeze_fs(). Not all of them do, however. Ext2 is
	 * one example. For those filesystems, there is no way to guarantee
	 * a filesystem-consistent snapshot.
	 */
	if (freeze_bdev(volume->bdev))
		return NULL;

	return volume;
}
#endif

/* see blksnap-compat.h for definitions of compatible vars THAW_VOL_* */
int thaw_volume(struct blksnap_volume *volume)
{
	THAW_VOL_RET;

#if THAW_BDEV_2_ARGS
	THAW_VOL_RET_LET thaw_bdev(volume->bdev, volume->frozen_sb);
#else
	THAW_VOL_RET_LET thaw_bdev(volume->bdev);
#endif
	return THAW_VOL_RET_VAR;
}

/* }}}1 */

