/*
 *     Copyright (C) 2017 Quest Software Inc.
 *
 *     This file is part of sockbd.
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License version 2 as
 *     published by the Free Software Foundation.
 */

/* {{{1 Preamble */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/miscdevice.h>
#include <linux/delay.h>

#include "../kernel/blksnap-compat.h"
#include "sockbd.h"

#ifndef SOCKBD_LICENSE
#  define SOCKBD_LICENSE "GPL"
#endif

#ifndef SOCKBD_AUTHOR
#  define SOCKBD_AUTHOR "Oleksandr Suvorov"
#endif

#ifndef SOCKBD_NAME
#  define SOCKBD_NAME "sockbdctl"
#endif

#ifndef SOCKBD_DESCR
#  define SOCKBD_DESCR "Socket to Block device driver"
#endif

#ifndef SOCKBD_VERSION
#  define SOCKBD_VERSION "undetermined version"
#endif

#define SOCKBD_IOCTL_LOCK_TRIES		3
#define SOCKBD_IOCTL_LOCK_SLEEP_MS	200

/* }}}1 */

/* {{{1 Static Globals */

/* The module can be disabled completely on the command line.
 */
static bool disabled = 0;
module_param(disabled, bool, 0444);
MODULE_PARM_DESC(disabled, "Disable this module completely");

/* Provide mutual exclusion for ioctl commands */
static DEFINE_MUTEX(sockbd_ioctl_lock);

/* Store the kernel version during compliation.  This is compared to the
 * running version at module load time to catch a mismatch.
 */
static char *compile_version = UTS_RELEASE;

/* Major number used for sockbd block devices. Ask the kernel to assign
 * an unused one and store it here.
 */
static int sockbd_major;
/* }}}1 */

/* {{{1 Public Interface */
/* {{{2 Ioctl Interface */

/* {{{3 Ioctl Routing */
static int __sockbd_ioctl(unsigned int cmd, struct sockbd_ioctl __user *user)
{
	/* Empty routing for awhile */
	return -ENOTTY;
}
/* }}}3 */

/* {{{3 Ioctl Locking */
static bool ioctl_trylock(void)
{
	unsigned int tries;
	
	for (tries = SOCKBD_IOCTL_LOCK_TRIES; tries > 0; tries--) {
		if (mutex_trylock(&sockbd_ioctl_lock))
			return true;
		msleep(SOCKBD_IOCTL_LOCK_SLEEP_MS);
	}
	return false;
}

static void ioctl_unlock(void)
{
	mutex_unlock(&sockbd_ioctl_lock);
}

static void ioctl_lock(void)
{
	mutex_lock(&sockbd_ioctl_lock);
}
/* }}}3 */

/* {{{3 Ioctl Main */

static long sockbd_ioctl(struct file *file,
			 unsigned int cmd,
			 unsigned long arg)
{
	int ret;

	/* Admin privileges are required to manipulate the kernel component */
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;
	
	if (_IOC_TYPE(cmd) != SOCKBD_IOCTL_MAGIC ||
	    _IOC_NR(cmd) > SOCKBD_IOCTL_CMD_LAST)
		return -ENOTTY;
	
	if (!ioctl_trylock())
		return -EBUSY;
	
	ret = __sockbd_ioctl(cmd, (struct sockbd_ioctl __user *) arg);
	
	ioctl_unlock();
	
	return (long) ret;
}
/* }}}3 */
/* }}}2 */
/* }}}1 */

/* {{{1 Misc Device */
static const struct file_operations sockbd_misc_fops = {
	.unlocked_ioctl	= sockbd_ioctl,
	.compat_ioctl	= sockbd_ioctl,
	.owner		= THIS_MODULE,
	.llseek		= noop_llseek,
};

static struct miscdevice sockbd_misc = {
	.minor	= MISC_DYNAMIC_MINOR,
	.name	= SOCKBD_NAME,
	.fops	= &sockbd_misc_fops,
};
/* }}}1 */

/* {{{1 Module Init/Exit */
/* should be called under ioctl_lock()
 */
static void __teardown(void)
{
	unregister_blkdev(sockbd_major, SOCKBD_NAME);
	misc_deregister(&sockbd_misc);
}

static void teardown(void)
{
	ioctl_lock();
	__teardown();
	ioctl_unlock();
}

/* For kernel versions that differ slightly it is possible to load a module
 * compiled for one version of the kernel into a non-matching version. If this
 * occurs, the module will load successfully but may exhibit strange behaviour.
 */
static int check_kernel_version(void)
{
	char *runtime_version = utsname()->release;
	
	if (strcmp(compile_version, runtime_version)) {
		printk(KERN_ERR "SockBD: ERROR: Kernel version mismatch (%s != %s)",
		       compile_version, runtime_version);
		return -EFAULT;
	}

	return 0;
}

static bool is_file_exist(const char *name)
{
	struct file *fp;
	
	fp = filp_open(name, O_RDONLY, 0);

	if (fp == ERR_PTR(-ENOENT))
		return false;
	
	if (!IS_ERR(fp))
		filp_close(fp, NULL);

	return true;
}

static int __init sockbd_init(void)
{
	int ret = 0;

	/* If the module should be disabled, load silently and do nothing
	 */
	if (disabled) {
		printk(KERN_INFO "SockBD: Disabled from command line\n");
		return 0;
	}

	/* Don't load module built for another kernel version */
	ret = check_kernel_version();
	if (ret)
		return ret;

	ioctl_lock();

	if (is_file_exist("/dev/" SOCKBD_NAME)) {
		ret = -EEXIST;
		goto unlock;
	}

	ret = misc_register(&sockbd_misc);
	if (ret) {
		printk(KERN_ERR "SockBD: ERROR: Can't register control device [%d]\n", ret);
		goto unlock;
	}

	sockbd_major = register_blkdev(0, SOCKBD_NAME);
	if (sockbd_major < 0) {
		ret = sockbd_major;
		printk(KERN_ERR "SockBD: ERROR: Can't register block device %s [%d]\n", SOCKBD_NAME, ret);
		goto error;
	}

	printk(KERN_INFO "SockBD %s loaded\n ", SOCKBD_VERSION);
	printk(KERN_INFO "SockBD Control block device MAJOR: %d\n", sockbd_major);

unlock:
	ioctl_unlock();

	return ret;
error:
	__teardown();
	goto unlock;
}

static void __exit sockbd_exit(void)
{
	if (disabled)
		return;
	teardown();
	printk(KERN_INFO "SockBD: unloaded\n");
}

module_init(sockbd_init);
module_exit(sockbd_exit);
/* }}}1 */

MODULE_AUTHOR(SOCKBD_AUTHOR);
MODULE_VERSION(SOCKBD_VERSION);
MODULE_LICENSE(SOCKBD_LICENSE);
MODULE_DESCRIPTION(SOCKBD_DESCR);

