#ifndef _SOCKBD_H
#define _SOCKBD_H

#include <linux/types.h>

#include "sockbd-version.h"

/* Version of ioctl interface.
 * Must be used to prevent using incompatible kernel module
 * and client
 */
#define SOCKBD_IOCTL_VERSION	1

/* Name of misc device to control sockbd devices */
#define SOCKBD_NAME	"sockbdctl"

/* Name prefix for sockbd block devices */
#define SOCKBD_MASK	"skbd"

/* Maximum size of sockbd device name */
#define SOCKBD_NAME_SIZE	128

/* Additional errno values
 * These should not collide with any system-defined values. The exit(2)
 * system call will mask out all but the lower 8 bits. To determine what value
 * you'll see in userspace: (value & 0377).
 */
#define EVERSION	700	/* 188 sockbd ioctl version mismatch */
#define ENOSOCK		701	/* 189 Can't connect to given socket */

/* NOTE: we can't calculate pointer size dynamically because we have to
 * use kernel's ptr size on both side of socket.
 */
#define PTR_SIZE	8

/**
 * Socket message interface
 */

/* NOTE: in future an agent can be run outside of local host,
 * so magics must be present to filter network requests from/to
 * exactly our sockbd/agent
 */
#define SOCKBD_MAGIC_REQ	0xCAB1E60D
#define SOCKBD_MAGIC_REP	0x60D0FBAD

/* sockbd socket command types */
typedef enum sockbd_command {
	SOCKBD_CMD_READ,	/* read from agent (in/out) */
	SOCKBD_CMD_WRITE,	/* write to agent  (out) */
	SOCKBD_CMD_DISC		/* discard request (in) */
} sockbd_command_t;

/* These structures could be send over network in the future,
 * so we have to use specific endian types
 */
typedef struct __attribute__((__packed__)) sockbd_request {
	__be32 magic;		/* magic number req/rep */
	__be64 __attribute__((aligned(8))) req_addr;
				/* struct request* address to
				   lookup queued request */
	__be32 cmd;		/* command, sockbd_command_t */
	__be64 __attribute__((aligned(8))) start;
				/* start sector of request */
	__be32 len;		/* size of requests in bytes */
} sockbd_request_t;

typedef struct __attribute__((__packed__)) sockbd_replay {
	__be32 magic;		/* magic number req/rep */
	char request[PTR_SIZE];	/* struct req* address to lookup request */
	__be32 err;		/* 0 - if ok, otherwise - errno */
} sockbd_replay_t;

/**
 * Ioctl interface
 */

/* sockbd device flags */
#define SOCKBD_DEV_READ_ONLY	(1 << 0)	/* read-only (in/out) */
#define SOCKBD_DEV_READY	(1 << 1)	/* ready to use (out) */

typedef struct sockbd_ioctl_empty {
} sockbd_ioctl_empty_t;

typedef struct sockbd_ioctl_device {
	char name[SOCKBD_NAME_SIZE];	/* name of sockbd device */
					/* input: detach, info */
					/* output: attach, list */
	struct socket *sock;		/* socket to connect device to */
					/* input: attach */
					/* output: info, list */
	__u32 flags;			/* device flags */
					/* input: attach */
					/* output: info, list */
	__u32 block_size;		/* block size of device */
					/* input: attach */
					/* output: info, list */
	__u64 size;			/* device size in bytes */
					/* input: attach */
					/* output: info, list */
} sockbd_ioctl_device_t;

typedef struct sockbd_ioctl_detach {
	char	name[SOCKBD_NAME_SIZE];	/* name of sockbd device */
					/* input */
} sockbd_ioctl_detach_t;

typedef struct sockbd_ioctl_list {
        __u32 ndevices;			/* output */
        sockbd_ioctl_device_t devices[0];
} sockbd_ioctl_list_t;

typedef struct sockbd_ioctl {
	__u32 version;	/* ABI version to check request compatibility */
	__u32 size;	/* total size of current sockbd_ioctl structure */
	union {
		sockbd_ioctl_empty_t	end_of_header;
		sockbd_ioctl_device_t	attach;
		sockbd_ioctl_detach_t	detach;
		sockbd_ioctl_device_t	info;
		sockbd_ioctl_list_t	list;
	};
} sockbd_ioctl_t;

#define SOCKBD_IOCTL_HEADER_SIZE offsetof(struct sockbd_ioctl, end_of_header)

enum {
	SOCKBD_IOCTL_CMD_ATTACH,
	SOCKBD_IOCTL_CMD_DETACH,
	SOCKBD_IOCTL_CMD_INFO,
	SOCKBD_IOCTL_CMD_LIST,
	__SOCKBD_IOCTL_CMD_LAST
};

/**
 * Ioctls definition
 */
#define SOCKBD_IOCTL_MAGIC 'S'
#define SOCKBD_IOCTL_CMD_FIRST SOCKBD_IOCTL_CMD_ATTACH
#define SOCKBD_IOCTL_CMD_LAST (__SOCKBD_IOCTL_CMD_LAST - 1)

#define SOCKBD_IOC_ATTACH \
	_IOR(SOCKBD_IOCTL_MAGIC, \
	     SOCKBD_IOCTL_CMD_ATTACH, sockbd_ioctl_t)

#define SOCKBD_IOC_DETACH \
	_IOR(SOCKBD_IOCTL_MAGIC, \
	     SOCKBD_IOCTL_CMD_DETACH, sockbd_ioctl_t)

#define SOCKBD_IOC_INFO \
	_IOR(SOCKBD_IOCTL_MAGIC, \
	     SOCKBD_IOCTL_CMD_INFO, sockbd_ioctl_t)

#define SOCKBD_IOC_LIST \
	_IOR(SOCKBD_IOCTL_MAGIC, \
	     SOCKBD_IOCTL_CMD_INFO, sockbd_ioctl_t)

#endif /* _SOCKBD_H */
