#include "blksnap-test.h"
#include "blksnap-test.c"

#include "blksnap-cl-buffer.h"
#include "blksnap-cl-buffer.c"

#include "blksnap-cl-thread.h"
#include "blksnap-cl-thread.c"

#include "blksnap-cl-store.h"
#include "blksnap-cl-store.c"

#include "blksnap-extent.h"
#include "blksnap-extent.c"

#include "blksnap-volume.h"
#include "blksnap-volume.c"

#include "blksnap-bio.h"
#include "blksnap-bio.c"

#include "blksnap-bio-queue.h"
#include "blksnap-bio-queue.c"

#include "blksnap-data.h"
#include "blksnap-data.c"

#include "blksnap-bitmap.h"
#include "blksnap-bitmap.c"

#include "blksnap-block.h"
#include "blksnap-block.c"

#include "blksnap-bio-submit.h"
#include "blksnap-bio-submit.c"

#include "blksnap-cow-tree.h"
#include "blksnap-cow-tree.c"

#include "blksnap-shared.c"

/*
 * This file is including unit test suite to perform unit and stress tests of
 * the changelog buffer. This buffer was designed to store changes on disk
 */

/* Common defines */
#define INITIAL_PAGE_COUNT	2
/* test_create_add_destroy defines */
#define ADD_BLOCK_COUNT		8003
#define ADD_ALLOC_PAGES		ADD_BLOCK_COUNT / (CL_PAGE_BUFFER_CAPACITY) + 1
#define ADD_OVERFLOW_COUNT	ADD_ALLOC_PAGES - INITIAL_PAGE_COUNT
#define ADD_SAVED_BLOCK_COUNT	ADD_BLOCK_COUNT
/* flush_call_count should be less then allocated pages for 1 because last page
 * is still not full and it will be written during destroy */
#define ADD_FLUSH_CALL_COUNT	ADD_ALLOC_PAGES - 1
/* Next parameter equals 1 because this is the last page, which was written
 * using direct call of page flush. All previous used flush_page_callback to
 * init new task in queue to perform page write */
#define ADD_PAGES_WRITTEN	1

/* test_add_and_flush defines */
#define FLUSH_BLOCK_COUNT	800
#define FLUSH_ALLOC_PAGES	\
	FLUSH_BLOCK_COUNT / (CL_PAGE_BUFFER_CAPACITY) + 1
#define FLUSH_SAVED_BLOCK_COUNT	FLUSH_BLOCK_COUNT
/* Next parameter should be equal allocated pages because all pages should be
 * flushed during destroy */
#define FLUSH_PAGES_WRITTEN	FLUSH_ALLOC_PAGES

static int flush_call_flag = 0;

static int flush_page_callback(struct blksnap_cl_page *page, void *data)
{
	flush_call_flag++;
	return 0;
}

static int writer_flag = 0;

static int writer_callback(sector_t *page_buffer, unsigned int len,
			   void *data)
{
	writer_flag++;
	return 0;
}


static void print_cl_stat(struct blksnap_cl_stat *stat)
{
/*	mu_print("Got statistics: size=%zu, allocated_pages=%u,"
		 " pages_written=%u, overflow_count=%u, error_count=%u,"
		 " saved_block_count=%u",
		 sizeof(*stat),
		 stat->allocated_pages, stat->pages_written,
		 stat->overflow_count, stat->error_count,
		 stat->saved_block_count); */
}

static test_ret_t cl_buffer_test_create_add_destroy(void)
{
	struct blksnap_cl_buffer *buffer;
	struct blksnap_cl_stat stat;
	int i = 0;
	int ret;

	mu_start(__func__);

	flush_call_flag = 0;

	buffer = blksnap_cl_buffer_init(flush_page_callback, &flush_call_flag,
				writer_callback, &writer_flag);
	mu_assert(!IS_ERR(buffer), "Create buffer");

	for (i = 0; i < ADD_BLOCK_COUNT; i++) {
		ret = blksnap_cl_buffer_add_block(buffer, (uint64_t) 1);
		mu_assert_equal(ret, 0, "Unable to add block [ret: %d]", ret);
	}

	blksnap_cl_get_stat(buffer, &stat);

	/*Check statistics */
	mu_assert_equal(stat.allocated_pages, ADD_ALLOC_PAGES,
			"Wrong value of stat.allocated_pages [%u != %lu]",
			stat.allocated_pages, ADD_ALLOC_PAGES);
	mu_assert_equal(stat.overflow_count, ADD_OVERFLOW_COUNT,
			"Wrong value of stat.overflow_count [%u != %lu",
			stat.overflow_count, ADD_OVERFLOW_COUNT);
	mu_assert_equal(stat.error_count, 0,
			"Got errors during adding blocks in buffer [%u]",
			stat.error_count);
	mu_assert_equal(stat.saved_block_count, ADD_SAVED_BLOCK_COUNT,
			"Wrong value of stat.saved_block_count [%u != %u]",
			stat.saved_block_count, ADD_SAVED_BLOCK_COUNT);
	print_cl_stat(&stat);
	mu_assert_equal(flush_call_flag, ADD_FLUSH_CALL_COUNT,
			"Wrong flush count [%d != %lu]",
			flush_call_flag, ADD_FLUSH_CALL_COUNT);

	blksnap_cl_buffer_destroy(buffer);

	blksnap_cl_get_stat(buffer, &stat);
	print_cl_stat(&stat);
	mu_assert_equal(stat.pages_written, ADD_PAGES_WRITTEN,
			"Wrong value of stat.pages_written [%u != %u]",
			stat.pages_written, ADD_PAGES_WRITTEN);

	mu_end(__func__);

	return 0;
}

static struct blksnap_cl_page *spage = NULL;

static int flush_page_callback2(struct blksnap_cl_page *page, void *data)
{
	spage = page;
	return 0;
}

static test_ret_t cl_buffer_test_add_and_flush(void)
{
	struct blksnap_cl_buffer *buffer;
	struct blksnap_cl_stat stat;
	int i = 0;
	int ret;

	mu_start(__func__);

	writer_flag = 0;
	flush_call_flag = 0;

	buffer = blksnap_cl_buffer_init(flush_page_callback2, &flush_call_flag,
				writer_callback, &writer_flag);
	mu_assert(!IS_ERR(buffer), "Create buffer");

	for (i = 0; i < FLUSH_BLOCK_COUNT; i++) {
		ret = blksnap_cl_buffer_add_block(buffer, (uint64_t) 1);
		mu_assert_equal(ret, 0,"Unable to add block [ret: %d]", ret);
	}

	blksnap_cl_get_stat(buffer, &stat);

	/* Check statistics */
	mu_assert_equal(stat.allocated_pages, FLUSH_ALLOC_PAGES,
			"Wrong value of stat.allocated_pages [%u != %lu]",
			stat.allocated_pages, FLUSH_ALLOC_PAGES);
	mu_assert_equal(stat.pages_written, 0,
			"Wrong initial value of stat.pages_written [%u]",
			stat.pages_written);
	mu_assert_equal(stat.error_count, 0,
			"Got errors during adding blocks in buffer [%u]",
			stat.error_count);
	mu_assert_equal(stat.saved_block_count, FLUSH_SAVED_BLOCK_COUNT,
			"Wrong value of stat.saved_block_count [%u != %u]",
			stat.saved_block_count, FLUSH_SAVED_BLOCK_COUNT);

	print_cl_stat(&stat);

	ret = blksnap_cl_page_flush(buffer, spage);
	mu_assert_equal(ret, 0, "blksnap_cl_page_flush [ret = %d]", ret);

	blksnap_cl_get_stat(buffer, &stat);
	print_cl_stat(&stat);

	mu_assert_equal(stat.pages_written, 1,
			"Wrong value of stat.pages_written after the first flush command [%u != %u]",
			stat.pages_written, 1);
	mu_assert_equal(writer_flag, 1,
			"Writer flag should have been incremented after first flush [%d != %d]",
			writer_flag, 1);

	blksnap_cl_buffer_destroy(buffer);

	blksnap_cl_get_stat(buffer, &stat);
	print_cl_stat(&stat);

	mu_assert_equal(stat.pages_written, FLUSH_PAGES_WRITTEN,
			"Wrong value of stat.pages_written [%u != %lu]",
			stat.pages_written, FLUSH_PAGES_WRITTEN);
	mu_assert_equal(writer_flag, FLUSH_PAGES_WRITTEN,
			"writer_flag was not updated during buffer destroy [%d != %lu]",
			writer_flag, FLUSH_PAGES_WRITTEN);

	mu_end(__func__);

	return 0;
}

static test_ret_t all_tests()
{
	mu_run_test(cl_buffer_test_create_add_destroy);
	mu_run_test(cl_buffer_test_add_and_flush);

	return 0;
}
