#include <linux/vmalloc.h>
#include <linux/rwsem.h>

#include "blksnap-test.h"
#include "blksnap-test.c"

#include "blksnap-rbtree-foreach.h"

#include "blksnap-shared.c"

#include "blksnap-volume.h"
#include "blksnap-volume.c"

#include "blksnap-bio.h"
#include "blksnap-bio.c"

#include "blksnap-bio-queue.h"
#include "blksnap-bio-queue.c"

#include "blksnap-data.h"
#include "blksnap-data.c"

#include "blksnap-extent.h"
#include "blksnap-extent.c"

#include "blksnap-block.h"
#include "blksnap-block.c"

#include "blksnap-bio-submit.h"
#include "blksnap-bio-submit.c"

#include "blksnap-cow-tree.h"
#include "blksnap-cow-tree.c"

#define BLOCKS_CNT	((unsigned long) 10000)

static struct blksnap_volume *create_volume(unsigned long nblocks)
{
	struct blksnap_volume *volume;
	int i, reminder;
	sector_t alloc_size;

	volume = kzalloc(sizeof(*volume), GFP_KERNEL);
	if (!volume) {
		WARN(1, "Failed to allocate %zu bytes for struct blksnap_volume",
		     sizeof(*volume));
		goto fail;
	}

	volume->cow_tree = RB_ROOT;
	volume->bdev = NULL;
	init_rwsem(&volume->tree_rw_lock);

	/* init bitmap for supporting cow_tree_insert() function */
	volume->bitmap.last_block = nblocks * PAGE_SECTORS;
	volume->bitmap.nchunks = div_u64_rem(nblocks, BITMAP_CHUNK_BLOCKS,
					     &reminder);
	volume->bitmap.nchunks += reminder ? 1 : 0;
	alloc_size = volume->bitmap.nchunks * sizeof(struct bitmap_chunk);

	volume->bitmap.vmalloc = false;
	volume->bitmap.chunks = kzalloc(alloc_size, GFP_KERNEL | __GFP_NOWARN);
	if (!volume->bitmap.chunks) {
		volume->bitmap.chunks = vzalloc(alloc_size);
		if (!volume->bitmap.chunks) {
			WARN(1, "Failed to allocate %zu bytes for bitmap chunks",
			     alloc_size);
			goto free_volume;
		}
		volume->bitmap.vmalloc = true;
	}

	for (i = 0; i < volume->bitmap.nchunks; i++)
		init_rwsem(&volume->bitmap.chunks[i].lock_alloc);

	return volume;
free_volume:
	kfree(volume);
fail:
	return NULL;
}

void free_volume(struct blksnap_volume *volume)
{
	int i;
	struct bitmap_chunk *chunk;

	for (i = 0; i < volume->bitmap.nchunks; i++) {
		chunk = &volume->bitmap.chunks[i];
		if (chunk->changed_blocks) {
			kfree(chunk->blocks);
			chunk->changed_blocks = 0;
			chunk->blocks = NULL;
		}
	}

	if (volume->bitmap.vmalloc)
		vfree(volume->bitmap.chunks);
	else
		kfree(volume->bitmap.chunks);

	kfree(volume);
}

bool bitmap_is_empty(struct blksnap_volume *volume)
{
	int i;
	bool ret = false;

	for (i = 0; i < volume->bitmap.nchunks; i++) {
		if (volume->bitmap.chunks[i].blocks) {
			mu_print("chunk %d is allocated", i);
			ret = true;
		}
	}

	return ret;
}

static test_ret_t rb_tree_test_init_fill_destroy(void)
{
	struct blksnap_volume *volume;
	struct blksnap_cow_block *blk, *next_blk;
	struct rb_node *node;
	unsigned long block_num, num_nodes;
	sector_t addr;

	mu_start(__func__);

	volume = create_volume(BLOCKS_CNT);
	mu_assert(volume, "Create volume");

	/* test empty tree */
	num_nodes = 0;
	for (node = rb_first(&volume->cow_tree); node; node = rb_next(node)) {
		num_nodes++;
	}
	mu_assert_equal(num_nodes, 0, "Tree is not empty [%lu]", num_nodes);

	/* test inserting nodes */
	num_nodes = 0;
	for (block_num = 0; block_num < BLOCKS_CNT; block_num++) {
		addr = block_num * PAGE_SECTORS;
		blk = cow_tree_insert(volume, addr);
		mu_assert(blk,"Failed to insert cow_block [%lu, %lu]",
			       addr, block_num);
		num_nodes++;
	}

	/* test if all fine with inserted values */
	num_nodes = 0;
	for (node = rb_first(&volume->cow_tree); node; node = rb_next(node)) {
		num_nodes++;
	}

	mu_assert_equal(num_nodes, BLOCKS_CNT,
			"Error walking tree with rb_next() [%lu != %lu]",
			num_nodes, BLOCKS_CNT);

	/* test of removing some nodes */
	num_nodes = 0;
	rbtree_safe_to_remove_foreach(blk, next_blk,
				      &volume->cow_tree, rb_node) {
		num_nodes++;
	}
	mu_assert_equal(num_nodes, BLOCKS_CNT,
			"Error walking tree with rbtree_safe_to_remove_foreach() [%lu != %lu]",
			num_nodes, BLOCKS_CNT);

	/* test emptying tree */
	num_nodes = 0;

	cow_tree_empty(volume);
	for (node = rb_first(&volume->cow_tree); node; node = rb_next(node)) {
		num_nodes++;
	}
	mu_assert_equal(num_nodes, 0, "Tree was emptied, but it is not empty [%lu]",
				      num_nodes);

	mu_assert(!bitmap_is_empty(volume), "Bitmap should be empty");
	free_volume(volume);

	mu_end(__func__);

	return 0;
}

/* insert a lot of nodes,
 * try to remove nodes with values multiple of 2,
 * then count removed and remained blocks
 */
static test_ret_t rb_tree_test_partially_nodes_remove(void)
{
	struct blksnap_volume *volume;
	struct blksnap_cow_block *blk, *next_blk;
	unsigned long block_num;
	unsigned long should_remain, should_remove;
	unsigned long num_nodes, num_remain, num_remove;
	struct rb_node *node;
	sector_t addr;

	mu_start(__func__);

	volume = create_volume(BLOCKS_CNT);
	mu_assert(volume, "Create volume");

	/* test inserting nodes */
	should_remove = should_remain = num_nodes = 0;
	for (block_num = 0; block_num < BLOCKS_CNT; block_num++) {
		if (block_num % 2)
			should_remain++;
		else
			should_remove++;
		addr = block_num * PAGE_SECTORS;
		blk = cow_tree_insert(volume, addr);
		mu_assert(blk, "Failed to insert cow_block [%lu]", addr);
		num_nodes++;
	}

	/* test if all is fine with inserted values */
	num_nodes = num_remain = num_remove = 0;
	for (node = rb_first(&volume->cow_tree); node; node = rb_next(node)) {
		blk = rb_entry(node, struct blksnap_cow_block, rb_node);
		num_nodes++;
		block_num = blk->addr / PAGE_SECTORS;
		if (block_num % 2)
			num_remain++;
		else
			num_remove++;
	}
	mu_assert_equal(num_nodes, BLOCKS_CNT,
			"Error walking tree with rb_next() (node) [%lu != %lu]",
			num_nodes, BLOCKS_CNT);
	mu_assert_equal(num_remain, should_remain,
			"Error walking tree with rb_next() (remain) [%lu != %lu]",
			num_remain, should_remain);
	mu_assert_equal(num_remove, should_remove,
			"Error walking tree with rb_next() (remove) [%lu != %lu]",
			num_remove, should_remove);

	/* test of removing some nodes */
	num_nodes = num_remain = num_remove = 0;
	rbtree_safe_to_remove_foreach(blk, next_blk,
				      &volume->cow_tree, rb_node) {
		num_nodes++;
		block_num = blk->addr / PAGE_SECTORS;
		if (block_num % 2) {
			num_remain++;
		} else {
			rb_erase(&blk->rb_node, &volume->cow_tree);
			kfree(blk);
			num_remove++;
		}
	}
	mu_assert_equal(num_nodes, BLOCKS_CNT,
			"Walking tree with rbtree_safe_to_remove_foreach() and remove nodes [%lu != %lu]",
			num_nodes, BLOCKS_CNT);
	mu_assert_equal(num_remain, should_remain,
			"Walking tree with rbtree_safe_to_remove_foreach() and remove nodes (remain) [%lu != %lu]",
			num_remain, should_remain);
	mu_assert_equal(num_remove, should_remove,
			"Walking tree with rbtree_safe_to_remove_foreach() and remove nodes (remove) [%lu != %lu]",
			num_remove, should_remove);

	/* test if all fine with remained nodes */
	num_nodes = num_remain = num_remove = 0;
	for (node = rb_first(&volume->cow_tree); node; node = rb_next(node)) {
		blk = rb_entry(node, struct blksnap_cow_block, rb_node);
		num_nodes++;
		block_num = blk->addr / PAGE_SECTORS;
		if (block_num % 2)
			num_remain++;
		else
			num_remove++;
	}
	mu_assert_equal(num_nodes, should_remain,
			"Walking tree with rb_next() after remove (node) [%lu != %lu]",
			num_nodes, should_remain);
	mu_assert_equal(num_remain, should_remain,
			"Walking tree with rb_next() after remove (remain) [%lu != %lu]",
			num_remain, should_remain);
	mu_assert_equal(num_remove, 0,
			"Walking tree with rb_next() after remove (remove) [%lu]",
			num_remove);

	/* test emtying tree */
	cow_tree_empty(volume);

	num_nodes = 0;
	for (node = rb_first(&volume->cow_tree); node; node = rb_next(node)) {
		blk = rb_entry(node, struct blksnap_cow_block, rb_node);
		num_nodes++;
	}
	mu_assert_equal(num_nodes, 0, "Tree is not empty [%lu]", num_nodes);

	mu_assert(bitmap_is_empty(volume), "Bitmap should be empty");
	free_volume(volume);

	mu_end(__func__);

	return 0;
}

static test_ret_t all_tests()
{
	mu_run_test(rb_tree_test_init_fill_destroy);
	mu_run_test(rb_tree_test_partially_nodes_remove);

	return 0;
}
