#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <stddef.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <linux/types.h>

#include <blksnap.h>
#include "bsctl-error.h"
#include "bsctl.h"
#include "bsctl-share.h"
#include "bsctl-ioctl.h"
#include "bsctl-chlog.h"

static struct blksnap_ioctl *send_ioctl_changelog(uint64_t offset,
						  uint32_t nextents,
						  const char *dev_path,
						  struct bsctl_error *err)
{
	size_t size;
	struct blksnap_ioctl *bsi;

	size = BSCTL_CALC_IOCTL_SIZE(struct blksnap_ioctl_changelog,
			nextents * sizeof(struct blksnap_ioctl_disk_extent));

	bsi = calloc(1, size);
	if (!bsi) {
		bsctl_error_set(err, errno, "ERROR: Failed to allocate memory for changelog ioctl");
		return NULL;
	}

	bsctl_init_ioctl(bsi);
	bsi->size = size;
	strncpy(bsi->changelog.name, dev_path, BLKSNAP_NAME_SIZE);
	bsi->changelog.nextents = nextents;
	bsi->changelog.offset = offset;

	if (bsctl_send_ioctl(BLKSNAP_IOC_CHANGELOG, bsi, dev_path, err)) {
		free(bsi);
		return NULL;
	}
	return bsi;
}

/* Returns zero both in case of 0-sized changelog and error, so *err should be
 * checked after call */
unsigned int bsctl_chlog_size(const char *dev_path, struct bsctl_error *err)
{
	unsigned int nblocks;
	struct blksnap_ioctl *bsi;

	/* By sending an ioctl with a 0 offset and
	 * 0 nextents, the ioctl will return the total size */
	bsi = send_ioctl_changelog(0, 0, dev_path, err);
	if (!bsi)
		return 0;

	nblocks = bsi->changelog.nblocks;
	free(bsi);

	return nblocks;
}

struct blksnap_ioctl *bsctl_chlog_get_chunk(uint64_t offset,
					    const char *dev_path,
					    struct bsctl_error *err)
{
	struct blksnap_ioctl *bsi;
	size_t max_nextents = elems_in_size(PAGE_SIZE,
					    sizeof(bsi->changelog),
					    sizeof(bsi->changelog.extents[0]));

	bsi = send_ioctl_changelog(offset, (uint64_t) max_nextents, dev_path,
				   err);
	if (!bsi) {
		return NULL;
	}

	if (!bsi->changelog.nextents) {
		free(bsi);
		return NULL;
	}

	return bsi;
}

/* an iterator of changelog chunks that accept a function pointer
 * to process given chunks */
struct changelog_stat *bsctl_chlog_process(const char *dev_path,
					   unsigned int total_blocks,
					   chlog_cb func, void *data,
					   struct bsctl_error *err)
{
	struct blksnap_ioctl *bsi;
	unsigned int i;
	unsigned long offset = 0;
	size_t max_nextents = elems_in_size(PAGE_SIZE,
					    sizeof(bsi->changelog),
					    sizeof(bsi->changelog.extents[0]));
	struct changelog_stat *stat = calloc(1, sizeof(struct changelog_stat));

	if (!stat) {
		bsctl_error_set(err, errno, "ERROR: Failed to allocate memory for changelog stats");
		return NULL;
	}

	bsi = bsctl_chlog_get_chunk(0, dev_path, err);
	if (!bsi) {
		free(stat);
		return NULL;
	}

	while (bsi) {
		/* save nextents to check later if
		 * chunk is not non-fullfilled */
		unsigned long nextents = bsi->changelog.nextents;

		stat->total_chunks++;
		stat->total_extents += nextents;
		for (i = 0; i < nextents; i++) {
			if (func(&(bsi->changelog.extents[i]), total_blocks,
				 data, err)) {
				free(bsi);
				free(stat);
				return NULL;
			}

			stat->total_blocks += bsi->changelog.extents[i].nblocks;
			offset = bsi->changelog.extents[i].start_sect
				+ (bsi->changelog.extents[i].nblocks - 1)
				  *  PAGE_SECTORS;
		}

		free(bsi);

		/* if current chunk is not full-filled
		 * - it's the last chunk we have */
		if (nextents < max_nextents)
			break;

		bsi = bsctl_chlog_get_chunk(offset, dev_path, err);
		if (!bsi) {
			free(stat);
			return NULL;
		}
	}
	return stat;
}
