#include <asm/types.h>
#include <dirent.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <inttypes.h>
#include <libgen.h>
#include <limits.h>
#include <linux/fs.h>
#include <linux/fiemap.h>
#include <math.h>
#include <mntent.h>
#include <poll.h>
#include <signal.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/statvfs.h>
#include <unistd.h>
#include <glob.h>

#include "blksnap.h"
#include "bsctl-error.h"
#include "bsctl.h"
#include "bsctl-share.h"
#include "bsctl-chlog.h"
#include "version.h"
#include "bsctl-log.h"

#define DEFAULT_LOG_LEVEL	LOG_NOTICE

struct store_size {
	double percent;
	uint64_t size;
};

struct bsctl_action {
	int action;
	int index;
	int log_pri;
	char dev_path[PATH_MAX];
	char opt_path[PATH_MAX];
	struct store_size store_size;
	int flags;
	enum map_method method;
	struct store_size buffer_size;
};

#define STATUS_LINE_FMT "%-20s %-10s %-8s %-5s %-5s %-4s %-8s %-4s\n"

#define show_line(fmt, args...) do { \
		fprintf(stdout, fmt, ##args); \
	} while(0)

enum tristate {
	BS_YES,
	BS_NO,
	BS_UNDEF
};

static inline enum tristate get_name_state(char * name)
{
	return name[0] == '\0' ? BS_UNDEF : BS_YES;
}

static inline enum tristate get_flag_state(struct bsctl_device_info *info,
	unsigned long flag)
{
	enum tristate state = get_name_state(info->device_name);
	if (state == BS_UNDEF) {
		return state;
	}
	return (info->flags & flag) ? BS_YES : BS_NO;
}

static void device_show_info(struct bsctl_device_info *info)
{
	const char *bs_name[] = {"Yes", "No", "-"};
	const char *dirty = bs_name[get_flag_state(info,
						   BLKSNAP_DEVICE_DIRTY)];
	const char *chlog = bs_name[get_flag_state(info,
						   BLKSNAP_DEVICE_CHLOG)];
	const char *data = bs_name[get_flag_state(info, BLKSNAP_DEVICE_DATA)];
	const char *metadata =
		bs_name[get_flag_state(info, BLKSNAP_DEVICE_METADATA)];
	const char *incr = bs_name[get_flag_state(info, BLKSNAP_DEVICE_INCR)];


	const char *device = get_name_state(info->device_name) == BS_UNDEF ?
		bs_name[BS_UNDEF] : info->device_name;
	const char *snapshot = get_name_state(info->snapshot_name) == BS_UNDEF ?
		bs_name[BS_UNDEF] : info->snapshot_name;

	show_line(STATUS_LINE_FMT, info->name, device, snapshot, dirty, chlog,
		  data, metadata, incr);
}

static void device_show_header(void)
{
	show_line(STATUS_LINE_FMT, "Filesystem", "Device", "Snapshot", "Dirty",
		  "ChLog", "Data", "Metadata", "Incr");
}

/* NOTE: str could be changed in this function in case it
 * contains percent
 */
static int parse_opt_size(char *str, struct store_size *ss)
{
	size_t len;

	if (!str || !ss)
		return EINVAL;
	len = strlen(str);
	if (!len)
		return EINVAL;
	/* Looking for % at the end of parameter */
	if (str[len - 1] == '%') {
		/* Getting data size in percents */
		str[len] = '\0';
		ss->percent = strtof(str, (char **) NULL);
		if (!ss->percent || ss->percent > 100.0) {
			ss->percent = 0;
			return ERANGE;
		}
	} else {
		/* Expecting that size is in bytes
		 * Using strtold for big amount */
		ss->size = strtold(str, (char **) NULL);
		if (!ss->size || ss->size == HUGE_VALL) {
			ss->size = 0;
			return ERANGE;
		}
	}
	return 0;
}

static int print_extents(struct blksnap_ioctl_disk_extent *extent,
			 unsigned int UNUSED(total_blocks),
			 void * UNUSED(data),
			 struct bsctl_error * UNUSED(err))
{
	fprintf(stdout, "%llu %zu\n", extent->start_sect, extent->nblocks);

	return 0;
}

static int print_extent_blocks(struct blksnap_ioctl_disk_extent *extent,
			       unsigned int UNUSED(total_blocks),
			       void * UNUSED(data),
			       struct bsctl_error * UNUSED(err))
{
	unsigned long blk;
	unsigned int  i;

	for (blk = extent->start_sect, i = 0;
	     i < extent->nblocks * PAGE_SECTORS;
	     i += PAGE_SECTORS) {
		fprintf(stdout, "%lu\n", blk + (unsigned long) i);
	}

	return 0;
}

static int drop_caches(struct bsctl_error *err)
{
	int fd, ret = 0;
	char value[] = "3\n";
	char cachefile[] = "/proc/sys/vm/drop_caches";

	sync();

	if ((fd = open(cachefile, O_WRONLY)) < 0) {
		ret = bsctl_err_set(err, errno, "failed to open %s [%d]",
				    cachefile, errno);
		return ret;
	}

	if (write(fd, value, sizeof(value) - 1) == -1)
		ret = bsctl_err_set(err, errno, "failed to write to %s [%d]",
				    cachefile, errno);

	close(fd);
	return ret;
}

__attribute__ ((noreturn))
static void usage_brief(const char *name)
{
	printf("Brief help: %s [[option]...] command [{device}]\n", name);

	printf(
"    -H|--help            full help\n"
"    -[vlD], -[adcisexSCk] {device}\n"
"    --[create|map|reset|check]-chlog-store {device}\n"
"    --[create|map|unmap|check]-data-store {device}\n"
"    --[map|update]-metadata-store {device}\n"
	);

	exit(EXIT_SUCCESS);
}

__attribute__ ((noreturn))
static void usage(const char *name, FILE *out)
{
	fprintf(out,
		"Usage: %s [ACTION [DEVICE]] [OPTIONS...]\n\n", name);

	fprintf(out,
"  General Options:\n"
"    -h                             show brief help\n"
"    --help                         show this help message\n"
"    -v, --version                  show version information\n"
"\n"
"  Setup Options:\n"
"    --store-size [size|percent]\n"
"        OPTIONS:\n"
"        [size|percent]             parameter can be given in bytes/percents\n"
"                                   100 > percent > 0\n"
"                                   size > 0\n"
"        EXAMPLES:\n"
"            --store-size 20%%      in percent of total device size\n"
"            --store-size 1000000   in byte\n"
"\n"
"    --source-device [DEV]\n"
"        OPTIONS:\n"
"        [DEV]                      should be a full path to block device\n"
"\n"
"    --file-path [PATH]\n"
"        OPTIONS:\n"
"        [PATH]                     path to a file to be mapped\n"
"\n"
"    -I, --incremental              tells snapshot should be incremental\n"
"\n"
"    -E, --extents                  tells changelog should gets in extents\n"
"\n"
"  Device Management:\n"
"    -a, --attach-to-device         attach to a block device\n"
"    -d, --detach-from-device       detach from a block device\n"
"    -c, --check-attached-device    check device attached to a block device\n"
"    -l, --list-devices             list currently attached devices\n"
"    -i, --show-device-info         show info of given device\n"
"\n"
"  Snapshot Control:\n"
"    -s, --start-snapshot           initiate a snapshot for a block device\n"
"    -e, --complete-snapshot        signal successful completion of a\n"
"                                   snapshot\n"
"    -x, --cancel-snapshot          cancel an existing snapshot and\n"
"                                   include block addresses in the next\n"
"                                   requested snapshot\n"
"\n"
"  Snapshot Information:\n"
"    -S, --show-snapshot            display the name of the snapshot device\n"
"    -C, --changelog                display the list of block addresses whose\n"
"                                   contents were modified since the last\n"
"                                   snapshot\n"
"\n"
"  Volume Restore:\n"
"        --restore-volume           initiate a volume restore from one block\n"
"                                   device to another\n"
"        --reset-volume             cancel an existing restore operation\n"
"\n"
"  On-Disk Storage:\n"
"        --create-chlog-store       create on-disk storage for block changelog\n"
"        --check-chlog-store        check changelog store for integrity\n"
"        --map-chlog-store          map block changelog store into kernel\n"
"        --reset-chlog-store        zero an exiting block changelog\n"
"\n"
"        --create-data-store        create on-disk storage for COW data\n"
"        --check-data-store         check data store for presence and\n"
"                                   integrity. \n"
"        --map-data-store           map COW snapshot data store into kernel\n"
"        --unmap-data-store         unmap COW snapshot data store from kernel\n"
"        --map-metadata-store       map metadata store into kernel\n"
"        --update-metadata-store    map metadata store and write value of\n"
"                                   superblock hash had been calculated during\n"
"                                   attach. Purposed for reabilitation volume\n"
"                                   after roll back.\n"
"  File Utilities:\n"
/* Undocumented features
 * "        --get-map-chlog-store      get store extents from module\n"
 * "        --get-map-data-store       get store extents from module\n"
 * "        --get-map-metadata-store   get store extents from module\n"
 */
"        --get-file-map             get file extents by system means or with\n"
"                                   module help (if former are failed).\n"
"                                   Output format: START_BLOCK BLOCKS_COUNT\n"
"                                   usage: [DEV_PATH] --file-path [FILE_PATH]"
"\n"
"  Debugging Utilities:\n"
"    -k, --dump-kernel              request kernel driver to log information\n"
"                                   pertaining to an attached device\n"
"    -D, --drop-caches              clear pagecache, dentries, and inodes\n"
"\n"
	);

	fprintf(out, "For more information see %s(8).\n", name);

	exit(out == stderr ? EINVAL : EXIT_SUCCESS);
}

__attribute__ ((noreturn))
static void version(char *name)
{
	bsctl_msg(LOG_NOTICE, "%s version %s (ioctl API %u, bsctl API %u)",
		basename(name),
		BLKSNAP_VERSION,
		BLKSNAP_IOCTL_VERSION,
		BSCTL_API_VERSION);
	exit(EXIT_SUCCESS);
}

__attribute__ ((noreturn))
static void api_version()
{
	bsctl_msg(LOG_NOTICE, "showing API version of bsctl: %u", BSCTL_API_VERSION);
	exit(EXIT_SUCCESS);
}

static int do_attach_device(struct bsctl_action *action)
{
	int ret = 0;
	struct bsctl_error err;

	bsctl_msg(LOG_INFO, "attaching device %s", action->dev_path);
	bsctl_err_reset(&err);
	ret = bsctl_attach(action->dev_path, &err);
	if (ret)
		bsctl_err(LOG_ERR, err.msg);
	else
		bsctl_msg(LOG_NOTICE, "device %s is attached", action->dev_path);

	return ret;
}

static int do_detach_device(struct bsctl_action *action)
{
	int ret = 0;
	struct bsctl_error err;

	bsctl_msg(LOG_INFO, "detaching device %s", action->dev_path);
	bsctl_err_reset(&err);
	ret = bsctl_detach(action->dev_path, &err);
	if (ret)
		bsctl_err(LOG_ERR, err.msg);
	else
		bsctl_msg(LOG_NOTICE, "device %s is detached", action->dev_path);

	return ret;
}

static int do_show_device_info(struct bsctl_action * UNUSED(action))
{
	int ret = 0;
	struct bsctl_device_info *info;
	struct bsctl_error err;

	bsctl_msg(LOG_INFO, "showing info for device %s", action->dev_path);
	bsctl_err_reset(&err);
	info = bsctl_device_info_get(action->dev_path, &err);
	if (info) {
		device_show_header();
		device_show_info(info);
		free(info);
	} else {
		bsctl_err(LOG_ERR, err.msg);
		ret = err.err;
	}

	return ret;
}

static int do_list_devices(struct bsctl_action * UNUSED(action))
{
	struct bsctl_device_list *info_list;
	struct bsctl_error err;
	int ret = 0;

	bsctl_msg(LOG_INFO, "showing info for all attached devices");
	bsctl_err_reset(&err);
	info_list = bsctl_device_list_get(&err);
	if (info_list) {
		int idx;
		device_show_header();
		for (idx = 0; idx < info_list->ndevices; idx++)
			device_show_info(&info_list->list[idx]);
		free(info_list);
	} else if (err.err) {
		bsctl_err(LOG_ERR, err.msg);
		ret = err.err;
	}

	return ret;
}

static int do_show_snapshot(struct bsctl_action *action)
{
	int ret = 0;
	char tmp_path[PATH_MAX];
	struct bsctl_error err;

	bsctl_msg(LOG_INFO, "showing snapshot name for device %s",
		  action->dev_path);
	bsctl_err_reset(&err);
	ret = bsctl_snapshot_get(action->dev_path, tmp_path, &err);
	if (ret)
		bsctl_err(LOG_ERR, err.msg);
	else
		bsctl_msg(LOG_NOTICE, "snapshot name for device %s is %s",
			  action->dev_path, tmp_path);
	return ret;
}

static int do_start_snapshot(struct bsctl_action *action)
{
	int ret = 0;
	char tmp_path[PATH_MAX];
	struct bsctl_error err;

	bsctl_msg(LOG_INFO, "starting snapshot for device %s",
		  action->dev_path);
	bsctl_err_reset(&err);
	ret = bsctl_snapshot_start(action->dev_path, tmp_path, action->flags,
				   &err);
	if (ret) {
		bsctl_err(LOG_ERR, err.msg);
	} else {
		bsctl_msg(LOG_ERR, "snapshot device is: %s", tmp_path);

		bsctl_msg(LOG_NOTICE, "snapshot %s for device %s is started",
			  tmp_path, action->dev_path);
	}
	return ret;
}

static int do_complete_snapshot(struct bsctl_action *action)
{
	int ret = 0;
	struct bsctl_error err;

	bsctl_msg(LOG_INFO, "completing snapshot for device %s",
		  action->dev_path);
	bsctl_err_reset(&err);
	ret = bsctl_snapshot_complete(action->dev_path, &err);
	if (ret)
		bsctl_err(LOG_ERR, err.msg);
	else
		bsctl_msg(LOG_NOTICE, "snapshot for device %s is completed",
			  action->dev_path);
	return ret;
}

static int do_cancel_snapshot(struct bsctl_action *action)
{
	int ret = 0;
	struct bsctl_error err;

	bsctl_msg(LOG_INFO, "cancelling snapshot for device %s",
		  action->dev_path);
	bsctl_err_reset(&err);
	ret = bsctl_snapshot_cancel(action->dev_path, &err);
	if (ret)
		bsctl_err(LOG_ERR, err.msg);
	else
		bsctl_msg(LOG_NOTICE, "snapshot for device %s is cancelled",
			  action->dev_path);
	return ret;
}

static int do_changelog(struct bsctl_action *action)
{
	struct changelog_stat *stat;
	struct bsctl_error err;
	int ret = 0;

	bsctl_msg(LOG_INFO, "showing changelog for device %s", action->dev_path);
	bsctl_err_reset(&err);
	if (action->flags & BLKSNAP_EXTENTS)
		stat = bsctl_chlog_process(action->dev_path, 0,
					   &print_extents, NULL,
					   &err);
	else
		stat = bsctl_chlog_process(action->dev_path, 0,
					   &print_extent_blocks, NULL,
					   &err);
	if (stat) {
		bsctl_msg(LOG_INFO, "total: chunks=%u extents=%u blocks=%u",
			  stat->total_chunks,
			  stat->total_extents,
			  stat->total_blocks);
		bsctl_msg(LOG_INFO, "changelog for device %s is displayed",
			  action->dev_path);
		free(stat);
	} else {
		bsctl_err(LOG_ERR, err.msg);
		ret = err.err;
	}

	return ret;
}

static int do_dump_kernel(struct bsctl_action *action)
{
	int ret = 0;
	struct bsctl_error err;

	bsctl_msg(LOG_INFO, "dumping kernel info to dmesg for device %s",
		  action->dev_path);
	bsctl_err_reset(&err);
	ret = bsctl_dump_kernel(action->dev_path, &err);
	if (ret)
		bsctl_err(LOG_ERR, err.msg);
	else
		bsctl_msg(LOG_NOTICE, "info for device %s is dumped to dmesg",
			  action->dev_path);
	return ret;
}

static int do_drop_caches(struct bsctl_action * UNUSED(action))
{
	int ret = 0;
	struct bsctl_error err;

	bsctl_msg(LOG_INFO, "dropping disk caches");
	bsctl_err_reset(&err);
	ret = drop_caches(&err);
	if (ret)
		bsctl_err(LOG_ERR, err.msg);
	else
		bsctl_msg(LOG_NOTICE, "disk caches are dropped");

	return ret;
}

static int do_create_data_store(struct bsctl_action *action)
{
	int ret = 0;
	struct bsctl_error err;

	bsctl_msg(LOG_INFO,
		  "creating data store for device %s (size=%llu percent=%.0f)",
		  action->dev_path,
		  (unsigned long long) action->store_size.size,
		  action->store_size.percent);
	bsctl_err_reset(&err);
	ret = bsctl_data_create(action->dev_path, action->store_size.size,
				action->store_size.percent, &err);
	if (ret)
		bsctl_err(LOG_ERR, err.msg);
	else
		bsctl_msg(LOG_NOTICE, "data store is created for device %s",
			  action->dev_path);

	return ret;
}

static int do_check_data_store(struct bsctl_action *action)
{
	int ret = 0;
	struct bsctl_error err;

	bsctl_msg(LOG_INFO, "checking data store for device %s", action->dev_path);
	bsctl_err_reset(&err);
	ret = __bsctl_data_check(action->dev_path, &action->store_size.size,
				 action->store_size.percent, &err);
	if (ret)
		bsctl_err(LOG_ERR, err.msg);
	else
		bsctl_msg(LOG_NOTICE,
			  "data store check with size %llu is passed for device %s",
			  (unsigned long long) action->store_size.size,
			  action->dev_path);

	return ret;
}

static int do_map_data_store(struct bsctl_action *action)
{
	int ret = 0;
	struct bsctl_error err;

	bsctl_msg(LOG_INFO, "mapping data store for device %s",
		  action->dev_path);
	bsctl_err_reset(&err);
	ret = bsctl_data_map_custom(action->dev_path, action->method,
				    action->buffer_size.size, &err);
	if (ret)
		bsctl_err(LOG_ERR, err.msg);
	else
		bsctl_msg(LOG_NOTICE, "data store for device %s is mapped",
			  action->dev_path);
	return ret;
}

static int do_unmap_data_store(struct bsctl_action *action)
{
	int ret = 0;
	struct bsctl_error err;

	bsctl_msg(LOG_INFO, "unmapping data store for device %s",
		  action->dev_path);
	bsctl_err_reset(&err);
	ret = bsctl_data_unmap(action->dev_path, &err);
	if (ret)
		bsctl_err(LOG_ERR, err.msg);
	else
		bsctl_msg(LOG_NOTICE, "data store for device %s is unmapped",
			  action->dev_path);
	return ret;
}

static int do_create_chlog_store(struct bsctl_action *action)
{
	int ret = 0;
	struct bsctl_error err;

	bsctl_msg(LOG_INFO, "creating changelog store for device %s",
		  action->dev_path);
	bsctl_err_reset(&err);
	ret = bsctl_chlog_create(action->dev_path, &err);
	if (ret)
		bsctl_err(LOG_ERR, err.msg);
	else
		bsctl_msg(LOG_NOTICE, "changelog store for device %s is created",
			  action->dev_path);
	return ret;
}

static int do_map_chlog_store(struct bsctl_action *action)
{
	int ret = 0;
	struct bsctl_error err;

	bsctl_msg(LOG_INFO, "mapping changelog store for device %s",
		  action->dev_path);
	bsctl_err_reset(&err);
	ret = bsctl_chlog_map_custom(action->dev_path, action->method,
				     action->buffer_size.size, &err);
	if (ret)
		bsctl_err(LOG_ERR, err.msg);
	else
		bsctl_msg(LOG_NOTICE, "changelog store for device %s is mapped",
			  action->dev_path);
	return ret;
}

static int do_check_chlog_store(struct bsctl_action *action)
{
	int ret = 0;
	struct bsctl_error err;

	bsctl_msg(LOG_INFO, "checking changelog store for device %s",
		  action->dev_path);
	bsctl_err_reset(&err);
	ret = bsctl_chlog_check(action->dev_path, &err);
	if (ret)
		bsctl_err(LOG_ERR, err.msg);
	else
		bsctl_msg(LOG_NOTICE,
			  "changelog store check is passed for device %s",
			  action->dev_path);
	return ret;
}

static int do_reset_chlog_store(struct bsctl_action *action)
{
	int ret = 0;
	struct bsctl_error err;

	bsctl_msg(LOG_INFO, "resetting changelog store for device %s",
		  action->dev_path);
	bsctl_err_reset(&err);
	ret = bsctl_chlog_reset(action->dev_path, &err);
	if (ret)
		bsctl_err(LOG_ERR, err.msg);
	else
		bsctl_msg(LOG_NOTICE, "changelog store for device %s is resetted",
			  action->dev_path);
	return ret;
}

static int do_map_metadata_store(struct bsctl_action *action)
{
	int ret = 0;
	struct bsctl_error err;

	bsctl_msg(LOG_INFO, "mapping metadata for device %s", action->dev_path);
	bsctl_err_reset(&err);
	ret = bsctl_metadata_map_custom(action->dev_path, action->method,
					action->buffer_size.size,
					0, &err);
	if (ret)
		bsctl_err(LOG_ERR, err.msg);
	else
		bsctl_msg(LOG_NOTICE, "metadata for device %s is mapped",
			  action->dev_path);
	return ret;
}

static int do_update_metadata_store(struct bsctl_action *action)
{
	int ret = 0;
	struct bsctl_error err;

	bsctl_msg(LOG_INFO, "updating metadata for device %s", action->dev_path);
	bsctl_err_reset(&err);
	ret = bsctl_metadata_map_custom(action->dev_path, action->method,
					action->buffer_size.size,
					BLKSNAP_STORE_UPDATE_METADATA, &err);
	if (ret)
		bsctl_err(LOG_ERR, err.msg);
	else
		bsctl_msg(LOG_NOTICE, "metadata for device %s is updated",
			  action->dev_path);
	return ret;
}

static void _print_map(struct bsctl_map *map)
{
	uint64_t i;
	for (i = 0; i < map->nextents; i++)
		printf("%llu, %llu\n", (unsigned long long)map->extents[i].start,
				(unsigned long long)map->extents[i].nsectors);
}

static int _get_map(const char* dev_path, enum blksnap_map_type map_type) {
	struct bsctl_error err;
	struct bsctl_map *map;

	bsctl_err_reset(&err);

	switch (map_type) {
	case BLKSNAP_MAP_TYPE_CHLOG:
		bsctl_msg(LOG_INFO, "getting map of chlog store for device %s",
			  dev_path);
		map = bsctl_chlog_get_map(dev_path, &err);
		break;
	case BLKSNAP_MAP_TYPE_DATA:
		bsctl_msg(LOG_INFO, "getting map of data store for device %s",
			  dev_path);
		map = bsctl_data_get_map(dev_path, &err);
		break;
	case BLKSNAP_MAP_TYPE_METADATA:
		bsctl_msg(LOG_INFO, "getting map of metadata store for device %s",
			  dev_path);
		map = bsctl_metadata_get_map(dev_path, &err);
		break;
	default:
		return EINVAL;
	}

	if (err.err)
		bsctl_err(LOG_ERR, err.msg);
	else
		_print_map(map);

	bsctl_map_free(map);
	return err.err;
}

static int do_get_map_chlog_store(struct bsctl_action *action)
{
	return _get_map(action->dev_path, BLKSNAP_MAP_TYPE_CHLOG);
}

static int do_get_map_data_store(struct bsctl_action *action)
{
	return _get_map(action->dev_path, BLKSNAP_MAP_TYPE_DATA);
}

static int do_get_map_metadata_store(struct bsctl_action *action)
{
	return _get_map(action->dev_path, BLKSNAP_MAP_TYPE_METADATA);
}

static int do_get_file_map(struct bsctl_action *action)
{
	uint64_t i;
	struct bsctl_aafmap *map;
	struct bsctl_error err;

	map = bsctl_aafmap_get(action->dev_path, action->opt_path, &err);

	if (!map) {
		bsctl_err(LOG_ERR, err.msg);
		return err.err;
	}

	for (i = 0; i < map->nextents; i++) {
		printf("%llu %llu\n",
		       (unsigned long long)map->extents[i].start_block,
		       (unsigned long long)map->extents[i].nblocks);
	}

	return 0;
}

static int do_restore_volume(struct bsctl_action *action)
{
	int ret = 0;
	struct bsctl_error err;

	bsctl_msg(LOG_INFO, "starting device %s restoration from %s",
		  action->dev_path, action->opt_path);
	bsctl_err_reset(&err);
	ret = bsctl_volume_restore(action->opt_path, action->dev_path, &err);
	if (ret)
		bsctl_err(LOG_ERR, err.msg);
	else
		bsctl_msg(LOG_NOTICE, "device %s restoration from %s is started",
			  action->dev_path, action->opt_path);
	return ret;
}

static int do_reset_volume(struct bsctl_action *action)
{
	int ret = 0;
	struct bsctl_error err;

	bsctl_msg(LOG_INFO, "reset device %s restoration", action->dev_path);
	bsctl_err_reset(&err);
	ret = bsctl_volume_reset(action->dev_path, &err);
	if (ret)
		bsctl_err(LOG_ERR, err.msg);
	else
		bsctl_msg(LOG_NOTICE, "device %s restoration is stopped",
			  action->dev_path);
	return ret;
}

static int do_check_device(struct bsctl_action *action)
{
	int ret = 0;
	struct bsctl_error err;

	bsctl_msg(LOG_INFO, "checking device %s", action->dev_path);
	bsctl_err_reset(&err);
	ret = bsctl_check(action->dev_path, &err);
	if (ret)
		bsctl_err(LOG_ERR, err.msg);
	else
		bsctl_msg(LOG_NOTICE, "device %s is checked", action->dev_path);

	return ret;
}

int main(int argc, char **argv)
{
	char *program_name = argv[0];
	int action, opt_ind = 0, ret = 0;
	struct bsctl_action act;

	/* init struct bsctl_action with default vaules */
	act.action = 0;
	act.log_pri = DEFAULT_LOG_LEVEL;
	act.flags = 0;
	act.method = MAP_AUTO;

	/* bufsize == 0 performs using default buffer size */
	act.buffer_size.size	= 0;
	act.buffer_size.percent	= 0;

	/* STORE_DATA_RATE is a part of value, so it's required to
	 * convert it to percents
	 */
	act.store_size.percent = STORE_DATA_RATE * 100.0;
	act.store_size.size    = 0;

	bsctl_set_log_priority(&act.log_pri);

	if (argc < 2) /* no parameters */
		usage_brief(program_name);

	/* When you add an option here, make sure to update usage() */
	const char *sopts = "TBZMIEFhHvVa:d:c:C:De:k:li:S:s:x:";
	struct option lopts[] = {
		/* log options */
		{ "debug",			0, &act.log_pri, LOG_DEBUG },
		{ "verbose",			0, &act.log_pri, LOG_INFO },
		{ "quiet",			0, &act.log_pri, LOG_EMERG },

		/* setup options */
		{ "store-size",			1, 0, 'T' },
		{ "buffer-size",		1, 0, 'B' },
		{ "source-device",		1, 0, 'Z' },
		{ "map-method",			1, 0, 'M' },
		{ "incremental",		0, 0, 'I' },
		{ "extents",			0, 0, 'E' },
		{ "file-path",			1, 0, 'F' },

		/* main commands with short option */
		{ "help",			0, 0, 'H' },
		{ "version",			0, 0, 'v' },
		{ "api-version",		0, 0, 'V' },
		{ "attach-to-device",		1, 0, 'a' },
		{ "detach-from-device",		1, 0, 'd' },
		{ "check-attached-device",	1, 0, 'c' },
		{ "show-device-info",		1, 0, 'i' },
		{ "list-devices",		0, 0, 'l' },
		{ "show-snapshot",		1, 0, 'S' },
		{ "start-snapshot",		1, 0, 's' },
		{ "complete-snapshot",		1, 0, 'e' },
		{ "cancel-snapshot",		1, 0, 'x' },
		{ "changelog",			1, 0, 'C' },
		{ "dump-kernel",		1, 0, 'k' },
		{ "drop-caches",		0, 0, 'D' },

		/* these commands do not have a short option */
		{ "create-data-store",		1, 0, '%' },
		{ "check-data-store",		1, 0, '&' },
		{ "map-data-store",		1, 0, '^' },
		{ "unmap-data-store",		1, 0, '>' },
		{ "get-map-data-store",		1, 0, '#' },
		{ "create-chlog-store",		1, 0, '1' },
		{ "check-chlog-store",		1, 0, '3' },
		{ "map-chlog-store",		1, 0, '2' },
		{ "reset-chlog-store",		1, 0, '4' },
		{ "get-map-chlog-store",	1, 0, '5' },
		{ "map-metadata-store",		1, 0, '<' },
		{ "update-metadata-store",	1, 0, '_' },
		{ "get-map-metadata-store",	1, 0, '$' },
		{ "get-file-map",		1, 0, '=' },
		{ "restore-volume",		1, 0, '~' },
		{ "reset-volume",		1, 0, '!' },

		{ NULL,				0, 0,  0  },
	};

	while ((action =
		getopt_long(argc, argv, sopts, lopts, &opt_ind)) != -1) {
		switch (action) {
		case 0: /* --debug, --verbose, --quiet */
			bsctl_set_log_priority(&act.log_pri);
			break;
		case 'T': /* --store-size */
			ret = parse_opt_size(optarg, &act.store_size);
			if (ret)
				bsctl_err(LOG_ERR, "invalid --store-size parameter");
			else
				bsctl_msg(LOG_DEBUG, "set store size to %.1f%% and %llu bytes",
					  act.store_size.percent,
					  (unsigned long long)
					  act.store_size.size);
			break;
		case 'B': /* --buffer-size */
			ret = parse_opt_size(optarg, &act.buffer_size);
			if (ret)
				bsctl_err(LOG_ERR, "invalid --buffer-size parameter");
			else
				bsctl_msg(LOG_DEBUG, "set buffer size to %.1f%% and %llu bytes",
					  act.buffer_size.percent,
					  (unsigned long long)
					  act.buffer_size.size);
			break;
		case 'M': /* --map-method */
			if (!strncmp("auto", optarg, strlen(optarg)))
					act.method = MAP_AUTO;
			else if (!strncmp("emulation", optarg, strlen(optarg)))
					act.method = MAP_EMULATION;
			else if (!strncmp("fiemap", optarg, strlen(optarg)))
					act.method = MAP_FIEMAP;
			else if (!strncmp("fibmap", optarg, strlen(optarg)))
					act.method = MAP_FIBMAP;
			break;
		case 'Z': /* --source-device */
		case 'F': /* --file-path */
			snprintf(act.opt_path, PATH_MAX, "%s", optarg);
			break;
		case 'I': /* --incremental */
			act.flags += BLKSNAP_START_INCR;
			break;
		case 'E': /* --extents */
			act.flags += BLKSNAP_EXTENTS;
			break;
		case 'h': /* -h */
			usage_brief(program_name);
			break;
		case 'H': /* --help */
			usage(program_name, stdout);
			break;
		case 'v': /* --version */
			version(program_name);
			break;
		case 'V': /* --api-version */
			api_version();
			break;
		/* all other options treat as command action.
		 * Just store what to do after process all options.
		 */
		default:
			if (act.action != 0) {
				bsctl_err(LOG_ERR, "%s allows one only action command per run",
					  program_name);
				usage(program_name, stderr);
			}
			act.action = action;
			act.index = opt_ind;
			/* GNU getopt_long() set optarg anytime so it is not needed to check it */
			snprintf(act.dev_path, PATH_MAX, "%s", optarg);
			break;
		}
	}

	switch (act.action) {
	case 'a': /* --attach-device */
		ret = do_attach_device(&act);
		break;
	case 'd': /* --detach-device */
		ret = do_detach_device(&act);
		break;
	case 'c': /* --check-attached-device */
		ret = do_check_device(&act);
		break;
	case 'i': /* --show-device-info */
		ret = do_show_device_info(&act);
		break;
	case 'l': /* --list-devices */
		ret = do_list_devices(&act);
		break;
	case 'S': /* --show-snapshot */
		ret = do_show_snapshot(&act);
		break;
	case 's': /* --start-snapshot */
		ret = do_start_snapshot(&act);
		break;
	case 'e': /* --complete-snapshot */
		ret = do_complete_snapshot(&act);
		break;
	case 'x': /* --cancel-snapshot */
		ret = do_cancel_snapshot(&act);
		break;
	case 'C': /* --changelog */
		ret = do_changelog(&act);
		break;
	case 'k': /* --dump-kernel */
		ret = do_dump_kernel(&act);
		break;
	case 'D': /* --drop-caches */
		ret = do_drop_caches(&act);
		break;
	case '%': /* --create-data-store */
		ret = do_create_data_store(&act);
		break;
	case '&': /* --check-data-store */
		ret = do_check_data_store(&act);
		break;
	case '^': /* --map-data-store */
		ret = do_map_data_store(&act);
		break;
	case '>': /* --unmap-data-store */
		ret = do_unmap_data_store(&act);
		break;
	case '#': /* get-map-data-store */
		ret = do_get_map_data_store(&act);
		break;
	case '1': /* --create-chlog-store */
		ret = do_create_chlog_store(&act);
		break;
	case '2': /* --map-chlog-store */
		ret = do_map_chlog_store(&act);
		break;
	case '3': /* --check-chlog-store */
		ret = do_check_chlog_store(&act);
		break;
	case '4': /* --reset-chlog-store */
		ret = do_reset_chlog_store(&act);
		break;
	case '5': /* --get-map-chlog-store */
		ret = do_get_map_chlog_store(&act);
		break;
	case '<': /* --map-metadata-store */
		ret = do_map_metadata_store(&act);
		break;
	case '_': /* --update-metadata-store */
		ret = do_update_metadata_store(&act);
		break;
	case '$': /* --get-map-metadata-store */
		ret = do_get_map_metadata_store(&act);
		break;
	case '=': /* --get-file-map */
		ret = do_get_file_map(&act);
		break;
	case '~': /* --restore-volume */
		ret = do_restore_volume(&act);
		break;
	case '!': /* --reset-volume */
		ret = do_reset_volume(&act);
		break;
	default: /* unknown action command */
		bsctl_err(LOG_ERR, "invalid action command \"--%s\"",
			  lopts[act.index].name);
		usage(program_name, stderr);
		break;
	}

	return ret;
}
