#define _FILE_OFFSET_BITS 64

#include <asm/types.h>
#include <assert.h>
#include <blkid/blkid.h>
#include <dirent.h>
#include <endian.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <glob.h>
#include <inttypes.h>
#include <libgen.h>
#include <limits.h>
#include <linux/fs.h>
#include <linux/fiemap.h>
#include <linux/hdreg.h>
#include <mntent.h>
#include <poll.h>
#include <pthread.h>
#include <signal.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/statvfs.h>
#include <sys/statfs.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>

#include <blksnap.h>
#include <blksnap-fs-types.h>

#include "bsctl-error.h"
#include "bsctl.h"
#include "bsctl-share.h"
#include "bsctl-ioctl.h"
#include "bsctl-chlog.h"
#include "fibmap.h"
#include "btrfs/log-to-phys.h"

#include "version.h"

#define FALSE			0
#define TRUE			(!FALSE)

#ifndef FS_NOCOW_FL
#define FS_NOCOW_FL		0x00800000
#endif

#define STORE_DIRNAME		".blksnap"

#define STORE_CHLOG_FILE	"changelog"
#define STORE_DATA_FILE		"data"
#define STORE_METADATA_FILE	"metadata"
#define STORE_METADATA_SIZE	PAGE_SIZE

#define FILE_PAGES(size)	((uint64_t) ((double) (size) / 4096.0 + 0.5))

#define DEVICE_WAIT_TIME	500000
#define DEVICE_WAIT_COUNT	5

/* For BTRFS we have to freeze fs manually as it isn't frozen
 * in freeze_bdev(). Unfortunately, sometimes BTRFS kernel module stucks
 * on FIFREEZE ioctl call and never finishes it. As workaround we are trying
 * to freeze BTRFS, wait for FREEZE_TIMEOUT and if FIFREEZE is not finished
 * in time, we call FITHAW ioctl and try to freeze BTRFS again. Repeat
 * this process FREEZE_MAX_TRIES time.
 */
#define FREEZE_MAX_TRIES 	10
#define FREEZE_TIMEOUT		2

/* Step size when creating store files. Optimal size is 350Mb */
#define WRITE_STEP	(2 * 1024 * 350)
#define CHUNK_SIZE	(SECTOR_SIZE * WRITE_STEP)

#define MIN_SEARCH_STEP 4096

/* Constants from XFS standard to define XFS filesystem version */
#define XFS_VERSION_BITS	0x000f
#define XFS_VERSION_5		5
#define XFS_VERSION_MAGIC	0x58465342

enum store_action {
	STORE_ALLOC,
	STORE_RESET,
	STORE_CHECK
};

/* Increase memory region and initiate added memory.
 * In case of error, old ptr requires to be freed.
 */
static void *recalloc(void *ptr, size_t old_size, size_t new_size,
		      struct bsctl_error *err)
{
	char *tmp;

	if (!ptr) {
		bsctl_error_set(err, EINVAL, "ERROR: Invalid pointer in %s",
				__func__);
		return NULL;
	}

	if (old_size == new_size)
		return ptr;

	if (old_size > new_size) {
		bsctl_error_set(err, EINVAL,
				"ERROR: Header (%llu) MUST be less than new_size (%llu) in %s",
				old_size, new_size, __func__);
		return NULL;
	}

	tmp = realloc(ptr, new_size);
	if (!tmp) {
		bsctl_error_set(err, errno,
				"ERROR: Failed to reallocate %llu memory",
				new_size);
		return NULL;
	}

	memset(tmp + old_size, 0, new_size - old_size);

	return tmp;
}

static struct mntent *get_mount_entry(const char *dev_path,
				      const char *mnt_entries_path,
				      struct mntent *mnt,
				      char *buf, int buflen)
{
	FILE *fp;
	struct mntent *ent = NULL;
	char *rpp;
	char resolved_path [PATH_MAX];
	char resolved_entry[PATH_MAX];

	fp = fopen(mnt_entries_path, "r");
	if (!fp)
		return NULL;

	rpp = realpath(dev_path, resolved_path);
	if (rpp)
		dev_path = rpp;

	while ((ent = getmntent_r(fp, mnt, buf, buflen))) {
		rpp = realpath(ent->mnt_fsname, resolved_entry);
		if (!rpp)
			rpp = ent->mnt_fsname;

		if (!strcmp(dev_path, rpp))
			break;
	}

	fclose(fp);
	return ent;
}

#define MNTENT_STRINGS_BUF_SIZE 4096

static int get_mount_point(const char *dev_path, char *mnt_buf, int mnt_len)
{
	struct mntent ent;
	char ent_str_buf[MNTENT_STRINGS_BUF_SIZE];

	if (!get_mount_entry(dev_path, "/proc/mounts",
			     &ent, ent_str_buf, sizeof(ent_str_buf)))
		/* Workaround for Btrfs v4.2. On the snapshot' start name of
		 * original device in /proc/mounts is replaced with name of
		 * snapshot device */
		if (!get_mount_entry(dev_path, "/etc/mtab",
				     &ent, ent_str_buf, sizeof(ent_str_buf)))
			return -ENOENT;

	strncpy(mnt_buf, ent.mnt_dir, mnt_len);

	return 0;
}

static bool is_dev_mounted_rw(const char *dev_path)
{
	char buf[4096];
	struct mntent ent;

	/* have no mount */
	if (!get_mount_entry(dev_path, "/proc/mounts", &ent, buf, sizeof(buf)))
		/* Workaround for Btrfs v4.2. On the snapshot' start name of
		 * original device in /proc/mounts is replaced with name of
		 * snapshot device */
		if (!get_mount_entry(dev_path, "/etc/mtab", &ent, buf,
				     sizeof(buf)))
			return false;

	if (hasmntopt(&ent, MNTOPT_RW))
		return true;

	return false;
}

static int check_xfs_version(const char *dev_path, struct bsctl_error *err)
{
	int fd, ret = 0;
	struct xfs_sb_t {
		uint32_t sb_magicnum;
		char reserved[96];
		uint16_t sb_versionnum;
	} sb;

	fd = open(dev_path, O_RDONLY);
	if (fd == -1) {
		return bsctl_error_set(err, errno,
				       "failed to open xfs journal (open, %s): %s",
				       dev_path, strerror(errno));
	}

	if (read(fd, &sb, sizeof(sb)) == -1) {
		ret = bsctl_error_set(err, errno,
				      "failed to read xfs journal (read, %s): %s",
				      dev_path, strerror(errno));
		goto clean;
	}

	if (sb.sb_magicnum != be32toh(XFS_VERSION_MAGIC)) {
		ret = bsctl_error_set(err, EBADF,
				      "xfs  magic number mismatched on device %s",
				       dev_path);
		goto clean;
	}

	ret = be16toh(sb.sb_versionnum) & XFS_VERSION_BITS;

clean:
	close(fd);
	return ret;
}


static __u8 setup_attach_flags(const char *dev_path, int fs_type, struct bsctl_error *err)
{
	__u8 flags = 0;
	bsctl_err_reset(err);

	if (fs_type == BLKSNAP_FS_BTRFS) {
		flags |= is_dev_mounted_rw(dev_path) ? BLKSNAP_ATTACH_MOUNT_RW
						     : 0;
	} else if (fs_type == BLKSNAP_FS_XFS) {
		int xfs_ver = check_xfs_version(dev_path, err);
		if (err->err)
			return flags;
		if (xfs_ver == XFS_VERSION_5)
			flags |= BLKSNAP_ATTACH_XFS_VERSION_5;
	}

	return flags;
}

static int validate_path_length(const char *path, struct bsctl_error *err)
{
	if (strlen(path) > BLKSNAP_NAME_SIZE - 1)
		return bsctl_error_set(err, EINVAL,
				       "ERROR: Module interface doesn't accept path longer than %d symbols",
				       BLKSNAP_NAME_SIZE - 1);

	return 0;
}

struct do_freeze_fs {
	int fd;
	const char *dev_path;
	struct bsctl_error *err;
};

static void *__do_freeze_fs(void *arg)
{
	int level = 1;
	struct do_freeze_fs *dffs = (struct do_freeze_fs *) arg;

	bsctl_err_reset(dffs->err);

	if (ioctl(dffs->fd, FIFREEZE, &level) == -1) {
		bsctl_error_set(dffs->err,
				errno,
				"FIFREEZE ioctl failed for device %s",
				dffs->dev_path);
	}
	pthread_exit(NULL);
}

static int freeze_fs(const char *dev_path, struct bsctl_error *err)
{
	int fd, level = 1, ret = 0, cur_try = 1;
	struct bsctl_error error;
	char mnt[PATH_MAX];
	struct do_freeze_fs dffs;
	pthread_t thread;
	struct timespec timeout;

	/* need defined err in __do_freeze_fs() */
	if (err)
		err = &error;

	bsctl_err_reset(err);

	/* if device is not mounted, we don't need to freeze it */
	if (get_mount_point(dev_path, mnt, sizeof(mnt)))
		return 0;

	fd = open(mnt, O_RDONLY);
	if (fd < 0)
		return bsctl_error_set(err, errno, "ERROR: Failed to open %s",
				       mnt);
	dffs.fd = fd;
	dffs.dev_path = dev_path;
	dffs.err = err;

retry:
	ret = pthread_create(&thread, NULL, __do_freeze_fs, (void *) &dffs);
	if (ret) {
		bsctl_err_set(err, errno,
			      "failed to create freeze fs thread for %s", dev_path);
		goto close_fd;
	}

	clock_gettime(CLOCK_REALTIME, &timeout);
	timeout.tv_sec += FREEZE_TIMEOUT;
	ret = pthread_timedjoin_np(thread, NULL, &timeout);
	if (!ret) {
		if (dffs.err->err)
			ret = dffs.err->err;
		goto close_fd;
	}

	if (ret == EINVAL) {
		bsctl_err_set(err, errno,
			      "ALERT: wrong format of timeout for %s", dev_path);
		pthread_join(thread, NULL);
		goto close_fd;
	}

	if (ret == ETIMEDOUT) {
		if (ioctl(fd, FITHAW, &level) == -1) {
			ret = bsctl_error_set(err, errno,
					"FITHAW ioctl failed for device %s",
					dev_path);
			goto close_fd;
		}
		pthread_join(thread, NULL);
		if (++cur_try <= FREEZE_MAX_TRIES)
			goto retry;
	}

close_fd:
	close(fd);

	return ret;
}

static int thaw_fs(const char *dev_path, struct bsctl_error *err)
{
	int fd;
	int ret = 0;
	int level = 1;
	char mnt[PATH_MAX];

	/* if device is not mounted, we don't need to freeze it */
	if (get_mount_point(dev_path, mnt, sizeof(mnt)))
		return 0;

	fd = open(mnt, O_RDONLY);
	if (fd < 0)
		return bsctl_error_set(err, errno, "ERROR: Failed to open %s", mnt);

	if (ioctl(fd, FITHAW, &level)  == -1)
		ret = bsctl_error_set(err, errno,"FITHAW ioctl failed for device %s",
				      dev_path);

	close(fd);

	return ret;
}

static int get_fstype(const char *dev_path)
{
	unsigned int fs_idx = 0, ret = BLKSNAP_FS_UNKNOWN;
	char *fs_type_str = blkid_get_tag_value(NULL, "TYPE", dev_path);

	if (!fs_type_str)
		return BLKSNAP_FS_UNKNOWN;

	for (fs_idx = 0;
	     fs_idx < sizeof(blksnap_fs_types) / sizeof(struct blksnap_fs);
	     fs_idx++) {
		if (!strcmp(fs_type_str, blksnap_fs_types[fs_idx].name)) {
			ret = blksnap_fs_types[fs_idx].fs_type;
			break;
		}
	}

	free(fs_type_str);
	return ret;
}

static bool is_fstype_supported(const enum blksnap_fs_type fstype)
{
	bool ret = true;
	FILE *fp = NULL;
	char *line = NULL;
	size_t len = 0;
	ssize_t nread = 0;
	const char *fs_name = fstype_to_string(fstype);

	fp = fopen(BLKSNAP_UNSUPPORTED_FS_TYPES_FILE, "r");
	if (fp) {
		while ((nread = getline(&line, &len, fp)) != -1) {
			if (nread > 0 && line[nread - 1] == '\n') {
				line[nread - 1] = '\0';
			}

			if (strcmp(fs_name, line) == 0) {
				ret = false;
				break;
			}
		}

		free(line);
		fclose(fp);
	}

	return ret;
}

/* 1. allocates requested size of memory.
 * 2. if fail, divide size to 2, then goto 2
 *
 * @size:	ptr to size of object to be allocated
 * @min_size:	minimum size of allocated object
 * @reset:	reset allocated memory (true/false)
 * @Return:	pointer to allocated memory, if fail - NULL
 */
static void *alloc_resize(size_t *psize, size_t min_size, bool reset) {
	void *ptr = NULL;
	size_t cur_size = *psize;

	while (!ptr) {
		if (reset)
			ptr = calloc(1, cur_size);
		else
			ptr = malloc(cur_size);

		if (ptr)
			break;

		if ((cur_size / 2) < min_size)
			break;

		cur_size /= 2;
		if (cur_size % min_size) {
			size_t multiplexer = cur_size / min_size;
			cur_size = multiplexer * min_size;
		}
	}

	*psize = cur_size;

	return ptr;
}

/* Write special signature to each sector of file to give a module chance to
 * catch every bio-request and find out file addresses
 */
static int _mark_every_sector(const char *file_path, int buf_size_arg,
			      struct bsctl_error *err)
{
	uint64_t *mark_data = NULL, *orig_data = NULL;
	off_t offset, size;
	struct stat statinfo;
	int fd, ret = 0;
	int objs_per_sector = SECTOR_SIZE / sizeof(uint64_t);
	size_t buf_size = (buf_size_arg) ? buf_size_arg : CHUNK_SIZE;
	size_t buf_objs;
	size_t index;

	fd = open(file_path, O_RDWR | O_SYNC);
	if (fd < 0)
		return bsctl_error_set(err, errno, "ERROR: Failed to open %s",
				       file_path);

	if (fstat(fd, &statinfo) < 0) {
		ret = bsctl_error_set(err, errno, "ERROR: Failed to get size of %s",
				      file_path);
		goto close_file;
	} else {
		size = statinfo.st_size;
	}

	mark_data = alloc_resize(&buf_size, SECTOR_SIZE, true);
	if (!mark_data) {
		ret = bsctl_error_set(err, ENOMEM,
				      "ERROR: Failed to allocate memory for mark data");
		goto close_file;
	}

	orig_data = malloc(buf_size);
	if (!orig_data) {
		ret = bsctl_error_set(err, ENOMEM,
				      "ERROR: Failed to allocate memory for original data");
		goto free_mark_data;
	}

	buf_objs = buf_size / sizeof(uint64_t);

	/* mark each "sector" in mark data with signature */
	for (index = 0; index < buf_objs; index += objs_per_sector) {
		mark_data[index] = BLKSNAP_MAP_SIGN;
	}

	if (lseek(fd, 0, SEEK_SET) < 0) {
		ret = bsctl_err_set(err, errno, "ERROR: failed to seek file %s",
				    file_path);
		goto free_all;
	}

	/* write signatures to each sector of file */
	for (offset = 0; offset < size; offset += buf_size) {
		bool error = false;
		/* adjust buf_size if there is a tail
		 * of file smaller than buf_size
		 */
		if (offset + (off_t) buf_size >= size)
			buf_size = size - offset;
		lseek(fd, offset, SEEK_SET);
		if (read(fd, orig_data, buf_size) != (ssize_t) buf_size) {
			ret = bsctl_err_set(err, errno,
					    "ERROR: failed to read file %s [%zu:%zu]",
					    file_path, offset, buf_size);
			goto free_all;
		}
		lseek(fd, offset, SEEK_SET);
		if (write(fd, mark_data, buf_size) != (ssize_t) buf_size) {
			ret = bsctl_err_set(err, errno, "ERROR: failed to mark file %s",
					    file_path);
			error = true;
		}
		lseek(fd, offset, SEEK_SET);
		if (write(fd, orig_data, buf_size) != (ssize_t) buf_size) {
			ret = bsctl_err_set(err, errno, "ERROR: failed to write file %s",
					    file_path);
			goto free_all;
		}
		if (error)
			break;
	}

free_all:
	fsync(fd);
	free(orig_data);
free_mark_data:
	free(mark_data);
close_file:
	close(fd);
	return ret;
}

/* Get file map by writing/catching signatures to each sector of file
 * Returned addresses are absolute (from device beginning, not volume)
 * @buf_size:	size of single write step to a file, CHUNK_SIZE by default/0
 * @extents:	for output, not preallocated, must be freed
 * @err:	valid pointer to bsctl_error, CAN'T be NULL
 * Return:	count of file extents, must be freed
 */
static int _probe_file_map(const char *dev_path,
			   const char *file_path,
			   int buf_size,
			   struct blksnap_ioctl_file_extent **extents,
			   struct bsctl_error *err)
{
	int ret;
	int nextents = 0;
	size_t ioctl_size;
	struct blksnap_ioctl *args, *tmp;

	args = calloc(1, sizeof(struct blksnap_ioctl));
	if (!args) {
		bsctl_error_set(err, errno, "ERROR: Failed to allocate memory in %s",
				__func__);
		return 0;
	}

	bsctl_init_ioctl(args);
	strncpy(args->grab.name, dev_path, BLKSNAP_NAME_SIZE);

	/* 1) Start grabbing file addresses */
	args->grab.cmd = BLKSNAP_GRAB_MAP_START;

	if (bsctl_send_ioctl(BLKSNAP_IOC_GRAB_MAP, args, dev_path, err))
		goto out;

	if (_mark_every_sector(file_path, buf_size, err))
		goto cancel_grabbing;

	/* 2) Finish grabbing. Receive extents count. Realloc ioctl. */
	args->grab.cmd = BLKSNAP_GRAB_MAP_END;

	ret = bsctl_send_ioctl(BLKSNAP_IOC_GRAB_MAP, args, dev_path, err);
	if (ret) {
		if (ret == EIO)
			goto out;
		else
			goto cancel_grabbing;
	}

	ioctl_size = BSCTL_CALC_IOCTL_SIZE(struct blksnap_ioctl_grab_map,
		args->grab.nextents * sizeof(struct blksnap_ioctl_file_extent));

	tmp = recalloc(args, args->size, ioctl_size, err);
	if (!tmp) {
		bsctl_error_set(err, errno, "ERROR: Realloc failed in %s",
				__func__);
		goto cancel_grabbing;
	}

	args = tmp;
	args->size = ioctl_size;

	/* 3) Fetch extents */
	args->grab.cmd = BLKSNAP_GRAB_MAP_FETCH_EXT;
	if (bsctl_send_ioctl(BLKSNAP_IOC_GRAB_MAP, args, dev_path, err))
		goto cancel_grabbing;

	/* 4) Copy extents to output */
	*extents = calloc(args->grab.nextents,
			  sizeof(struct blksnap_ioctl_file_extent));
	if (!*extents) {
		bsctl_error_set(err, errno,
				"ERROR: Failed to allocated extents in %s", __func__);
		goto cancel_grabbing;
	}

	nextents = args->grab.nextents;
	memcpy(*extents, args->grab.extents,
	       nextents * sizeof(struct blksnap_ioctl_file_extent));

out:
	free(args);
	return nextents;

cancel_grabbing:
	args->grab.cmd = BLKSNAP_GRAB_MAP_CANCEL;
	bsctl_send_ioctl(BLKSNAP_IOC_GRAB_MAP, args, dev_path, err);
	goto out;
}


/* Manual emulation of fiemap ioctl
 * Get absolute addresses of file via writing/catching signatures to each
 * sector of file and convert them as if they were get via fiemap ioctl:
 * - start and length of extents are in bytes
 * - addresses are relative to volume start
 * Return:	fiemap structure, must be freed
 */
static struct fiemap *_fiemap_emulation(const char *dev_path,
					const char *file_path,
					int buf_size,
					struct bsctl_error *err)
{
	int i, nextents;
	uint64_t volume_start_sect;
	struct fiemap *result_fiemap = NULL;
	struct blksnap_ioctl_file_extent *extents = NULL;

	nextents = _probe_file_map(dev_path, file_path, buf_size,
				   &extents, err);
	if (!nextents)
		return NULL;

	result_fiemap = calloc(1, sizeof(struct fiemap) +
			       nextents * sizeof(struct fiemap_extent));
	if (!result_fiemap) {
		bsctl_error_set(err, errno, "ERROR: failed to allocate fiemap in %s",
				__func__);
		free(extents);
		return NULL;
	}

	/* probe_file_map returns absolute addresses from device start,
	 * but fiemap ioctl returns relative addresses
	 */
	volume_start_sect = bsctl_get_volume_start_sect(dev_path);

	result_fiemap->fm_mapped_extents = nextents;
	for (i = 0; i < nextents; i++) {
		result_fiemap->fm_extents[i].fe_physical =
			(extents[i].start_sect - volume_start_sect)
			<< SECTOR_SHIFT;
		result_fiemap->fm_extents[i].fe_length =
			extents[i].nsectors << SECTOR_SHIFT;
	}

	free(extents);
	return result_fiemap;
}

static struct fiemap *_fiemap_get(const char *file_path,
				  struct bsctl_error *err)
{
	int fd;
	struct fiemap *fiemap = NULL;
	struct fiemap *result_fiemap = NULL;
	struct fiemap *fm_tmp;
	uint32_t extent_num = 0;
	uint64_t start = 0, last_ext_end = 0, end;
	struct stat statinfo;

	fd = open(file_path, O_RDONLY | O_SYNC);

	if (fd < 0) {
		bsctl_error_set(err, errno, "ERROR: Failed to open %s",
				file_path);
		return NULL;
	}

	if (fstat(fd, &statinfo) < 0)
		end = UINT64_MAX;
	else
		end = statinfo.st_size;

	fiemap = malloc(sizeof(struct fiemap));
	if (!fiemap) {
		bsctl_error_set(err, errno,
				"ERROR: Failed to allocate memory for fiemap");
		goto fail_cleanup;
	}

	/* initial nullifying alloc */
	result_fiemap = calloc(1, sizeof(struct fiemap));
	if (!result_fiemap) {
		bsctl_error_set(err, errno,
				"ERROR: Failed to allocate memory for fiemap");
		goto fail_cleanup;
	}

	/* In regular case that loop is meaningless, but it seems like XFS 
	 * filesystem has incorrect implementation of fiemap ioctl and
	 * returns extents for only one block-group at a time, so we need
	 * to handle it manually, starting the next fiemap call from the end
	 * of the last extent */
	while (start < end) {
		uint32_t idx;

		memset(fiemap, 0, sizeof(struct fiemap));
		fiemap->fm_start = start;
		fiemap->fm_length = end;
		fiemap->fm_flags = FIEMAP_FLAG_SYNC;

		if (ioctl(fd, FS_IOC_FIEMAP, fiemap) == -1) {
			bsctl_error_set(err, errno,
					"FS_IOC_FIEMAP ioctl failed");
			goto fail_cleanup;
		}

		if (fiemap->fm_mapped_extents == 0) {
			break;
		}

		/* We have two fiemaps: the working one (fiemap) and the
		 * result one (result fiemap), so on each iteration we
		 * reallocating the working one for it to correspond to
		 * current number of extents, and the resulting one to hold
		 * all the extents for the whole file */
		fm_tmp = realloc(fiemap, sizeof(struct fiemap)
					  + (sizeof(struct fiemap_extent)
					     * fiemap->fm_mapped_extents));
		if (!fm_tmp) {
			bsctl_error_set(err, errno, "failed to realloc fiemap");
			goto fail_cleanup;
		} else {
			fiemap = fm_tmp;
		}

		memset(fiemap->fm_extents, 0,
		       sizeof(struct fiemap_extent)
		       * fiemap->fm_mapped_extents);
		fiemap->fm_extent_count = fiemap->fm_mapped_extents;
		fiemap->fm_mapped_extents = 0;
		fiemap->fm_start = start;
		fiemap->fm_length = end;
		fiemap->fm_flags = FIEMAP_FLAG_SYNC;

		if (ioctl(fd, FS_IOC_FIEMAP, fiemap) == -1) {
			bsctl_error_set(err, errno,
					"FS_IOC_FIEMAP ioctl failed");
			goto fail_cleanup;
		}

		fm_tmp = realloc(result_fiemap,
				 sizeof(struct fiemap)
				 + (sizeof(struct fiemap_extent)
				    * (extent_num + fiemap->fm_mapped_extents)
				   )
				);

		if (!fm_tmp) {
			bsctl_error_set(err, errno,
					"failed to realloc result fiemap");
			goto fail_cleanup;
		} else {
			result_fiemap = fm_tmp;
		}

		for (idx = 0; idx < fiemap->fm_mapped_extents; idx++) {
			uint64_t estart = fiemap->fm_extents[idx].fe_physical;
			uint64_t elen = fiemap->fm_extents[idx].fe_length;

			if (!extent_num || estart != last_ext_end) {
				memcpy(result_fiemap->fm_extents +
						extent_num++,
				       fiemap->fm_extents + idx,
				       sizeof(struct fiemap_extent));
				goto next_extent;
			}

			result_fiemap->fm_extents[extent_num - 1].fe_length +=
				elen;
next_extent:
			last_ext_end = estart + elen;
			start = fiemap->fm_extents[idx].fe_logical + elen;
		}

		if (fiemap->fm_extents[fiemap->fm_mapped_extents - 1].fe_flags &
		    FIEMAP_EXTENT_LAST) {
			break;
		}
	}

	if (!result_fiemap) {
		bsctl_error_set(err, EIO,
				"failed to map store via FS_IOC_FIEMAP");
		goto fail_cleanup;
	}

	result_fiemap->fm_mapped_extents = extent_num;

	close(fd);
	free(fiemap);
	return result_fiemap;

fail_cleanup:
	if (result_fiemap)
		free(result_fiemap);
	if (fiemap)
		free(fiemap);
	close(fd);
	return NULL;
}

static struct fiemap *_get_file_map(struct mapping_params *params,
				    struct bsctl_error *err)
{
	struct fiemap *fiemap;

	switch (params->method) {
	case MAP_FIEMAP:
		fiemap = _fiemap_get(params->file_path, err);
		goto btrfs_convert;
	case MAP_FIBMAP:
		fiemap = bsctl_get_fiemap_via_fibmap(params->file_path, err);
		goto btrfs_convert;
	case MAP_EMULATION:
		return _fiemap_emulation(params->dev_path, params->file_path,
					 params->bufsize, err);
	case MAP_AUTO:
		fiemap = _fiemap_get(params->file_path, err);
		if (fiemap)
			goto btrfs_convert;

		bsctl_err_reset(err);
		fiemap = bsctl_get_fiemap_via_fibmap(params->file_path, err);
		if (fiemap)
			goto btrfs_convert;

		bsctl_err_reset(err);
		return _fiemap_emulation(params->dev_path, params->file_path,
					 params->bufsize, err);
	default:
		bsctl_error_set(err, EINVAL, "ERROR: misusage of %s", __func__);
		return NULL;
	}

btrfs_convert:
	/* Btrfs fiemap/fibmap ioctls return addresses are needed to be
	 * converted with usage of the Btrfs chunk list.
	 * See details in comments above the log_to_phys func definition.
	 */
	if (get_fstype(params->dev_path) == BLKSNAP_FS_BTRFS && fiemap) {
		if (bsctl_btrfs_logical_to_physical(params->dev_path, fiemap,
						    err)) {
			free(fiemap);
			fiemap = NULL;
		}
	}
	return fiemap;
}

static int map_xfs_journal(const char *dev_path, uint64_t start,
			   struct blksnap_ioctl_file_extent ext[],
			   __u32 *nextents,
			   struct bsctl_error *err)
{
	int fd, ret = 0;
	uint32_t sect_in_block_log;
	struct xfs_sb_t {
		uint32_t sb_magicnum;
		char reserved0[44];
		uint64_t logstart;
		char reserved1[28];
		uint32_t agblocks;
		char reserved2[8];
		uint32_t logblocks;
		char reserved3[20];
		uint8_t blocklog;
		char reserved4[3];
		uint8_t agblklog;
	} sb;

	fd = open(dev_path, O_RDONLY);
	if (fd == -1) {
		return bsctl_error_set(err, errno,
				       "failed to map xfs journal (open, %s): %s",
				       dev_path, strerror(errno));
	}

	if (read(fd, &sb, sizeof(sb)) == -1) {
		ret = bsctl_error_set(err, errno,
				      "failed to map xfs journal (read, %s): %s",
				      dev_path, strerror(errno));
		goto clean;
	}

	if (sb.sb_magicnum != be32toh(0x58465342)) {
		ret = bsctl_error_set(err, EBADF,
				      "failed to map xfs journal: magic number mismatched on device %s",
				      dev_path);
		goto clean;
	}

	sb.logstart = be64toh(sb.logstart);
	sb.agblocks = be32toh(sb.agblocks);
	sb.logblocks = be32toh(sb.logblocks);
	sect_in_block_log = sb.blocklog - 9;

	if (!sb.logstart) {
		/* Partition has external log. It's not fatal since we can use
		 * ag0 metadata (with less reliability).
		 */
		goto clean;
	}

	*nextents = 1;
	if (!ext)
		goto clean;

	ext[0].start_sect =
		start + (uint64_t)
		(
		  ((uint64_t)(sb.logstart >> sb.agblklog) * sb.agblocks
		    + (sb.logstart & (((uint32_t)1 << sb.agblklog) - 1))
		  ) << sect_in_block_log
		);
	ext[0].nsectors = sb.logblocks << sect_in_block_log;
clean:
	close(fd);
	return ret;
}

static __u32 fstype_to_hashblocks(const char *dev_path,
				  enum blksnap_fs_type fstype,
				  struct blksnap_ioctl_file_extent ext[])
{
	/* all extent maps present in kernel module should be with absolute
	 * offset from the start of whole disk
	 */
	uint64_t volume_start_sect = bsctl_get_volume_start_sect(dev_path);
	uint32_t nextents = 0;

	switch (fstype) {
	case BLKSNAP_FS_BTRFS:
		nextents = 1;
		if (!ext)
			break;
		/* btrfs has superblock in 128th sector */
		ext[0].start_sect = volume_start_sect + 128;
		ext[0].nsectors = PAGE_SECTORS;
		break;
	case BLKSNAP_FS_XFS:
		if (map_xfs_journal(dev_path, volume_start_sect, ext,
				    &nextents, 0)) {
			nextents = 1;
			if (!ext)
				break;
			/* absolute offset */
			ext[0].start_sect = volume_start_sect;
			ext[0].nsectors = PAGE_SECTORS * 2;
		}
		break;
	default:
		nextents = 1;
		if (!ext)
			break;
		/* absolute offset */
		ext[0].start_sect = volume_start_sect;
		ext[0].nsectors = PAGE_SECTORS;
	}

	return nextents;
}

int bsctl_api_version(struct bsctl_error * UNUSED(err))
{
	return BSCTL_API_VERSION;
}

int bsctl_attach(const char *dev_path, struct bsctl_error *err)
{
	struct blksnap_ioctl *args;
	enum blksnap_fs_type fstype;
	size_t ioctl_size;
	int ret = 0, nextents;

	ret = validate_path_length(dev_path, err);
	if (ret)
		return ret;

	fstype = get_fstype(dev_path);
	if (fstype == BLKSNAP_FS_UNKNOWN)
		return bsctl_error_set(err, EINVAL,
				       "ERROR: Unknown filesystem on device %s",
				       dev_path);

	if (!is_fstype_supported(fstype))
		return bsctl_error_set(err, EINVAL,
				       "ERROR: Unsupported filesystem on device %s",
				       dev_path);

	/* idle call to know count of extents for hashing */
	nextents = fstype_to_hashblocks(dev_path, fstype, NULL);

	ioctl_size = BSCTL_CALC_IOCTL_SIZE(struct blksnap_ioctl_attach,
			  nextents * sizeof(struct blksnap_ioctl_file_extent));

	args = calloc(1, ioctl_size);
	if (!args)
		return bsctl_error_set(err, errno,
				       "ERROR: Insufficient memory to attach %s",
				       dev_path);

	bsctl_init_ioctl(args);
	args->size = ioctl_size;
	strncpy(args->attach.name, dev_path, BLKSNAP_NAME_SIZE);
	args->attach.fstype = fstype;

	args->attach.nextents = fstype_to_hashblocks(dev_path,
						     args->attach.fstype,
						     args->attach.hashblocks);

	args->attach.flags = setup_attach_flags(dev_path, args->attach.fstype, err);
	if (err->err)
		return err->err;

	ret = bsctl_send_ioctl(BLKSNAP_IOC_ATTACH, args, dev_path, err);

	free(args);
	return ret;
}

int bsctl_detach(const char *dev_path, struct bsctl_error *err)
{
	struct blksnap_ioctl *args;
	int ret = 0;

	ret = validate_path_length(dev_path, err);
	if (ret)
		return ret;

	args = calloc(1, sizeof(struct blksnap_ioctl));
	if (!args)
		return bsctl_error_set(err, errno,
				       "ERROR: Insufficient memory to detach %s",
				       dev_path);

	bsctl_init_ioctl(args);
	strncpy(args->detach.name, dev_path, BLKSNAP_NAME_SIZE);

	ret = bsctl_send_ioctl(BLKSNAP_IOC_DETACH, args, dev_path, err);

	free(args);
	return ret;
}

/* We could simply memcpy the structures, but with following approach in case
 * ioctl structures got chaged, we'll receive compile error, that could prevent
 * some dangerous errors */
static void _copy_device_info_from_ioctl(struct bsctl_device_info *dst,
					 struct blksnap_ioctl_device *src)
{
	strncpy(dst->name, src->name, BLKSNAP_NAME_SIZE);
	strncpy(dst->device_name, src->device_name, BLKSNAP_NAME_SIZE);
	strncpy(dst->snapshot_name, src->snapshot_name, BLKSNAP_NAME_SIZE);
	strncpy(dst->link_name, src->link_name, BLKSNAP_NAME_SIZE);
	dst->flags = src->flags;
}

struct bsctl_device_list *bsctl_device_list_get(struct bsctl_error *err)
{
	struct blksnap_ioctl *args, *tmp;
	struct bsctl_device_list *list = NULL;
	uint32_t i;
	size_t header = sizeof(struct blksnap_ioctl);
	int rc = 0;
	struct bsctl_error local_err;

	if (!err)
		err = &local_err;

	bsctl_err_reset(err);

	args = calloc(1, header);
	if (!args) {
		bsctl_error_set(err, errno, "ERROR: Failed to allocate memory for bsi");
		return NULL;
	}

	bsctl_init_ioctl(args);

	for (i = 0; i < DEVICE_WAIT_COUNT; i++) {
		args->list.ndevices = BLKSNAP_DL_REQ;

		if (bsctl_send_ioctl(BLKSNAP_IOC_LIST, args, NULL, err))
			goto clean;

		if (!args->list.ndevices)
			goto clean;

		args->size = BSCTL_CALC_IOCTL_SIZE(struct blksnap_ioctl_list,
					args->list.ndevices
					* sizeof(struct blksnap_ioctl_device));

		tmp = recalloc(args, header, args->size, err);
		if (!tmp) {
			bsctl_error_set(err, err->err, "ERROR: Failed to allocate memory for ioctl device list");
			goto clean;
		}

		args = tmp;

		rc = bsctl_send_ioctl(BLKSNAP_IOC_LIST, args, NULL, err);

		if (!rc)
			break;

		if (rc != EAGAIN) {
			bsctl_error_set(err, rc, "ERROR: Failed to perform ioctl");
			goto clean;
		}
	}

	if (rc)
		goto clean;

	list = calloc(1,
		      sizeof(struct bsctl_device_list) +
		      (args->list.ndevices *
		       sizeof(struct bsctl_device_info)));
	if (!list) {
		bsctl_error_set(err, errno, "ERROR: Failed to allocate memory for outgoing device list");
		goto clean;
	}

	list->ndevices = args->list.ndevices;
	for (i = 0; i < list->ndevices; ++i) {
		_copy_device_info_from_ioctl(&list->list[i],
					     &args->list.devices[i]);
	}

clean:
	free(args);
	return list;
}

void bsctl_device_list_free(struct bsctl_device_list *list)
{
	free(list);
}

static int _get_path_statfs(const char *file_path,
			    struct statfs *fsinfo,
			    struct bsctl_error *err)
{
	int fd, ret = 0;

	fd = open(file_path, O_RDONLY);
	if (fd < 0)
		return bsctl_error_set(err, errno, "ERROR: failed to open %s",
				       file_path);

	if (fstatfs(fd, fsinfo))
		ret = bsctl_error_set(err, errno, "ERROR: failed statfs on %s",
				      file_path);

	close(fd);

	return ret;
}

/* Attention: for using fiemap_emulation functionality, device must be
 * attached to module.
 */
struct bsctl_aafmap *bsctl_aafmap_get(const char *dev_path,
				      const char *file_path,
				      struct bsctl_error *err)
{
	int i, nextents;
	blksize_t block_size;
	struct statfs fsinfo;
	struct fiemap *fiemap;
	struct mapping_params params;
	struct bsctl_aafmap *aafmap = NULL;

	bzero(&params, sizeof(params));
	params.dev_path = dev_path;
	params.method = MAP_AUTO;
	strncpy(params.file_path, file_path, sizeof(params.file_path));

	fiemap = _get_file_map(&params, err);
	if (!fiemap)
		return NULL;

	/* convert fiemap to aafmap */
	if (_get_path_statfs(file_path, &fsinfo, err))
		goto free_fiemap;

	block_size = fsinfo.f_bsize;
	nextents = fiemap->fm_mapped_extents;

	aafmap = calloc(1, sizeof(struct bsctl_aafmap) +
			   nextents * sizeof(struct bsctl_aaf_extent));
	if (!aafmap) {
		bsctl_error_set(err, errno, "ERROR: Failed to allocate aafmap");
		goto free_fiemap;
	}

	aafmap->nextents = nextents;

	for (i = 0; i < nextents; i++) {
		aafmap->extents[i].start_block =
			fiemap->fm_extents[i].fe_physical / block_size;
		aafmap->extents[i].nblocks =
			fiemap->fm_extents[i].fe_length / block_size;
	}

free_fiemap:
	free(fiemap);
	return aafmap;
}

void bsctl_aafmap_free(struct bsctl_aafmap *map)
{
	free(map);
}

struct bsctl_device_info *bsctl_device_info_get(const char *dev_path,
						struct bsctl_error *err)
{
	struct bsctl_device_info *info = NULL;
	struct blksnap_ioctl *args = calloc(1, sizeof(struct blksnap_ioctl));

	if (!args) {
		bsctl_error_set(err, errno,
				"ERROR: Failed to allocate memory for device ioctl");
		return NULL;
	}

	if (validate_path_length(dev_path, err))
		goto clean;

	bsctl_init_ioctl(args);
	strncpy(args->device.name, dev_path, BLKSNAP_NAME_SIZE);
	if (bsctl_send_ioctl(BLKSNAP_IOC_INFO, args, dev_path, err)) {
		goto clean;
	}

	info = calloc(1, sizeof(struct bsctl_device_info));
	if (!info) {
		bsctl_error_set(err, errno,
				"ERROR: Failed to allocate memory for device info");
		goto clean;
	}

	_copy_device_info_from_ioctl(info, &args->device);

clean:
	free(args);
	return info;
}

void bsctl_device_info_free(struct bsctl_device_info *info)
{
	free(info);
}

int bsctl_snapshot_get(const char *dev_path, char *snap,
		       struct bsctl_error *err)
{
	struct blksnap_ioctl *args;
	int ret;

	if (!snap)
		return bsctl_error_set(err, EINVAL,
				       "ERROR: snap parameter in %s() should not be NULL",
				       __func__);

	if ((ret = validate_path_length(dev_path, err)))
		return ret;

	args = calloc(1, sizeof(struct blksnap_ioctl));
	if (!args)
		return bsctl_error_set(err, errno,
				       "ERROR: Insufficient memory to get snapshot device of %s",
				       dev_path);

	bsctl_init_ioctl(args);
	strncpy(args->show.name, dev_path, BLKSNAP_NAME_SIZE);

	if ((ret = bsctl_send_ioctl(BLKSNAP_IOC_SHOW, args, dev_path, err)))
		goto clean;

	snprintf(snap, PATH_MAX, "/dev/%s", args->show.snapshot_name);

clean:
	free(args);
	return ret;
}

bool _fs_wait(const char *device) {
	bool ret = false;
	int tries = 0;

	do {
		char *data;
		if ((data = blkid_get_tag_value(NULL, "TYPE", device))) {
			ret = true;
			free(data);
			break;
		}
		usleep(DEVICE_WAIT_TIME);
	} while (tries++ < DEVICE_WAIT_COUNT);

	return ret;
}

static int _wait_for_device(const char *dev_path,
			    int fstype,
			    struct bsctl_error *err)
{
	struct stat statinfo;
	unsigned int tries = 0;

	bool res = _fs_wait(dev_path);

	if (fstype == BLKSNAP_FS_BTRFS && !res)
		goto err;

	do {
		if (stat(dev_path, &statinfo) == 0)
			return 0;

		usleep(DEVICE_WAIT_TIME);
		tries++;

	} while (tries < DEVICE_WAIT_COUNT);

err:
	return bsctl_error_set(err, ETIMEDOUT,
			       "ERROR: Waited for device %s for too long",
			       dev_path);
}

int bsctl_snapshot_start(const char *dev_path, char *snap, int32_t flags,
			 struct bsctl_error *err)
{
	struct blksnap_ioctl *args;
	char snap_buf[PATH_MAX];
	enum blksnap_fs_type fstype;
	int ret;
	int xfs_version = 0;

	if (!snap)
		snap = snap_buf;

	if ((ret = validate_path_length(dev_path, err)))
		return ret;

	args = calloc(1, sizeof(struct blksnap_ioctl));
	if (!args)
		return bsctl_error_set(err, errno,
				       "ERROR: Insufficient memory to start snapshot of %s",
				       dev_path);

	bsctl_init_ioctl(args);
	strncpy(args->start.name, dev_path, BLKSNAP_NAME_SIZE);
	args->start.flags = flags;

	/* btrfs doesn't registed its superblock in bdev
	 * so freeze_bdev() doesn't call freeze_super().
	 * We need to do it ourselves.
	 */
	fstype = get_fstype(dev_path);
	if (fstype == BLKSNAP_FS_BTRFS) {
		if ((ret = freeze_fs(dev_path, err)))
			goto out;
	}

	if (fstype == BLKSNAP_FS_XFS) {
		xfs_version = check_xfs_version(dev_path, err);
		if (xfs_version == XFS_VERSION_5) {
			if ((ret = freeze_fs(dev_path, err)))
				goto out;
		}
	}

	ret = bsctl_send_ioctl(BLKSNAP_IOC_START, args, dev_path, err);

	if (fstype == BLKSNAP_FS_BTRFS ||
	   (fstype == BLKSNAP_FS_XFS && xfs_version == XFS_VERSION_5))
		thaw_fs(dev_path, err);

	if (ret)
		goto out;

	snprintf(snap, PATH_MAX, "/dev/%s", args->start.snapshot_name);

	ret = _wait_for_device(snap, get_fstype(dev_path), err);
out:
	free(args);
	return ret;
}

int bsctl_snapshot_complete(const char *dev_path, struct bsctl_error *err)
{
	struct blksnap_ioctl *args;
	enum blksnap_fs_type fstype;
	int ret;
	int xfs_version = 0;

	if ((ret = validate_path_length(dev_path, err)))
		return ret;

	args = calloc(1, sizeof(struct blksnap_ioctl));
	if (!args)
		return bsctl_error_set(err, errno,
				       "ERROR: Insufficient memory to complete snapshot of %s",
				       dev_path);

	bsctl_init_ioctl(args);
	strncpy(args->stop.name, dev_path, BLKSNAP_NAME_SIZE);

	/* btrfs doesn't registed its superblock in bdev
	 * so freeze_bdev() doesn't call freeze_super().
	 * We need to do it ourselves.
	 */
	fstype = get_fstype(dev_path);
	if (fstype == BLKSNAP_FS_BTRFS) {
		if ((ret = freeze_fs(dev_path, err)))
			goto out;
	}

	if (fstype == BLKSNAP_FS_XFS) {
		xfs_version = check_xfs_version(dev_path, err);
		if (xfs_version  == XFS_VERSION_5) {
			if ((ret = freeze_fs(dev_path, err)))
				goto out;
		}
	}

	ret = bsctl_send_ioctl(BLKSNAP_IOC_STOP, args, dev_path, err);

	if (fstype == BLKSNAP_FS_BTRFS ||
	   (fstype == BLKSNAP_FS_XFS && xfs_version == XFS_VERSION_5))
		thaw_fs(dev_path, err);

out:
	free(args);
	return ret;
}

int bsctl_snapshot_cancel(const char *dev_path, struct bsctl_error *err)
{
	struct blksnap_ioctl *args;
	enum blksnap_fs_type fstype;
	int ret;
	int xfs_version = 0;

	if ((ret = validate_path_length(dev_path, err)))
		return ret;

	args = calloc(1, sizeof(struct blksnap_ioctl));
	if (!args)
		return bsctl_error_set(err, errno,
				       "ERROR: Insufficient memory to cancel snapshot of %s",
				       dev_path);

	bsctl_init_ioctl(args);
	strncpy(args->stop.name, dev_path, BLKSNAP_NAME_SIZE);
	args->stop.merge = 1;

	/* btrfs doesn't register its superblock in bdev
	 * so freeze_bdev() doesn't call freeze_super().
	 * We need to do it ourselves.
	 */
	fstype = get_fstype(dev_path);
	if (fstype == BLKSNAP_FS_BTRFS) {
		if ((ret = freeze_fs(dev_path, err)))
			goto out;
	}

	if (fstype == BLKSNAP_FS_XFS) {
		xfs_version = check_xfs_version(dev_path, err);
		if (xfs_version  == XFS_VERSION_5) {
			if ((ret = freeze_fs(dev_path, err)))
				goto out;
		}
	}

	ret = bsctl_send_ioctl(BLKSNAP_IOC_STOP, args, dev_path, err);

	if (fstype == BLKSNAP_FS_BTRFS ||
	   (fstype == BLKSNAP_FS_XFS && xfs_version == XFS_VERSION_5))
		thaw_fs(dev_path, err);

out:
	free(args);
	return ret;
}

static int _fill_changelog(struct blksnap_ioctl_disk_extent *extent,
			   unsigned int UNUSED(total_blocks),
			   void *data, struct bsctl_error *err)
{
	struct bsctl_chlog *chlog, **user_data = data;
	uint64_t nextents;

	if (!data)
		return bsctl_error_set(err, EINVAL,
				       "ERROR: %s(): data should not be NULL",
				       __func__);

	chlog = *user_data;
	nextents = chlog ? chlog->nextents + 1 : 1;

	chlog = realloc(chlog, sizeof(*chlog) +
			nextents * sizeof(struct bsctl_chlog_extent));

	if (!chlog) {
		if (*user_data) {
			free(*user_data);
			*user_data = NULL;
		}
		return bsctl_error_set(err, errno,
				       "ERROR: Failed to allocate memory for changelog chunk");
	}

	*user_data = chlog;

	chlog->nextents = nextents;
	chlog->extents[nextents - 1].start = extent->start_sect;
	chlog->extents[nextents - 1].len = extent->nblocks;

	return 0;
}

struct bsctl_chlog *bsctl_chlog_get(const char *dev_path,
				    struct bsctl_error *err)
{
	struct changelog_stat *cl_stat;
	struct bsctl_chlog *chlog;
	unsigned int total_blocks;

	if (!(chlog = calloc(1, sizeof(struct bsctl_chlog)))) {
		bsctl_error_set(err, errno,
				"ERROR: Failed to allocate memory for changelog");
		return NULL;
	}

	if (!(total_blocks = bsctl_chlog_size(dev_path, err)))
		goto fail;

	if (!(cl_stat = bsctl_chlog_process(dev_path, total_blocks,
					    &_fill_changelog, &chlog, err)))
		goto fail;

	free(cl_stat);
	return chlog;
fail:
	free(chlog);
	return NULL;
}

void bsctl_chlog_free(struct bsctl_chlog *chlog)
{
	free(chlog);
}

/* set OFF CoW for directory (.blksnap as usual) */
static int set_cow_off(const char *dir_path, struct bsctl_error *err)
{
	int fd, ret = 0;
	const int flags = FS_NOCOW_FL;

	fd = open(dir_path, O_RDONLY);
	if (fd == -1)
		return bsctl_error_set(err, errno, "ERROR: Failed to open %s",
				       dir_path);

	if (ioctl(fd, FS_IOC_SETFLAGS, &flags) == -1) {
		ret = bsctl_error_set(err, errno, "ERROR: ioctl FS_IOC_SETFLAGS failed on file %s",
				      dir_path);
		goto close;
	}
close:
	close(fd);
	return ret;
}

bool fs_has_cow_feature(const char *dev_path)
{
	if (get_fstype(dev_path) == BLKSNAP_FS_BTRFS)
		return true;
	else
		return false;
}

static int _check_directory(const char *dev_path,
			    const char *dir_path,
			    struct bsctl_error *err)
{
	struct stat statinfo;

	if (stat(dir_path, &statinfo) == -1) {
		if (mkdir(dir_path, 0755) == -1)
			return bsctl_error_set(err, errno,
					       "ERROR: Failed to create dir %s",
					       dir_path);
		if (fs_has_cow_feature(dev_path))
			return set_cow_off(dir_path, err);
	} else if (!S_ISDIR(statinfo.st_mode))
		return bsctl_error_set(err, ENOTDIR,
				       "ERROR: %s is not a directory", dir_path);
	return 0;
}

static int _fallocate(int UNUSED(fd), off_t UNUSED(offset), off_t UNUSED(len))
{
#if NO_FALLOCATE
	return -1;
#else
	return fallocate(fd, 0, offset, len);
#endif
}

/* _validate_filesize: gets required file size
 * and returns max possible size allowed on
 * target volume filesystem if it less then
 * required size.
 */
static uint64_t _validate_filesize(const char *file_path, uint64_t size)
{
	int fd;
	uint64_t ret = 0, filesize = size, step = size;
	char tmpfname[PATH_MAX];

	snprintf(tmpfname, PATH_MAX, "%s-XXXXXX", file_path);
	if ((fd = mkstemp(tmpfname)) == -1)
		return 0;

	unlink(tmpfname);

	do {
		step = (step + 1) / 2;
		if (ftruncate(fd, filesize) < 0) {
			filesize -= step;
		} else {
			ret = filesize;
			if (ret == size)
				break;
			filesize += step;
		}
	} while(step > MIN_SEARCH_STEP);

	close(fd);
	return ret;
}

static bool _file_is_allocated(int fd, uint64_t size, struct bsctl_error *err)
{
	struct stat st;
	bool is_allocated, is_correct_size;

	if (fstat(fd, &st) == -1) {
		bsctl_error_set(err, errno, "ERROR: Failed to get fstat (%s)",
				strerror(errno));
		return false;
	}

	is_correct_size = (uint64_t) st.st_size == size;
	if (!is_correct_size) {
		bsctl_err_set(err, ESIZESTORE,
				"ERROR: File has an incorrect size");
		return false;
	}

	is_allocated = (is_correct_size &&
		(st.st_size + SECTOR_SIZE - 1) / SECTOR_SIZE <= st.st_blocks);

	if (!is_allocated) {
		bsctl_err_set(err, ENOTALLOC,
				"ERROR: File is not allocated");
		return false;
	}

	return true;
}

/* external function as it is used from bsrw */
int bsctl_file_reset(int fd, uint64_t size, struct bsctl_error *err)
{
	int ret = 0;
	char *chunk;
	size_t chunk_size = CHUNK_SIZE;

	if (!(chunk = alloc_resize(&chunk_size, SECTOR_SIZE, true)))
		return bsctl_error_set(err, errno,
				       "ERROR: Failed to allocate memory for file resetting chunk");

	for (; size > chunk_size; size -= chunk_size) {
		if (write(fd, chunk, chunk_size) == -1) {
			ret = bsctl_error_set(err, errno,
					      "Write failed while file resetting");
			goto out;
		}
	}

	if (size > 0 && write(fd, chunk, size) == -1) {
		ret = bsctl_error_set(err, errno,
				      "Write failed while file resetting");
		goto out;
	}
out:
	free(chunk);
	return ret;
}

static int __store_file_allocate(int fd, uint64_t size, unsigned char reset,
				 struct bsctl_error *err)
{
	if (ftruncate(fd, size) < 0)
		return bsctl_error_set(err, errno, "Failed to truncate file");

	if (_file_is_allocated(fd, size, err)) {
		if (!reset)
			return 0;
	} else {
		if (_fallocate(fd, 0, size) == 0 && !reset)
			return 0;
	}

	return bsctl_file_reset(fd, size, err);
}

static int _store_file_check(const char *file_path, uint64_t size,
			     struct bsctl_error *err)
{
	int ret = 0, fd, oflags = O_RDONLY;

	fd = open(file_path, oflags, S_IRUSR);
	if (fd == -1)
		return bsctl_error_set(err, errno,
				       "WARNING: Failed to open file %s: %s",
				       file_path, strerror(errno));

	if (!_file_is_allocated(fd, size, err))
		ret = err->err;

	close(fd);
	return ret;
}

static int _store_file_allocate(const char *file_path, uint64_t size,
				bool reset, struct bsctl_error *err)
{
	int fd, oflags = O_RDWR | O_SYNC | O_CREAT;
	int ret = 0;

	fd = open(file_path, oflags, S_IRUSR);
	if (fd == -1)
		return bsctl_error_set(err, errno, "Failed to open file '%s'",
				       file_path);

	ret = __store_file_allocate(fd, size, reset, err);
	if (ret)
		goto close;

	/* Creating files (especially big ones) on Btrfs volumes (especially
	 * fresh ones) makes chunk tree grow, but since it is lazy to flush on
	 * disk, flushing chunk tree needs to be forced by sync call.
	 * Inconsistent chunk tree might cause wrong translation from logical
	 * addresses on stores mapping (if at least some store addresses are
	 * disposed on the new chunks).
	 */
	ret = fsync(fd);
close:
	close(fd);
	return ret;
}

static int _store_action(const char *file_path, uint64_t size,
			 enum store_action act, struct bsctl_error *err)
{

	char *fnames = NULL;
	uint64_t filesize;
	unsigned long nfiles, cur_file = 0;
	int ret = 0;

	filesize = _validate_filesize(file_path, size);
	if (!filesize)
		return bsctl_error_set(err, ENOENT,
				       "Failed to determine filesystem filesize for %s",
				       file_path);

	nfiles = (size + filesize - 1) / filesize;

	fnames = calloc(nfiles, PATH_MAX);
	if (!fnames)
		return bsctl_error_set(err, errno,
				       "Failed to allocate memory for store filenames");

	for (; cur_file < nfiles; ++cur_file, size -= filesize) {
		char *cur_fname = fnames + (PATH_MAX * cur_file);
		if (cur_file == 0)
			strncpy(cur_fname, file_path, PATH_MAX);
		else
			snprintf(cur_fname, PATH_MAX, "%s-%lu",
				 file_path, cur_file);

		filesize = size > filesize ? filesize : size;

		switch(act) {
		case STORE_ALLOC:
			ret = _store_file_allocate(cur_fname, filesize,
						   false, err);
			if (ret) {
				for (; cur_file > 0; --cur_file) {
					cur_fname = fnames
						    + (PATH_MAX * cur_file);
					unlink(cur_fname);
				}
			}
			break;
		case STORE_RESET:
			ret = _store_file_allocate(cur_fname, filesize,
						   true, err);
			break;
		case STORE_CHECK:
			ret = _store_file_check(cur_fname, filesize, err);
			break;
		}
	}

	if (fnames)
		free(fnames);

	return ret;
}

dev_t bsctl_get_rdev(const char *dev_path, struct bsctl_error *err)
{
	struct stat statinfo;

	if (stat(dev_path, &statinfo) == -1) {
		bsctl_error_set(err, errno, "stat failed on %s", dev_path);
		return 0;
	}

	if (!S_ISBLK(statinfo.st_mode))
		bsctl_error_set(err, ENOTBLK, "%s: not a block device",
				dev_path);

	return statinfo.st_rdev;
}

/* Backup scenario in case if device does not support ioctl calls
 * Return 0 on success and error code otherwise.
 */
static int _get_field_from_sysfs(const char *dev_path, const char *field,
				 uint64_t *out, struct bsctl_error *err)
{
	FILE *fp;
	char *devicename;
	char path_resolved[PATH_MAX], filename[PATH_MAX];
	int ret = 0, i = 0;

	if (realpath(dev_path, path_resolved) == NULL)
		return bsctl_error_set(err, errno,
				       "Could not resolve path \"%s\"",
				       dev_path);

	devicename = basename(path_resolved);
	snprintf(filename, sizeof(filename), "/sys/block/%s/%s",
		 devicename, field);

	fp = fopen(filename, "r");
	if (!fp) {
		char *parentname, tmp[PATH_MAX];

		strncpy(tmp, devicename, sizeof(tmp));
		parentname = strtok(tmp, "0123456789");
		if (!parentname)
			return bsctl_error_set(err, ENOENT,
					       "Could not determine a major number for device \"%s\"",
					       devicename);
		snprintf(filename, sizeof(filename), "/sys/block/%s/%s/%s",
			 parentname, devicename, field);
		fp = fopen(filename, "r");
		if (!fp) {
			strncpy(tmp, devicename, sizeof(tmp));
			for (i = strlen(tmp) - 1; i >= 0 ; i--) {
				if (tmp[i] >= '0' && tmp[i] <= '9') {
					tmp[i] = '\0';
				} else {
					if (tmp[i] == 'p')
						tmp[i] = '\0';
					break;
				}
			}

			parentname = tmp;
			if (!parentname)
				return bsctl_error_set(err, ENOENT,
						       "Could not determine a major number for device \"%s\"",
						       devicename);
			snprintf(filename, sizeof(filename), "/sys/block/%s/%s/%s",
				 parentname, devicename, field);
			fp = fopen(filename, "r");
			if (!fp)
				return bsctl_error_set(err, errno,
						       "Could not find a field \"%s\" for  device \"%s\" (%s): %s",
						       field, devicename, filename,
						       strerror(errno));
		}
	}

	fscanf(fp, "%" PRIu64, out);
	fclose(fp);

	return ret;
}

static int _run_ro_ioctl(const char *dev_path, int cmd, void *data)
{
	int fd, ret = 0;

	fd = open(dev_path, O_RDONLY);
	if (fd == -1)
		return errno;

	if (ioctl(fd, cmd, data) == -1)
		ret = errno;

	close(fd);
	return ret;
}

/* @path: path of any file within the mounted FS */
int bsctl_get_device_freespace(const char *path, uint64_t *free_bytes,
			       struct bsctl_error *err)
{
	struct statvfs fs_info;

	if (statvfs(path, &fs_info) ==  -1)
		return bsctl_error_set(err, errno,
				       "ERROR: Failed to get filesystem information for path %s",
				       path);

	*free_bytes = fs_info.f_bfree * fs_info.f_bsize;

	return 0;
}

/* returns device size in sectors. So to get real size of device
 * in bytes use size * SECTOR_SIZE
 */
uint64_t bsctl_get_device_size(const char *dev_path, struct bsctl_error *err)
{
	uint64_t size = 0;

	if (_run_ro_ioctl(dev_path, BLKGETSIZE64, &size)) {
		if (_get_field_from_sysfs(dev_path, "size", &size, err))
			/* could not determine device size, error stored */
			return 0;
	} else {
		size = size / SECTOR_SIZE;
	}

	return size;
}

uint64_t bsctl_get_volume_start_sect(const char *dev_path)
{
	struct hd_geometry geo;
	uint64_t start = 0;

	if (_run_ro_ioctl(dev_path, HDIO_GETGEO, &geo)) {
		if (_get_field_from_sysfs(dev_path, "start", &start, 0))
			/* Assuming that start is 0 if there is no entry in
			 * /sys/ directory */
			start = 0;
	} else {
		start = geo.start;
	}

	return start;
}

static int _store_create(const char *file_path, uint64_t size,
			 struct bsctl_error *err)
{
	return _store_action(file_path, size, STORE_ALLOC, err);
}

static int _store_reset(const char *file_path, uint64_t size,
			struct bsctl_error *err)
{
	return _store_action(file_path, size, STORE_RESET, err);
}

static int _store_check(const char *file_path, uint64_t size,
			struct bsctl_error *err)
{
	return _store_action(file_path, size, STORE_CHECK, err);
}

/* Return 0 if sum of fiemap extents size is equal to file size */
static bool _check_file_map_size(const struct fiemap *fiemap,
				 const char *file_path,
				 struct bsctl_error *err)
{
	int ret = 0;
	uint32_t i;
	off_t fiemap_size;
	off_t file_size_aligned;
	struct stat statinfo;
	struct statvfs stvfs;

	if (stat(file_path, &statinfo) < 0)
		return bsctl_error_set(err, errno, "ERROR: Failed to get size of %s",
				       file_path);

	if (statvfs(file_path, &stvfs) < 0)
		return bsctl_error_set(err, errno, "ERROR: Failed to get FS block size of %s",
				       file_path);

	file_size_aligned = statinfo.st_size;
	if (file_size_aligned % stvfs.f_bsize)
		file_size_aligned += (stvfs.f_bsize
				      - (file_size_aligned % stvfs.f_bsize));

	for (i = 0, fiemap_size = 0; i < fiemap->fm_mapped_extents; i++)
		fiemap_size += fiemap->fm_extents[i].fe_length;

	if (fiemap_size != file_size_aligned)
		return bsctl_err_set(err, EINVAL,
				     "%s: fiemap size and file size are mismatched (%llu vs %llu)",
				     file_path, fiemap_size, file_size_aligned);

	return ret;
}

#define STORE_GLOB "%s*"

static int _store_map(struct mapping_params *params, struct bsctl_error *err)
{
	uint32_t e;
	uint64_t volume_start_sect;
	size_t ioctl_size;
	struct fiemap **fiemap = NULL;
	size_t files_count;
	struct blksnap_ioctl *args = NULL;
	struct bsctl_error local_error;
	char glob_pattern[PATH_MAX];
	glob_t store_glob;
	size_t f, max_extents_per_page, extent_n = 0;
	int ret = 0;

	if (!err)
		err = &local_error;

	bsctl_err_reset(err);

	if ((ret = validate_path_length(params->dev_path, err)))
		return ret;

	snprintf(glob_pattern, sizeof(glob_pattern), STORE_GLOB,
		 params->file_path);

	ret = glob(glob_pattern, 0, NULL, &store_glob);
	if (ret)
		return bsctl_error_set(err, ENOENT,
				       "ERROR: No store found on glob "
				       STORE_GLOB " (glob errcode: %d)",
				       params->file_path, ret);

	files_count = store_glob.gl_pathc;
	fiemap = calloc(files_count, sizeof(struct fiemap *));
	if (!fiemap) {
		ret = bsctl_error_set(err, errno, "ERROR: Failed to allocate memory for fiemap");
		goto glob_free_exit;
	}

	for (f = 0; f < files_count; ++f) {
		strncpy(params->file_path, store_glob.gl_pathv[f], PATH_MAX);

		fiemap[f] = _get_file_map(params, err);
		if (!fiemap[f]) {
			ret = err->err;
			files_count = f;
			goto fiemap_free_exit;
		}

		if (_check_file_map_size(fiemap[f], params->file_path, err)) {
			ret = err->err;
			goto fiemap_free_exit;
		}
	}

	/* all extent maps present in kernel module should be with absolute
	 * offset from the start of whole disk
	 */
	volume_start_sect = bsctl_get_volume_start_sect(params->dev_path);

	max_extents_per_page =
		elems_in_size(PAGE_SIZE,
			      BLKSNAP_IOCTL_HEADER_SIZE
			      + sizeof(struct blksnap_ioctl_map),
			      sizeof(struct blksnap_ioctl_file_extent));

	ioctl_size = BSCTL_CALC_IOCTL_SIZE(struct blksnap_ioctl_map,
		max_extents_per_page * sizeof(struct blksnap_ioctl_file_extent));

	args = calloc(1, ioctl_size);
	if (!args) {
		ret = bsctl_error_set(err, errno, "ERROR: Failed to allocate memory for map ioctl");
		goto fiemap_free_exit;
	}

	bsctl_init_ioctl(args);
	args->size = ioctl_size;

	strncpy(args->map.name, params->dev_path, BLKSNAP_NAME_SIZE);

	args->map.type = params->type;
	args->map.npages = params->npages;
	args->map.nextents = 0;
	args->map.flags = BLKSNAP_STORE_FIRST | params->flags;

	for (f = 0; f < files_count; ++f) {
		for (e = 0; e < fiemap[f]->fm_mapped_extents; e++) {
			/* absolute offset */
			args->map.extents[extent_n].start_sect =
				volume_start_sect
				+ fiemap[f]->fm_extents[e].fe_physical
				  / SECTOR_SIZE;
			args->map.extents[extent_n].nsectors =
				fiemap[f]->fm_extents[e].fe_length
				/ SECTOR_SIZE;
			extent_n++;
			if (extent_n >= max_extents_per_page) {
				args->map.nextents = extent_n;
				if ((ret = bsctl_send_ioctl(BLKSNAP_IOC_MAP,
							    args,
							    params->dev_path,
							    err))) {
					goto out;
				}
				extent_n = 0;
				args->map.flags = BLKSNAP_STORE_NEXT;
			}
		}
	}

	args->map.nextents = extent_n;
	args->map.flags |= BLKSNAP_STORE_LAST;
	ret = bsctl_send_ioctl(BLKSNAP_IOC_MAP, args, params->dev_path, err);

out:
	if (ret == EEXIST) {
		switch (params->type) {
		case BLKSNAP_MAP_TYPE_CHLOG:
			bsctl_err_set(err, ret, "changelog store is already mapped on device %s",
				      params->dev_path);
			break;
		case BLKSNAP_MAP_TYPE_DATA:
			bsctl_err_set(err, ret, "datastore is already mapped on device %s",
				      params->dev_path);
			break;
		case BLKSNAP_MAP_TYPE_METADATA:
			bsctl_err_set(err, ret, "metadata is already mapped on device %s",
				      params->dev_path);
			break;
		}
	}
	free(args);

fiemap_free_exit:
	for (f = 0; f < files_count; ++f) {
		free(fiemap[f]);
	}
	free(fiemap);

glob_free_exit:
	globfree(&store_glob);

	return ret;
}

static int _get_path(const char *dev_path, const char *store_file,
		     char *filepath, size_t filepath_size,
		     char *mnt, size_t mnt_size,
		     struct bsctl_error *err)
{
	char filedir[PATH_MAX], mntbuf[PATH_MAX];
	int ret = 0;

	if (!mnt || !mnt_size) {
		mnt = mntbuf;
		mnt_size = sizeof(mntbuf);
	}

	ret = get_mount_point(dev_path, mnt, mnt_size);
	if (ret)
		return bsctl_error_set(err, ret,
				       "ERROR: Failed to find mount point for %s",
				       dev_path);

	snprintf(filedir, PATH_MAX, "%s/%s", mnt, STORE_DIRNAME);

	ret = _check_directory(dev_path, filedir, err);
	if (ret)
		return ret;

	snprintf(filepath, filepath_size, "%s/%s", filedir, store_file);

	return ret;
}

static uint64_t _store_calculate_size(const char *dev_path, double percents,
				      struct bsctl_error *err)
{
	uint64_t dev_sectors;

	dev_sectors = bsctl_get_device_size(dev_path, err);

	return dev_sectors * SECTOR_SIZE * percents;
}

static uint64_t _chlog_get_size(const char *dev_path, struct bsctl_error *err)
{
	return PAGE_SIZE *
	       FILE_PAGES(_store_calculate_size(dev_path, STORE_CHLOG_RATE, err));
}

static int _check_chlog_store(const char *dev_path, const char *file_path,
			      struct bsctl_error *err)
{
	return _store_check(file_path, _chlog_get_size(dev_path, err), err);
}

int bsctl_chlog_create(const char *dev_path, struct bsctl_error *err)
{
	uint64_t chlog_size, free_space;
	char mnt[PATH_MAX], file_path[PATH_MAX];
	int ret = 0;

	if ((ret = _get_path(dev_path, STORE_CHLOG_FILE,
			     file_path, sizeof(file_path),
			     mnt, sizeof(mnt),
			     err)))
		return ret;

	chlog_size = _chlog_get_size(dev_path, err);

	if (!chlog_size)
		return bsctl_error_set(err, ERANGE,
				       "Invalid changelog store size (%llu)",
				       (unsigned long long) chlog_size);

	ret = bsctl_get_device_freespace(mnt, &free_space, err);
	if (ret)
		return ret;

	if (chlog_size > free_space)
		return bsctl_error_set(err, ENOSPC,
				       "Not enough space on device %s (%llu bytes needed)",
				       dev_path,
				       (unsigned long long) chlog_size);

	return _store_reset(file_path, chlog_size, err);
}

int bsctl_chlog_reset(const char *dev_path, struct bsctl_error *err)
{
	char file_path[PATH_MAX];
	int ret;

	if ((ret = _get_path(dev_path, STORE_CHLOG_FILE,
			     file_path, sizeof(file_path),
			     NULL, 0, err)))
		return ret;

	return _store_reset(file_path, _chlog_get_size(dev_path, err), err);
}

static inline bool _is_data_store_mapped(__u32 flags)
{
	return flags & BLKSNAP_DEVICE_DATA;
}

static int _set_data_store_size(const char *dev_path, uint64_t *size,
				double percent, struct bsctl_error *err)
{
	int ret = 0;

	if (*size)
		goto out;

	/* if there is no value or percentage specified, use default value */
	if (!percent) {
		*size = _store_calculate_size(dev_path, STORE_DATA_RATE, err);
		goto out;
	}

	if (percent >= 100)
		goto err_range;

	*size = _store_calculate_size(dev_path, percent / 100.0, err);

	if (!(*size))
		goto err_range;

	goto out;

err_range:
	ret = bsctl_error_set(err, ERANGE,
			      "invalid data store size for device %s", dev_path);
out:
	return ret;
}

/* If size > 0, percent is ignored. If both are 0, use default percent value */
int bsctl_data_create(const char *dev_path, uint64_t size, double percent,
		      struct bsctl_error *err)
{
	char mnt[PATH_MAX], file_path[PATH_MAX];
	struct bsctl_device_info *device;
	struct bsctl_error local_error;
	uint64_t free_space;
	int ret;

	if (!err)
		err = &local_error;

	bsctl_err_reset(&local_error);

	device = bsctl_device_info_get(dev_path, err);

	if (!device) {
		/* it's ok if ioctl returns ENODEV, create-data-store
		 * command could be called for not attached device
		 */
		if (err->err != ENODEV)
			return err->err;
	} else {
		if (device && _is_data_store_mapped(device->flags)) {
			free(device);
			return bsctl_error_set(err,
					       BLKSNAP_ERROR_TO_USER(EHASDATA),
					       "ERROR: device %s has already mapped data store",
					       dev_path);
		}
		free(device);
	}

	if ((ret = _get_path(dev_path, STORE_DATA_FILE,
			     file_path, sizeof(file_path),
			     mnt, sizeof(mnt),
			     err)))
		return ret;

	errno = 0;
	ret = _set_data_store_size(dev_path, &size, percent, err);
	if (ret)
		return ret;

	ret = bsctl_get_device_freespace(mnt, &free_space, err);
	if (ret)
		return ret;

	if (size > free_space) {
		return bsctl_error_set(err, ENOSPC,
				       "not enough space on device %s (%llu bytes needed)",
				       dev_path, size);
	}

	return _store_create(file_path, size, err);
}

int __bsctl_data_check(const char *dev_path, uint64_t *size, double percent,
		     struct bsctl_error *err)
{
	char file_path[PATH_MAX];
	int ret;

	if ((ret = _get_path(dev_path, STORE_DATA_FILE,
			     file_path, sizeof(file_path),
			     NULL, 0,
			     err)))
		return ret;

	ret = _set_data_store_size(dev_path, size, percent, err);
	if (ret)
		return ret;

	return _store_check(file_path, *size, err);
}

/* A wrapper for libbsctl to avoid marshalling size parameter */
int bsctl_data_check(const char *dev_path, uint64_t size, double percent,
		     struct bsctl_error *err)
{
	return __bsctl_data_check(dev_path, &size, percent, err);
}

int bsctl_chlog_check(const char *dev_path, struct bsctl_error *err)
{
	char file_path[PATH_MAX];
	int ret;

	if ((ret = _get_path(dev_path, STORE_CHLOG_FILE,
			     file_path, sizeof(file_path),
			     NULL, 0,
			     err)))
		return ret;

	return _check_chlog_store(dev_path, file_path, err);
}

int bsctl_chlog_map_custom(const char *dev_path, enum map_method method,
			   ssize_t bufsize, struct bsctl_error *err)
{
	struct mapping_params map_args;
	int ret;

	if ((ret = _get_path(dev_path, STORE_CHLOG_FILE,
			     map_args.file_path, sizeof(map_args.file_path),
			     NULL, 0, err)))
		return ret;

	/* Return error if changelog doesn't exist or has wrong size */
	if ((ret = _check_chlog_store(dev_path, map_args.file_path, err)))
		return ret;

	map_args.dev_path	= dev_path;
	map_args.npages		= 0;
	map_args.flags		= 0;
	map_args.method		= method;
	map_args.bufsize	= bufsize;
	map_args.type		= BLKSNAP_MAP_TYPE_CHLOG;

	return _store_map(&map_args, err);
}

int bsctl_chlog_map(const char *dev_path, struct bsctl_error *err)
{
	return bsctl_chlog_map_custom(dev_path, MAP_AUTO, 0, err);
}

int bsctl_data_map_custom(const char *dev_path, enum map_method method,
			  ssize_t bufsize, struct bsctl_error *err)
{
	struct mapping_params map_args;
	int ret = 0;

	if ((ret = _get_path(dev_path, STORE_DATA_FILE,
			     map_args.file_path, sizeof(map_args.file_path),
			     NULL, 0, err)))
		return ret;

	map_args.dev_path	= dev_path;
	map_args.npages		= STORE_DATA_NPAGES;
	map_args.flags		= 0;
	map_args.bufsize	= bufsize;
	map_args.method		= method;
	map_args.type		= BLKSNAP_MAP_TYPE_DATA;

	return _store_map(&map_args, err);
}

int bsctl_data_map(const char *dev_path, struct bsctl_error *err)
{
	return bsctl_data_map_custom(dev_path, MAP_AUTO, 0, err);
}

int bsctl_data_unmap(const char *dev_path, struct bsctl_error *err)
{
	struct blksnap_ioctl *args;
	int ret;

	if ((ret = validate_path_length(dev_path, err)))
		return ret;

	args = calloc(1, sizeof(struct blksnap_ioctl));
	if (!args)
		return bsctl_error_set(err, errno,
				       "ERROR: Insufficient memory to unmap data store on %s",
				       dev_path);

	bsctl_init_ioctl(args);
	strncpy(args->unmap.name, dev_path, BLKSNAP_NAME_SIZE);
	args->unmap.type = BLKSNAP_UNMAP_TYPE_DATA;

	ret = bsctl_send_ioctl(BLKSNAP_IOC_UNMAP, args, dev_path, err);

	free(args);
	return ret;
}

int bsctl_metadata_map_custom(const char *dev_path, enum map_method method,
			      ssize_t bufsize, uint32_t flags,
			      struct bsctl_error *err)
{
	struct mapping_params map_arg;
	int ret = 0;

	if ((ret = _get_path(dev_path, STORE_METADATA_FILE,
			     map_arg.file_path, sizeof(map_arg.file_path),
			     NULL, 0, err)))
		return ret;

	if (_store_file_check(map_arg.file_path, STORE_METADATA_SIZE, err)) {
		/* reset option is needed to avoid tail-packed blocks reveals
		 * during mapping file consists of uninitialized blocks
		 */
		if ((ret = _store_file_allocate(map_arg.file_path,
						STORE_METADATA_SIZE,
						true, err))) {
			return ret;
		}
	}

	map_arg.dev_path	= dev_path;
	map_arg.npages		= 0;
	map_arg.flags		= flags;
	map_arg.bufsize		= bufsize;
	map_arg.method		= method;
	map_arg.type		= BLKSNAP_MAP_TYPE_METADATA;

	return _store_map(&map_arg, err);
}

int bsctl_metadata_map(const char *dev_path, struct bsctl_error *err)
{
	return bsctl_metadata_map_custom(dev_path, MAP_AUTO, PAGE_SIZE, 0, err);
}

int bsctl_metadata_update(const char *dev_path, struct bsctl_error *err)
{
	return bsctl_metadata_map_custom(dev_path, MAP_AUTO, PAGE_SIZE,
					 BLKSNAP_STORE_UPDATE_METADATA, err);
}

static struct bsctl_map *_request_file_map(const char *dev_path,
					   enum blksnap_map_type map_type,
					   struct bsctl_error *err)
{
	struct bsctl_map *map = NULL;
	struct blksnap_ioctl *args, *tmp;
	uint32_t i;

	/* 1) Request number of store extents */
	args = calloc(1, sizeof(struct blksnap_ioctl));
	if (!args) {
		bsctl_error_set(err, errno, "ERROR: Failed to allocate memory in %s",
				__func__);
		return NULL;
	}

	bsctl_init_ioctl(args);

	strncpy(args->map.name, dev_path, BLKSNAP_NAME_SIZE);
	args->map.type = map_type;

	if (bsctl_send_ioctl(BLKSNAP_IOC_GET_MAP, args, dev_path, err))
		goto fail_cleanup;

	/* 2) Request store extents */
	args->size = BSCTL_CALC_IOCTL_SIZE(struct blksnap_ioctl_map,
		args->map.nextents * sizeof(struct blksnap_ioctl_file_extent));

	tmp = recalloc(args, sizeof(struct blksnap_ioctl), args->size, err);
	if (!tmp) {
		bsctl_error_set(err, errno, "ERROR: failed to reallocate memory in %s",
				__func__);
		goto fail_cleanup;
	}
	args = tmp;

	if (bsctl_send_ioctl(BLKSNAP_IOC_GET_MAP, args, dev_path, err))
		goto fail_cleanup;

	/* 3) Copy extents to map structure */
	map = calloc(1, sizeof(struct bsctl_map)
		        + (args->map.nextents * sizeof(struct bsctl_extent)));
	if (!map) {
		bsctl_error_set(err, errno, "ERROR: calloc failed in %s",
				__func__);
		goto fail_cleanup;
	}
	map->nextents = args->map.nextents;
	for (i = 0; i < args->map.nextents; i++) {
		map->extents[i].start = args->map.extents[i].start_sect;
		map->extents[i].nsectors = args->map.extents[i].nsectors;
	}

	free(args);
	return(map);
fail_cleanup:
	free(args);
	return NULL;
}

struct bsctl_map *bsctl_chlog_get_map(const char *dev_path, struct bsctl_error *err)
{
	return _request_file_map(dev_path, BLKSNAP_MAP_TYPE_CHLOG, err);
}

struct bsctl_map *bsctl_data_get_map(const char *dev_path, struct bsctl_error *err)
{
	return _request_file_map(dev_path, BLKSNAP_MAP_TYPE_DATA, err);}

struct bsctl_map *bsctl_metadata_get_map(const char *dev_path,
				         struct bsctl_error *err)
{
	return _request_file_map(dev_path, BLKSNAP_MAP_TYPE_METADATA, err);
}

void bsctl_map_free(struct bsctl_map *map)
{
	free(map);
}

int bsctl_dump_kernel(const char *dev_path, struct bsctl_error *err)
{
	struct blksnap_ioctl *args;
	int ret;

	if ((ret = validate_path_length(dev_path, err)))
		return ret;

	args = calloc(1, sizeof(struct blksnap_ioctl));
	if (!args)
		return bsctl_error_set(err, errno,
				       "ERROR: Insufficient memory to dump kernel structures for %s",
				       dev_path);

	bsctl_init_ioctl(args);
	strncpy(args->dump.name, dev_path, BLKSNAP_NAME_SIZE);

	ret = bsctl_send_ioctl(BLKSNAP_IOC_DUMP, args, dev_path, err);

	free(args);
	return ret;
}

int bsctl_volume_restore(const char *source, const char *target,
			 struct bsctl_error *err)
{
	struct blksnap_ioctl *args;
	int ret;

	if ((ret = validate_path_length(source, err)) ||
	    (ret = validate_path_length(target, err)))
		return ret;

	args = calloc(1, sizeof(struct blksnap_ioctl));
	if (!args)
		return bsctl_error_set(err, errno,
				       "ERROR: Insufficient memory to restore %s",
				       target);

	bsctl_init_ioctl(args);

	strncpy(args->restore.source_name, source, BLKSNAP_NAME_SIZE);
	strncpy(args->restore.target_name, target, BLKSNAP_NAME_SIZE);

	ret = bsctl_send_ioctl(BLKSNAP_IOC_RESTORE, args, target, err);

	free(args);
	return ret;
}

int bsctl_volume_reset(const char *dev_path, struct bsctl_error *err)
{
	struct blksnap_ioctl *args;
	int ret;

	if ((ret = validate_path_length(dev_path, err)))
		return ret;

	args = calloc(1, sizeof(struct blksnap_ioctl));
	if (!args)
		return bsctl_error_set(err, errno,
				       "ERROR: Insufficient memory to reset %s",
				       dev_path);

	bsctl_init_ioctl(args);
	strncpy(args->reset.name, dev_path, BLKSNAP_NAME_SIZE);

	ret = bsctl_send_ioctl(BLKSNAP_IOC_RESET, args, dev_path, err);

	free(args);
	return ret;
}

int bsctl_check(const char *dev_path, struct bsctl_error *err)
{
	struct blksnap_ioctl *args;
	int ret = 0;

	ret = validate_path_length(dev_path, err);
	if (ret)
		return ret;

	args = calloc(1, sizeof(struct blksnap_ioctl));
	if (!args)
		return bsctl_error_set(err, errno,
				       "ERROR: Insufficient memory to check %s",
				       dev_path);

	bsctl_init_ioctl(args);
	strncpy(args->check.name, dev_path, BLKSNAP_NAME_SIZE);

	ret = bsctl_send_ioctl(BLKSNAP_IOC_CHECK, args, dev_path, err);

	free(args);
	return ret;
}

int bsctl_version(char *buf, size_t len)
{
	snprintf(buf, len, "libbsctl version: %s (API version %u)",
		 BLKSNAP_VERSION, BLKSNAP_IOCTL_VERSION);
	return 0;
}
