#define _FILE_OFFSET_BITS 64

#include <asm/types.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <libgen.h>
#include <limits.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/file.h>
#include <unistd.h>
#include <pthread.h>
#include <math.h>
#include <stdbool.h>

#include "blksnap.h"
#include "bsctl-error.h"
#include "bsctl.h"
#include "bsctl-share.h"
#include "bsctl-log.h"
#include "bsctl-chlog.h"
#include "version.h"

/* Version of CLI (Command Line Interface).
 * DON'T FORGET TO UPDATE VERSION VALUE FOR ANY COMMAND LINE INTERFACE CHANGE
 *
 * Version 1 is the version without --cli-version command
 *
 * Version 2 includes:
 *  new command: -c --cli-version
 *  new option:  -h --hide-progress
 */
#define CLI_VERSION	2

#define FILE_MODE (S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)

/* flags to be passed to copy procedure and workers */
#define CF_INCREMENTAL		(1 << 0)
#define CF_SEPARATE		(1 << 1)
#define CF_HIDE_PROGRESS	(1 << 2)
#define CF_NO_SPACE_CHECK	(1 << 3)
#define CF_NO_WRITE		(1 << 4)

typedef struct copy_snapshot {
	char blkdev[BLKSNAP_NAME_SIZE + 1];
	char snapdev[BLKSNAP_NAME_SIZE + 1];
	char snapfile[BLKSNAP_NAME_SIZE + 1];
	int ifd;
	int ofd;
	int workers;
	int flags;
	sector_t *blocks;
	struct bsctl_error err;
} copy_snapshot_t;

struct worker_args {
	const char *snapdev;
	const char *snapfile;
	int         ifd;
	int         ofd;
	/* Numbers of changed blocks' sectors. Not for whole-volume baseimage */
	const sector_t *blocks;
	/* Number of the start/end block if (blocks == NULL).
	 * Index of the start/end block in blocks[] otherwise.
	 */
	sector_t start;
	sector_t end;
	uint64_t dev_sectors;
	/* Copy each block in a separate file? */
	int flags;
	struct bsctl_error err;
};

#define PROGRESS_REFRESH_RATE_USEC	40000

#ifdef BSRWAAFS
#include "aaerror.h"
#include "aafs.h"
#define BITS_PER_BYTE 8

struct bitmap_iter {
	sector_t nblock;
	sector_t addr;
	sector_t bmap_nbyte;
	int bmap_nbit;
};

/* Step to next bitmap block */
void inc_iter(aafs_t *map, struct bitmap_iter *iter)
{
	iter->nblock++;
	iter->addr += map->bsize;
	iter->bmap_nbyte = iter->nblock / BITS_PER_BYTE;
	iter->bmap_nbit = iter->nblock % BITS_PER_BYTE;
}

void set_iter_addr(aafs_t *map, struct bitmap_iter *iter, sector_t addr)
{
	iter->nblock = addr / map->bsize;
	iter->addr = addr;
	iter->bmap_nbyte = iter->nblock / BITS_PER_BYTE;
	iter->bmap_nbit = iter->nblock % BITS_PER_BYTE;
}

/* Gather array of numbers of occupied blocks
 * Return number of occupied blocks
 */
uint64_t aafs_get_occupied_blocks(const char *dev_path,
				  sector_t **blocks,
				  struct bsctl_error *err)
{
	aafs_t fs_info;
	struct aaerror aaerr;
	sector_t nblocks_occupied = 0;
	bool occupied;
	sector_t *blocks_list = NULL;
	sector_t block_list_counter = 0;
	struct bitmap_iter iter;

	aaerror_reset(&aaerr);

	/* Get bitmap of FS native-sized blocks */
	bzero(&fs_info, sizeof(fs_info));
	if (aafs_acquire_v2(&fs_info, dev_path, false, &aaerr)) {
		bsctl_error_set(err, aaerr.err,
				"Failed to acquire aafs fs_info: '%s'",
				aaerr.msg);
		return 0;
	}

	/* Calculate number of occupied blksnap-sized blocks */
	bzero(&iter, sizeof(iter));
	while (iter.nblock < fs_info.bcount) {
		occupied = fs_info.data[iter.bmap_nbyte]
			   & ((unsigned char)1 << iter.bmap_nbit);
		if (occupied) {
			if (fs_info.bsize >= PAGE_SIZE) {
				nblocks_occupied += fs_info.bsize / PAGE_SIZE;
			} else {
				nblocks_occupied++;
				sector_t next_page_addr =
					((iter.addr >> PAGE_SHIFT) + 1)
					<< PAGE_SHIFT;
				set_iter_addr(&fs_info, &iter, next_page_addr);
				continue;
			}
		}
		inc_iter(&fs_info, &iter);
	}

	/* Gather array of occupied blksnap-sized blocks */
	blocks_list = malloc(nblocks_occupied * sizeof(sector_t));
	if (!blocks_list) {
		bsctl_error_set(err, ENOMEM, "failed to allocate block list [%llu]",
				nblocks_occupied * sizeof(sector_t));
		goto fs_info_free;
	}

	bzero(&iter, sizeof(iter));
	while (iter.nblock < fs_info.bcount) {
		occupied = fs_info.data[iter.bmap_nbyte]
			   & ((unsigned char)1 << iter.bmap_nbit);
		if (occupied) {
			if (fs_info.bsize >= PAGE_SIZE) {
				int i;
				int blk_nblocks = fs_info.bsize / PAGE_SIZE;
				for (i = 0; i < blk_nblocks; i++) {
					blocks_list[block_list_counter++] =
						(iter.addr + i * PAGE_SIZE)
						>> SECTOR_SHIFT;
				}
			} else {
				sector_t curr_nblock = iter.addr >> PAGE_SHIFT;
				blocks_list[block_list_counter++] =
					curr_nblock << PAGE_SECTORS_SHIFT;
				set_iter_addr(&fs_info, &iter,
					      (curr_nblock + 1) << PAGE_SHIFT);
				continue;
			}
		}
		inc_iter(&fs_info, &iter);
	}

	*blocks = blocks_list;

fs_info_free:
	aafs_destroy(&fs_info);
	return nblocks_occupied;
}
#endif /* BSRWAAFS */

struct block_counter {
	uint32_t blocks_done;
	pthread_mutex_t mutex;
};

static struct block_counter block_counter = {0, PTHREAD_MUTEX_INITIALIZER};

static void *show_progress(void *arg)
{
	const uint64_t blocks_total = *((uint64_t *) arg);
	uint32_t percent = 0;

	setvbuf(stdout, NULL, _IONBF, 0);

	while (percent < 100) {
		percent = (200 * block_counter.blocks_done + 1) /
			  (blocks_total * 2);
		printf(" %8u / %llu (%u%%)\r", block_counter.blocks_done,
		       (unsigned long long) blocks_total, percent);
		usleep(PROGRESS_REFRESH_RATE_USEC);
	}

	printf("\n");

	pthread_exit(NULL);
}

static int copy_block(int ifd, off_t i_off,
		      int ofd, off_t o_off,
		      size_t size,
		      char *buf,
		      int flags,
		      struct bsctl_error *err)
{
	ssize_t rlen, wlen;

	rlen = pread(ifd, buf, size, i_off);
	if (rlen != (ssize_t) size && errno)
		return bsctl_error_set(err, errno, "input file read at %llu failed (%llu/%llu): %s",
				       (unsigned long long) i_off, rlen, size,
				       strerror(errno));

	/* skip manipulations with files if --no-write option given */
	if (flags & CF_NO_WRITE)
		goto out;

	wlen = pwrite(ofd, buf, rlen, o_off);
	if (wlen != rlen) {
		if (errno)
			return bsctl_error_set(err, errno,
					       "output file write at %llu failed (%llu/%llu): %s",
					       (unsigned long long) o_off,
					       wlen, rlen, strerror(errno));
		else
			return bsctl_error_set(err, EIO,
					       "unexpected write behavior at off %llu (%llu/%llu)",
					       (unsigned long long) o_off,
					       wlen, rlen);
	}

out:
	return 0;
}

/* returns fd if success, otherwise - negative error code */
static int open_separate_file(const char *filename, sector_t addr_sect,
			      struct bsctl_error *err)
{
	char fullname[PATH_MAX];
	int size, fd;

	size = snprintf(fullname, PATH_MAX, "%s.%llu",
			filename, (unsigned long long) addr_sect);
	if (size <= 0) {
		bsctl_error_set(err, errno,
				"could not store full name of separated file %s",
				filename);
		return -err->err;
	}

	fd = open(fullname, O_WRONLY | O_CREAT | O_TRUNC, FILE_MODE);
	if (fd < 0) {
		bsctl_error_set(err, errno, "failed to open file %s", fullname);
		return -err->err;
	}

	return fd;
}

int fd_to_filename(int fd, char *filename, size_t size)
{
	char path[PATH_MAX];

	if (snprintf(path, sizeof(path), "/proc/self/fd/%d", fd) < 0)
		return errno;

	if (readlink(path, filename, size) < 0)
		return errno;

	return 0;
}

void *copy_blocks(void *arg)
{
	uint64_t idx;
	sector_t pos_sect;
	int ifd, ofd;
	off_t offset;
	size_t size;
	char *buf;
	struct worker_args *args = (struct worker_args *) arg;

	buf = malloc(PAGE_SIZE);
	if (!buf) {
		bsctl_error_set(&args->err, errno,
				"%s: Failed to allocate %lu bytes for I/O buffer",
				__func__, PAGE_SIZE);
		goto exit;
	}

	ifd = args->ifd;

	if (!(args->flags & CF_SEPARATE)) {
		ofd = args->ofd;
	}

	for (idx = args->start; idx <= args->end; idx++) {
		if (args->blocks != NULL)
			pos_sect = args->blocks[idx];
		else
			pos_sect = idx * PAGE_SECTORS;

		offset = pos_sect << SECTOR_SHIFT;
		size = (args->dev_sectors - pos_sect >= PAGE_SECTORS) ?
				PAGE_SIZE :
				(args->dev_sectors - pos_sect) * SECTOR_SIZE;

		if ((args->flags & CF_SEPARATE)) {
			ofd = open_separate_file(args->snapfile, pos_sect,
						 &args->err);
			if (ofd < 0)
				goto free_buf;

			if (copy_block(ifd, offset,
				       ofd, 0,
				       size, buf,
				       args->flags, &args->err)) {
				close(ofd);
				goto free_buf;
			}

			if (close(ofd) == -1) {
				bsctl_error_set(&args->err, errno, "ERROR: Failed to close separate output file for block with sector #%lu [%d, '%s']",
						pos_sect, errno, strerror(errno));
				goto free_buf;
			}
		} else {
			if (copy_block(ifd, offset,
				       ofd, offset,
				       size, buf,
				       args->flags, &args->err)) {
				goto free_buf;
			}
		}

		if (!(args->flags & CF_HIDE_PROGRESS)) {
			pthread_mutex_lock(&block_counter.mutex);
			block_counter.blocks_done++;
			pthread_mutex_unlock(&block_counter.mutex);
		}
	}
free_buf:
	free(buf);
exit:
	if (args->err.err)
		bsctl_err(LOG_ERR, args->err.msg);
	pthread_exit(NULL);
}

/* It's forbidden to clear previous errors */
static void wait_for_workers(pthread_t *tids, struct worker_args *worker_args,
			     int workers, struct bsctl_error *err)
{
	int worker_idx;

	for (worker_idx = 0; worker_idx < workers; worker_idx++) {
		struct worker_args *worker = &worker_args[worker_idx];

		if (pthread_join(tids[worker_idx], NULL)) {
			if (!err->err)
				bsctl_err_set(err, errno,
					      "failed to join thread");
			else
				bsctl_err(LOG_ERR, "failed to join thread");
			return;
		}
		if (worker->err.err) {
			bsctl_err(LOG_ERR, err->msg);
			if (!err->err)
				bsctl_error_set(err, worker->err.err,
						"one of workers has failed");
		}
	}
}

/* as this variable initialize once and will not been
 * access concurrently, we could use this weird implementation
 */
static sector_t block_idx = 0;
int fill_array_of_blocks(struct blksnap_ioctl_disk_extent *extent,
			 unsigned int UNUSED(total_blocks),
			 void *data, struct bsctl_error * UNUSED(err))
{
	sector_t *blocks = data;
	unsigned int offset;

	for (offset = 0;
		offset < extent->nblocks * PAGE_SECTORS;
		offset += PAGE_SECTORS) {
		blocks[block_idx++] = extent->start_sect
				      + (unsigned long) offset;
	}

	return 0;
}

/* Get blocks are needed to be copied:
 * Incremental snapshot: request changelog from the kernel module
 * Classic base snapshot: calculate volume size to copy whole volume
 * Aafs base snapshot: request occupied blocks from aafs lib
 * Return: count of changed blocks
 */
sector_t get_copy_task(copy_snapshot_t *pcopy)
{
	sector_t nblocks;

	if (pcopy->flags & CF_INCREMENTAL) {
		/* request list of changed blocks from kernel module */
		struct changelog_stat *stat;

		nblocks = bsctl_chlog_size(pcopy->blkdev, &pcopy->err);
		if (pcopy->err.err)
			return 0;

		pcopy->blocks = malloc(nblocks * sizeof(sector_t));
		if (!pcopy->blocks) {
			bsctl_error_set(&pcopy->err, errno,
					"could not allocate %llu bytes to store changelog",
					nblocks * sizeof(sector_t));
			return 0;
		}

		stat = bsctl_chlog_process(pcopy->blkdev, nblocks,
					   &fill_array_of_blocks,
					   (void *) pcopy->blocks, &pcopy->err);
		if (!stat)
			free(pcopy->blocks);
		else
			free(stat);
	} else {
#ifdef BSRWAAFS
		/* request list of blocks from aafslib */
		nblocks = aafs_get_occupied_blocks(pcopy->snapdev,
						   &pcopy->blocks, &pcopy->err);
#else
		/* request volume size for copying it wholly */
		uint64_t dev_sectors = bsctl_get_device_size(pcopy->blkdev,
							     &pcopy->err);
		nblocks = (uint64_t) ceil((float) dev_sectors
					  / (float) PAGE_SECTORS);
#endif
	}

	return nblocks;
}

static int copy_snap(copy_snapshot_t *pcopy)
{
	int i, ret;
	sector_t nblocks_todo, start, end, blocks_per_worker;
	pthread_t tids[pcopy->workers], progress_bar_tid;
	struct worker_args worker_args[pcopy->workers];
	sector_t dev_sectors;

	dev_sectors = bsctl_get_device_size(pcopy->blkdev, &pcopy->err);
	if (!dev_sectors)
		return pcopy->err.err;

	nblocks_todo = get_copy_task(pcopy);
	if (pcopy->err.err)
		return pcopy->err.err;

	/* truncate image file for base snapshot */
	if (!(pcopy->flags & CF_INCREMENTAL || pcopy->flags & CF_NO_WRITE)) {
		struct stat stat_out;

		if (fstat(pcopy->ofd, &stat_out) == -1)
			return bsctl_error_set(&pcopy->err, errno,
					       "ERROR: fstat(%s) failed",
					       pcopy->snapfile);
		if (S_ISREG(stat_out.st_mode)) {
			ret = bsctl_file_reset(pcopy->ofd,
					       dev_sectors << SECTOR_SHIFT,
					       &pcopy->err);
			if (ret == -1)
				return ret;
		}
	}

	if (!(pcopy->flags & CF_HIDE_PROGRESS)) {
		ret = pthread_create(&progress_bar_tid, NULL, show_progress,
				     (void *) &nblocks_todo);
		if (ret) {
			bsctl_err_set(&pcopy->err, errno,
				      "failed to create thread in %s: %s",
				      __func__, strerror(errno));
			goto free_blocks;
		}
	}

	blocks_per_worker = (sector_t) ceil((float) nblocks_todo /
					    (float) pcopy->workers);

	bsctl_msg(LOG_DEBUG,
		  "Device sectors: %llu; Blocks to copy: %llu; Workers: %d; Blocks per workers: %llu",
		  dev_sectors, nblocks_todo, pcopy->workers, blocks_per_worker);

	memset(worker_args, 0, sizeof(struct worker_args) * pcopy->workers);
	for (i = 0; i < pcopy->workers; i++) {
		start = i * blocks_per_worker;
		if (start + blocks_per_worker <= nblocks_todo)
			end = start + blocks_per_worker - 1;
		else
			end = nblocks_todo - 1;

		bsctl_msg(LOG_DEBUG,
			  "Worker #%u started with blocks %srange [%llu, %llu]",
			  i, pcopy->blocks ? "index " : "",
			  (unsigned long long) start, (unsigned long long) end);

		worker_args[i].snapdev = pcopy->snapdev;
		worker_args[i].snapfile = pcopy->snapfile;
		worker_args[i].ifd = pcopy->ifd;
		worker_args[i].ofd = pcopy->ofd;
		worker_args[i].blocks = pcopy->blocks;
		worker_args[i].start = start;
		worker_args[i].end = end;
		worker_args[i].dev_sectors = dev_sectors;
		worker_args[i].flags = pcopy->flags;

		ret = pthread_create(&tids[i], NULL, copy_blocks,
				     (void *) &worker_args[i]);
		if (ret) {
			bsctl_err_set(&pcopy->err, errno,
				      "failed to create worker thread [%d]: %s",
				      i, strerror(errno));
			pcopy->workers = i;
			goto wait_workers;
		}
		bsctl_msg(LOG_DEBUG, "Worker thread #%u started with tid %lu",
			  i, tids[i]);
	}

wait_workers:
	wait_for_workers(tids, worker_args, pcopy->workers, &pcopy->err);
	if (!(pcopy->flags & CF_HIDE_PROGRESS))
		pthread_join(progress_bar_tid, NULL);
free_blocks:
	free(pcopy->blocks);
	return pcopy->err.err;
}

__attribute__ ((noreturn))
static void usage(char *name, FILE *out, int ret)
{
	fprintf(out, "Usage: %s [OPTION] [DEVICE] [FILE]\n\n", basename(name));

	fprintf(out,
"    -h, --help                     show this help message\n"
"    -v, --version                  show blksnap version\n"
"    -c, --cli-version              show CLI (command line interface) version\n"
"    -i, --incremental              perform incremental snapshot (base snapshot\n"
"                                   is assumed otherwise)\n"
"    -w, --workers                  perform getting snapshot with N parallel\n"
"                                   workers (default - 1)\n"
"    -s, --separated                if perform incremental snapshot - write\n"
"                                   data to separated files FILE.SECTOR_ADDR\n"
"                                   (default - write data to single file FILE\n"
"                                   with real offset of each sector)\n"
"    -p, --hide-progress            do not show progress indicator\n"
"    -n, --no-space-check           don't check free space on target device\n"
"    -d, --no-write                 dry-run (don't write to target file\n"
"\n"
	);

	fprintf(out, "For more information see %s(8).\n", name);

	exit(ret);
}

__attribute__ ((noreturn))
static void version(char *name)
{
	fprintf(stdout, "%s version %s\n", basename(name), BLKSNAP_VERSION);
	exit(EXIT_SUCCESS);
}

__attribute__ ((noreturn))
static void cli_version(void)
{
	fprintf(stdout, "%u\n", CLI_VERSION);
	exit(EXIT_SUCCESS);
}

void init_copy_snap(copy_snapshot_t *pcopy)
{
	bzero(pcopy, sizeof(*pcopy));
	pcopy->workers = 1;	/* default - 1 copy worker */
	pcopy->flags = 0;	/* default - 0 */
}

int main(int argc, char **argv)
{
	int c, ret = 0, log_pri = LOG_INFO;
	copy_snapshot_t copy;

	init_copy_snap(&copy);
	bsctl_err_reset(&copy.err);
	bsctl_set_log_priority(&log_pri);

	const char *sopts = "hivw:spcnd";
	struct option lopts[] = {
		/* log options */
		{ "debug",		0, &log_pri, LOG_DEBUG },
		{ "verbose",		0, &log_pri, LOG_INFO },
		{ "quiet",		0, &log_pri, LOG_EMERG },

		/* main commands */
		{ "help",		0, 0, 'h' },
		{ "version",		0, 0, 'v' },
		{ "cli-version",	0, 0, 'c' },
		{ "incremental",	0, 0, 'i' },
		{ "workers",		1, 0, 'w' },
		{ "separated",		0, 0, 's' },
		{ "no-space-check",	0, 0, 'n' },
		{ "no-write",		0, 0, 'd' },

		/* output options */
		{ "hide-progress",	0, 0, 'p' },
		{ NULL,			0, 0,  0  }
	};

	while ((c = getopt_long(argc, argv, sopts, lopts, NULL)) != -1) {
		switch (c) {
		case 0: /* --debug, --verbose, --quiet */
			bsctl_set_log_priority(&log_pri);
			break;
		case 'h':
			usage(argv[0], stdout, 0);
			break;
		case 'i':
			copy.flags |= CF_INCREMENTAL;
			break;
		case 'v':
			version(argv[0]);
			break;
		case 'c':
			cli_version();
			break;
		case 'w':
			copy.workers = atoi(optarg);
			break;
		case 's':
			copy.flags |= CF_SEPARATE;
			break;
		case 'p':
			copy.flags |= CF_HIDE_PROGRESS;
			break;
		case 'n':
			copy.flags |= CF_NO_SPACE_CHECK;
			break;
		case 'd':
			copy.flags |= CF_NO_WRITE;
			break;
		default:
			usage(argv[0], stderr, EINVAL);
			break;
		}
	}

	if (argc - optind != 2) {
		bsctl_err(LOG_ERR, "ERROR: unexpected number of arguments");
		usage(argv[0], stderr, EINVAL);
	}

	strncpy(copy.blkdev, argv[optind++], BLKSNAP_NAME_SIZE);
	strncpy(copy.snapfile, argv[optind++], BLKSNAP_NAME_SIZE);

	if (copy.workers < 1) {
		bsctl_err(LOG_ERR,
			  "ERROR: --workers must be greater or equal to 1");
		usage(argv[0], stderr, ERANGE);
	}

	/* verify valid block device */
	bsctl_get_rdev(copy.blkdev, &copy.err);
	ret = copy.err.err;
	if (ret)
		goto out;

	ret = bsctl_snapshot_get(copy.blkdev, copy.snapdev, &copy.err);
	if (ret)
		goto out;

	bsctl_msg(LOG_DEBUG, "SnapShot Device for source %s: %s",
		  copy.blkdev, copy.snapdev);

	copy.ifd = open(copy.snapdev, O_RDONLY);
	if (copy.ifd < 0) {
		ret = errno;
		bsctl_err(LOG_ERR, "ERROR: could not open file %s: %s",
			  copy.snapdev, strerror(ret));
		goto out;
	}

	if (!(copy.flags & CF_SEPARATE)) {
		int open_flags = O_RDWR | O_CREAT;
		open_flags |= (!(copy.flags & CF_INCREMENTAL)) ? O_TRUNC : 0;

		copy.ofd = open(copy.snapfile, open_flags, FILE_MODE);
		if (copy.ofd < 0) {
			ret = errno;
			bsctl_err(LOG_ERR, "ERROR: could not open file %s", copy.snapfile);
			goto close_ifd;
		}
	}

	bsctl_msg(LOG_NOTICE,
		  "Performing %s snapshot from %s (source: %s) to %s with %u %s",
		  (copy.flags & CF_INCREMENTAL) ? "incremental" : "base",
		  copy.snapdev, copy.blkdev, copy.snapfile, copy.workers,
		  (copy.workers > 1) ? "workers" : "worker");

	bsctl_msg(LOG_DEBUG, "BlkDev: %s SnapDev: %s SnapFile: %s",
		  copy.blkdev, copy.snapdev, copy.snapfile);

	ret = copy_snap(&copy);

	if (!(copy.flags & CF_SEPARATE)) {
		if (close(copy.ofd) == -1)
			ret = bsctl_error_set(&copy.err, errno,
					      "ERROR: failed to close %s [%d, '%s']",
					      copy.snapfile, errno,
					      strerror(errno));
	}
close_ifd:
	close(copy.ifd);
out:
	if (ret)
		bsctl_err(LOG_ERR, copy.err.msg);

	return ret;
}
