#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <search.h>
#include "bsctl-error.h"
#include "bsctl-share.h"
#include "btrfs/address-tree.h"

struct chunk {
	uint64_t logical;
	uint64_t physical;
	uint64_t bytes;
};

static void *_alloc_chunk(struct bsctl_error *err)
{
	struct chunk *ptr;

	ptr = malloc(sizeof(struct chunk));
	if (!ptr)
		bsctl_error_set(err, errno, "malloc failed: %s", strerror(errno));

	return ptr;
}

static int __cmp(const void *a, const void *b)
{
	const struct chunk *entry;
	const struct chunk *ins;

	ins = (struct chunk*)a;
	entry = (struct chunk*)b;

	if (ins->logical >= entry->logical &&
	    ins->logical < entry->logical + entry->bytes)
		return 0;

	if (ins->logical < entry->logical)
		return -1;
	else if (ins->logical > entry->logical)
		return 1;

	return 0;
}

uint64_t bsctl_logical_to_physical(void *root, uint64_t logical)
{
	void *val;
	struct chunk search = {logical, 0, 0};
	struct chunk *found;
	uint64_t offset;

	val = tfind((void *)&search, &root, __cmp);
	if (val == NULL)
		return logical;

	found = *(struct chunk**)val;
	offset = logical - found->logical;

	return found->physical + offset;
}

int bsctl_insert_chunk(void **root,
		       uint64_t l_bytenr, uint64_t p_bytenr, uint64_t len,
		       struct bsctl_error *err)
{
	int ret = 0;
	void *val;
	struct chunk *chunk;

	chunk = _alloc_chunk(err);
	if (chunk == NULL)
		return ENOMEM;

	chunk->logical = l_bytenr;
	chunk->physical = p_bytenr;
	chunk->bytes = len;
	val = tsearch((void *) chunk, root, __cmp);

	if (val == NULL) {
		ret = bsctl_error_set(err, errno, "tsearch failed: %s",
				      strerror(errno));
		free(chunk);
		return ret;
	}
	if ((*(struct chunk **) val) != chunk)
		free(chunk);

	return ret;
}

void bsctl_free_address_tree(void *root)
{
	tdestroy(root, free);
}

static void __print_action(const void *nodep,
			   const VISIT which,
			   const int UNUSED(level))
{
	struct chunk *entry;

	switch (which) {
	case preorder:
		break;
	case postorder:
		entry = *(struct chunk**)nodep;
		printf("logical = %llu, physical = %llu, length = %llu\n",
		       (unsigned long long) entry->logical,
		       (unsigned long long) entry->physical,
		       (unsigned long long) entry->bytes);
		break;
	case endorder:
		break;
	case leaf:
		entry = *(struct chunk**)nodep;
		printf("logical = %llu, physical = %llu, length = %llu\n",
		       (unsigned long long) entry->logical,
		       (unsigned long long) entry->physical,
		       (unsigned long long) entry->bytes);
		break;
	}
}

void bsctl_debug_print_addr_tree(void *root)
{
	twalk(root, __print_action);
}
