#ifndef _BSCTL_BTRFS_H
#define _BSCTL_BTRFS_H

#include <assert.h>
#include <stddef.h>

#define BTRFS_MAGIC 0x4D5F53665248425FULL /* ascii _BHRfS_M, no null */

#define BTRFS_CHUNK_ITEM_KEY	228

#define BTRFS_SUPER_INFO_OFFSET	(64 * 1024)

struct btrfs_key {
	uint64_t objectid;
	uint8_t type;
	uint64_t offset;
} __attribute__ ((__packed__));

#define BTRFS_FSID_SIZE	16
#define BTRFS_UUID_SIZE	16
#define BTRFS_CSUM_SIZE	32

struct btrfs_stripe {
	uint64_t devid;
	uint64_t offset;
	uint8_t dev_uuid[BTRFS_UUID_SIZE];
} __attribute__ ((__packed__));


struct btrfs_chunk {
	/* size of this chunk in bytes */
	uint64_t length;

	/* objectid of the root referencing this chunk */
	uint64_t owner;

	uint64_t stripe_len;
	uint64_t type;

	/* optimal io alignment for this chunk */
	uint32_t io_align;

	/* optimal io width for this chunk */
	uint32_t io_width;

	/* minimal io size for this chunk */
	uint32_t sector_size;

	/* 2^16 stripes is quite a lot, a second limit is the size of a single
	* item in the btree
	*/
	uint16_t num_stripes;

	/* sub stripes only matter for raid10 */
	uint16_t sub_stripes;
	struct btrfs_stripe stripe;
	/* additional stripes go here */
} __attribute__ ((__packed__));

#define BTRFS_SYSTEM_CHUNK_ARRAY_SIZE 2048
#define BTRFS_LABEL_SIZE 256

struct btrfs_super_block {
	uint8_t csum[BTRFS_CSUM_SIZE];
	/* the first 3 fields must match struct btrfs_header */
	uint8_t fsid[BTRFS_FSID_SIZE];    /* FS specific uuid */
	uint64_t bytenr; /* this block number */
	uint64_t flags;

	/* allowed to be different from the btrfs_header from here own down */
	uint64_t magic;
	uint64_t generation;
	uint64_t root;
	uint64_t chunk_root;
	uint64_t log_root;

	uint64_t log_root_transid;
	uint64_t total_bytes;
	uint64_t bytes_used;
	uint64_t root_dir_objectid;
	uint64_t num_devices;
	uint32_t sectorsize;
	uint32_t nodesize;
	uint32_t leafsize;
	uint32_t stripesize;
	uint32_t sys_chunk_array_size;
	uint64_t chunk_root_generation;
	uint64_t compat_flags;
	uint64_t compat_ro_flags;
	uint64_t incompat_flags;
	uint16_t csum_type;
	uint8_t root_level;
	uint8_t chunk_root_level;
	uint8_t log_root_level;
	uint8_t dev_item[98];

	char label[BTRFS_LABEL_SIZE];

	uint64_t cache_generation;
	uint64_t uuid_tree_generation;

	uint64_t reserved[30];
	uint8_t sys_chunk_array[BTRFS_SYSTEM_CHUNK_ARRAY_SIZE];
	uint8_t super_roots[672];
} __attribute__ ((__packed__));

struct btrfs_header {
	/* these first four must match the super block */
	uint8_t csum[BTRFS_CSUM_SIZE];
	uint8_t fsid[BTRFS_FSID_SIZE]; /* FS specific uuid */
	uint64_t bytenr; /* which block this node is supposed to live in */
	uint64_t flags;

	/* allowed to be different from the super from here on down */
	uint8_t chunk_tree_uuid[BTRFS_UUID_SIZE];
	uint64_t generation;
	uint64_t owner;
	uint32_t nritems;
	uint8_t level;
} __attribute__ ((__packed__));

struct btrfs_item {
	struct btrfs_key key;
	uint32_t offset;
	uint32_t size;
} __attribute__ ((__packed__));

struct btrfs_leaf {
	struct btrfs_header header;
	struct btrfs_item items[];
} __attribute__ ((__packed__));

struct btrfs_key_ptr {
	struct btrfs_key key;
	uint64_t blockptr;
	uint64_t generation;
} __attribute__ ((__packed__));

struct btrfs_node {
	struct btrfs_header header;
	struct btrfs_key_ptr ptrs[];
} __attribute__ ((__packed__));

/* In-RAM container of btrfs node */
struct node_buffer {
	uint32_t len;
	char data[];
};

static inline uint32_t btree_node_size(struct btrfs_super_block *super,
				       int level)
{
	if (level == 0)
		return super->leafsize;
	return super->nodesize;
}

static inline unsigned long btrfs_chunk_item_size(int num_stripes)
{
	assert(num_stripes);
	return sizeof(struct btrfs_chunk) +
		sizeof(struct btrfs_stripe) * (num_stripes - 1);
}

static void get_from_node_buffer(struct node_buffer *nb, void *dst,
				 uint32_t offset, uint32_t len)
{
	memcpy(dst, nb->data + offset, len);
}

#define read_nb_member(nb, ptr, type, member, result) (		\
	get_from_node_buffer(nb, (char *)(result),		\
			     ((unsigned long)(ptr)) +		\
			     offsetof(type, member),		\
			     sizeof(((type *)0)->member)))

static inline uint32_t btrfs_header_nritems(struct node_buffer* nb)
{
	struct btrfs_header *h = (struct btrfs_header *)nb->data;
	return h->nritems;
}

static inline uint8_t btrfs_header_level(struct node_buffer *nb)
{
	struct btrfs_header *h = (struct btrfs_header *)nb->data;
	return h->level;
}

static inline uint64_t btrfs_key_blockptr(struct node_buffer *nb,
					  struct btrfs_key_ptr *s)
{
	unsigned long offset = (unsigned long)s;
	struct btrfs_key_ptr *p = (struct btrfs_key_ptr *) (nb->data + offset);
	return p->blockptr;
}

/* exact fields from node items */
static inline uint64_t btrfs_node_blockptr(struct node_buffer *eb, int nr)
{
	unsigned long ptr;
	ptr = offsetof(struct btrfs_node, ptrs) +
	      sizeof(struct btrfs_key_ptr) * nr;
	return btrfs_key_blockptr(eb, (struct btrfs_key_ptr *)ptr);
}

static inline int btrfs_is_leaf(struct node_buffer *eb)
{
	return (btrfs_header_level(eb) == 0);
}

static inline unsigned long btrfs_item_nr_offset(int nr)
{
	return offsetof(struct btrfs_leaf, items) +
			sizeof(struct btrfs_item) * nr;
}

static inline struct btrfs_item *btrfs_item_nr(int nr)
{
	return (struct btrfs_item *)btrfs_item_nr_offset(nr);
}

static inline void btrfs_item_key(struct node_buffer *nb,
				  struct btrfs_key *disk_key, int nr)
{
	struct btrfs_item *item = btrfs_item_nr(nr);
	read_nb_member(nb, item, struct btrfs_item, key, disk_key);
}

static inline uint32_t btrfs_item_offset(struct node_buffer *nb,
					 struct btrfs_item *s)
{
	unsigned long offset = (unsigned long)s;
	struct btrfs_item *p = (struct btrfs_item *) (nb->data + offset);
	return p->offset;
}

static inline uint32_t btrfs_item_offset_nr(struct node_buffer *nb, int nr)
{
	return btrfs_item_offset(nb, btrfs_item_nr(nr));
}

static inline unsigned long btrfs_leaf_data()
{
	return offsetof(struct btrfs_leaf, items);
}

#define btrfs_item_ptr_offset(leaf, slot)	\
	((unsigned long)(btrfs_leaf_data() +	\
	btrfs_item_offset_nr(leaf, slot)))

#endif
