#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include "bsctl-error.h"
#include "btrfs/format.h"
#include "btrfs/address-tree.h"
#include "btrfs/io.h"

static struct node_buffer *__alloc_nb(uint32_t size, struct bsctl_error *err)
{
	struct node_buffer *nb;

	nb = calloc(1, sizeof(struct node_buffer) + size);
	if (!nb) {
		bsctl_error_set(err, errno, "Node buffer allocation failed: %s",
				strerror(errno));
		return NULL;
	}

	nb->len = size;
	return nb;
}

static int __read_safe(int fd, void *buffer, uint32_t size, uint64_t bytenr,
		       struct bsctl_error *err)
{
	int r;
	uint8_t *buf = (uint8_t*) buffer;

	while (size && (r = pread64(fd, buf, size, bytenr))) {
		if (r == -1) {
			if (errno == EINTR)
				continue;
			return bsctl_error_set(err, errno,
				"Failed to read Btrfs device from address %llu: %s",
				(unsigned long long)bytenr, strerror(errno));
		}
		size -= r;
		bytenr += r;
		buf += r;
	}

	return 0;
}

int bsctl_read_btrfs_superblock(int fd, struct btrfs_super_block *super,
				struct bsctl_error *err)
{
	int ret;
	uint64_t bytenr = BTRFS_SUPER_INFO_OFFSET;

	ret = __read_safe(fd, (uint8_t*)super, sizeof(struct btrfs_super_block),
			bytenr, err);
	if (ret)
		return ret;

	if (super->magic != BTRFS_MAGIC
	    || super->bytenr != BTRFS_SUPER_INFO_OFFSET) {
		return bsctl_error_set(err, EBADF,
				"Btrfs superblock appears to be corrupted");
	}

	return 0;
}

int bsctl_read_node(int fd, struct node_buffer **nb,
		    uint64_t logical_offset, uint32_t node_size,
		    void *addr_tree, struct bsctl_error *err)
{
	int ret = 0;
	uint64_t physical_offset;

	physical_offset = bsctl_logical_to_physical(addr_tree, logical_offset);
	*nb = __alloc_nb(node_size, err);
	if (*nb == NULL)
		return ENOMEM;

	ret = __read_safe(fd, (*nb)->data, node_size, physical_offset, err);
	if (ret)
		free(*nb);

	return ret;
}
