#include <stdint.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <linux/fiemap.h>

#include "bsctl-error.h"
#include "btrfs/format.h"
#include "btrfs/io.h"
#include "btrfs/address-tree.h"
#include "btrfs/log-to-phys.h"

/* Walks through superblock->sys_chunk_array to build rb-tree of chunks.
 * These chunks contains B-tree of chunks - that solves egg problem.
 */
static int __collect_init_address_tree(
			struct btrfs_super_block *super,
			void **addr_tree,
			int (*bsctl_insert_chunk)(void **root, uint64_t logical,
						  uint64_t physical,
						  uint64_t len,
						  struct bsctl_error *err),
			struct bsctl_error *err)
{
	int ret = 0;
	int num_stripes;
	uint8_t *ptr;
	uint32_t cur = 0;
	struct btrfs_key disk_key;
	struct btrfs_chunk *chunk;

	ptr = super->sys_chunk_array;

	while (cur < super->sys_chunk_array_size) {
		/* we can't use pointers assignment because of CentOS 6 gcc
		 * error "breaking strict-aliasing rules" on dereferencing
		 * disk_key->type
		 */
		memcpy(&disk_key, ptr, sizeof(disk_key));
		ptr += sizeof(disk_key);
		cur += sizeof(disk_key);

		if (disk_key.type == BTRFS_CHUNK_ITEM_KEY) {
			chunk = (struct btrfs_chunk *)ptr;
			ret = bsctl_insert_chunk(addr_tree, disk_key.offset,
						 chunk->stripe.offset,
						 chunk->length,
						 err);
			if (ret)
				break;

			num_stripes = chunk->num_stripes;
			ptr += btrfs_chunk_item_size(num_stripes);
			cur += btrfs_chunk_item_size(num_stripes);
		} else {
			bsctl_error_set(err, EBADF,
					"ERROR: Unexpected data in superblock");
			ret = EBADF;
			break;
		}
	}

	return ret;
}

static int __handle_chunk_leaf(struct node_buffer *leaf,
			int (*bsctl_insert_chunk)(void **root, uint64_t logical,
						  uint64_t physical,
						  uint64_t len,
						  struct bsctl_error *err),
			void **addr_tree, struct bsctl_error *err)
{
	int ret = 0;
	uint32_t i;
	uint32_t nr_items = btrfs_header_nritems(leaf);

	for (i = 0; i < nr_items && !ret; i++) {
		struct btrfs_key key;
		struct btrfs_chunk chunk_item;

		btrfs_item_key(leaf, &key, i);

		if (key.type != BTRFS_CHUNK_ITEM_KEY)
			continue;
		get_from_node_buffer(leaf, &chunk_item,
				     btrfs_item_ptr_offset(leaf, i),
				     sizeof(chunk_item));
		ret = bsctl_insert_chunk(addr_tree, key.offset,
					 chunk_item.stripe.offset,
					 chunk_item.length,
					 err);
	}

	return ret;
}

static int __walk_chunk_tree(
			int fd, struct btrfs_super_block *super,
			struct node_buffer *node,
			void **addr_tree,
			int (*bsctl_insert_chunk)(void **root, uint64_t logical,
						  uint64_t physical,
						  uint64_t len,
						  struct bsctl_error *err),
			struct bsctl_error *err)
{
	int ret = 0;
	uint32_t i, nr, size;

	if (btrfs_is_leaf(node))
		return __handle_chunk_leaf(node, bsctl_insert_chunk, addr_tree,
					   err);

	size = btree_node_size(super, btrfs_header_level(node) - 1);
	nr = btrfs_header_nritems(node);
	for (i = 0; i < nr && !ret; i++) {
		struct node_buffer *next;
		ret = bsctl_read_node(fd, &next, btrfs_node_blockptr(node, i), size,
				*addr_tree, err);
		if (ret)
			return ret;
		ret = __walk_chunk_tree(fd, super, next, addr_tree,
					bsctl_insert_chunk, err);
		free(next);
	}

	return ret;
}

/*
 * Btrfs fiemap/fibmap implementation doesn't assume that consumer is
 * going to threat files directly, by addresses. So, they give logical
 * addresses in terms of Btrfs. But we need to have physical addresses to
 * write/read to file to/from from module.
 * To calculate physical addresses from logical, we need to know the map
 * of Btrfs chunks. It can be read from the Btrfs on-disk chunk tree. It
 * also can be printed by btrfs-debug-tree command. Every chunk has
 * logical offset, physical offset (they can be equal) and length.
 *
 * if:
 *   log_addr - value got from fiemap ioctl
 *   phys_addr - value are needed for module
 *   chunk - chunk [logical_start, length] got from chunk tree for which
 *   this condition is right:
 *   chunk.logical_start <= log_addr <= chunk.logical_start + chunk.length
 * then:
 * phys_addr = log_addr + (chunk.physical_start - chunk.logical_start)
 */
int bsctl_btrfs_logical_to_physical(const char *dev_path,
				    struct fiemap *fiemap,
				    struct bsctl_error *err)
{
	int ret;
	int fd;
	size_t i;
	struct fiemap_extent *extent;
	struct btrfs_super_block super;
	void *addr_tree = NULL;

	fd = open(dev_path, O_RDONLY);
	if (fd == -1)
		return bsctl_error_set(err, errno, "Failed to open %s: %s",
				       dev_path, strerror(errno));

	ret = bsctl_read_btrfs_superblock(fd, &super, err);
	if (ret)
		return ret;

	/* build address tree of chunks contained extents with chunk tree */
	ret = __collect_init_address_tree(&super, &addr_tree,
					  bsctl_insert_chunk, err);
	if (ret)
		goto clean;

	/* read root node of chunk tree */
	struct node_buffer *chunk_tree_root;
	ret = bsctl_read_node(fd, &chunk_tree_root, super.chunk_root,
			btree_node_size(&super, super.chunk_root_level),
			addr_tree, err);
	if (ret)
		goto clean;

	/* build whole address tree */
	ret = __walk_chunk_tree(fd, &super, chunk_tree_root, &addr_tree,
				bsctl_insert_chunk, err);
	free(chunk_tree_root);
	if (ret)
		goto clean;

	/* translate fiemap addresses */
	for (i = 0; i < fiemap->fm_mapped_extents; i++) {
		extent = &fiemap->fm_extents[i];
		extent->fe_physical = bsctl_logical_to_physical(addr_tree,
							   extent->fe_physical);
	}

clean:
	bsctl_free_address_tree(addr_tree);
	close(fd);
	return ret;
}
