#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/fiemap.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <err.h>
#include "bsctl-error.h"
#include "fibmap.h"

struct file_extent {
	__u64 byte_offset;
	__u64 first_block;
	__u64 last_block;
	__u64 block_count;
};

static int add_extent(struct fiemap **fiemap, struct file_extent ext,
		      unsigned int blksize, struct bsctl_error *err)
{
	struct fiemap_extent *extent;
	struct fiemap *new_fiemap = NULL;

	(*fiemap)->fm_mapped_extents++;
	new_fiemap = realloc(*fiemap, sizeof(struct fiemap)
				      + ((*fiemap)->fm_mapped_extents
					 * sizeof(struct fiemap_extent)));
	if (new_fiemap == NULL) {
		return bsctl_error_set(err, errno,
				      "%s: realloc failed", __func__);
	}

	extent = &new_fiemap->fm_extents[new_fiemap->fm_mapped_extents - 1];
	extent->fe_logical  = ext.byte_offset;
	extent->fe_physical = ext.first_block * blksize;
	extent->fe_length   = ext.block_count * blksize;

	*fiemap = new_fiemap;
	return 0;
}

struct fiemap *bsctl_get_fiemap_via_fibmap(const char *file_path,
					   struct bsctl_error *err)
{
	int fd;
	unsigned int blksize;
	unsigned long num_blocks;
	__u64 blk_idx, hole = ~0ULL;
	struct file_extent ext;
	struct fiemap *fiemap = NULL;
	struct stat statinfo;

	fd = open(file_path, O_RDONLY | O_SYNC);
	if (fd < 0) {
		bsctl_error_set(err, errno, "ERROR: Failed to open %s",
				file_path);
		return NULL;
	}

	if (ioctl(fd, FIGETBSZ, &blksize) == -1) {
		bsctl_error_set(err, errno, "%s: unable to determine block size",
				__func__);
		goto clean_up;
	};

	if (fstat(fd, &statinfo)) {
		bsctl_error_set(err, errno, "ERROR: Failed to fstat %s in %s",
				file_path, __func__);
		goto clean_up;
	}

	fiemap = calloc(1, sizeof(struct fiemap));
	if (!fiemap) {
		bsctl_error_set(err, errno, "ERROR: calloc failed in %s", __func__);
		goto clean_up;
	}
	num_blocks = (statinfo.st_size + blksize - 1) / blksize;
	memset(&ext, 0, sizeof(ext));

	for (blk_idx = 0; blk_idx < num_blocks; blk_idx++) {
		unsigned int blknum = blk_idx;
		__u64 blknum64;

		if (ioctl(fd, FIBMAP, &blknum) == -1) {
			bsctl_error_set(err, errno, "FIBMAP ioctl failed");
			goto clean_up;
		}

		if (!blknum) {
			bsctl_error_set(err, ESPIPE,
					"FIBMAP returned address 0. Maybe file is not allocated.");
			goto clean_up;
		}
		blknum64 = blknum;
		if (blk_idx && blknum64 == (ext.last_block + 1)) {
			/* Continuation of extent */
			ext.last_block = blknum64 ? blknum64 : hole;
			ext.block_count++;
		} else {
			/* New extent: add previous extent to fiemap and re-init
			 * the extent record.
			 */
			if (blk_idx)
				if (add_extent(&fiemap, ext, blksize, err))
					goto clean_up;
			ext.first_block = blknum64;
			ext.last_block  = blknum64 ? blknum64 : hole;
			ext.block_count = 1;
			ext.byte_offset = blk_idx * blksize;
		}
	}
	if (add_extent(&fiemap, ext, blksize, err))
		goto clean_up;

	close(fd);
	return fiemap;

clean_up:
	if (fiemap)
		free(fiemap);
	close(fd);
	return NULL;
}
